﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

using Windows.UI;
using Windows.UI.Input;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Media.Animation;
using Windows.UI.Xaml.Shapes;

using UIPoint = Windows.Foundation.Point;

using PDFPoint = pdftron.PDF.Point;
using PDFRect = pdftron.PDF.Rect;
using PDFPage = pdftron.PDF.Page;
using PDFDouble = pdftron.Common.DoubleRef;
using pdftron.PDF;
using pdftron.Common;

namespace pdftron.PDF.Tools
{
    /// <summary>
    /// This class is the base class for several shape creation classes that need clicks to determine
    /// vertices vs drag and drop, e.g., Polyline, Polygon, Cloud and etc.
    /// </summary>
    public abstract class AdvancedShapeCreate : SimpleShapeCreate
    {
        #region Members / Properties

        #endregion

        #region Constructor

        public AdvancedShapeCreate(PDFViewCtrl pdfViewControl, ToolManager toolManager) 
            : base(pdfViewControl, toolManager)
        {
            if (pdfViewControl == null)
            {
                throw new ArgumentNullException("PDFViewCtrl");
            }

            if (toolManager == null)
            {
                throw new ArgumentNullException("ToolManager");
            }
        }

        #endregion

        #region Event Handlers
        
        internal override bool PointerPressedHandler(object sender, PointerRoutedEventArgs e)
        {
            try
            {
                if ((e.Pointer.PointerDeviceType == Windows.Devices.Input.PointerDeviceType.Mouse) && (!e.GetCurrentPoint(null).Properties.IsLeftButtonPressed)
                   && (e.Pointer.PointerDeviceType != Windows.Devices.Input.PointerDeviceType.Touch))
                {
                    mSimpleShapeCanceledAtPress = true;
                    return false;
                }

                base.PointerPressedHandler(sender, e);

                if (!mStartDrawing)
                {
                    InitializeDrawing();
                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }

            return false;
        }        

        internal override bool PointerMovedHandler(object sender, PointerRoutedEventArgs e)
        {
            try
            {
                if (e.Pointer.PointerId != mPointerID)
                {
                    return false;
                }

                UIPoint currentPosition = e.GetCurrentPoint(mViewerCanvas).Position;

                m_cnPt2.x = currentPosition.X;
                m_cnPt2.y = currentPosition.Y;

                TrimToPageBounds(m_cnPt2);

                // Size the canvas according to the two points
                if (mStartDrawing)
                {
                    this.Width = Math.Abs(m_cnPt2.x - m_cnPt1.x);
                    this.Height = Math.Abs(m_cnPt2.y - m_cnPt1.y);
                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }

            return false;
        }
        
        internal override bool PointerReleasedHandler(object sender, PointerRoutedEventArgs e)
        {
            mPointerDown = false;
            return false;
        }        

        internal override bool PointerCaptureLostHandler(object sender, PointerRoutedEventArgs e)
        {
            return base.PointerCaptureLostHandler(sender, e);
        }

        internal override bool PointerCanceledHandler(object sender, PointerRoutedEventArgs e)
        {
            return base.PointerCanceledHandler(sender, e);
        }

        internal override bool TappedHandler(object sender, TappedRoutedEventArgs e)
        {
            if (!mStartDrawing)
            {
                InitializeDrawing();
            }
            return false;
        }

        internal override bool DoubleTappedHandler(object sender, DoubleTappedRoutedEventArgs e)
        {
            return false;
        }

        internal override bool RightTappedHandler(Object sender, RightTappedRoutedEventArgs e)
        {
            return false;
        }

        internal override bool KeyDownAction(object sender, KeyRoutedEventArgs e)
        {
            return base.KeyDownAction(sender, e);
        }

        #endregion

        #region Methods

        private void InitializeDrawing()
        {
            if (m_cnPt1 == null)
                return;

            mStartDrawing = true;
            mViewerCanvas.Children.Add(this);
            this.SetValue(Canvas.LeftProperty, m_cnPt1.x);
            this.SetValue(Canvas.TopProperty, m_cnPt1.y);
            mRenderTransform = new CompositeTransform();
            this.RenderTransform = mRenderTransform;
            mMirrorX = false;
            mMirrorY = false;
            this.Opacity = mOpacity;
        }

        internal override void OnClose()
        {
            base.OnClose();
        }

        internal override bool OnScale()
        {
            return base.OnScale();
        }

        internal override bool OnSize()
        {
            return base.OnSize();
        }

        private void TrimToPageBounds(PDFPoint point)
        {
            // want to make sure we're within our bounding box.
            if (mPageCropOnClient != null)
            {
                if (point.x < mPageCropOnClient.x1)
                {
                    point.x = mPageCropOnClient.x1;
                }
                if (point.x > mPageCropOnClient.x2)
                {
                    point.x = mPageCropOnClient.x2;
                }
                if (point.y < mPageCropOnClient.y1)
                {
                    point.y = mPageCropOnClient.y1;
                }
                if (point.y > mPageCropOnClient.y2)
                {
                    point.y = mPageCropOnClient.y2;
                }
            }
        }

        internal void TrimToPageBounds(PDFPoint point, double drawThickness)
        {
            // we need to restrict the border further, since the line can stick out by halfthickness
            double halfThickness = drawThickness / 2;
            if (mPageCropOnClient != null)
            {
                if (point.x < mPageCropOnClient.x1 + halfThickness)
                {
                    point.x = mPageCropOnClient.x1 + halfThickness;
                }
                if (point.x > mPageCropOnClient.x2 - halfThickness)
                {
                    point.x = mPageCropOnClient.x2 - halfThickness;
                }
                if (point.y < mPageCropOnClient.y1 + halfThickness)
                {
                    point.y = mPageCropOnClient.y1 + halfThickness;
                }
                if (point.y > mPageCropOnClient.y2 - halfThickness)
                {
                    point.y = mPageCropOnClient.y2 - halfThickness;
                }
            }
        }

        protected PDFRect GetAdvancedShapeBBox(List<PDFPoint> pointList)
        {
            try
            {
                if ((pointList == null) || (pointList.Count == 0))
                {
                    return new PDFRect(0, 0, 0, 0);
                }

                double minX = Double.MaxValue;
                double minY = Double.MaxValue;
                double maxX = Double.MinValue;
                double maxY = Double.MinValue;

                foreach (PDFPoint point in pointList)
                {
                    if (point.x < minX)
                    {
                        minX = point.x; 
                    }

                    if (point.y < minY)
                    {
                        minY = point.y;
                    }

                    if (point.x > maxX)
                    {
                        maxX = point.x;
                    }

                    if (point.y > maxY)
                    {
                        maxY = point.y;
                    }
                }

                //computes the bounding box in page space.
                PDFDouble x1 = new PDFDouble(minX);
                PDFDouble y1 = new PDFDouble(minY);
                PDFDouble x2 = new PDFDouble(maxX);
                PDFDouble y2 = new PDFDouble(maxY);

                mPDFView.ConvAnnotationCanvasPtToScreenPt(x1, y1);
                mPDFView.ConvAnnotationCanvasPtToScreenPt(x2, y2);
                mPDFView.ConvScreenPtToPagePt(x1, y1, mDownPageNumber);
                mPDFView.ConvScreenPtToPagePt(x2, y2, mDownPageNumber);

                PDFRect rect;
                rect = new PDFRect(x1.Value, y1.Value, x2.Value, y2.Value);
                rect.Normalize();
                return rect;
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                return null;
            }
        }

        /// <summary>
        ///  Must be implemented by the derived classes in order to create and push back the right type of annotation
        /// </summary>
        internal new virtual void Finish()
        {

        }

        #endregion

        #region Events
        public delegate void PointCreatedDelegate(int total);
        /// <summary>
        /// This event is invoked when a new point is created and passes the total number of points.
        /// </summary>
        public event PointCreatedDelegate PointCreated;

        protected virtual void OnPointCreated(int totalPts)
        {
            PointCreated?.Invoke(totalPts);
        }
        #endregion
    }
}
