using System;
using System.Collections.Generic;
using System.Threading.Tasks;
using System.Text;
using System.Runtime.InteropServices.WindowsRuntime;

using Windows.UI;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Shapes;
using Windows.Data.Xml.Dom;
using Windows.Storage;
using Windows.Storage.Streams;
using Windows.UI.Input;

using UIPoint = Windows.Foundation.Point;
using UIRect = Windows.Foundation.Rect;
using UIPopup = Windows.UI.Xaml.Controls.Primitives.Popup;

using PDFPoint = pdftron.PDF.Point;
using PDFRect = pdftron.PDF.Rect;
using PDFPage = pdftron.PDF.Page;
using PDFDouble = pdftron.Common.DoubleRef;

using pdftron.PDF.Annots;
using pdftron.PDF.Tools.Controls;
using pdftron.SDF;
using pdftron.PDF.OCG;
using pdftron.PDF.Tools.Utilities;
using pdftron.PDF.Tools.Commands;
using pdftron.FDF;
using pdftron.PDF.Tools.Controls.Multimedia;
using pdftron.PDF.Tools.Controls.ViewModels;
using static pdftron.PDF.Tools.UtilityFunctions;


namespace pdftron.PDF.Tools
{
    /// <summary>
    ///
    ///
    /// Note: Currently, PointerEvents when the textbox for editing FreeText annotations is open do not fire
    /// Therefore, we will allow regular scroll and zoom, but we won't allow resize while in text edit mode.
    /// </summary>
    public class AnnotEdit : Tool
    {
        internal delegate void AnnotPropertyChangedInHelperModeDelegate();
        internal event AnnotPropertyChangedInHelperModeDelegate AnnotPropertyChangedInHelperMode;

        protected Canvas mViewerCanvas;
        protected AnnotType mType;
        protected IMarkup mMarkup;
        protected ITextMarkup mTextMarkup;

        // For color picker
        internal FixedSizedBoxPopup mBoxPopup;
        protected Color mSelectedColor; // The color we're currently working with
        protected bool mUseColor = false;
        protected const int e_stroke = 0;
        protected const int e_fill = 1;
        protected const int e_text_stroke = 2;
        protected const int e_text_fill = 3;
        protected const int e_highlight = 4;
        protected const int e_text_markup = 5;
        protected int mEffectiveBrush = -1;

        // text editing
        protected bool mIsEditingText = false;
        protected double mFontSize = 11;
        protected bool mKeyboardOpen = false;
        protected double mKeyboardTranslateOffset = 0;
        protected bool mAnnotIsSticky = false;
        protected bool mAnnotIsSignature = false;
        protected bool mAnnotIsStamper = false;
        protected bool mAnnotIsTextMarkup = false;
        protected UIPopup mTextPopup;
        protected RichEditBox mEditTextBox;
        protected AnnotationPropertiesDisplay mPopup;
        internal KeyboardHelper mKeyboardHelper;
        protected bool mSaveTextOnExit = false;

        // position of the selection tool
        protected double mLeft, mRight, mTop, mBottom, mHorizontalMiddle, mVerticalMiddle;
        protected double mOldLeft, mOldRight, mOldTop, mOldBottom;
        protected UIPoint[] mCtrlPointLocations; // Canvas coordinates

        // Drawing the selection tool
        protected Rectangle mAnnotBBoxRectangle;
        protected TranslateTransform[] mCtrlPointTransforms; // Positions Control Points
        protected Path[] mControlPointShapes;
        protected SolidColorBrush mSelectionBackgroundColor;
        protected double mCtrlPointRadius;
        protected SolidColorBrush mCtrlPointBorderBrush;
        protected double mCtrlPointBorderThickness;
        protected SolidColorBrush mCtrlPointFillBrush;

        // Manipulation
        protected const int MOVE_WIDGET_CODE = -2; // Code to flag we are moving the entire widget
        protected static int CONTROL_POINT_FOR_MOVING = 8;
        protected int mNumberCtrlPts = CONTROL_POINT_FOR_MOVING;
        protected int mEffectiveCtrlPoint = 8; // 8 for translate, -1 for none.
        protected bool mIsManipulated = false;
        protected bool mStartMoving = false;
        protected Rect mPageCropOnClient;
        protected int mPointerID = -1;
        protected Point mPt1, mPt2;
        protected bool mAnnotIsScalable = true;
        protected bool mAnnotIsMovable = true;
        protected bool mMaintainAspectRatio = false;
        protected double mMinAnnotationSize;

        protected PDFRect mHelperModePositionRect = null;

        protected const int e_ll = 0;	//lower left control point
        protected const int e_lm = 1;	//lower middle
        protected const int e_lr = 2;	//lower right
        protected const int e_mr = 3;	//middle right
        protected const int e_ur = 4;	//upper left
        protected const int e_um = 5;	//upper middle
        protected const int e_ul = 6;	//upper left
        protected const int e_ml = 7;	//middle left
        protected const int START_MOVING_THRESHOLD = 10; // The distance we need to drag before the annot starts moving

        protected double mAspectRatio = 0;

        // Notes
        protected StackPanel mNoteDisplay = null;
        protected TextBox mNoteTextBox = null;

        protected Dictionary<string, string> mMenuTitles;

        // NOTE: mExtraContextmenu is need for things like multi-selecting annotations and
        // clicking opacity which requires a second context menu to display opacity items. 
        // The original context menu (mContextMenu) is displayed after an opacity is selected
        // and the mExtraContextMenu is disposed of.
        protected ContextMenuControl mExtraContextMenu;

        public AnnotEdit(PDFViewCtrl ctrl, ToolManager tMan)
            : base(ctrl, tMan)
        {
            mNextToolMode = ToolType.e_annot_edit;
            mToolMode = ToolType.e_annot_edit;
            mSelectedColor = new Color();

            mPt1 = new Point();
            mPt2 = new Point();

            // appearance of the selection widget
            mSelectionBackgroundColor = new SolidColorBrush(Color.FromArgb(128, 128, 128, 125));
            mCtrlPointFillBrush = new SolidColorBrush(Color.FromArgb(255, 150, 180, 255));
            mCtrlPointBorderBrush = new SolidColorBrush(Color.FromArgb(255, 200, 200, 200));
            mCtrlPointBorderThickness = 2;
            mCtrlPointRadius = 7.5;

            mMinAnnotationSize = mCtrlPointRadius * 2; // Min annotation size is 2 x Control point size

            // register for the keyboard callback
            mKeyboardHelper = new KeyboardHelper();
            mKeyboardHelper.SubscribeToKeyboard(true);
        }

        #region Public Properties
        /// <summary>
        /// Returns true if the current selected annotation (FreeText) is getting it's text edited
        /// </summary>
        public bool IsEditingText { get => mIsEditingText; }
        #endregion

        /// <summary>
        /// Select the annotation passed for editing
        /// </summary>
        /// <param name="annotation">the annotation to be selected</param>
        /// <param name="pageNumber">the page number where the annotation lives</param>
        internal void SelectAnnotation(IAnnot annotation, int pageNumber)
        {
            if ((annotation == null) || (pageNumber < 1))
                return;

            bool isRtReply = false;
            try
            {
                isRtReply = UtilityFunctions.IsReplyTypeReply(annotation);
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }

            if (isRtReply)
                return;

            mAnnot = annotation;

            try
            {
                mPDFView.DocLockRead();

                PopulateMenu();

                mAnnotPageNum = pageNumber;
                mPage = mPDFView.GetDoc().GetPage(mAnnotPageNum);

                mJustSwitchedFromAnotherTool = false;
                CreateAppearance();
                EnableScrolling();

            }
            catch (Exception ex)
            {
                AppCenterHandlerBase.CURRENT.TrackError(ex);
            }
            finally
            {
                mPDFView.DocUnlockRead();
            }
        }

        internal override void OnCreate()
        {
            try
            {
                mPDFView.DocLockRead();
                PopulateMenu(); // Can't be done in constructor, since mAnnot hasn't been set.
                InitializeCanvasAndGeometry();
            }
            catch (Exception) { }
            finally
            {
                mPDFView.DocUnlockRead();
            }
            this.IsHitTestVisible = false;

            mToolManager.RaiseSelectionChangedEvent(mAnnot, mAnnotPageNum);

        }

        protected virtual void InitializeCanvasAndGeometry()
        {
            mViewerCanvas = mPDFView.GetAnnotationCanvas();

            if (mViewerCanvas != null && !mViewerCanvas.Children.Contains(this))
            {
                mViewerCanvas.Children.Add(this);
            }

            // Figure out selection appearance according to type of annot
            mAnnotBBoxRectangle = new Rectangle();
            mAnnotBBoxRectangle.Stroke = mCtrlPointBorderBrush;
            mAnnotBBoxRectangle.StrokeDashArray = new DoubleCollection() { 2, 2 };
            mAnnotBBoxRectangle.StrokeThickness = 2;

            // The size of the gray rectangle is accomplished by scaling it
            mAnnotBBoxRectangle.Width = 1;
            mAnnotBBoxRectangle.Height = 1;

            this.Children.Add(mAnnotBBoxRectangle);

            if (mAnnot?.GetAnnotType() == AnnotType.e_Stamp)
            {
                mMaintainAspectRatio = true;
            }

            mNumberCtrlPts = CONTROL_POINT_FOR_MOVING; // reset to default

            if (mAnnotIsScalable)
            {
                mCtrlPointTransforms = new TranslateTransform[mNumberCtrlPts];
                mCtrlPointLocations = new UIPoint[mNumberCtrlPts];
                mControlPointShapes = new Path[mNumberCtrlPts];

                for (int i = 0; i < mNumberCtrlPts; i++)
                {
                    mCtrlPointTransforms[i] = new TranslateTransform();
                    mCtrlPointLocations[i] = new UIPoint();
                    mControlPointShapes[i] = new Path();

                    if (mMaintainAspectRatio && i % 2 != 0)
                    {
                        mControlPointShapes[i].Visibility = Visibility.Collapsed;
                        mCtrlPointLocations[i].X = -100;
                        continue;
                    }

                    EllipseGeometry geom = new EllipseGeometry();
                    //geom.Center = mCtrlPointLocations[i];
                    geom.RadiusX = mCtrlPointRadius;
                    geom.RadiusY = mCtrlPointRadius;
                    mControlPointShapes[i].Data = geom;
                    mControlPointShapes[i].Stroke = mCtrlPointBorderBrush;
                    mControlPointShapes[i].StrokeThickness = mCtrlPointBorderThickness;
                    mControlPointShapes[i].Fill = mCtrlPointFillBrush;
                    mControlPointShapes[i].RenderTransform = mCtrlPointTransforms[i];
                    this.Children.Add(mControlPointShapes[i]);
                }
            }
        }

        internal override void OnClose()
        {
            mToolManager.RaiseSelectionChangedEvent(null, 0);
            if (mSaveTextOnExit)
            {
                SaveFreeText();
            }
            if(mExtraContextMenu != null)
            {
                HideContextMenu(mExtraContextMenu, true);
                mExtraContextMenu = null;
            }
            if (mContextMenu != null)
            {
                HideContextMenu(mContextMenu, true);
                mContextMenu = null;
            }
            if (mBoxPopup != null)
            {
                mBoxPopup.Hide();
                mBoxPopup = null;
                EnableScrolling();
            }
            if (mTextPopup != null)
            {
                mTextPopup.IsOpen = false;
                mTextPopup = null;
            }

            if (mViewerCanvas != null && mViewerCanvas.Children.Contains(this))
            {
                mViewerCanvas.Children.Remove(this);
            }
            EnableScrolling();
            base.OnClose();
            DelayedUnsubscribe();
        }

        // This delays the unsubscribing of the keyboard, so that the UI will have time to invoke the callbacks it needs.
        protected async void DelayedUnsubscribe()
        {
            await Task.Delay(50);
            if (mEditTextBox != null)
            {
                mKeyboardHelper.RemoveShowingHandler(mEditTextBox);
                mKeyboardHelper.SetHidingHandler(null);
            }
            if (mNoteTextBox != null)
            {
                mKeyboardHelper.RemoveShowingHandler(mNoteTextBox);
            }
            mKeyboardHelper.SubscribeToKeyboard(false);
        }

        internal void CreateAppearance(bool showMenu = true)
        {
            SetPosition();

            if (mAnnotIsSticky && showMenu)
            {
                HandleNote();
            }
            else if (mToolManager.IsPopupMenuEnabled && showMenu)
            {
                PositionContextMenu(base.mContextMenu);
                ShowContextMenu(base.mContextMenu);
            }
        }

        internal override bool OnSize()
        {
            SetPosition();
            PositionTextEditingPopup();
            return false;
        }

        internal override bool OnScale()
        {
            SetPosition();

            PositionContextMenu(base.mContextMenu);
            if (mIsShowingCommandMenu)
            {
                ShowContextMenu(base.mContextMenu);
            }
            base.OnScale();
            return false;
        }

        internal override bool OnPageNumberChanged(int current_page, int num_pages)
        {
            mNextToolMode = ToolType.e_pan;
            return false;
        }

        internal override bool TappedHandler(object sender, TappedRoutedEventArgs e)
        {
            if (mJustSwitchedFromAnotherTool) // this should be the only entry point into the tool - Except for stylus select button
            {
                mJustSwitchedFromAnotherTool = false;
                if (mAnnot == null)
                {
                    return false;
                }
                CreateAppearance();
                EnableScrolling();
            }
            else if (mBoxPopup != null) // e.g. color picker is open
            {
                mBoxPopup.Hide();
                mBoxPopup = null;
                EnableScrolling();
            }
            else
            {
                // check bounds
                double x = e.GetPosition(mViewerCanvas).X;
                double y = e.GetPosition(mViewerCanvas).Y;

                // tap on annot
                if (x >= mLeft && x <= mRight && y >= mTop && y <= mBottom)
                {
                    if (mAnnotIsTextMarkup && mToolManager.IsPopupMenuEnabled)
                    {
                        PositionContextMenu(mContextMenu);
                        ShowContextMenu(mContextMenu);
                    }
                }
                else
                {
                    mNextToolMode = ToolType.e_pan;
                    if (mIsEditingText)
                    {
                        SaveFreeText();
                    }
                }
            }
            return true;
        }

        internal override bool DoubleClickedHandler(object sender, PointerRoutedEventArgs e)
        {
            // might want to open notes here.
            return true;
        }

        internal override bool RightTappedHandler(object sender, Windows.UI.Xaml.Input.RightTappedRoutedEventArgs e)
        {
            if (e.PointerDeviceType == Windows.Devices.Input.PointerDeviceType.Pen)
            {
                return true;
            }
            // If we remove a menu, don't open app bar.
            if (mExtraContextMenu != null)
            {
                HideContextMenu(mExtraContextMenu, true);
                mExtraContextMenu = null;
                e.Handled = true;
            }
            if (mIsShowingCommandMenu) // we removed a menu, don't open app bar.
            {
                HideContextMenu(mContextMenu);
                e.Handled = true;
            }
            return false;
        }

        internal override bool HoldingHandler(object sender, HoldingRoutedEventArgs e)
        {
            if (mEffectiveCtrlPoint == -1) // would have been set to something else if we had hit a ctrl point or the square
            {
                mNextToolMode = ToolType.e_pan;
                return false;
            }
            return true;
        }

        /// <summary>
        /// Updates the locations of control points and transforms
        /// </summary>
        protected virtual void CreateSelectionAppearance()
        {
            // update rectangle
            mAnnotBBoxRectangle.SetValue(Canvas.LeftProperty, mLeft - mPageCropOnClient.x1);
            mAnnotBBoxRectangle.SetValue(Canvas.TopProperty, mTop - mPageCropOnClient.y1);

            mAnnotBBoxRectangle.Width = mRight - mLeft;
            mAnnotBBoxRectangle.Height = mBottom - mTop;

            if (mAnnotIsScalable)
            {
                // update mCtrlPointLocations
                mCtrlPointLocations[e_ll].X = mLeft;
                mCtrlPointLocations[e_ll].Y = mBottom;
                mCtrlPointLocations[e_lr].X = mRight;
                mCtrlPointLocations[e_lr].Y = mBottom;
                mCtrlPointLocations[e_ur].X = mRight;
                mCtrlPointLocations[e_ur].Y = mTop;
                mCtrlPointLocations[e_ul].X = mLeft;
                mCtrlPointLocations[e_ul].Y = mTop;

                if (!mMaintainAspectRatio)
                {
                    mCtrlPointLocations[e_lm].X = mHorizontalMiddle;
                    mCtrlPointLocations[e_lm].Y = mBottom;
                    mCtrlPointLocations[e_mr].X = mRight;
                    mCtrlPointLocations[e_mr].Y = mVerticalMiddle;
                    mCtrlPointLocations[e_um].X = mHorizontalMiddle;
                    mCtrlPointLocations[e_um].Y = mTop;
                    mCtrlPointLocations[e_ml].X = mLeft;
                    mCtrlPointLocations[e_ml].Y = mVerticalMiddle;
                }

                for (int i = 0; i < 8; i++)
                {
                    // skip middle ctrl points when maintaining aspect ratio
                    if (mMaintainAspectRatio && i % 2 != 0)
                        continue;

                    mCtrlPointTransforms[i].X = mCtrlPointLocations[i].X - mPageCropOnClient.x1;
                    mCtrlPointTransforms[i].Y = mCtrlPointLocations[i].Y - mPageCropOnClient.y1;
                }

                // if maintaining aspect ratio, calculate aspect ratio (only calculate once)
                if (mMaintainAspectRatio && mAspectRatio == 0)
                {
                    double height = mBottom - mTop;
                    double width = mRight - mLeft;
                    mAspectRatio = height / width;
                }
            }
        }

        protected virtual void SetPosition()
        {
            // Place the tool's canvas on top of the page
            mPageCropOnClient = BuildPageBoundBoxOnClient(mAnnotPageNum);
            this.SetValue(LeftProperty, mPageCropOnClient.x1);
            this.SetValue(TopProperty, mPageCropOnClient.y1);
            this.Width = mPageCropOnClient.x2 - mPageCropOnClient.x1;
            this.Height = mPageCropOnClient.y2 - mPageCropOnClient.y1;

            BuildAnnotBBox();

            if ((mToolMode != ToolType.e_annot_edit) && (mToolMode != ToolType.e_advanced_shape_edit))
            {
                return;
            }

            // Convert to screen space
            PDFDouble x1 = new PDFDouble(mAnnotBBox.x1);
            PDFDouble y1 = new PDFDouble(mAnnotBBox.y1);
            PDFDouble x2 = new PDFDouble(mAnnotBBox.x2);
            PDFDouble y2 = new PDFDouble(mAnnotBBox.y2);

            // Get coordinates of two opposite points in screen space
            mPDFView.ConvPagePtToScreenPt(x1, y1, mAnnotPageNum);
            mPDFView.ConvPagePtToScreenPt(x2, y2, mAnnotPageNum);
            mPDFView.ConvScreenPtToAnnotationCanvasPt(x1, y1);
            mPDFView.ConvScreenPtToAnnotationCanvasPt(x2, y2);

            mLeft = mOldLeft = Math.Min(x1.Value, x2.Value);
            mRight = mOldRight = Math.Max(x1.Value, x2.Value);
            mTop = mOldTop = Math.Min(y1.Value, y2.Value);
            mBottom = mOldBottom = Math.Max(y1.Value, y2.Value);
            mHorizontalMiddle = (mLeft + mRight) / 2.0;
            mVerticalMiddle = (mTop + mBottom) / 2.0;

            CreateSelectionAppearance();

            if (mIsShowingCommandMenu)
            {
                HideContextMenu(base.mContextMenu);
                PositionContextMenu(base.mContextMenu);
                ShowContextMenu(base.mContextMenu);
            }
        }

        /// <summary>
        /// Adds in the menu options according to the type of annotation
        /// </summary>
        protected virtual void PopulateMenu()
        {
            if (mToolManager == null || mToolManager.ReadOnlyMode)
                return;

            if (mAnnot != null)
            {
                mMenuTitles = null;
                try
                {
                    mMenuTitles = new Dictionary<string, string>();
                    //locks the document first as accessing annotation/doc information isn't thread safe.
                    mPDFView.DocLockRead();
                    mType = mAnnot.GetAnnotType();
                    mMarkup = (IMarkup)mAnnot;
                    mMenuTitles["done"] = ResourceHandler.GetString("EditOption_Done");

                    mAnnotIsSticky = (mType == AnnotType.e_Text);
                    mAnnotIsTextMarkup = (mType == AnnotType.e_Highlight ||
                                            mType == AnnotType.e_Underline ||
                                            mType == AnnotType.e_StrikeOut ||
                                            mType == AnnotType.e_Squiggly);

                    if ((mAnnot.IsMarkup() && (mType != AnnotType.e_FreeText && mType != AnnotType.e_Sound))
                        || mType == AnnotType.e_Text)
                    {
                        bool addNote = true;
                        if (mAnnot.GetAnnotType() == AnnotType.e_Stamp)
                        {
                            pdftron.SDF.Obj sigObj = mAnnot.GetSDFObj();
                            pdftron.SDF.Obj stampObj = mAnnot.GetSDFObj();

                            sigObj = sigObj.FindObj(ToolManager.SignatureAnnotationIdentifyingString);
                            stampObj = stampObj.FindObj(ToolManager.StamperAnnotationIdentifyingString);

                            addNote = false;

                            if (sigObj != null)
                            {
                                mAnnotIsSignature = true;
                                mMenuTitles["flatten"] = ResourceHandler.GetString("EditOption_Flatten");
                            }
                            else if (stampObj != null)
                            {
                                mAnnotIsStamper = true;
                                mMenuTitles["flatten"] = ResourceHandler.GetString("EditOption_Flatten");
                                mMenuTitles["rotate"] = ResourceHandler.GetString("EditOption_Rotate");
                            }
                        }
                        if (addNote)
                        {
                            if (mAnnotIsSticky)
                            {
                                mMenuTitles["Note"] = ResourceHandler.GetString("EditOption_Note");
                            }
                            else if (UtilityFunctions.DoesAnnotHaveContent(mMarkup))
                            {
                                mMenuTitles["Note"] = ResourceHandler.GetString("EditOption_ViewNote");
                            }
                            else
                            {
                                mMenuTitles["Note"] = ResourceHandler.GetString("EditOption_AddNote");
                            }
                        }
                    }

                    if (mType == AnnotType.e_Sound)
                    {
                        mMenuTitles["play"] = ResourceHandler.GetString("AudioControl_PlayAudio");
                    }

                    if (mType == AnnotType.e_Line)
                    {
                        if (MeasureUtils.IsRulerMeasure(mAnnot))
                            mMenuTitles["calibrate"] = ResourceHandler.GetString("EditOption_Calibrate");
                    }

                    if (mAnnot.IsMarkup() && !mToolManager.ReadOnlyMode &&
                        mType != AnnotType.e_Stamp && mType != AnnotType.e_Caret && mType != AnnotType.e_FileAttachment && mType != AnnotType.e_Sound)
                    {
                        mMenuTitles["Style"] = ResourceHandler.GetString("EditOptiob_Style");
                    }

                    if (mAnnotIsTextMarkup)
                    {
                        mMenuTitles["copy to clipboard"] = ResourceHandler.GetString("EditOption_CopyAnnot");
                    }
                    else if (mType != AnnotType.e_Sound)
                    {
                        mMenuTitles["copy annot to clipboard"] = ResourceHandler.GetString("EditOption_CopyAnnot");
                        mMenuTitles["duplicate annot"] = ResourceHandler.GetString("EditOption_DuplicateAnnot");
                        mMenuTitles["cut annot"] = ResourceHandler.GetString("EditOption_CutAnnot");
                    }
                     
                    if (mType == AnnotType.e_FreeText && !mToolManager.ReadOnlyMode)
                    {
                        mMenuTitles["edit text"] = ResourceHandler.GetString("EditOption_EditText");
                    }

                    if (mAnnotIsSticky || mToolManager.ReadOnlyMode)
                    {
                        mAnnotIsScalable = false;
                    }
                    if (mAnnotIsTextMarkup && !mToolManager.ReadOnlyMode)
                    {
                        mAnnotIsScalable = false;
                        mAnnotIsMovable = false;
                        mMenuTitles["Type"] = ResourceHandler.GetString("EditOption_TextMarkupType");
                    }

                    mMarkup = (IMarkup)mAnnot;
                    if (mAnnotIsTextMarkup)
                    {
                        mTextMarkup = (ITextMarkup)mAnnot;
                    }

                    if (!mToolManager.ReadOnlyMode)
                    {
                        mMenuTitles["delete"] = ResourceHandler.GetString("EditOption_Delete");
                    }
                }
                catch (Exception ex)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                }
                finally
                {
                    mPDFView.DocUnlockRead();
                }


                if (mMenuTitles != null && !mAnnotIsSticky) // if sticky, we don't create menu until it's needed
                {
                    base.mContextMenu = new ContextMenuControl(mPDFView, mMenuTitles);
                    base.mContextMenu.ItemClicked += this.OnCommandMenuClicked;
                }
            }
        }

        internal override bool PointerPressedHandler(object sender, PointerRoutedEventArgs e)
        {
            if (mPointerID != -1)
            {
                return false; // ignore additional pointer presses
            }
            if (e.Pointer.PointerDeviceType == Windows.Devices.Input.PointerDeviceType.Mouse && !e.GetCurrentPoint(null).Properties.IsLeftButtonPressed)
            {
                return false; // If mouse, make sure it's the left button
            }
            if (mIsEditingText)
            {
                // do nothing, handled by tap
                return false;
            }

            // Hide menus
            HideContextMenu(mExtraContextMenu, true);
            HideContextMenu(mContextMenu, true);
            mExtraContextMenu = null;
            mContextMenu = null;

            if (mBoxPopup != null) // e.g. color picker is open
            {
                return false;
            }
            if (mAnnotIsTextMarkup) // markup can't be moved
            {
                return false;
            }

            mPointerID = (int)e.Pointer.PointerId;
            mStartMoving = false; // reset whether we're moving or not.

            PointerPoint cn_downPoint = e.GetCurrentPoint(mViewerCanvas);

            if (e.Pointer.PointerDeviceType == Windows.Devices.Input.PointerDeviceType.Pen)
            {
                if (cn_downPoint.Properties.IsEraser)
                {
                    EndCurrentTool(ToolType.e_pan);
                    return false;
                }
                if (mJustSwitchedFromAnotherTool)
                {
                    if (cn_downPoint.Properties.IsBarrelButtonPressed)
                    {
                        CreateAppearance(false);
                    }
                    else
                    {
                        EndCurrentTool(ToolType.e_pan);
                        return false;
                    }
                }
            }

            // see if we hit anything interesting in mViewerCanvas
            GetControlPoint(cn_downPoint.Position);
            if (mToolManager.ReadOnlyMode)
            {
                mEffectiveCtrlPoint = -1;
            };

            if (mEffectiveCtrlPoint >= 0)
            {
                if (mJustSwitchedFromAnotherTool)
                {
                    mEffectiveCtrlPoint = CONTROL_POINT_FOR_MOVING; // always do a move the first time
                }
                DisableScrolling();
            }

            if (mEffectiveCtrlPoint == MOVE_WIDGET_CODE)
                DisableScrolling();

            mViewerCanvas.CapturePointer(e.Pointer);

            mJustSwitchedFromAnotherTool = false;
            return false;
        }

        protected virtual void GetControlPoint(UIPoint p)
        {
            double x = p.X;
            double y = p.Y;
            mEffectiveCtrlPoint = -1;

            // see if we're close to a ctrl point
            if (mAnnotIsScalable)
            {
                double thresh = Math.Pow(mCtrlPointRadius * 2.25, 2); // cheaper than sqrt
                double shortest_dist = thresh + 1;
                for (int i = 0; i < mNumberCtrlPts; ++i)
                {
                    // skip middle ctrl points when maintaining aspect ratio
                    if (mMaintainAspectRatio && i % 2 != 0)
                        continue;

                    double s = mCtrlPointLocations[i].X;
                    double t = mCtrlPointLocations[i].Y;

                    double dist = (x - s) * (x - s) + (y - t) * (y - t);
                    if (dist <= thresh && (dist < shortest_dist || shortest_dist < 0))
                    {
                        mEffectiveCtrlPoint = i;
                        shortest_dist = dist;
                    }
                }
            }

            // if we didn't get a ctrl point, check bounds
            if (mEffectiveCtrlPoint == -1)
            {
                if (x >= mLeft && x <= mRight && y >= mTop && y <= mBottom)
                {
                    mEffectiveCtrlPoint = MOVE_WIDGET_CODE;
                }
            }

            // means we're manipulating the widget
            if ((mEffectiveCtrlPoint >= 0) || (mEffectiveCtrlPoint == MOVE_WIDGET_CODE))
            {
                mPt1.x = x;
                mPt1.y = y;
                mIsManipulated = true;
                mOldLeft = mLeft;
                mOldRight = mRight;
                mOldTop = mTop;
                mOldBottom = mBottom;
            }
        }

        internal override bool PointerMovedHandler(object sender, PointerRoutedEventArgs e)
        {
            if (!mIsManipulated || mToolManager.ReadOnlyMode)
            {
                return false;
            }
            mPt2.x = e.GetCurrentPoint(mViewerCanvas).Position.X;
            mPt2.y = e.GetCurrentPoint(mViewerCanvas).Position.Y;
            if (!mStartMoving) // ensure we actually move a little before we start with updates and stuff
            {
                if ((Math.Abs(mPt2.x - mPt1.x) > START_MOVING_THRESHOLD)
                    || (Math.Abs(mPt2.y - mPt1.y) > START_MOVING_THRESHOLD))
                {
                    mStartMoving = true;
                }
                else
                {
                    return false;
                }
            }

            if (mToolMode != ToolType.e_annot_edit)
            {
                return false;
            }

            CheckBoundsOnPointerMoved();
            return true;
        }

        protected void CheckBoundsOnPointerMoved()
        {
            // check bounds
            switch (mEffectiveCtrlPoint)
            {
                case MOVE_WIDGET_CODE: // regular move
                    mLeft = mOldLeft + mPt2.x - mPt1.x;
                    mLeft = Math.Max(mLeft, mPageCropOnClient.x1);
                    mLeft = Math.Min(mLeft, mPageCropOnClient.x2 + mOldLeft - mOldRight);
                    mRight = mLeft + mOldRight - mOldLeft;
                    mTop = mOldTop + mPt2.y - mPt1.y;
                    mTop = Math.Max(mTop, mPageCropOnClient.y1);
                    mTop = Math.Min(mTop, mPageCropOnClient.y2 + mOldTop - mOldBottom);
                    mBottom = mTop + mOldBottom - mOldTop;
                    break;
                case e_ll:
                    // don't allow x expansion when y can't expand anymore if maintaining aspect ratio
                    if (!mMaintainAspectRatio || !(mBottom >= mPageCropOnClient.y2 && mPt2.x <= mPt1.x && mMaintainAspectRatio))
                    {
                        var calcLeft = mOldLeft + mPt2.x - mPt1.x;
                        if ((mRight - calcLeft) > mMinAnnotationSize)
                            mLeft = calcLeft;
                    }

                    mLeft = Math.Max(mLeft, mPageCropOnClient.x1);
                    mLeft = Math.Min(mLeft, mRight);

                    if (!mMaintainAspectRatio)
                    {
                        var calcBottom = mOldBottom + mPt2.y - mPt1.y;
                        if ((calcBottom - mTop) > mMinAnnotationSize)
                            mBottom = calcBottom;
                    }
                    // don't allow y expansion when the x axis can't expand anymore if maintaining aspect ratio
                    else if (mPt2.x >= 0)
                    {
                        mBottom = mOldBottom - (mPt2.x - mPt1.x) * mAspectRatio;
                    }

                    mBottom = Math.Max(mBottom, mTop);
                    mBottom = Math.Min(mBottom, mPageCropOnClient.y2);
                    break;
                case e_lm:
                    if (!mMaintainAspectRatio)
                    {
                        var calcBottom = mOldBottom + mPt2.y - mPt1.y;

                        // Ensure min allowed annotation size
                        if ((calcBottom - mTop) > mMinAnnotationSize)
                        {
                            mBottom = calcBottom;
                            mBottom = Math.Max(mBottom, mTop);
                            mBottom = Math.Min(mBottom, mPageCropOnClient.y2);
                        }
                    }
                    break;
                case e_lr:
                    // don't allow x expansion when y can't expand anymore if maintaining aspect ratio
                    if (!mMaintainAspectRatio || !(mBottom >= mPageCropOnClient.y2 && mPt2.x >= mPt1.x && mMaintainAspectRatio))
                    {
                        var calcRight = mOldRight + mPt2.x - mPt1.x;

                        // Ensure min allowed annotation size
                        if ((calcRight - mLeft) > mMinAnnotationSize)
                            mRight = calcRight;
                    }

                    mRight = Math.Max(mRight, mOldLeft);
                    mRight = Math.Min(mRight, mPageCropOnClient.x2);
                    if (!mMaintainAspectRatio)
                    {
                        var calcBottom = mOldBottom + mPt2.y - mPt1.y;

                        // Ensure min allowed annotation size
                        if ((calcBottom - mTop) > mMinAnnotationSize)
                            mBottom = calcBottom;
                    }
                    else if (mPt2.x <= mPageCropOnClient.x2)
                    {
                        mBottom = mOldBottom + (mPt2.x - mPt1.x) * mAspectRatio;
                    }
                    mBottom = Math.Max(mBottom, mTop);
                    mBottom = Math.Min(mBottom, mPageCropOnClient.y2);
                    break;
                case e_mr:
                    if (!mMaintainAspectRatio)
                    {
                        var calcMedRight = mOldRight + mPt2.x - mPt1.x;

                        // Ensure min allowed annotation size
                        if ((calcMedRight - mLeft) > mMinAnnotationSize)
                        {
                            mRight = calcMedRight;
                            mRight = Math.Max(mRight, mOldLeft);
                            mRight = Math.Min(mRight, mPageCropOnClient.x2);
                        }
                    }
                    break;
                case e_ur:
                    // don't allow x expansion when y can't expand anymore if maintaining aspect ratio
                    if (!mMaintainAspectRatio || !(mTop <= mPageCropOnClient.y1 && mPt2.x >= mPt1.x && mMaintainAspectRatio))
                    {
                        var calcRight = mOldRight + mPt2.x - mPt1.x;

                        // Ensure min allowed annotation size
                        if ((calcRight - mLeft) > mMinAnnotationSize)
                            mRight = calcRight;
                    }

                    mRight = Math.Max(mRight, mOldLeft);
                    mRight = Math.Min(mRight, mPageCropOnClient.x2);
                    if (!mMaintainAspectRatio)
                    {
                        var calcTop = mOldTop + mPt2.y - mPt1.y;

                        // Ensure min allowed annotation size
                        if ((mBottom - calcTop) > mMinAnnotationSize)
                            mTop = calcTop;
                    }
                    else if (mPt2.x <= mPageCropOnClient.x2)
                    {
                        mTop = mOldTop - (mPt2.x - mPt1.x) * mAspectRatio;
                    }
                    mTop = Math.Max(mTop, mPageCropOnClient.y1);
                    mTop = Math.Min(mTop, mBottom);
                    break;
                case e_um:
                    if (!mMaintainAspectRatio)
                    {
                        var calcTop = mOldTop + mPt2.y - mPt1.y;

                        // Ensure min allowed annotation size
                        if ((mBottom - calcTop) > mMinAnnotationSize)
                        {
                            mTop = calcTop;
                            mTop = Math.Max(mTop, mPageCropOnClient.y1);
                            mTop = Math.Min(mTop, mBottom);
                        }
                    }
                    break;
                case e_ul:
                    // don't allow x expansion when y can't expand anymore if maintaining aspect ratio
                    if (!mMaintainAspectRatio || !(mTop <= mPageCropOnClient.y1 && mPt2.x <= mPt1.x && mMaintainAspectRatio))
                    {
                        var calcMedLeft = mOldLeft + mPt2.x - mPt1.x;

                        // Ensure min allowed annotation size
                        if ((mRight - calcMedLeft) > mMinAnnotationSize)
                            mLeft = calcMedLeft;
                    }

                    mLeft = Math.Max(mLeft, mPageCropOnClient.x1);
                    mLeft = Math.Min(mLeft, mRight);
                    if (!mMaintainAspectRatio)
                    {
                        var calcTop = mOldTop + mPt2.y - mPt1.y;

                        // Ensure min allowed annotation size
                        if ((mBottom - calcTop) > mMinAnnotationSize)
                            mTop = calcTop;
                    }
                    else if (mPt2.x >= 0)
                    {
                        mTop = mOldTop + (mPt2.x - mPt1.x) * mAspectRatio;
                    }
                    mTop = Math.Max(mTop, mPageCropOnClient.y1);
                    mTop = Math.Min(mTop, mBottom);
                    break;
                case e_ml:
                    if (!mMaintainAspectRatio)
                    {
                        var calcMedLeft = mOldLeft + mPt2.x - mPt1.x; // calculate new Left point

                        // Ensure min allowed annotation size
                        if ((mRight - calcMedLeft) > mMinAnnotationSize)
                        {
                            mLeft = calcMedLeft;
                            mLeft = Math.Max(mLeft, mPageCropOnClient.x1);
                            mLeft = Math.Min(mLeft, mRight);
                        }
                    }
                    break;
            }
            mHorizontalMiddle = (mLeft + mRight) / 2;
            mVerticalMiddle = (mTop + mBottom) / 2;
            CreateSelectionAppearance();
        }

        internal override bool PointerReleasedHandler(object sender, PointerRoutedEventArgs e)

        {
            if (mPointerID < 0 || e.Pointer.PointerId != mPointerID)
            {
                return false;
            }
            mPointerID = -1;
            mEffectiveCtrlPoint = -1;
            mViewerCanvas.ReleasePointerCaptures();

            if ((mToolMode != ToolType.e_annot_edit) && (mToolMode != ToolType.e_advanced_shape_edit))
            {
                return false;
            }

            if (mIsManipulated)
            {
                if (mStartMoving)
                {
                    try
                    {
                        mPDFView.DocLock(true);

                        // points in Screen space
                        PDFDouble x1 = new PDFDouble(mLeft);
                        PDFDouble y1 = new PDFDouble(mTop);
                        PDFDouble x2 = new PDFDouble(mRight);
                        PDFDouble y2 = new PDFDouble(mBottom);

                        // Get coordinates of two opposite points in screen space
                        mPDFView.ConvAnnotationCanvasPtToScreenPt(x1, y1);
                        mPDFView.ConvAnnotationCanvasPtToScreenPt(x2, y2);
                        mPDFView.ConvScreenPtToPagePt(x1, y1, mAnnotPageNum);
                        mPDFView.ConvScreenPtToPagePt(x2, y2, mAnnotPageNum);

                        Rect newAnnotRect = null;
                        if (mAnnot.GetFlag(AnnotFlag.e_no_zoom))
                        {
                            Rect annotRect = mAnnot.GetRect();
                            newAnnotRect = new Rect(x1.Value, y1.Value - annotRect.Height(), x1.Value + annotRect.Width(), y1.Value);
                        }
                        else
                        {
                            newAnnotRect = new Rect(x1.Value, y1.Value, x2.Value, y2.Value);
                        }
                        newAnnotRect.Normalize();

                        /* NOTE: annotation specific logic to make sure we can resize, move, and create control points properly
                         * for a Square annotation that had the border changed to cloudy using .SetBorderEffect()*/
                        if (mAnnot.IsMarkup() && mAnnot.GetAnnotType() == AnnotType.e_Square)
                        {
                            var markup = new Markup(mAnnot);
                            mAnnot.SetRect(newAnnotRect);
                            markup.SetContentRect(newAnnotRect);
                        }
                        else
                        {
                            mAnnot.Resize(newAnnotRect);
                        }

                        if (mType == AnnotType.e_Line
                            || mType == AnnotType.e_Circle
                            || mType == AnnotType.e_Square
                            || mType == AnnotType.e_Polyline
                            || mType == AnnotType.e_Polygon
                            || mType == AnnotType.e_Ink
                            || mType == AnnotType.e_FreeText)
                        {
                            mAnnot.RefreshAppearance();
                        }

                        BuildAnnotBBox();

                        mAnnot.SetDateToNow();

                        mPDFView.UpdateWithAnnot(mAnnot, mAnnotPageNum);

                        // Update the viewer at the location of the old rectangle
                        x1 = new PDFDouble(mOldLeft);
                        y1 = new PDFDouble(mOldTop);
                        x2 = new PDFDouble(mOldRight);
                        y2 = new PDFDouble(mOldBottom);
                        mPDFView.ConvAnnotationCanvasPtToScreenPt(x1, y1);
                        mPDFView.ConvAnnotationCanvasPtToScreenPt(x2, y2);

                        Rect updateRect = new Rect(x1.Value, y1.Value, x2.Value, y2.Value);
                        mPDFView.Update(updateRect);

                        mToolManager.RaiseAnnotationEditedEvent(mAnnot, mAnnotPageNum);

                    }
                    catch (Exception ex)
                    {
                        Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(
                            Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY,
                            Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                    }
                    finally
                    {
                        mPDFView.DocUnlock();
                    }
                }

                // Our new values now become our old values for the next manipulation
                mOldLeft = mLeft;
                mOldRight = mRight;
                mOldTop = mTop;
                mOldBottom = mBottom;

                if (mToolManager.IsPopupMenuEnabled && !mIsInHelperMode)
                {
                    if (mContextMenu == null)
                    {
                        mContextMenu = new ContextMenuControl(mPDFView, mMenuTitles);
                        mContextMenu.ItemClicked += OnCommandMenuClicked;
                    }
                    PositionContextMenu(base.mContextMenu);
                    ShowContextMenu(base.mContextMenu);
                }
            }
            // reset values concerned with moving the widget
            mStartMoving = false;
            mIsManipulated = false;
            EnableScrolling();
            return false;
        }

        internal override bool PointerCaptureLostHandler(object sender, PointerRoutedEventArgs e)
        {
            if (mEffectiveCtrlPoint >= 0) // we were moving the widget, just abort
            {
                mNextToolMode = ToolType.e_pan;
            }
            mPointerID = -1;
            return true;
        }

        internal override bool PointerCanceledHandler(object sender, Windows.UI.Xaml.Input.PointerRoutedEventArgs e)
        {
            mNextToolMode = ToolType.e_pan;
            return true;
        }

        internal override bool OnViewChanged(Object sender, ScrollViewerViewChangedEventArgs e)
        {
            // hide menus if we scroll or zoom
            HideContextMenu(mExtraContextMenu, true);
            mExtraContextMenu = null;
            if (mIsShowingCommandMenu)
            {
                HideContextMenu(mContextMenu);
            }
            base.OnViewChanged(sender, e);
            return false;
        }

        internal override bool KeyDownAction(object sender, KeyRoutedEventArgs e)
        {
            if (mPointerID >= 0) // if we're currently moving the widget, don't allow key presses
            {
                return false;
            }
            if (e.Key == Windows.System.VirtualKey.Escape)
            {
                if (mBoxPopup != null)
                {
                    mBoxPopup.Hide();
                    mBoxPopup = null;
                    mPDFView.Focus(Windows.UI.Xaml.FocusState.Pointer);
                    return true;
                }
                else if (mExtraContextMenu != null)
                {
                    HideContextMenu(mExtraContextMenu, true);
                    mExtraContextMenu = null;
                    return true;
                }
                else if (mIsShowingCommandMenu)
                {
                    HideContextMenu(mContextMenu);
                    return true;
                }
                else
                {
                    mNextToolMode = ToolType.e_pan;
                }
                return false;
            }
            if (mBoxPopup != null)
            {
                return false;
            }

            if (e.Key == Windows.System.VirtualKey.Delete) // delete annotation
            {
                HideContextMenu(mContextMenu);
                DeleteAnnotationOnKeyDown();
                mNextToolMode = ToolType.e_pan;
                return true;
            }

            return base.KeyDownAction(sender, e);
        }

        /// <summary>
        /// Delete annotation on delete key pressed
        /// </summary>
        protected virtual void DeleteAnnotationOnKeyDown()
        {
            try
            {
                mPDFView.DocLock(true); // Document lock and cancel rendering

                mToolManager.RaiseAnnotationPreRemovedEvent(mAnnot, mAnnotPageNum);

                // Perform the Annotation removal and update PDFViewer
                mPage = mPDFView.GetDoc().GetPage(mAnnotPageNum);
                mAnnot = AnnotUtils.SafeDeleteAnnotAndUpdate(mPDFView, mPage, mAnnot, mAnnotPageNum);

                // make sure to raise removed event after mPDFView.UpdateWithAnnot()
                mToolManager.RaiseAnnotationRemovedEvent(mAnnot, mAnnotPageNum);
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                mPDFView.DocUnlock();
                mAnnot = null;
            }
        }

        internal override bool CloseOpenDialog()
        {
            bool closed = false;
            if (mExtraContextMenu != null)
            {
                HideContextMenu(mExtraContextMenu, true);
                mExtraContextMenu = null;
                closed = true;
            }
            else if (mBoxPopup != null)
            {
                mBoxPopup.Hide();
                mBoxPopup = null;
                EnableScrolling();
                closed = true;
            }
            else if (mTextPopup != null)
            {
                mTextPopup.IsOpen = false;
                mTextPopup = null;
                closed = true;
            }
            else if (mIsShowingCommandMenu)
            {
                HideContextMenu(mContextMenu);
                closed = true;
            }
            return closed;
        }

        #region Handlers For Popup Menu

        internal override async void OnCommandMenuClicked(string title)
        {
            HideContextMenu(base.mContextMenu);

            if (title.Equals("done", StringComparison.OrdinalIgnoreCase))
            {
                CreateNewTool(ToolType.e_pan);
            }
            else if (title.Equals("delete", StringComparison.OrdinalIgnoreCase))
            {
                DeleteAnnotation();
                CreateNewTool(ToolType.e_pan);
            }
            else if (title.Equals("note", StringComparison.OrdinalIgnoreCase))
            {
                HandleNote();
            }
            else if (title.Equals("thickness", StringComparison.OrdinalIgnoreCase))
            {
                HandleThickness();
            }
            else if (title.Equals("type", StringComparison.OrdinalIgnoreCase))
            {
                HandleType();
            }
            else if (title.Equals("stroke color", StringComparison.OrdinalIgnoreCase))
            {
                mEffectiveBrush = e_stroke;
                ShowColorPicker();
            }
            else if (title.Equals("fill color", StringComparison.OrdinalIgnoreCase))
            {
                if (mAnnot.GetAnnotType() == AnnotType.e_FreeText)
                {
                    mEffectiveBrush = e_text_fill;
                    ShowColorPicker();
                }
                else
                {
                    mEffectiveBrush = e_fill;
                    ShowColorPicker();
                }
            }
            else if (title.Equals("opacity", StringComparison.OrdinalIgnoreCase))
            {
                HandleOpacity();
            }
            else if (title.Equals("color", StringComparison.OrdinalIgnoreCase))
            {
                AnnotType aType = mAnnot.GetAnnotType();
                if (aType == AnnotType.e_Line || aType == AnnotType.e_Ink || aType == AnnotType.e_Polyline)
                {
                    mEffectiveBrush = e_stroke;
                    ShowColorPicker();
                }
                else if (aType == AnnotType.e_Highlight)
                {
                    mEffectiveBrush = e_highlight;
                    ShowColorPicker();
                }
                else
                {
                    mEffectiveBrush = e_text_markup;
                    ShowColorPicker();
                }
            }
            else if (title.Equals("edit text", StringComparison.OrdinalIgnoreCase))
            {
                HideContextMenu(base.mContextMenu);
                //mCommandMenu = null; // we don't want this anymore
                HandleEditText();
            }
            else if (title.Equals("text Color", StringComparison.OrdinalIgnoreCase))
            {
                mEffectiveBrush = e_text_stroke;
                ShowColorPicker();
            }
            else if (title.Equals("font size", StringComparison.OrdinalIgnoreCase))
            {
                HandleFontSize();
            }
            else if (title.Equals("copy to clipboard", StringComparison.OrdinalIgnoreCase))
            {
                HandleCopy();
            }
            else if (title.Equals("copy annot to clipboard", StringComparison.OrdinalIgnoreCase))
            {
                HandleAnnotCopy();
            }
            else if (title.Equals("paste annot from clipboard", StringComparison.OrdinalIgnoreCase))
            {
                await HandleAnnotPasteAsync();
            }
            else if (title.Equals("duplicate annot", StringComparison.OrdinalIgnoreCase))
            {
                await HandleDuplicateAnnotAsync();
            }
            else if (title.Equals("cut annot"))
            {
                HandleCutAnnot();
            }
            else if (title.Equals("style", StringComparison.OrdinalIgnoreCase))
            {
                HandleStyle(true);
            }
            else if (title.Equals("flatten", StringComparison.OrdinalIgnoreCase))
            {
                HandleFlatten();
            }
            else if (title.Equals("rotate", StringComparison.OrdinalIgnoreCase))
            {
                HandleRotate();
            }
            else if (title.Equals("calibrate", StringComparison.OrdinalIgnoreCase))
            {
                await HandleCalibrationAsync();
            }
            else if (title.Equals("play", StringComparison.OrdinalIgnoreCase))
            {
                await HandlePlaySoundAnnotAsync();
            }
            else
            {
                throw new Exception("The option doesn't exist");
            }
        }

        /// <summary>
        /// Prepares and shows the color picker
        /// </summary>
        protected void ShowColorPicker()
        {
            GetCurrentColor();
            bool canChooseEmpty = false;
            if (mEffectiveBrush == e_fill || mEffectiveBrush == e_text_fill)
            {
                canChooseEmpty = true;
            }

            Utilities.ColorPickerSimple cp = new Utilities.ColorPickerSimple();
            cp.AllowEmpty = canChooseEmpty;
            cp.VisibleColors = Utilities.ColorPickerSimple.ColorsToSelect.eight;
            if (mEffectiveBrush == e_highlight)
            {
                cp.VisibleColors = Utilities.ColorPickerSimple.ColorsToSelect.highlight;
            }
            cp.ColorSelected += cps_ColorSelected;

            mBoxPopup = new FixedSizedBoxPopup(mPDFView, cp, mToolManager.PopupTargetInternal, true, false);
            mBoxPopup.PreferAbove = true;
            mBoxPopup.Show(GetBoxPopupRect());

            // This callback allows the tool to catch the escape key from the color picker
            mBoxPopup.KeyDown += ColorPickerKeyDown;
        }

        void cps_ColorSelected(Color color)
        {
            if (color.A == 0)
            {
                ColorPickerHandler(false, true, color);
            }
            else
            {
                ColorPickerHandler(false, false, color);
            }
        }

        protected void ColorPickerKeyDown(object sender, KeyRoutedEventArgs e)
        {
            if (e.Key == Windows.System.VirtualKey.Escape)
            {
                ColorPickerHandler(true, false, Colors.Black);
            }
        }

        /// <summary>
        /// Delegate for the ColorPicker. Receives the color and saves the choice in the roaming settings
        /// </summary>
        /// <param name="canceled">If the user clicked on cancel</param>
        /// <param name="empty">If the selected Empty Color</param>
        /// <param name="color">The selected Color</param>
        protected void ColorPickerHandler(bool canceled, bool empty, Color color)
        {
            mBoxPopup.Hide();
            mBoxPopup = null;
            ShowContextMenu(base.mContextMenu);
            EnableScrolling();
            if (canceled)
            {
                return;
            }
            else
            {
                UpdateColorPickerSettings(empty, color);
            }
            SetCurrentAnnotColor();
        }

        protected void UpdateColorPickerSettings(bool empty, Color color)
        {
            mUseColor = !empty;
            if (mUseColor)
            {
                mSelectedColor = color;
            }

            SettingsColor sc = new SettingsColor() { R = color.R, G = color.G, B = color.B, Use = mUseColor };
            switch (mEffectiveBrush)
            {
                case (e_stroke):
                    if (!mUseColor)
                    {
                        SettingsColor oldColor;
                        if (Settings.HasMarkupStrokeColor)
                        {
                            oldColor = Settings.MarkupStrokeColor;
                        }
                        else
                        {
                            oldColor = new SettingsColor() { R = mSelectedColor.R, G = mSelectedColor.G, B = mSelectedColor.B };
                        }
                        sc.R = oldColor.R;
                        sc.G = oldColor.G;
                        sc.B = oldColor.B;
                    }
                    Settings.MarkupStrokeColor = sc;
                    break;
                case (e_fill):
                    Settings.MarkupFillColor = sc;
                    break;
                case e_text_stroke:
                    Settings.TextStrokeColor = sc;
                    break;
                case (e_text_fill):
                    Settings.TextFillColor = sc;
                    break;
                case (e_highlight):
                    Settings.HighlightColor = sc;
                    break;
                case (e_text_markup):
                    Settings.TextMarkupColor = sc;
                    break;
            }
        }


        /// <summary>
        /// Gets the selected color of the current annotation
        /// </summary>
        protected void GetCurrentColor()
        {
            ColorPt color = new ColorPt();
            int r = 0;
            int g = 0;
            int b = 0;
            mSelectedColor.A = 255;

            switch (mEffectiveBrush)
            {
                case e_stroke:
                case e_text_fill:
                case e_highlight:
                case e_text_markup:
                    color = mAnnot.GetColorAsRGB();
                    if (mAnnot.GetColorCompNum() == 0)
                    {
                        mSelectedColor.A = 0;
                    }
                    r = (int)(color.Get(0) * 255 + 0.5);
                    g = (int)(color.Get(1) * 255 + 0.5);
                    b = (int)(color.Get(2) * 255 + 0.5);
                    mSelectedColor.R = (byte)r;
                    mSelectedColor.G = (byte)g;
                    mSelectedColor.B = (byte)b;
                    break;
                case e_fill:
                    if (mMarkup != null)
                    {
                        color = mMarkup.GetInteriorColor();
                        if (mMarkup.GetInteriorColorCompNum() == 0)
                        {
                            mSelectedColor.A = 0;
                        }
                        r = (int)(color.Get(0) * 255 + 0.5);
                        g = (int)(color.Get(1) * 255 + 0.5);
                        b = (int)(color.Get(2) * 255 + 0.5);
                        mSelectedColor.R = (byte)r;
                        mSelectedColor.G = (byte)g;
                        mSelectedColor.B = (byte)b;
                    }
                    break;
                case e_text_stroke:
                    //pdftron.PDF.Annots.FreeText ft = new pdftron.PDF.Annots.FreeText(mAnnot.GetSDFObj());
                    pdftron.PDF.Annots.FreeText ft = (FreeText)mAnnot;
                    color = ft.GetTextColor();
                    r = (int)(color.Get(0) * 255 + 0.5);
                    g = (int)(color.Get(1) * 255 + 0.5);
                    b = (int)(color.Get(2) * 255 + 0.5);
                    mSelectedColor.R = (byte)r;
                    mSelectedColor.G = (byte)g;
                    mSelectedColor.B = (byte)b;
                    break;
            }
        }

        /// <summary>
        /// Sets the selected color of the current annotation
        /// </summary>
        protected virtual void SetCurrentAnnotColor()
        {
            try
            {
                mPDFView.DocLock(true);
                ColorPt color = new ColorPt(mSelectedColor.R / 255.0, mSelectedColor.G / 255.0, mSelectedColor.B / 255.0);
                switch (mAnnot.GetAnnotType())
                {
                    case AnnotType.e_Circle:
                    case AnnotType.e_Square:
                    case AnnotType.e_Line:
                    case AnnotType.e_Ink:
                    case AnnotType.e_Highlight:
                    case AnnotType.e_Underline:
                    case AnnotType.e_Squiggly:
                    case AnnotType.e_StrikeOut:
                    case AnnotType.e_Polygon:
                    case AnnotType.e_Polyline:
                        //pdftron.PDF.Annots.Markup markup = new pdftron.PDF.Annots.Markup(mAnnot.GetSDFObj());
                        if (mEffectiveBrush == e_stroke || mEffectiveBrush == e_highlight || mEffectiveBrush == e_text_markup)
                        {
                            if (mUseColor)
                            {
                                mMarkup.SetColor(color, 3);
                            }
                            else
                            {
                                ColorPt emptyColor = new ColorPt(0, 0, 0, 0);
                                mMarkup.SetColor(emptyColor, 0); // using 0 color points makes it transparent
                            }
                        }
                        else if (mEffectiveBrush == e_fill)
                        {
                            if (mUseColor)
                            {
                                mMarkup.SetInteriorColor(color, 3);
                            }
                            else
                            {
                                ColorPt emptyColor = new ColorPt(0, 0, 0, 0);
                                mMarkup.SetInteriorColor(emptyColor, 0);
                            }
                        }
                        mMarkup.RefreshAppearance();

                        mAnnot.SetDateToNow();

                        mPDFView.UpdateWithAnnot(mAnnot, mAnnotPageNum);
                        mToolManager.RaiseAnnotationEditedEvent(mAnnot, mAnnotPageNum);
                        break;

                    case AnnotType.e_FreeText:
                        pdftron.PDF.Annots.FreeText ft = (FreeText)mAnnot;

                        if (mEffectiveBrush == e_text_stroke)
                        {
                            if (mUseColor)
                            {
                                ft.SetTextColor(color, 3);
                            }
                            else
                            {
                                ColorPt emptyColor = new ColorPt(0, 0, 0, 0);
                                ft.SetTextColor(emptyColor, 0);
                            }
                        }

                        if (mEffectiveBrush == e_text_fill)
                        {
                            if (mUseColor)
                            {
                                ft.SetColor(color, 3);
                            }
                            else
                            {
                                ColorPt emptyColor = new ColorPt(0, 0, 0, 0);
                                ft.SetColor(emptyColor, 0);
                            }
                        }
                        ft.RefreshAppearance();

                        mAnnot.SetDateToNow();

                        mPDFView.UpdateWithAnnot(mAnnot, mAnnotPageNum);
                        mToolManager.RaiseAnnotationEditedEvent(mAnnot, mAnnotPageNum);
                        break;
                }
            }
            catch (Exception) { }
            finally
            {
                mPDFView.DocUnlock();
            }
        }

        #region Opacity
        protected void HandleOpacity()
        {
            Dictionary<string, string> opacityDict = new Dictionary<string, string>();
            opacityDict["25%"] = string.Format(ResourceHandler.GetString("EditOpacity_Format"), 25);
            opacityDict["50%"] = string.Format(ResourceHandler.GetString("EditOpacity_Format"), 50);
            opacityDict["75%"] = string.Format(ResourceHandler.GetString("EditOpacity_Format"), 75);
            opacityDict["100%"] = string.Format(ResourceHandler.GetString("EditOpacity_Format"), 100);
            mExtraContextMenu = new ContextMenuControl(mPDFView, opacityDict);
            mExtraContextMenu.ItemClicked += OnOpacitySelected;

            PositionContextMenu(mExtraContextMenu);
            ShowContextMenu(mExtraContextMenu);
        }

        protected virtual void OnOpacitySelected(string title)
        {
            HideContextMenu(mExtraContextMenu, true);
            mExtraContextMenu = null;
            double opacity = -1;
            if (title.Equals("0%", StringComparison.OrdinalIgnoreCase))
            {
                opacity = 0;
            }
            else if (title.Equals("25%", StringComparison.OrdinalIgnoreCase))
            {
                opacity = 0.25;
            }
            else if (title.Equals("50%", StringComparison.OrdinalIgnoreCase))
            {
                opacity = 0.5;
            }
            else if (title.Equals("75%", StringComparison.OrdinalIgnoreCase))
            {
                opacity = 0.75;
            }
            else if (title.Equals("100%", StringComparison.OrdinalIgnoreCase))
            {
                opacity = 1.0;
            }
            else
            {
                throw new Exception("The option doesn't exist");
            }

            if (opacity < 0 || opacity > 1)
            {
                return;
            }

            EnableScrolling();

            UpdateOpacitySettingsAndSetOpacity(opacity);
            ShowContextMenu(mContextMenu);
        }

        protected virtual void UpdateOpacitySettingsAndSetOpacity(double opacity)
        {
            // Update settings
            AnnotType type = mAnnot.GetAnnotType();

            if (type == AnnotType.e_Underline || type == AnnotType.e_StrikeOut || type == AnnotType.e_Squiggly)
            {
                Settings.TextMarkupOpacity = opacity;
            }
            else if (type == AnnotType.e_Highlight)
            {
                Settings.HightlightOpacity = opacity;
                //roamingSettings.Values["Tools:HighlightOpacity"] = opacity;
            }
            else if (mAnnot.IsMarkup())
            {
                Settings.MarkupOpacity = opacity;
            }

            SetOpacity(opacity);
        }

        protected void SetOpacity(double opacity)
        {
            try
            {
                mPDFView.DocLock(true);
                mMarkup.SetOpacity(opacity);
                mMarkup.RefreshAppearance();
                mPDFView.UpdateWithAnnot(mAnnot, mAnnotPageNum);
                mToolManager.RaiseAnnotationEditedEvent(mAnnot, mAnnotPageNum);
            }
            catch (Exception )
            {

            }
            finally
            {
                mPDFView.DocUnlock();
            }
        }

        #endregion Opacity

        #region Thickness
        protected void HandleThickness()
        {
            Dictionary<string, string> thicknessDict = new Dictionary<string, string>();
            thicknessDict["0.5 pt"] = string.Format(ResourceHandler.GetString("EditThickness_Format"), 0.5);
            thicknessDict["1 pt"] = string.Format(ResourceHandler.GetString("EditThickness_Format"), 1);
            thicknessDict["1.5 pt"] = string.Format(ResourceHandler.GetString("EditThickness_Format"), 1.5);
            thicknessDict["3 pt"] = string.Format(ResourceHandler.GetString("EditThickness_Format"), 3);
            thicknessDict["5 pt"] = string.Format(ResourceHandler.GetString("EditThickness_Format"), 5);
            thicknessDict["7 pt"] = string.Format(ResourceHandler.GetString("EditThickness_Format"), 7);
            thicknessDict["9 pt"] = string.Format(ResourceHandler.GetString("EditThickness_Format"), 9);

            mExtraContextMenu = new ContextMenuControl(mPDFView, thicknessDict);
            mExtraContextMenu.ItemClicked += OnThicknessSelected;

            PositionContextMenu(mExtraContextMenu);
            ShowContextMenu(mExtraContextMenu);
        }

        protected void OnThicknessSelected(string title)
        {
            double thickness = -1;
            if (title.Equals("0.5 pt", StringComparison.OrdinalIgnoreCase))
            {
                thickness = 0.5;
            }
            else if (title.Equals("1 pt", StringComparison.OrdinalIgnoreCase))
            {
                thickness = 1.0;
            }
            else if (title.Equals("1.5 pt", StringComparison.OrdinalIgnoreCase))
            {
                thickness = 1.5;
            }
            else if (title.Equals("3 pt", StringComparison.OrdinalIgnoreCase))
            {
                thickness = 3.0;
            }
            else if (title.Equals("5 pt", StringComparison.OrdinalIgnoreCase))
            {
                thickness = 5.0;
            }
            else if (title.Equals("7 pt", StringComparison.OrdinalIgnoreCase))
            {
                thickness = 7.0;
            }
            else if (title.Equals("9 pt", StringComparison.OrdinalIgnoreCase))
            {
                thickness = 9.0;
            }
            else
            {
                throw new Exception("The option doesn't exist");
            }

            HideContextMenu(mExtraContextMenu, true);
            mExtraContextMenu = null;
            EnableScrolling();
            if (thickness < 0)
            {
                return;
            }

            // Update settings
            AnnotType type = mAnnot.GetAnnotType();
            if (type == AnnotType.e_Underline || type == AnnotType.e_StrikeOut || type == AnnotType.e_Squiggly)
            {
                Settings.TextMarkupThickness = thickness;
            }
            else
            {
                Settings.MarkupStrokeThickness = thickness;
            }
            SetThickness(thickness);
            ShowContextMenu(mContextMenu);
        }


        protected void SetThickness(double thickness)
        {
            try
            {
                mPDFView.DocLock(true);
                AnnotBorderStyle bs = mMarkup.GetBorderStyle();
                bs.width = thickness;
                mMarkup.SetBorderStyle(bs);
                mMarkup.RefreshAppearance();
                mPDFView.UpdateWithAnnot(mAnnot, mAnnotPageNum);
                mToolManager.RaiseAnnotationEditedEvent(mAnnot, mAnnotPageNum);
            }
            catch (Exception)
            {

            }
            finally
            {
                mPDFView.DocUnlock();
            }

        }

        #endregion Thickness

        #region Text Markup Type

        protected void HandleType()
        {
            Dictionary<string, string> typeDict = new Dictionary<string, string>();
            AnnotType annotType = mAnnot.GetAnnotType();
            if (annotType != AnnotType.e_Highlight)
            {
                typeDict["Highlight"] = ResourceHandler.GetString("TextSelect_Option_Highlight");
            }
            if (annotType != AnnotType.e_Underline)
            {
                typeDict["Underline"] = ResourceHandler.GetString("TextSelect_Option_Underline");
            }
            if (annotType != AnnotType.e_StrikeOut)
            {
                typeDict["StrikeOut"] = ResourceHandler.GetString("TextSelect_Option_Strikeout");
            }
            if (annotType != AnnotType.e_Squiggly)
            {
                typeDict["Squiggly"] = ResourceHandler.GetString("TextSelect_Option_Squiggly");
            }

            mExtraContextMenu = new ContextMenuControl(mPDFView, typeDict);
            mExtraContextMenu.ItemClicked += OnTypeSelected;

            PositionContextMenu(mExtraContextMenu);
            ShowContextMenu(mExtraContextMenu);
        }

        protected void OnTypeSelected(string title)
        {
            HideContextMenu(mExtraContextMenu, true);
            mExtraContextMenu = null;
            EnableScrolling();

            SetSubType(title);
            ShowContextMenu(mContextMenu);
        }


        protected void SetSubType(string subType)
        {
            try
            {
                mPDFView.DocLock(true);
                mAnnot.GetSDFObj().PutName("Subtype", subType);
                mAnnot.RefreshAppearance();
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                mPDFView.DocUnlock();
            }

            mPDFView.UpdateWithAnnot(mAnnot, mAnnotPageNum);
            mToolManager.RaiseAnnotationEditedEvent(mAnnot, mAnnotPageNum);
        }
        #endregion Text Markup Type

        #region Copy

        protected void HandleCopy()
        {
            string text = UtilityFunctions.GetSelectedTextOnPage(mPDFView, mAnnotPageNum);

            UtilityFunctions.CopySelectedTextToClipBoard(text);
        }

        protected void HandleAnnotCopy()
        {
            Utilities.AppCenterHandlerBase.CURRENT.SendEvent("Clipboard", new Dictionary<string, string>() { ["Command"] = "Copy Annotation - Popup Menu" });
            AnnotationClipboardHelper.CopyAnnotToClipboard(mPDFView, mAnnot);
        }
        #endregion Copy

        #region Paste
        protected async Task HandleAnnotPasteAsync()
        {
            if (AnnotationClipboardHelper.IsAnnotToPaste)
            {
                Utilities.AppCenterHandlerBase.CURRENT.SendEvent("Clipboard", new Dictionary<string, string>() { ["Command"] = "Paste Annotation - Popup Menu" });

                var currentPage = mPDFView.GetCurrentPage();

                bool hasPasted = await AnnotationClipboardHelper.PasteAnnotFromClipboardAsync(mPDFView, currentPage);
                if (hasPasted == false)
                    return;

                try
                {
                    var doc = mPDFView.GetDoc();
                    if (doc.GetPageCount() >= currentPage)
                    {
                        var page = doc.GetPage(currentPage);
                        var annot = page.GetAnnot(page.GetNumAnnots() - 1);
                        mToolManager.RaiseAnnotationAddedEvent(annot, currentPage);
                        mToolManager.SelectAnnotation(annot, currentPage);
                    }
                }
                catch
                {
                    //Swallow exception
                }
            }
        }
        #endregion

        #region Duplicate
        protected async Task HandleDuplicateAnnotAsync()
        {
            Utilities.AppCenterHandlerBase.CURRENT.SendEvent("Clipboard", new Dictionary<string, string>() { ["Command"] = "Duplicate Annotation - Popup Menu" });
            HandleAnnotCopy();
            await HandleAnnotPasteAsync();
        }
        #endregion

        #region Cut Annot
        protected void HandleCutAnnot()
        {
            Utilities.AppCenterHandlerBase.CURRENT.SendEvent("Clipboard", new Dictionary<string, string>() { ["Command"] = "Cut Annotation - Popup Menu" });
            
            HandleAnnotCopy();
            DeleteAnnotation();
            CreateNewTool(ToolType.e_pan);
        }
        #endregion

        #region Notes


        protected void HandleNote()
        {
            try
            {
                mPDFView.DocLockRead();
                if (mMarkup != null && mPDFView.IsEnabled)
                {
                    Utilities.NoteDialogBase noteDialog = null;
                    bool canDelete = mAnnotIsSticky || UtilityFunctions.DoesAnnotHaveContent(mMarkup);
                    string contents = UtilityFunctions.GetAnnotContent(mMarkup);
                    if (mToolManager.IsOnPhone)
                    {
                        noteDialog = new Utilities.NoteDialogPhone(contents, canDelete, false, mPDFView, mToolManager);
                    }
                    else
                    {
                        PDFDouble x = new PDFDouble((mLeft + mRight) / 2);
                        PDFDouble y = new PDFDouble((mTop + mBottom) / 2);
                        mPDFView.ConvAnnotationCanvasPtToScreenPt(x, y);

                        noteDialog = new Utilities.NoteDialogWindows(contents, canDelete, false, mPDFView, mToolManager, new UIPoint(x.Value, y.Value));
                    }
                    noteDialog.NoteClosed += NoteDialog_NoteClosed;
                }
            }
            catch (Exception) { }
            finally
            {
                mPDFView.DocUnlockRead();
            }
            return;
        }

        private void NoteDialog_NoteClosed(Utilities.NoteDialogBase.NotedialogResults result, string newText)
        {
            bool endTool = false;
            if (result != Utilities.NoteDialogBase.NotedialogResults.Canceled)
            {
                try
                {
                    mPDFView.DocLock(true);

                    if (result == Utilities.NoteDialogBase.NotedialogResults.Changed)
                    {
                        if (UtilityFunctions.DoesAnnotHavePopup(mMarkup))
                        {
                            Annots.Popup pop = mMarkup.GetPopup();
                            pop.SetContents(newText);
                        }
                        else
                        {
                            Annots.Popup pop = UtilityFunctions.AddPopupToAnnot(mPDFView.GetDoc(), mMarkup, newText);
                            mPage = mPDFView.GetDoc().GetPage(mAnnotPageNum);
                            mPage.AnnotPushBack(pop);
                        }

                        if (!mAnnotIsSticky)
                        {
                            if (UtilityFunctions.DoesAnnotHaveContent(mMarkup))
                            {
                                mMenuTitles["Note"] = ResourceHandler.GetString("EditOption_ViewNote");
                            }
                            else
                            {
                                mMenuTitles["Note"] = ResourceHandler.GetString("EditOption_AddNote");
                            }
                            base.mContextMenu = new ContextMenuControl(mPDFView, mMenuTitles);
                            base.mContextMenu.ItemClicked += this.OnCommandMenuClicked;
                        }
                        mAnnot.SetDateToNow();
                        mToolManager.RaiseAnnotationEditedEvent(mAnnot, mAnnotPageNum);
                    }

                    if (result == Utilities.NoteDialogBase.NotedialogResults.Deleted)
                    {
                        if (mAnnotIsSticky)
                        {
                            mToolManager.RaiseAnnotationPreRemovedEvent(mAnnot, mAnnotPageNum);

							// Perform the Annotation removal and update PDFViewer
                            mPage = mPDFView.GetDoc().GetPage(mAnnotPageNum);
                            mAnnot = AnnotUtils.SafeDeleteAnnotAndUpdate(mPDFView, mPage, mAnnot, mAnnotPageNum);

							// make sure to raise removed event after mPDFView.UpdateWithAnnot()
                            mToolManager.RaiseAnnotationRemovedEvent(mAnnot, mAnnotPageNum);

                            endTool = true;
                            EndCurrentTool(ToolType.e_pan);
                        }
                        else
                        {
                            mMarkup.SetContents("");
                            if (UtilityFunctions.DoesAnnotHavePopup(mMarkup))
                            {
                                Annots.Popup pop = mMarkup.GetPopup();
                                pop.SetContents("");
                            }
                            mToolManager.RaiseAnnotationEditedEvent(mAnnot, mAnnotPageNum);

                            mMenuTitles["Note"] = ResourceHandler.GetString("EditOption_AddNote");
                            base.mContextMenu = new ContextMenuControl(mPDFView, mMenuTitles);
                            base.mContextMenu.ItemClicked += this.OnCommandMenuClicked;
                        }
                    }
                }
                catch (Exception e)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(e));
                }
                finally
                {
                    mPDFView.DocUnlock();
                }
            }

            if (!endTool)
            {
                if (mContextMenu == null)
                {
                    mContextMenu = new ContextMenuControl(mPDFView, mMenuTitles);
                    mContextMenu.ItemClicked += this.OnCommandMenuClicked;
                }
                PositionContextMenu(mContextMenu);
                ShowContextMenu(mContextMenu);
            }
        }

        void noteDialog_NoteClosed2(bool wasDeleted)
        {
            if (wasDeleted)
            {
                EndCurrentTool(ToolType.e_pan);
            }
            else if (mToolManager.IsPopupMenuEnabled)
            {
                if (mContextMenu == null)
                {
                    mContextMenu = new ContextMenuControl(mPDFView, mMenuTitles);
                    mContextMenu.ItemClicked += OnCommandMenuClicked;
                }
                PositionContextMenu(mContextMenu);
                ShowContextMenu(mContextMenu);
            }
        }

        protected void NoteKeyDown(object sender, KeyRoutedEventArgs e)

        {
            if (e.Key == Windows.System.VirtualKey.Escape)
            {
                mBoxPopup.Hide();
                mBoxPopup = null;
            }
        }


        /// <summary>
        /// Translates the typing area if necessary in order to keep the text area above virtual keyboard
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        protected void NoteCustomKeyboardHandler(object sender, Windows.UI.ViewManagement.InputPaneVisibilityEventArgs e)
        {
            double RemainingHeight = e.OccludedRect.Y; // Y value of keyboard top (thus, also height of rest).
            double TotalHeight = mPDFView.ActualHeight;
            double KeyboardHeight = e.OccludedRect.Height; // height of keyboard

            e.EnsuredFocusedElementInView = true;
            mKeyboardOpen = true;

            // figure out how much of the text box will be covered.
            UIRect textRect = GetElementRect(mNoteDisplay);
            UIRect viewRect = GetElementRect(mToolManager.PopupTargetInternal);
            double bottomMargin = RemainingHeight - textRect.Bottom - 10;
            double topMargin = textRect.Top - viewRect.Top;

            // figure out translation
            if (bottomMargin < 0)
            {
                mKeyboardTranslateOffset = -bottomMargin;
                if (mKeyboardTranslateOffset > topMargin)
                {
                    mKeyboardTranslateOffset = topMargin;
                }
                Windows.UI.Xaml.Media.Animation.Storyboard sb = KeyboardTranslate(mNoteDisplay, 0, -mKeyboardTranslateOffset, TimeSpan.FromMilliseconds(733));
                sb.Begin();
            }
        }

        /// <summary>
        /// Translates the PDFViewCtrl back to offset 0
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        protected void NoteInputPaneHiding(Windows.UI.ViewManagement.InputPane sender, Windows.UI.ViewManagement.InputPaneVisibilityEventArgs e)
        {
            mKeyboardOpen = false;
            if (mKeyboardTranslateOffset != 0)
            {
                Windows.UI.Xaml.Media.Animation.Storyboard sb = KeyboardTranslate(mNoteDisplay, -mKeyboardTranslateOffset, 0, TimeSpan.FromMilliseconds(367));
                sb.Begin();
            }
        }

        #endregion Notes

        #region Text Edit

        protected void HandleEditText()
        {
            if (mToolManager.IsOnPhone)
            {
                HandleEditTextPhone();
            }
            else
            {
                HandleEditTextWindows();
            }
        }

        protected void HandleEditTextWindows()
        {
            mIsEditingText = true;
            this.Children.Clear();

            mEditTextBox = new RichEditBox();
            mEditTextBox.AcceptsReturn = true;
            mEditTextBox.TextWrapping = Windows.UI.Xaml.TextWrapping.Wrap;

            /* Note: Version Adaptive code
             * SelectionFlyout is present in build version 10.0.17763 */
            var isPropPresent = Windows.Foundation.Metadata.ApiInformation.IsPropertyPresent("Windows.UI.Xaml.Controls.RichEditBox", "SelectionFlyout");
            if (isPropPresent)
            {
                mEditTextBox.SelectionFlyout = null; // Note: Set it to null to avoid displaying Popup Text edit options (Bold, Italic) b/c we don't support it
            }

            // Add two pixels per side, to make sure it cover the entire annotation
            mEditTextBox.Width = mRight - mLeft + 4;
            mEditTextBox.Height = mBottom - mTop + 4;
            mEditTextBox.KeyDown += FreeTextPopupKeyDown;

            pdftron.PDF.Annots.FreeText ft = (FreeText)mAnnot;

            bool isRichText = UtilityFunctions.IsRichFreeText(ft);

            // fetch text color
            int colorPts = ft.GetTextColorCompNum();
            ColorPt color = ft.GetTextColor();
            mSelectedColor.A = (byte)255;
            if (colorPts == 1) // greyscale
            {
                int r = (int)(color.Get(0) * 255 + 0.5);
                int g = (int)(color.Get(0) * 255 + 0.5);
                int b = (int)(color.Get(0) * 255 + 0.5);
                mSelectedColor.R = (byte)r;
                mSelectedColor.G = (byte)g;
                mSelectedColor.B = (byte)b;
            }
            else if (colorPts == 3)
            {
                int r = (int)(color.Get(0) * 255 + 0.5);
                int g = (int)(color.Get(1) * 255 + 0.5);
                int b = (int)(color.Get(2) * 255 + 0.5);
                mSelectedColor.R = (byte)r;
                mSelectedColor.G = (byte)g;
                mSelectedColor.B = (byte)b;
            }
            else // default to red
            {
                mSelectedColor = Colors.Red;
            }

            // Fetch background color
            Color BackgroundColor = Colors.White; // default
            colorPts = ft.GetColorCompNum();
            color = ft.GetColorAsRGB();
            if (colorPts == 3)
            {
                int r = (int)(color.Get(0) * 255 + 0.5);
                int g = (int)(color.Get(1) * 255 + 0.5);
                int b = (int)(color.Get(2) * 255 + 0.5);
                BackgroundColor.R = (byte)r;
                BackgroundColor.G = (byte)g;
                BackgroundColor.B = (byte)b;
            }
            mFontSize = ft.GetFontSize();
            if (mFontSize <= 0)
            {
                mFontSize = 11;
            }

            mEditTextBox.FontSize = mFontSize * mPDFView.GetZoom();

            var controlStyles = new ResourceDictionary()
            {
                Source = new Uri("ms-appx:///pdftron.PDF.Tools/Controls/Resources/ControlStyles.xaml", UriKind.RelativeOrAbsolute)
            };

            Windows.UI.Xaml.Style style = controlStyles["RichEditBoxWithCustomColors"] as Style;

            SolidColorBrush textBrush = new SolidColorBrush(mSelectedColor);

            if (colorPts <= 0)
            {
                // For transparent backgrounds, the PDF already has this text displayed on the document so
                // we need to use a white or black background (depending on brightness of font) to cover it
                if (mSelectedColor.R + mSelectedColor.G + mSelectedColor.B > 500)
                {
                    BackgroundColor.R = 0;
                    BackgroundColor.G = 0;
                    BackgroundColor.B = 0;
                }
            }
            SolidColorBrush fillBrush = new SolidColorBrush(BackgroundColor);

            //If RC exists, use colors from RC
            if (isRichText == false)
            {
                style.Setters.Add(new Windows.UI.Xaml.Setter(RichEditBox.ForegroundProperty, textBrush));
            }
            style.Setters.Add(new Windows.UI.Xaml.Setter(RichEditBox.BackgroundProperty, fillBrush));

            mEditTextBox.Style = style;
            mEditTextBox.BorderThickness = new Thickness(0);

            UtilityFunctions.FontBundle bundle = UtilityFunctions.GetFontFromAnnot(ft);
            if (bundle != null)
            {
                mEditTextBox.FontFamily = new FontFamily(bundle.FontUri);
                mEditTextBox.FontWeight = bundle.FontWeight;
                mEditTextBox.FontStyle = bundle.FontStyle;
            }

            if (isRichText)
            {
                string textPlain = ft.GetContents();
                string textHTML = UtilityFunctions.GetRichFreeText(ft);
                string textRTF = "";
                ConvertRTFAndHTML convertRTFAndHTML = new ConvertRTFAndHTML();
                if (string.IsNullOrEmpty(textHTML))
                {
                    textRTF = textPlain;
                }
                else
                {
                    textRTF = convertRTFAndHTML.ConvertHTMLToRTF(textHTML);
                }

                mEditTextBox.Document.SetText(Windows.UI.Text.TextSetOptions.FormatRtf | Windows.UI.Text.TextSetOptions.ApplyRtfDocumentDefaults, textRTF);
            }
            else
            {
                mEditTextBox.Document.SetText(Windows.UI.Text.TextSetOptions.ApplyRtfDocumentDefaults, ft.GetContents());
            }

            mEditTextBox.Loaded += (s, e) =>
                {
                    mEditTextBox.Focus(Windows.UI.Xaml.FocusState.Programmatic); // can't set focus until loaded
                    string outString = "";
                    mEditTextBox.Document.GetText(Windows.UI.Text.TextGetOptions.UseCrlf, out outString);
                    mEditTextBox.Document.Selection.StartPosition = outString.Length;
                };

            Border border = new Border();
            border.Width = mEditTextBox.Width;
            border.Height = mEditTextBox.Height;
            border.Background = fillBrush;

            mTextPopup = new Windows.UI.Xaml.Controls.Primitives.Popup();
            mTextPopup.Width = mEditTextBox.Width + (mFontSize * 0.2);
            mTextPopup.Height = mEditTextBox.Height + (mFontSize);
            PositionTextEditingPopup();

            border.Child = mEditTextBox;

            mTextPopup.Child = border;
            mTextPopup.IsOpen = true;

            DisableScrolling();

            mSaveTextOnExit = true;

            mKeyboardHelper.AddShowingHandler(mEditTextBox, new InputPaneShowingHandler(CustomKeyboardHandler));
            mKeyboardHelper.SetHidingHandler(new InputPaneHidingHandler(InputPaneHiding));

            HandleStyle(false);
        }

        protected void PositionTextEditingPopup()
        {
            if (mTextPopup != null)
            {
                UIRect offsetRect = GetElementRect(mPDFView);

                PDFDouble x = new PDFDouble(mLeft);
                PDFDouble y = new PDFDouble(mTop);
                mPDFView.ConvAnnotationCanvasPtToScreenPt(x, y);

                mTextPopup.HorizontalOffset = x.Value + offsetRect.X;
                mTextPopup.VerticalOffset = y.Value + offsetRect.Y;
            }
        }

        private void FreeTextPopupKeyDown(object sender, Windows.UI.Xaml.Input.KeyRoutedEventArgs e)
        {
            if (e.Key == Windows.System.VirtualKey.Escape)
            {
                mSaveTextOnExit = false;
                CreateNewTool(ToolType.e_pan);
                if (mPopup != null)
                {
                    mPopup.IsPopupOpen = false;
                }
            }
        }

        pdftron.PDF.Tools.Utilities.NoteDialogPhone _NoteDialog = null;
        protected void HandleEditTextPhone()
        {

            string contents = string.Empty;
            try
            {
                mPDFView.DocLockRead();
                pdftron.PDF.Annots.FreeText ft = (FreeText)mAnnot;
                contents = ft.GetContents();
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                mPDFView.DocUnlockRead();
            }

            _NoteDialog = new Utilities.NoteDialogPhone(contents, false, false, mPDFView, mToolManager);
            _NoteDialog.NoteClosed += FreeTextNoteDialog_Closed;
        }

        private void FreeTextNoteDialog_Closed(NoteDialogBase.NotedialogResults result, string newText)
        {
            if (result == NoteDialogBase.NotedialogResults.Changed)
            {
                SaveFreeText(_NoteDialog.AnnotText);
                mToolManager.RaiseAnnotationEditedEvent(mAnnot, mAnnotPageNum);
            }

            PositionContextMenu(mContextMenu);
            ShowContextMenu(mContextMenu);
        }

        protected void SaveFreeText(string text = null)
        {
            mSaveTextOnExit = false;
            pdftron.PDF.Annots.FreeText ft = (FreeText)mAnnot;
            string outString = text;
            if (outString == null)
            {
                if (mEditTextBox== null)
                {
                    return;
                }
                mEditTextBox.Document.GetText(Windows.UI.Text.TextGetOptions.UseCrlf, out outString);
            }
            bool deleted = false;
            if (outString == "")
            {
                DeleteAnnotation();
                CreateNewTool(ToolType.e_pan);
                deleted = true;
            }
            else
            {
                try
                {
                    mPDFView.DocLock(true);
                    if (UtilityFunctions.IsRichFreeText(ft) || mToolManager.IsRichTextEnabled)
                    {
                        ConvertRTFAndHTML converter = new ConvertRTFAndHTML();
                        string outStrRTF = "";
                        mEditTextBox.Document.GetText(Windows.UI.Text.TextGetOptions.FormatRtf, out outStrRTF);
                        string richTextHTML = converter.ConvertRTFToHTML(outStrRTF);

                        ft.GetSDFObj().PutString(UtilityFunctions.KeyRichContent, richTextHTML);
                    }
                    ft.SetContents(outString);

                    string fontName = UtilityFunctions.GetFontNameFromAnnot(ft);
                    if (!string.IsNullOrEmpty(fontName))
                    {
                        UtilityFunctions.FontBundle bundle = UtilityFunctions.FontBundle.CreateFromPDFTronName(fontName);
                        if (bundle != null)
                        {
                            UtilityFunctions.SetFontFromBundle(ft, mPDFView.GetDoc(), bundle);
                            UtilityFunctions.GenerateDSEntry(ft, bundle);
                        }
                        else
                        {
                            UtilityFunctions.SetFreeTextDefaultAppearanceWithoutFont(ft);
                        }
                    }

                    ft.RefreshAppearance();
                    mAnnot.SetDateToNow();
                }
                catch (Exception ex)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                }
                finally
                {
                    mSaveTextOnExit = false;
                    mPDFView.DocUnlock();
                    if (mPopup != null)
                    {
                        mPopup.IsPopupOpen = false;
                    }
                }
            }
            if (!deleted)
            {
                mPDFView.UpdateWithAnnot(mAnnot, mAnnotPageNum);
                mToolManager.RaiseAnnotationEditedEvent(mAnnot, mAnnotPageNum);
            }
        }

        /// <summary>
        /// Translates the entire PDFViewCtrl if necessary in order to keep the text area above virtual keyboard
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        protected void CustomKeyboardHandler(object sender, Windows.UI.ViewManagement.InputPaneVisibilityEventArgs e)
        {
            double RemainingHeight = e.OccludedRect.Y; // Y value of keyboard top (thus, also height of rest).
            double TotalHeight = mPDFView.ActualHeight;
            double KeyboardHeight = e.OccludedRect.Height; // height of keyboard

            e.EnsuredFocusedElementInView = true;
            mKeyboardOpen = true;

            // figure out how much of the text box will be covered.
            UIRect textRect = GetElementRect(mEditTextBox);
            UIRect viewRect = GetElementRect(mToolManager.PopupTargetInternal);
            double bottomMargin = RemainingHeight - textRect.Bottom - 10;
            double topMargin = textRect.Top - viewRect.Top;

            // figure out translation
            if (bottomMargin < 0)
            {
                mKeyboardTranslateOffset = -bottomMargin;
                if (mKeyboardTranslateOffset > topMargin)
                {
                    mKeyboardTranslateOffset = topMargin;
                }
                Windows.UI.Xaml.Media.Animation.Storyboard sb = KeyboardTranslate(mPDFView, 0, -mKeyboardTranslateOffset, TimeSpan.FromMilliseconds(733));
                Windows.UI.Xaml.Media.Animation.Storyboard sbText = KeyboardTranslate(mEditTextBox, 0, -mKeyboardTranslateOffset, TimeSpan.FromMilliseconds(733));
                sb.Begin();
                sbText.Begin();
            }
        }

        /// <summary>
        /// Translates the PDFViewCtrl back to offset 0
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        protected void InputPaneHiding(Windows.UI.ViewManagement.InputPane sender, Windows.UI.ViewManagement.InputPaneVisibilityEventArgs e)
        {
            mKeyboardOpen = false;
            if (mKeyboardTranslateOffset != 0)
            {
                Windows.UI.Xaml.Media.Animation.Storyboard sb = KeyboardTranslate(mPDFView, -mKeyboardTranslateOffset, 0, TimeSpan.FromMilliseconds(367));
                Windows.UI.Xaml.Media.Animation.Storyboard sbText = KeyboardTranslate(mEditTextBox, -mKeyboardTranslateOffset, 0, TimeSpan.FromMilliseconds(367));
                sb.Begin();
                sbText.Begin();
            }
        }

        protected UIRect GetElementRect(Windows.UI.Xaml.FrameworkElement element)
        {
            GeneralTransform elementtransform = element.TransformToVisual(null);
            UIRect rect = elementtransform.TransformBounds(new UIRect(new UIPoint(0, 0), new Windows.Foundation.Size(element.ActualWidth, element.ActualHeight)));
            return rect;
        }
        #endregion Text Edit

        #region Font Size

        protected void HandleFontSize()
        {
            Dictionary<string, string> fontsizeDict = new Dictionary<string, string>();
            fontsizeDict["8"] = string.Format(ResourceHandler.GetString("EditFontSize_Format"), 8);
            fontsizeDict["11"] = string.Format(ResourceHandler.GetString("EditFontSize_Format"), 11);
            fontsizeDict["16"] = string.Format(ResourceHandler.GetString("EditFontSize_Format"), 16);
            fontsizeDict["24"] = string.Format(ResourceHandler.GetString("EditFontSize_Format"), 24);
            fontsizeDict["36"] = string.Format(ResourceHandler.GetString("EditFontSize_Format"), 36);

            mExtraContextMenu = new ContextMenuControl(mPDFView, fontsizeDict);
            mExtraContextMenu.ItemClicked += OnFontSizeSelected;
            PositionContextMenu(mExtraContextMenu);

            //var eventArgs = new ShowingContextMenuEventArgs
            //{
            //    ToolType = ToolType.e_annot_edit,
            //    AnnotType = mAnnot.GetAnnotType(),
            //    Annotation = mAnnot,
            //    PageNumber = mPDFView.GetCurrentPage()
            //};
            //mToolManager.RaiseOnShowContextMenu(mExtraContextMenu, eventArgs);

            ShowContextMenu(mExtraContextMenu);
        }

        protected void OnFontSizeSelected(string title)
        {
            HideContextMenu(mExtraContextMenu, true);
            mExtraContextMenu = null;
            double fontSize = -1;
            if (title.Equals("8", StringComparison.OrdinalIgnoreCase))
            {
                fontSize = 8;
            }
            else if (title.Equals("11", StringComparison.OrdinalIgnoreCase))
            {
                fontSize = 11;
            }
            else if (title.Equals("16", StringComparison.OrdinalIgnoreCase))
            {
                fontSize = 16;
            }
            else if (title.Equals("24", StringComparison.OrdinalIgnoreCase))
            {
                fontSize = 24;
            }
            else if (title.Equals("36", StringComparison.OrdinalIgnoreCase))
            {
                fontSize = 36;
            }
            else
            {
                throw new Exception("The option doesn't exist");
            }

            if (fontSize < 0)
            {
                return;
            }

            // Update settings
            Settings.FontSize = fontSize;

            //SetFontSize(new pdftron.PDF.Annots.FreeText(mAnnot.GetSDFObj()), fontSize);
            SetFontSize((FreeText)mAnnot, fontSize);
            ShowContextMenu(mContextMenu);
        }

        /// <summary>
        /// This function sets the font size of the free text annotation
        /// </summary>
        /// <param name="ft">The Free Text Annotation</param>
        /// <param name="fontSize">The size of the font</param>
        protected void SetFontSize(pdftron.PDF.Annots.FreeText ft, double fontSize)
        {
            try
            {
                mPDFView.DocLock(true);
                ft.SetFontSize(fontSize);
                ft.RefreshAppearance();
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                mPDFView.DocUnlock();
            }
            mPDFView.UpdateWithAnnot(mAnnot, mAnnotPageNum);
            mToolManager.RaiseAnnotationEditedEvent(mAnnot, mAnnotPageNum);
        }

        #endregion Font Size

        #region AnnotationProperties

        private void HandleStyle(bool isLightDismissed)
        {
            AnnotationPropertyDialog dialog = new AnnotationPropertyDialog();

            SetDialogProperties(dialog.ViewModel);

            try
            {
                mPDFView.DocLockRead();
                ToolType toolType = UtilityFunctions.GetToolTypeFromAnnotType(mAnnot);

                if (mIsEditingText && mToolManager.IsRichTextEnabled && toolType == ToolType.e_text_annot_create)
                {
                    dialog.ItemTag = "EditText";
                }
                dialog.ViewModel.AssociatedTool = toolType;
            }
            catch (Exception) { }
            finally
            {
                mPDFView.DocUnlockRead();
            }

            UIRect positionRect = GetBoxPopupRect();
            UIRect pdfViewCtrlRect = UtilityFunctions.GetElementRect(mPDFView);
            UIRect bound = Window.Current.Bounds;

            double shiftBackHalfLeftNavigationPane = 0;
            if (pdfViewCtrlRect.Left > 0)
            {
                shiftBackHalfLeftNavigationPane = pdfViewCtrlRect.Left * 0.5;
            }

            positionRect.X += pdfViewCtrlRect.Left - shiftBackHalfLeftNavigationPane;
            positionRect.Y += pdfViewCtrlRect.Top;

            double left = positionRect.Left + pdfViewCtrlRect.Left- shiftBackHalfLeftNavigationPane;
            double right = positionRect.Right + pdfViewCtrlRect.Left- shiftBackHalfLeftNavigationPane;
            double top = positionRect.Top + pdfViewCtrlRect.Top;
            double bottom = positionRect.Bottom + pdfViewCtrlRect.Top;

            if (left < 0)
            {
                left = 0;
            }
            if (top < 0)
            {
                top = 0;
            }
            if (right > bound.Width)
            {
                right = bound.Width;
            }
            if (right <= left)
            {
                right = left + 1;
            }
            if (bottom > bound.Height)
            {
                bottom = bound.Height;
            }
            if (bottom <= top)
            {
                bottom = top + 1;
            }

            if (mIsEditingText && mToolManager.IsRichTextEnabled || !mIsEditingText)
            {
                mPopup = AnnotationPropertiesDisplay.CreateAnnotationPropertiesDisplay(
                    new UIRect(left, top, right - left, bottom - top), dialog,
                    AnnotationPropertiesDisplay.Placement.PreferBeside);
                mPopup.IsPopupLightDismissed = isLightDismissed;
                mPopup.PlacePopup();
                mPopup.StyleChangeRequested += DisplayStyleChangeRequested;
            }
        }

        protected async void HandleFlatten()
        {
            Context mContext = mPDFView.GetOCGContext();

            ContentDialog dialog = new ContentDialog
            {
                PrimaryButtonText = ResourceHandler.GetString("Stamper_Flatten_Dialog_Flatten"),
                CloseButtonText = ResourceHandler.GetString("Stamper_Flatten_Dialog_Cancel")
            };

            if (mAnnotIsSignature)
            {
                dialog.Title = ResourceHandler.GetString("Signature_Flatten_Dialog_Title");
                dialog.Content = ResourceHandler.GetString("Signature_Flatten_Dialog_Content");
            }
            else if (mAnnotIsStamper)
            {
                dialog.Title = ResourceHandler.GetString("Stamper_Flatten_Dialog_Title");
                dialog.Content = ResourceHandler.GetString("Stamper_Flatten_Dialog_Content");
            }
            else if ((mGroupAnnotations != null) && (mGroupAnnotations.Count > 0))
            {
                dialog.Title = ResourceHandler.GetString("Group_Flatten_Dialog_Title");
                dialog.Content = ResourceHandler.GetString("Group_Flatten_Dialog_Content");
            }

            dialog.PrimaryButtonCommand = new RelayCommand((_) =>
            {
                try
                {
                    mPDFView.DocLock(true);

                    Page page = mPDFView.GetDoc().GetPage(mAnnotPageNum);
                    FlattenAnnotation(page);
                    EndCurrentTool(ToolType.e_pan);

                    mToolManager.RaiseAnnotationEditedEvent(mAnnot, mAnnotPageNum);
                }
                catch (Exception ex)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                }
                finally
                {
                    mPDFView.DocUnlock();
                }
            });

            dialog.CloseButtonCommand = new RelayCommand((_) =>
            {
                ShowContextMenu(mContextMenu);
            });


            await DialogHelper.ShowContentDialogAsync(dialog);
        }

        protected virtual void FlattenAnnotation(PDFPage page)
        {
            mAnnot.Flatten(page);
            mPDFView.UpdateWithAnnot(mAnnot, mAnnotPageNum);
        }

        private void HandleRotate()
        {
            if (!mAnnotIsStamper)
            {
                return;
            }

            Obj stampObj = mAnnot.GetSDFObj();
            Obj rotationObj = stampObj.FindObj(ToolManager.StamperRotationAnnotationIdentifyingString);
            int rotation = 0;
            if ((rotationObj != null) && (rotationObj.IsNumber()))
            {
                rotation = (int)rotationObj.GetNumber();
            }

            PageRotate pageRotation = Page.AddRotations(Page.DegreeToRotation(rotation), PageRotate.e_270);
            rotation = Page.RotationToDegree(pageRotation);
            if (mMarkup != null)
            {
                mMarkup.SetRotation(rotation);
            }

            stampObj.PutNumber(ToolManager.StamperRotationAnnotationIdentifyingString, rotation);
            mAnnot.RefreshAppearance();

            mAspectRatio = 1 / mAspectRatio;

            double width = mCtrlPointLocations[e_ur].X - mCtrlPointLocations[e_ul].X;
            double height = mCtrlPointLocations[e_lr].Y - mCtrlPointLocations[e_ur].Y;
            double delta = (height - width) / 2;

            mCtrlPointLocations[e_ul].X = mCtrlPointLocations[e_ul].X + delta * -1;
            mCtrlPointLocations[e_ul].Y = mCtrlPointLocations[e_ul].Y + delta;
            mCtrlPointLocations[e_ll].X = mCtrlPointLocations[e_ll].X + delta * -1;
            mCtrlPointLocations[e_ll].Y = mCtrlPointLocations[e_ll].Y + delta * -1;
            mCtrlPointLocations[e_lr].X = mCtrlPointLocations[e_lr].X + delta;
            mCtrlPointLocations[e_lr].Y = mCtrlPointLocations[e_lr].Y + delta * -1;
            mCtrlPointLocations[e_ur].X = mCtrlPointLocations[e_ur].X + delta;
            mCtrlPointLocations[e_ur].Y = mCtrlPointLocations[e_ur].Y + delta;

            // bound rotation to within PDFViewCtrl
            if (mCtrlPointLocations[e_ul].X < 0)
            {
                double offset = mCtrlPointLocations[e_ul].X * -1;
                mCtrlPointLocations[e_ul].X = 0;
                mCtrlPointLocations[e_ll].X = 0;
                mCtrlPointLocations[e_lr].X += offset;
                mCtrlPointLocations[e_ur].X += offset;
            }
            else if (mCtrlPointLocations[e_ur].X > mPageCropOnClient.x2)
            {
                double offset = mCtrlPointLocations[e_ur].X - mPageCropOnClient.x2;
                mCtrlPointLocations[e_ur].X = mPageCropOnClient.x2;
                mCtrlPointLocations[e_lr].X = mPageCropOnClient.x2;
                mCtrlPointLocations[e_ll].X -= offset;
                mCtrlPointLocations[e_ul].X -= offset;
            }
            else if (mCtrlPointLocations[e_ul].Y < 0)
            {
                double offset = mCtrlPointLocations[e_ul].Y * -1;
                mCtrlPointLocations[e_ul].Y = 0;
                mCtrlPointLocations[e_ur].Y = 0;
                mCtrlPointLocations[e_lr].Y += offset;
                mCtrlPointLocations[e_ll].Y += offset;
            }
            else if (mCtrlPointLocations[e_ll].Y > mPageCropOnClient.y2)
            {
                double offset = mCtrlPointLocations[e_ll].Y - mPageCropOnClient.y2;
                mCtrlPointLocations[e_ll].Y = mPageCropOnClient.y2;
                mCtrlPointLocations[e_lr].Y = mPageCropOnClient.y2;
                mCtrlPointLocations[e_ul].Y -= offset;
                mCtrlPointLocations[e_ur].Y -= offset;
            }

            mLeft = mCtrlPointLocations[e_ll].X;
            mRight = mCtrlPointLocations[e_lr].X;
            mTop = mCtrlPointLocations[e_ul].Y;
            mBottom = mCtrlPointLocations[e_ll].Y;

            CreateSelectionAppearance();

            UpdateAnnotSize();

            ShowContextMenu(mContextMenu);
        }

        async private Task HandleCalibrationAsync()
        {
            var calDialog = new Controls.AnnotationPropertiesSubControls.CalibrationDialog(mAnnot, mPDFView);
            await DialogHelper.ShowContentDialogAsync(calDialog);
        }

        /// <summary>
        /// Handle playing a sound annotation
        /// </summary>
        /// <remarks>
        /// This call will read the stream from a Sound Annot and parse it to save saved
        /// on a wave audio file on disk, then use this file to play using <c>AudioGraph</c>
        /// using lossless PCM format
        /// </remarks>
        private async Task HandlePlaySoundAnnotAsync()
        {
            var soundAnnot = mAnnot as pdftron.PDF.Annots.Sound;
            if (soundAnnot == null)
                return;

            try
            {
                var soundStream = soundAnnot.GetSoundStream();
                if (soundStream == null || soundStream.IsStream() == false)
                    return;

                // Get stream from Sound Annotation and parse it to bytes
                pdftron.Filters.IFilter source = soundStream.GetDecodedStream();
                pdftron.Filters.FilterReader reader = new pdftron.Filters.FilterReader(source);
                
                byte[] resultBytes;
                using (var memoryStream = new System.IO.MemoryStream())
                {
                    byte[] buffer = new byte[1024];
                    uint bytesRead = reader.Read(buffer);

                    uint bytesSize = bytesRead;

                    memoryStream.Write(buffer, 0, buffer.Length);
                    while (bytesRead > 0)
                    {
                        bytesRead = reader.Read(buffer);

                        memoryStream.Write(buffer, 0, (int)bytesRead);
                    }

                    resultBytes = memoryStream.ToArray();
                }

                if (resultBytes == null)
                    return;

                /* 
                 * Now we swap the bytes back to get the original lossless PCM audio format
                 * which is playable using AudioGraph
                 */
                var swapBytes = UtilityFunctions.SwapByteArray(resultBytes);
               // var memStream = new System.IO.MemoryStream(swapBytes);
                //var finalAudioStream = await UtilityFunctions.ConvertToRandomAccessStreamAsync(memStream);
                
                if (swapBytes == null || swapBytes.Length <= 0)
                    return;

                var soundFile = await DocumentManager.SaveAudioWaveToDiskAsync(swapBytes);

                AudioGraphPlayer audioGraphPlayer = new AudioGraphPlayer();
                if (await audioGraphPlayer.InitAudioGraphAsync())
                {
                    await audioGraphPlayer.LoadFileAsync(soundFile);
                    audioGraphPlayer.Play();
                }
            } 
            catch (Exception ex)
            {
                AppCenterHandlerBase.CURRENT.TrackError(ex);
            }
        }

        private void DisplayStyleChangeRequested(Controls.ViewModels.AnnotStyleType annotStyleType, object val)
        {
            SetAnnotPropertiesFromDialog(annotStyleType, val);
        }

        private void SetDialogProperties(Controls.ViewModels.AnnotationPropertiesViewModel viewModel)
        {
            try
            {
                mPDFView.DocLockRead();
                SettingsColor primary = GetPrimaryColor();
                viewModel.MainColor = primary;
                if (HasSecondaryColor())
                {
                    SettingsColor secondary = GetSecondaryColor();
                    viewModel.FillColor = secondary;
                }

                if (mType == AnnotType.e_Line || mType == AnnotType.e_Ink || mType == AnnotType.e_Square
                    || mType == AnnotType.e_Circle || mType == AnnotType.e_Polygon || mType == AnnotType.e_Polyline
                    || mType == AnnotType.e_Underline || mType == AnnotType.e_Squiggly || mType == AnnotType.e_StrikeOut)
                {
                    AnnotBorderStyle bs = mAnnot.GetBorderStyle();
                    viewModel.Thickness = bs.width;
                }

                if (mAnnotIsTextMarkup || mType == AnnotType.e_Line || mType == AnnotType.e_Ink
                    || mType == AnnotType.e_Square || mType == AnnotType.e_Circle || mType == AnnotType.e_Polygon
                    || mType == AnnotType.e_Polyline || mType == AnnotType.e_FreeText)
                {
                    viewModel.Opacity = mMarkup.GetOpacity();
                }

                if (mType == AnnotType.e_FreeText)
                {
                    pdftron.PDF.Annots.FreeText ft = (FreeText)mAnnot;
                    viewModel.FontSize = ft.GetFontSize();

                    UtilityFunctions.FontBundle bundle = UtilityFunctions.GetFontFromAnnot(ft);
                    if (bundle == null)
                    {
                        bundle = UtilityFunctions.FontBundle.CreateNullFont();
                    }
                    viewModel.FontName = bundle;
                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                mPDFView.DocUnlockRead();
            }
        }

        private async void SetAnnotPropertiesFromDialog(AnnotStyleType annotStyleType, object val)
        {
            try
            {
                mPDFView.DocLock(true);
                switch (annotStyleType)
                {
                    case AnnotStyleType.FillColor:
                        if (HasSecondaryColor())
                        {
                            if (val is SettingsColor)
                            {
                                var fillColor = (SettingsColor)val;
                                SetSecondaryColor(fillColor);
                            }
                        }
                        break;
                    case AnnotStyleType.FontName:
                        if (mType == AnnotType.e_FreeText)
                        {
                            FontBundle fontBundle = val as FontBundle;

                            if (fontBundle == null)
                            {
                                return;
                            }

                            FreeText ft = (FreeText)mAnnot;
                            FontBundle bundleFromProperties = fontBundle;

                            if (bundleFromProperties != null && !string.IsNullOrEmpty(bundleFromProperties.DisplayName))
                            {
                                FontBundle currentFontBundle = GetFontFromAnnot(ft);
                                if (currentFontBundle == null || !currentFontBundle.Equals(bundleFromProperties))
                                {
                                    SetFontFromBundle(ft, mPDFView.GetDoc(), bundleFromProperties);
                                    GenerateDSEntry(ft, currentFontBundle);
                                    SetFreeTextDefaultAppearanceWithFont(ft);
                                }
                            }
                        }
                        break;
                    case AnnotStyleType.FontSize:
                        if (mType == AnnotType.e_FreeText)
                        {
                            FreeText ft = (FreeText)mAnnot;
                            var fontSize = System.Convert.ToDouble(val);
                            ft.SetFontSize(fontSize);
                        }
                        break;
                    case AnnotStyleType.Icon:
                        if (mType == AnnotType.e_Text)
                        {
                            string icon = val as string;
                            if (string.IsNullOrEmpty(icon))
                            {
                                return;
                            }

                            Text text = new Text(mAnnot.GetSDFObj());
                            text.SetIcon(icon);
                            mAnnot = text;
                            text.RefreshAppearance();
                            await StickyNoteCreate.RefreshCustomStickyNoteAppearance(mAnnot, mPDFView);
                            mPDFView.UpdateWithAnnot(text, mPDFView.GetCurrentPage());

                            mToolManager.RaiseAnnotationEditedEvent(mAnnot, mAnnotPageNum);
                            RaisePropertyChangedInHelperMode();
                            return;
                        }
                        break;
                    case AnnotStyleType.MainColor:
                        if (val is SettingsColor)
                        {
                            SettingsColor mainColor = (SettingsColor)val;
                            SetPrimaryColor(mainColor);

                            if (mType == AnnotType.e_FreeText)
                            {
                                if (mIsEditingText && !FormatRichText.IsEditTextBoxNull(mEditTextBox))
                                {
                                    FormatRichText.ColorSelectedText(mEditTextBox.Document.Selection, ConvertToColor(mainColor));
                                    mEditTextBox.Focus(FocusState.Pointer);
                                    return;
                                }

                                FreeText ft = (FreeText)mAnnot;

                                // If there is RichText, replace color
                                if (IsRichFreeText(ft))
                                {
                                    string textHTML = UtilityFunctions.GetRichFreeText(ft);

                                    if (!string.IsNullOrEmpty(textHTML))
                                    {
                                        HTMLUtilities htmlUtilities = new HTMLUtilities();
                                        string newRC = htmlUtilities.ReplacePrimaryColor(textHTML, ConvertToColor(mainColor));
                                        ft.GetSDFObj().PutString(KeyRichContent, newRC);
                                    }
                                }
                            }
                            else if (mType == AnnotType.e_Text)
                            {
                                pdftron.PDF.Annots.Text text = new Text(mAnnot.GetSDFObj());
                                text.RefreshAppearance();
                                await StickyNoteCreate.RefreshCustomStickyNoteAppearance(mAnnot, mPDFView);
                                mPDFView.UpdateWithAnnot(text, mPDFView.GetCurrentPage());

                                mToolManager.RaiseAnnotationEditedEvent(mAnnot, mAnnotPageNum);
                                RaisePropertyChangedInHelperMode();
                                return;
                            }
                        }
                        break;
                    case AnnotStyleType.Opacity:
                        if (mAnnotIsTextMarkup || mType == AnnotType.e_Line || mType == AnnotType.e_Ink
                            || mType == AnnotType.e_Square || mType == AnnotType.e_Circle || mType == AnnotType.e_Polygon
                            || mType == AnnotType.e_Polyline || mType == AnnotType.e_FreeText)
                        {
                            var opacity = System.Convert.ToDouble(val);
                            mMarkup.SetOpacity(opacity);
                        }
                        break;
                    case AnnotStyleType.Thickness:
                        if (mType == AnnotType.e_Line || mType == AnnotType.e_Ink || mType == AnnotType.e_Square
                            || mType == AnnotType.e_Circle || mType == AnnotType.e_Polygon || mType == AnnotType.e_Polyline
                            || mType == AnnotType.e_Underline || mType == AnnotType.e_Squiggly || mType == AnnotType.e_StrikeOut)
                        {
                            var thickness = System.Convert.ToDouble(val);
                            AnnotBorderStyle bs = mMarkup.GetBorderStyle();
                            bs.width = thickness;
                            mMarkup.SetBorderStyle(bs);
                        }
                        break;
                    case AnnotStyleType.FormatText:

                        string formatText = val as string;

                        if (string.IsNullOrEmpty(formatText))
                        {
                            return;
                        }

                        if (mIsEditingText && mType == AnnotType.e_FreeText && !FormatRichText.IsEditTextBoxNull(mEditTextBox)
                            && !FormatRichText.IsSelectionNullOrEmpty(mEditTextBox.Document.Selection))
                        {
                            string selection = mEditTextBox.Document.Selection.Text;
                            switch (formatText)
                            {
                                case "Bold":
                                    mEditTextBox.Focus(FocusState.Pointer);
                                    FormatRichText.BoldSelectedText(mEditTextBox.Document.Selection);                                  
                                    break;
                                case "Italic":
                                    mEditTextBox.Focus(FocusState.Pointer);
                                    FormatRichText.ItalicSelectedText(mEditTextBox.Document.Selection);    
                                    break;
                                case "Underline":
                                    mEditTextBox.Focus(FocusState.Pointer);
                                    FormatRichText.UnderlineSelectedText(mEditTextBox.Document.Selection);
                                    
                                    break;
                                case "Strikethrough":
                                    mEditTextBox.Focus(FocusState.Pointer);
                                    FormatRichText.StrikethroughSelectedText(mEditTextBox.Document.Selection);
                                    break;
                                default: break;
                            }
                            return;
                        }
                        break;
                    default: break;
                }
            }
            catch (Exception ex)
            {
                AnalyticsHandlerBase.CURRENT.SendEvent(AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                mPDFView.DocUnlock();
            }
            mAnnot.RefreshAppearance();
            mPDFView.UpdateWithAnnot(mAnnot, mAnnotPageNum);
            mToolManager.RaiseAnnotationEditedEvent(mAnnot, mAnnotPageNum);

            RaisePropertyChangedInHelperMode();
        }

        private SettingsColor GetPrimaryColor()
        {
            int r;
            int g;
            int b;
            if (mType == AnnotType.e_FreeText)
            {
                pdftron.PDF.Annots.FreeText ft = (FreeText)mAnnot;
                ColorPt textColor = ft.GetTextColor();
                r = (int)(textColor.Get(0) * 255 + 0.5);
                g = (int)(textColor.Get(1) * 255 + 0.5);
                b = (int)(textColor.Get(2) * 255 + 0.5);
                return new SettingsColor((byte)r, (byte)g, (byte)b, true);
            }

            ColorPt color = mAnnot.GetColorAsRGB();
            r = (int)(color.Get(0) * 255 + 0.5);
            g = (int)(color.Get(1) * 255 + 0.5);
            b = (int)(color.Get(2) * 255 + 0.5);
            return new SettingsColor((byte)r, (byte)g, (byte)b, true);
        }

        private void SetPrimaryColor(SettingsColor color)
        {
            ColorPt mainColor = new ColorPt(color.R / 255.0, color.G / 255.0, color.B / 255.0);
            if (mType == AnnotType.e_FreeText)
            {
                pdftron.PDF.Annots.FreeText ft = (FreeText)mAnnot;
                ft.SetTextColor(mainColor, 3);
            }
            else
            {
                mAnnot.SetColor(mainColor, 3);
            }
        }

        private bool HasSecondaryColor()
        {
            switch (mType)
            {
                case AnnotType.e_Square:
                case AnnotType.e_Circle:
                case AnnotType.e_Polygon:
                case AnnotType.e_FreeText:
                case AnnotType.e_Text:
                    return true;
            }
            return false;
        }

        private SettingsColor GetSecondaryColor()
        {
            int r;
            int g;
            int b;
            if (mType == AnnotType.e_FreeText)
            {
                pdftron.PDF.Annots.FreeText ft = (FreeText)mAnnot;
                if (ft.GetColorCompNum() == 0)
                {
                    return new SettingsColor(0, 0, 0, false);
                }
                ColorPt color = ft.GetColorAsRGB();
                r = (int)(color.Get(0) * 255 + 0.5);
                g = (int)(color.Get(1) * 255 + 0.5);
                b = (int)(color.Get(2) * 255 + 0.5);
                return new SettingsColor((byte)r, (byte)g, (byte)b, true);
            }
            if (mMarkup.GetInteriorColorCompNum() == 0)
            {
                return new SettingsColor(0, 0, 0, false);
            }
            ColorPt fillColor = mMarkup.GetInteriorColor();
            r = (int)(fillColor.Get(0) * 255 + 0.5);
            g = (int)(fillColor.Get(1) * 255 + 0.5);
            b = (int)(fillColor.Get(2) * 255 + 0.5);
            return new SettingsColor((byte)r, (byte)g, (byte)b, true);
        }

        private void SetSecondaryColor(SettingsColor color)
        {
            ColorPt fillColor = new ColorPt(color.R / 255.0, color.G / 255.0, color.B / 255.0);
            ColorPt emptyColor = new ColorPt(0, 0, 0, 0);
            if (mType == AnnotType.e_FreeText)
            {
                pdftron.PDF.Annots.FreeText ft = (FreeText)mAnnot;
                if (color.Use)
                {
                    ft.SetColor(fillColor, 3);
                }
                else
                {

                    mMarkup.SetColor(emptyColor, 0);
                }
            }
            else
            {
                if (color.Use)
                {
                    mMarkup.SetInteriorColor(fillColor, 3);
                }
                else
                {
                    mMarkup.SetInteriorColor(emptyColor, 0);
                }
            }
        }

        #endregion AnnotationProperties

        #endregion Handlers For Popup Menu

        #region Helper Functions

        [Obsolete("ShowMenu is deprecated, please use ShowContextMenu with ContextMenuControl instead.")]
        internal void ShowMenu(PopupCommandMenu m)
        {
            if (m != null)
            {
                var eventArgs = new ShowingPopupMenuEventArgs
                {
                    ToolType = ToolMode,
                    AnnotType = mAnnot.GetAnnotType(),
                    Annotation = mAnnot,
                    PageNumber = mPDFView.GetCurrentPage()
                };
                mToolManager.RaiseOnShowMenu(m, eventArgs);

                m.Show();
                DisableScrolling();
            }
        }

        [Obsolete("HideMenu is deprecated, please use HideContextMenu with ContextMenuControl instead.")]
        internal void HideMenu(PopupCommandMenu m)
        {
            if (m != null)
            {
                m.Hide();
            }
        }

        internal void ShowContextMenu(ContextMenuControl context)
        {
            if (context != null)
            {
                var eventArgs = new ShowingContextMenuEventArgs
                {
                    ToolType = ToolMode,
                    AnnotType = mAnnot.GetAnnotType(),
                    Annotation = mAnnot,
                    PageNumber = mPDFView.GetCurrentPage()
                };
                mToolManager.RaiseOnShowContextMenu(context, eventArgs);

                context.Show();
                DisableScrolling();
            }
        }

        /// <summary>
        /// Hide the Context Menu
        /// </summary>
        /// <param name="context"></param>
        /// <param name="shouldDispose">
        /// Set this to true if the context parameter should be unsubscribed from events and disposed of.
        /// If context will be set to null after calling this method, then shouldDispose should be set to true.
        /// </param>
        internal void HideContextMenu(ContextMenuControl context, bool shouldDispose = false)
        {
            if (context != null)
            {
                if(shouldDispose)
                    ResetContextMenuEvents(context);
                context.Hide(shouldDispose);
            }
        }

        [Obsolete("PositionMenu is deprecated, please use PositionContextMenu with ContextMenuControl instead.")]
        internal virtual void PositionMenu(PopupCommandMenu m)
        {
            if (m != null)
            {
                if (mHelperModePositionRect != null)
                {
                    m.TargetSquare(mHelperModePositionRect.x1, mHelperModePositionRect.y1, mHelperModePositionRect.x2, mHelperModePositionRect.y2);
                }
                else
                {
                    PDFDouble x1 = new PDFDouble(mLeft);
                    PDFDouble y1 = new PDFDouble(mTop);
                    PDFDouble x2 = new PDFDouble(mRight);
                    PDFDouble y2 = new PDFDouble(mBottom);
                    mPDFView.ConvAnnotationCanvasPtToScreenPt(x1, y1);
                    mPDFView.ConvAnnotationCanvasPtToScreenPt(x2, y2);

                    m.TargetSquare(x1.Value, y1.Value, x2.Value, y2.Value);
                }
            }
        }

        internal virtual void PositionContextMenu(ContextMenuControl context)
        {
            if (context != null)
            {
                if (mHelperModePositionRect != null)
                {
                    context.TargetSquare(mHelperModePositionRect.x1, mHelperModePositionRect.y1, mHelperModePositionRect.x2, mHelperModePositionRect.y2);
                }
                else
                {
                    PDFDouble x1 = new PDFDouble(mLeft);
                    PDFDouble y1 = new PDFDouble(mTop);
                    PDFDouble x2 = new PDFDouble(mRight);
                    PDFDouble y2 = new PDFDouble(mBottom);
                    mPDFView.ConvAnnotationCanvasPtToScreenPt(x1, y1);
                    mPDFView.ConvAnnotationCanvasPtToScreenPt(x2, y2);

                    context.TargetSquare(x1.Value, y1.Value, x2.Value, y2.Value);
                }
            }
        }

        internal virtual UIRect GetBoxPopupRect()
        {
            if (mHelperModePositionRect != null)
            {
                return new UIRect(mHelperModePositionRect.x1, mHelperModePositionRect.y1, mHelperModePositionRect.Width(), mHelperModePositionRect.Height());
            }

            PDFDouble x1 = new PDFDouble(mLeft);
            PDFDouble y1 = new PDFDouble(mTop);
            mPDFView.ConvAnnotationCanvasPtToScreenPt(x1, y1);

            return new UIRect(x1.Value, y1.Value, mRight - mLeft, mBottom - mTop);
        }

        internal void SetAnnotEditAsHelperTool(IAnnot annot, int pageNumber)
        {
            try
            {
                mPDFView.DocLockRead();
                mAnnot = annot;
                mAnnotPageNum = pageNumber;
                PopulateMenu();
                mViewerCanvas = mPDFView.GetAnnotationCanvas();
                mIsInHelperMode = true;
            }
            catch (Exception) { }
            finally
            {
                mPDFView.DocUnlockRead();
            }
        }


        internal void HelperModeShowMenu(PDFRect rect)
        {
            mHelperModePositionRect = rect;
            mContextMenu.TargetSquare(rect.x1, rect.y1, rect.x2, rect.y2);
            rect = ConvertFromScreenRectToCanvasRect(rect);
            mLeft = rect.x1;
            mTop = rect.y1;
            mRight = rect.x2;
            mBottom = rect.y2;
            ShowContextMenu(mContextMenu);
        }

        internal void HelperModeHideMenu()
        {
            if (mIsShowingCommandMenu)
            {
                HideContextMenu(mContextMenu);
            }
            if (mExtraContextMenu != null)
            {
                HideContextMenu(mExtraContextMenu);
            }
        }

        protected void RaisePropertyChangedInHelperMode()
        {
            if (mIsInHelperMode && AnnotPropertyChangedInHelperMode != null)
            {
                AnnotPropertyChangedInHelperMode();
            }
        }

        #endregion Helper Functions

        private async void UpdateAnnotSize()
        {
            Rect oldUpdateRect = null;
            try
            {
                mPDFView.DocLock(true);
                // Compute the new annotation position
                PDFDouble x1 = new PDFDouble(mCtrlPointLocations[e_ul].X);
                PDFDouble y1 = new PDFDouble(mCtrlPointLocations[e_ul].Y);
                PDFDouble x2 = new PDFDouble(mCtrlPointLocations[e_lr].X);
                PDFDouble y2 = new PDFDouble(mCtrlPointLocations[e_lr].Y);

                mPDFView.ConvAnnotationCanvasPtToScreenPt(x1, y1);
                mPDFView.ConvAnnotationCanvasPtToScreenPt(x2, y2);
                mPDFView.ConvScreenPtToPagePt(x1, y1, mAnnotPageNum);
                mPDFView.ConvScreenPtToPagePt(x2, y2, mAnnotPageNum);

                Rect newAnnotRect = new Rect(x1.Value, y1.Value, x2.Value, y2.Value);
                newAnnotRect.Normalize();

                Rect r = mAnnot.GetRect();
                PDFDouble rx1 = new PDFDouble(r.x1);
                PDFDouble ry1 = new PDFDouble(r.y1);
                PDFDouble rx2 = new PDFDouble(r.x2);
                PDFDouble ry2 = new PDFDouble(r.y2);

                mPDFView.ConvPagePtToScreenPt(rx1, ry1, mAnnotPageNum);
                mPDFView.ConvPagePtToScreenPt(rx2, ry2, mAnnotPageNum);

                oldUpdateRect = new Rect(rx1.Value, ry1.Value, rx2.Value, ry2.Value);
                oldUpdateRect.Normalize();

                if (mAnnot.GetAnnotType() != AnnotType.e_Stamp && mAnnot.GetAnnotType() != AnnotType.e_Text)
                {
                    mAnnot.RefreshAppearance();
                }

                mAnnot.Resize(newAnnotRect);

                if (mAnnot.GetAnnotType() == AnnotType.e_Text)
                {
                    await StickyNoteCreate.RefreshCustomStickyNoteAppearance(mAnnot, mPDFView);
                }
                else if (mAnnot.GetAnnotType() != AnnotType.e_Stamp)
                {
                    mAnnot.RefreshAppearance();
                }

                BuildAnnotBBox();
            }
            catch (Exception)
            {

            }
            finally
            {
                mPDFView.DocUnlock();
            }
            if (oldUpdateRect != null)
            {
                mPDFView.Update(oldUpdateRect);
            }
            mPDFView.UpdateWithAnnot(mAnnot, mAnnotPageNum);

            mToolManager.RaiseAnnotationEditedEvent(mAnnot, mAnnotPageNum);
        }

        /// <summary>
        /// TODO: unified all Delete Annotation into 1 method
        /// </summary>
        private void DeleteAnnotation()
        {
            mViewerCanvas.Children.Clear();
            try
            {
                mPDFView.DocLock(true);
                mToolManager.RaiseAnnotationPreRemovedEvent(mAnnot, mAnnotPageNum);

                // Perform the Annotation removal and update PDFViewer
                mPage = mPDFView.GetDoc().GetPage(mAnnotPageNum);
                mAnnot = AnnotUtils.SafeDeleteAnnotAndUpdate(mPDFView, mPage, mAnnot, mAnnotPageNum);

                mToolManager.RaiseAnnotationRemovedEvent(mAnnot, mAnnotPageNum);
            }
            catch (Exception e)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(e));
            }
            finally
            {
                mPDFView.DocUnlock();
                mAnnot = null;
            }
        }

        private void ResetContextMenuEvents(ContextMenuControl context)
        {
            if(context != null)
            {
                context.ItemClicked -= OnCommandMenuClicked;
                context.ItemClicked -= OnOpacitySelected;
                context.ItemClicked -= OnThicknessSelected;
                context.ItemClicked -= OnTypeSelected;
                context.ItemClicked -= OnFontSizeSelected;
            }
        }
    }
}
