﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace pdftron.PDF.Tools
{
    /*
     * https://www.pdftron.com/documentation/android/guides/annotation/sync/
     * NOTES: 1. Each annot has to have an unique ID
     *        2. On each modification setDateNow must be called for the change to be effective
     *        3. Core updates on seconds, so make sure to give 1 sec on every change
     */

    /// <summary>
    /// This class is responsible for generating local annotation changes XFDF string and
    /// merging remote annotation changes XFDF string
    /// </summary>
    public class AnnotManager
    {
        pdftron.PDF.ExternalAnnotManager mExternalAnnotManager;
        pdftron.PDF.PDFViewCtrl mPDFViewCtrl;
        ToolManager mToolManager;

        public enum AnnotationAction
        {
            ADD,
            MODIFY,
            DELETE,
            UNDO,
            REDO
        }

        /// <summary>
        /// Initializes annotation manager for annotation syncing
        /// </summary>
        /// <param name="toolManager"> ToolManager that handles all annotation events </param>
        /// <param name="user_id"> the unique identifier of the current user </param>
        public AnnotManager(ToolManager toolManager, string user_id)
        {
            mToolManager = toolManager;            
            mPDFViewCtrl = toolManager.GetPDFViewCtrl();
            mExternalAnnotManager = mPDFViewCtrl.EnableExternalAnnotManager(user_id);

            if (mExternalAnnotManager == null)
                throw new NullReferenceException("ExternalAnnotManagre can't be null");
        }

        #region Events

        public class AnnotLocalChangedEventArgs : EventArgs
        {
            public AnnotationAction Action { get; set; }
            public string XFDFCommand { get; set; }
            public string XFDFJSON { get; set; }
        }

        public class AnnotRemoteChangedEventArgs : EventArgs
        {
            public string XFDFIncoming { get; set; }
        }

        /// <summary>
        /// Delegate that handles Local annotation change
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"> Local change event args </param>
        public delegate void AnnotLocalChangedEventHandler(object sender, AnnotLocalChangedEventArgs e);

        /// <summary>
        /// Delegate that handles Remote annotation change
        /// </summary>
        /// <param name="sender"> Sender object </param>
        /// <param name="e"> Remote change event args </param>
        public delegate void AnnotRemoteChangedEventHandler(object sender, AnnotRemoteChangedEventArgs e);

        /// <summary>
        /// Raised when a local change is performed
        /// </summary>
        public event AnnotLocalChangedEventHandler LocalChanged;

        /// <summary>
        /// Used when annotation change event is received from remote service
        /// </summary>
        public event AnnotRemoteChangedEventHandler RemoteChanged;

        /// <summary>
        /// Used internally to raise local annotation change event
        /// </summary>
        /// <param name="action"> annotation action type </param>
        internal void RaiseLocalChange(AnnotationAction action)
        {
            var annotManagerArgs = new AnnotManager.AnnotLocalChangedEventArgs
            {
                Action = action,
                XFDFCommand = mExternalAnnotManager.GetLastXFDF(),
                XFDFJSON = mExternalAnnotManager.GetLastJSON()
            };

            LocalChanged?.Invoke(this, annotManagerArgs);
        }

        /// <summary>
        /// Used when annotation change event is received from remote service
        /// </summary>
        /// <param name="incoming"> incoming the XFDF command string </param>
        public void RaiseRemoteChange(string incoming)
        {
            bool shouldUnlock = false;

            try
            {
                shouldUnlock = true;
                mPDFViewCtrl.DocLock(true);
                mExternalAnnotManager.MergeXFDF(incoming);
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(
                    Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY,
                    Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                if (shouldUnlock)
                    mPDFViewCtrl.DocUnlock();
            }

            var eventArgs = new AnnotRemoteChangedEventArgs { XFDFIncoming = incoming };

            RemoteChanged?.Invoke(this, eventArgs);
        }

        #endregion
    }
}
