﻿using pdftron.Common;
using pdftron.PDF;
using pdftron.PDF.Tools;
using pdftron.PDF.Tools.Collections;
using pdftron.PDF.Tools.Utilities;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Diagnostics;
using System.Diagnostics.CodeAnalysis;
using System.IO;
using System.Linq;
using System.Text.RegularExpressions;
using System.Threading;
using System.Threading.Tasks;
using Windows.Foundation;
using Windows.Storage;
using Windows.Storage.AccessCache;

namespace pdftron.PDF.Tools.Collections
{
    /// <summary>
    /// This class represents a collection of recent items the user has accessed.
    /// </summary>
    public sealed class RecentItemsData : INotifyPropertyChanged
    {
        private static readonly RecentItemsData _instance = new RecentItemsData();
        private SemaphoreSlim _loadingSemaphore;

        static RecentItemsData()
        {
        }

        [SuppressMessage("Simplification", "RCS1085:Use auto-implemented property.", Justification = "<Pending>")]
        public static RecentItemsData Instance => _instance;

        public event PropertyChangedEventHandler PropertyChanged;

        private void RaisePropertyChanged(string property) => PropertyChanged?.Invoke(this, new PropertyChangedEventArgs(property));

        public ObservableCollection<RecentItem> RecentFiles { get; private set; }

        public ObservableCollection<RecentItem> RecentFolders { get; private set; }

        public static bool IsReady { get; private set; }

        private SemaphoreSlim LoadingSemaphore => _loadingSemaphore ?? (_loadingSemaphore = new SemaphoreSlim(1));

        /// <summary>
        /// Creates a new instance of RecentItemsData.
        /// This function will return immediately, so there is no guarantee that the list of recent items
        /// has been populated.
        /// </summary>
        /// <param name="parent">The owner of the recent items list. Used by this class to invoke the dispatcher</param>
        private RecentItemsData()
        {
        }

        /// <summary>
        /// This functions will always wait until the RecentItems list is populated, and will then return an instance of RecentItemsData.
        /// This should always be the way to get an instance of this class.
        /// </summary>
        /// <param name="parent">The owner of the recent items list. Used by this class to invoke the dispatcher</param>
        /// <returns>An instance of RecentItem guaranteed to have the list of recent items populated.</returns>
        public async Task GetItemSourceAsync()
        {
            await LoadingSemaphore.WaitAsync().ConfigureAwait(false);
            try
            {
                if (!IsReady && DocumentManager.IsReady)
                {
                    if (RecentFiles == null)
                    {
                        DocumentManager docManager = DocumentManager.Instance;
                        RecentFiles = await _instance.GetRecentDocumentsAsync(docManager, StorageItemTypes.File);
                        RecentFolders = await _instance.GetRecentDocumentsAsync(docManager, StorageItemTypes.Folder);
                    }

                    IsReady = true;
                }
            }
            catch (Exception e)
            {
                Debug.WriteLine("Error: " + e.ToString());
            }
            finally
            {
                LoadingSemaphore.Release();
            }
        }

        private async Task<ObservableCollection<RecentItem>> GetRecentDocumentsAsync(DocumentManager docManager, StorageItemTypes type)
        {
            try
            {
                ObservableCollection<RecentItem> recentItemsList = new ObservableCollection<RecentItem>();

                foreach (AccessListEntry entry in StorageApplicationPermissions.MostRecentlyUsedList.Entries)
                {
                    RecentItem item = null;
                    try
                    {
                        RecentDocumentProperties props = RecentDocumentProperties.FromString(entry.Metadata);

                        // If we don't have a file vs folder as we expect, then move on
                        if ((type == StorageItemTypes.Folder && !props.IsFolder)
                            || (type == StorageItemTypes.File && props.IsFolder))
                        {
                            continue;
                        }

                        item = new RecentItem(props, entry.Token);

                        string tempFilePath = null;
                        if (!string.IsNullOrEmpty(item.DocumentPath))
                        {
                            tempFilePath = await docManager.GetRecentThumbnailPathAsync(item.DocumentPath);
                        }

                        if (!string.IsNullOrEmpty(tempFilePath))
                        {
                            // This means we're using a temporary file.
                            item.ThumbnailLocation = RecentlyUsedCache.GetBitmapPathIfExists(tempFilePath);
                        }

                        if (type == StorageItemTypes.Folder)
                            item.Properties.ThumbnailLocation = "ms-appx:///Assets/DocumentPage/FolderIcon.png";

                        recentItemsList.Add(item);
                    }
                    catch (Exception)
                    {
                        Debug.WriteLine("Failed to load entry from XML string");
                    } // likely because it's an old entry that hasn't been updated yet

                    if (item == null)
                    {
                        try
                        {
                            string thumbPath = string.Empty;
                            string[] filePathAndFlags = entry.Metadata.Split('/');
                            if (filePathAndFlags.Length > 0)
                            {
                                string tempFilePath = await docManager.GetRecentThumbnailPathAsync(filePathAndFlags[0]);
                                if (!string.IsNullOrEmpty(tempFilePath))
                                {
                                    // This means we're using a temporary file.
                                    thumbPath = RecentlyUsedCache.GetBitmapPathIfExists(tempFilePath);
                                }
                                else
                                {
                                    // this means we're using a
                                    thumbPath = RecentlyUsedCache.GetBitmapPathIfExists(filePathAndFlags[0]);
                                }
                            }

                            bool isProtected = FindOptionInFlags("L", filePathAndFlags);
                            if (type == StorageItemTypes.Folder)
                            {
                                thumbPath = "ms-appx:///Assets/DocumentPage/FolderIcon.png";
                            }
                            else if (isProtected)
                            {
                                thumbPath = "ms-appx:///Assets/DocumentPage/FilePlaceHolder_pdf_locked.png";
                            }
                            else if (string.IsNullOrEmpty(thumbPath))
                            {
                                thumbPath = "ms-appx:///Assets/DocumentPage/FilePlaceHolder_pdf.png";
                            }

                            item = new RecentItem(Path.GetFileName(filePathAndFlags[0]), thumbPath, entry.Token, filePathAndFlags[0], isProtected)
                            {
                                LastPageNumber = GetPageNumberFromMetadataString(entry.Metadata)
                            };

                            recentItemsList.Add(item);
                        }
                        catch
                        {
                            //Swallow exception
                        }
                    }
                }

                return recentItemsList;
            }
            catch
            {
                //Swallow exception
            }

            return new ObservableCollection<RecentItem>();
        }

        /// <summary>
        /// Retrieves the StorageFile associated with the item
        /// </summary>
        /// <param name="item"></param>
        /// <returns></returns>
        public async Task<RecentDocumentProperties> GetRecentFileAsync(RecentItem item)
        {
            await LoadingSemaphore.WaitAsync(1);

            try
            {
                IStorageItem file = null;
                if (item.Properties.IsFolder)
                {
                    file = await StorageApplicationPermissions.MostRecentlyUsedList.GetFolderAsync(item.Token);
                }
                else
                {
                    file = await StorageApplicationPermissions.MostRecentlyUsedList.GetFileAsync(item.Token, AccessCacheOptions.SuppressAccessTimeUpdate);
                }

                item.Properties.File = file;
                return item.Properties;
            }
            finally
            {
                LoadingSemaphore.Release();
            }
        }

        /// <summary>
        /// Checks to see if there are any new thumbnails available.
        /// </summary>
        public async Task UpdateThumbnailLocations()
        {
            foreach (RecentItem item in RecentFiles)
            {
                string tempFilePath = null;
                if (!string.IsNullOrEmpty(item.DocumentPath))
                    tempFilePath = await DocumentManager.Instance.GetRecentThumbnailPathAsync(item.DocumentPath);

                if (!string.IsNullOrEmpty(tempFilePath))
                {
                    // This means we're using a temporary file.
                    item.ThumbnailLocation = RecentlyUsedCache.GetBitmapPathIfExists(tempFilePath);
                }
            }
        }

        /// <summary>
        /// Updates the recent list with a file that was selected from the recent list
        /// This means the only work that has to be done is to move it to the top of the list
        /// </summary>
        /// <param name="file"></param>
        /// <param name="isProtected"></param>
        /// <returns>The token for the file</returns>
        public string UpdateWithRecentFile(IStorageItem file, bool isProtected = false, string recentFileToken = null)
        {
            bool isFolder = file.IsOfType(StorageItemTypes.Folder);
            RecentDocumentProperties props = new RecentDocumentProperties(null, 1, isProtected, isFolder)
            {
                FilePath = file.Path
            };

            if (UtilityFunctions.IsOneDrive(props.FilePath))
            {
                props.AdditonalIcon = AdditionalIcons.OneDrive;
            }
            else
            {
                props.AdditonalIcon = AdditionalIcons.None;
            }

            string token;
            string metaData = props.MakeString();

            if (string.IsNullOrEmpty(recentFileToken) == false
                && StorageApplicationPermissions.MostRecentlyUsedList.ContainsItem(recentFileToken))
            {
                token = recentFileToken;
            }
            else
            {
                token = StorageApplicationPermissions.MostRecentlyUsedList.Add(file, metaData);
            }    

            int listIndex = FindTokenInList(token);
            if (listIndex >= 0 && listIndex < RecentFiles.Count)
            {
                if (RecentFiles[listIndex].LastPageNumber > 0)
                {
                    props = RecentFiles[listIndex].Properties;
                    metaData = props.MakeString();
                    StorageApplicationPermissions.MostRecentlyUsedList.AddOrReplace(token, file, metaData);
                }

                if (listIndex < RecentFiles.Count)
                    RecentFiles.RemoveAt(listIndex);
            }

            RecentItem item = new RecentItem(props, token);

            if (file.IsOfType(StorageItemTypes.File))
            {
                RecentFiles.Insert(0, item);
            }
            else
            {
                RecentFolders.Insert(0, item);
            }

            return token;
        }

        /// <summary>
        /// Updates the recent list with a file picked from somewhere other than the recent list
        /// </summary>
        /// <param name="file"></param>
        /// <param name="isProtected"></param>
        /// <returns>The token for the file</returns>
        public string UpdateWithNewFile(IStorageItem file, bool isProtected = false, string token = null)
        { 
            return UpdateWithRecentFile(file, isProtected, token); 
        }

        /// <summary>
        /// Creates a recent item from a storage item without adding it to the future access list
        /// </summary>
        /// <param name="file"></param>
        /// <returns>The recent item created</returns>
        public static RecentItem CreateNewPinnedItem(IStorageItem file)
        {
            bool isFolder = file.IsOfType(StorageItemTypes.Folder);
            RecentDocumentProperties props = new RecentDocumentProperties(null, 1, false, isFolder)
            {
                FilePath = file.Path,
                File = file
            };

            if (UtilityFunctions.IsOneDrive(props.FilePath))
            {
                props.AdditonalIcon = AdditionalIcons.OneDrive;
            }
            else
            {
                props.AdditonalIcon = AdditionalIcons.None;
            }

            return new RecentItem(props, string.Empty);
        }

        /// <summary>
        /// Removes all items from recentItems from the list of recent items
        /// </summary>
        /// <param name="recentItems"></param>
        public void RemoveItems(IList<RecentItem> recentItems)
        {
            foreach (RecentItem recentItem in recentItems)
            {
                try
                {
                    StorageApplicationPermissions.MostRecentlyUsedList.Remove(recentItem.Token);
                    RecentFiles.Remove(recentItem);
                    RecentlyUsedCache.RemoveDocument(recentItem.DocumentPath);
                }
                catch
                {
                    //Swallow exception
                }
            }
        }

        /// <summary>
        /// Will remove the first recent item which references the storage file if there is any.
        /// </summary>
        /// <param name="file"></param>
        public async Task RemoveItemAsync(StorageFile file)
        {
            await LoadingSemaphore.WaitAsync(1);
            try
            {
                foreach (RecentItem item in RecentFiles)
                {
                    try
                    {
                        StorageFile fileToCheck = await StorageApplicationPermissions.MostRecentlyUsedList.GetFileAsync(item.Token);
                        if (fileToCheck != null && UtilityFunctions.AreFilesEqual(fileToCheck, file))
                        {
                            StorageApplicationPermissions.MostRecentlyUsedList.Remove(item.Token);
                            RecentFiles.Remove(item);
                            RecentlyUsedCache.RemoveDocument(item.DocumentPath);

                            RaisePropertyChanged("RecentItems");

                            return;
                        }
                    }
                    catch
                    {
                        //Swallow exception
                    }
                }
            }
            finally
            {
                LoadingSemaphore.Release();
            }
        }

        /// <summary>
        /// Removes the last entered item in the collection.
        /// </summary>
        public void RemoveLastItem()
        {
            if (RecentFiles.Count > 0)
            {
                StorageApplicationPermissions.MostRecentlyUsedList.Remove(RecentFiles[0].Token);
                RecentFiles.RemoveAt(0);
            }
        }

        /// <summary>
        /// Clears the recent list
        /// </summary>
        public void ClearList()
        {
            // clear the cache
            foreach (RecentItem recentItem in RecentFiles)
            {
                RecentlyUsedCache.RemoveDocument(recentItem.DocumentPath);
            }

            StorageApplicationPermissions.MostRecentlyUsedList.Clear();
            RecentFiles.Clear();
        }

        /// <summary>
        /// Clears all StorageFiles in the recent list
        /// </summary>
        public void ClearFilesList()
        {
            // clear the cache
            foreach (RecentItem recentItem in RecentFiles)
            {
                if (!recentItem.Properties.IsFolder)
                {
                    RecentlyUsedCache.RemoveDocument(recentItem.DocumentPath);
                    // got a crash where MRU was removing a token in RecentFiles that it didn't contain. Protecting against it for now.
                    if (StorageApplicationPermissions.MostRecentlyUsedList.ContainsItem(recentItem.Token))
                        StorageApplicationPermissions.MostRecentlyUsedList.Remove(recentItem.Token);
                }
            }

            RecentFiles.Clear();
        }

        public async Task UpdatePageAsync(string token, int pageNumber)
        {
            await LoadingSemaphore.WaitAsync(1);

            try
            {
                if (StorageApplicationPermissions.MostRecentlyUsedList.Entries.Count == 0)
                    return;

                AccessListEntry entry = StorageApplicationPermissions.MostRecentlyUsedList.Entries[0];
                if (!entry.Token.Equals(token))
                    return;

                string metadata = entry.Metadata;

                Regex pageNumberRegex = new Regex(@"/p\d+");
                string pageNumberString = "/p" + pageNumber;

                Match match = pageNumberRegex.Match(metadata);
                if (match.Success)
                {
                    metadata = pageNumberRegex.Replace(metadata, pageNumberString);
                }
                else
                {
                    metadata += pageNumberString;
                }

                entry.Metadata = metadata;
                StorageFile fileToUpdate = await StorageApplicationPermissions.MostRecentlyUsedList.GetFileAsync(token);
                _ = StorageApplicationPermissions.MostRecentlyUsedList.Add(fileToUpdate, metadata);
            }
            finally
            {
                LoadingSemaphore.Release();
            }
        }

        public async Task UpdatePropertiesAsync(string token, PDFViewCtrl ctrl, bool isReflow)
        {
            await LoadingSemaphore.WaitAsync();
            try
            {
                if (StorageApplicationPermissions.MostRecentlyUsedList.Entries.Count == 0)
                    return;

                AccessListEntry entry = StorageApplicationPermissions.MostRecentlyUsedList.Entries[0];
                if (!entry.Token.Equals(token))
                    return;

                string metadata = entry.Metadata;
                RecentDocumentProperties properties = RecentDocumentProperties.FromString(metadata);
                properties.PagePresentationMode = ctrl.GetPagePresentationMode();
                properties.PageRotation = ctrl.GetRotation();
                properties.PageNumber = ctrl.GetCurrentPage();
                properties.Zoom = ctrl.GetZoom();
                properties.HorizontalScrollPos = (int)ctrl.GetHScrollPos();
                properties.VerticalScrollPos = (int)ctrl.GetVScrollPos();
                properties.IsInReflowMode = isReflow;

                metadata = properties.MakeString();

                entry.Metadata = metadata;
                StorageFile fileToUpdate = await StorageApplicationPermissions.MostRecentlyUsedList.GetFileAsync(token);
                _ = StorageApplicationPermissions.MostRecentlyUsedList.Add(fileToUpdate, metadata);

                if (!string.IsNullOrWhiteSpace(token))
                {
                    foreach (RecentItem item in RecentFiles)
                    {
                        if (token.Equals(item.Token))
                        {
                            item.Properties = properties;
                        }
                    }
                }
            }
            finally
            {
                LoadingSemaphore.Release();
            }
        }

        public int GetPageNumberFromToken(string token)
        {
            if (StorageApplicationPermissions.MostRecentlyUsedList.Entries.Count == 0)
                return -1;

            AccessListEntry entry = StorageApplicationPermissions.MostRecentlyUsedList.Entries[0];
            if (!entry.Token.Equals(token))
                return -1;

            return GetPageNumberFromMetadataString(entry.Metadata);
        }

        public async Task<RecentDocumentProperties> GetPropertiesForFileIfInList(StorageFile file)
        {
            await LoadingSemaphore.WaitAsync(1);
            try
            {
                foreach (RecentItem item in RecentFiles.ToList())
                {
                    try
                    {
                        StorageFile fileToCheck = await StorageApplicationPermissions.MostRecentlyUsedList.GetFileAsync(item.Token);
                        if (fileToCheck != null && UtilityFunctions.AreFilesEqual(fileToCheck, file))
                            return item.Properties;
                    }
                    catch
                    {
                        //Swallow exception
                    }
                }
            }
            finally
            {
                LoadingSemaphore.Release();
            }

            return null;
        }

        /// <summary>
        /// Returns the index of the list item which contains token
        /// </summary>
        /// <param name="token"></param>
        /// <returns></returns>
        private int FindTokenInList(string token)
        {
            if (RecentFiles == null)
                return -1;

            int listSize = RecentFiles.Count;
            for (int i = 0; i < listSize; i++)
            {
                RecentItem item = RecentFiles[i];
                if (item.Token.Equals(token, StringComparison.Ordinal))
                    return i;
            }

            return -1;
        }

        private bool FindOptionInFlags(string option, string[] flags)
        {
            int numOptions = flags.Length;
            for (int i = 1; i < numOptions; i++)
            {
                if (flags[i].Equals(option, StringComparison.OrdinalIgnoreCase))
                    return true;
            }

            return false;
        }

        private int GetPageNumberFromMetadataString(string metadata)
        {
            int result = -1;

            try
            {
                Regex pageNumberRegex = new Regex(@"/p\d+");
                Match match = pageNumberRegex.Match(metadata);

                if (match.Success)
                {
                    Regex justNMumberRegex = new Regex(@"\d+");
                    Match pgMatch = justNMumberRegex.Match(match.Groups[0].ToString());
                    if (pgMatch.Success)
                    {
                        int.TryParse(pgMatch.Groups[0].ToString(), out result);
                    }
                }
            }
            catch
            {
                //Swallow exception
            }

            return result;
        }
    }
}
