﻿using Microsoft.UI.Xaml.Controls;
using pdftron.PDF.Tools.Controls.Resources;
using pdftron.PDF.Tools.Controls.Viewer;
using pdftron.PDF.Tools.Controls.ViewModels;
using System;
using System.Collections.Generic;
using Windows.Foundation;
using Windows.Foundation.Metadata;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Controls.Primitives;
using Windows.UI.Xaml.Data;
using Windows.UI.Xaml.Media;

// The User Control item template is documented at http://go.microsoft.com/fwlink/?LinkId=234236

namespace pdftron.PDF.Tools.Controls
{
    /// <summary>
    /// The AnnotationToolbar works with a ToolManager to allow quick selection of different tools.
    /// The toolbar shows a list of buttons which prompts the associated ToolManager to switch to that tool.
    /// </summary>
    public sealed partial class AnnotationCommandBar : AnnotationToolbarBase
    {
        private const int  DEFAULT_SECONDARY_BUTTON_WIDTH = 120;
        private const int DEFAULT_SECONDARY_BUTTON_HEIGHT = 40;

        /// <summary>
        /// It holds the list of buttons that will be displayed on the OverflowFlyoutPanel by clicking
        /// on the extra / extend button (ellipsis / See more button)
        /// </summary>
        private Dictionary<string, Tuple<AppBarTabButton, bool>> _SecondaryButtons = new Dictionary<string, Tuple<AppBarTabButton, bool>>();

        // Notifies to ViewerPage to expand tab command bar if tools command bar expands
        public delegate void AppBarEventHandler(bool isOpen);
        public event AppBarEventHandler AppBarChanged;

        /// <summary>
        /// Allows user to externally control whether the app bar is expanded or compact (since view model is not exposed)
        /// </summary>
        static DependencyProperty IsAppBarOpenProperty =
          DependencyProperty.Register("IsAppBarOpen", typeof(bool),
          typeof(AnnotationCommandBar), new PropertyMetadata(false));

        public bool IsAppBarOpen
        {
            get
            {
                return (bool)(GetValue(IsAppBarOpenProperty));
            }
            set
            {
                SetValue(IsAppBarOpenProperty, value);
                ViewModel.IsMainAppBarOpen = value;
            }
        }

        public bool IsInkOpen
        {
            get { return ViewModel.ShowInkToolbar; }
        }

        public bool IsPolygonSaveOpen
        {
            get { return ViewModel.ShowPolygonSaveToolbar; }
        }

        public bool IsMeasureOpen
        {
            get { return ViewModel.ShowMeasureToolbar; }
        }

        /// <summary>
        /// Creates a new AnnotationToolbar that is not associated with a ToolManager.
        /// Use the ToolManager property to make the AnnotationToolbar Work with a ToolManager.
        /// </summary>
        public AnnotationCommandBar() : base(null)
        {
            InitializeComponent();
            Init();
        }

        /// <summary>
        /// Creates a new AnnotationToolbar that is not associated with a ToolManager.
        /// Use the ToolManager property to make the AnnotationToolbar Work with a ToolManager.
        /// </summary>
        /// <param name="viewModel">The annotation commmand bar view model</param>
        public AnnotationCommandBar(AnnotationToolbarViewModel viewModel) : base(viewModel)
        {
            InitializeComponent();
            Init();
        }

        /// <summary>
        /// Creates a new AnnotationToolbar that works with the ToolManager.
        /// </summary>
        /// <param name="ctrl"></param>
        /// <param name="toolManager"></param>
        public AnnotationCommandBar(ToolManager toolManager) : base(null)
        {
            InitializeComponent();
            ToolManager = toolManager;
            Init();
        }

        private void Init()
        {
            MINIMUM_BUTTON_WIDTH = (double)this.Resources["AnnotationButtonWidth"];
            MINIMUM_BUTTON_WIDTH_SMALL_SCREEN = (double)this.Resources["AnnotationButtonSmallScreenWidth"];
            ExtraPaddingForInkToolbar = (double)this.Resources["InkToolBarButtonLeftPadding"];

            this.MainAppBar.Opening += MainAppBar_Opening;
            this.MainAppBar.Closing += MainAppBar_Closing;

            ToolbarButtonRanking = new ButtonBase[ToolbarButtonMainStack.Children.Count];

            int index = 0;

            // Highest priority: Will always be displayed
            // INVESTIGATE: it seems like the order is being changed when toolbar buttons go to the seconday menu
            ToolbarButtonRanking[index++] = ToolbarCloseButton;
            ToolbarButtonRanking[index++] = ExpandButton;
            ToolbarButtonRanking[index++] = UndoButton;
            ToolbarButtonRanking[index++] = RedoButton;
            ToolbarButtonRanking[index++] = AnnotPanButton;
            ToolbarButtonRanking[index++] = AnnotStickyNoteButton;
            ToolbarButtonRanking[index++] = AnnotHighlightButton;
            ToolbarButtonRanking[index++] = AnnotFreeHandButton;
            ToolbarButtonRanking[index++] = AnnotFreeTextButton;
            ToolbarButtonRanking[index++] = AnnotEraserButton;
            ToolbarButtonRanking[index++] = AnnotSignatureButton;
            ToolbarButtonRanking[index++] = AnnotStrikeoutButton;
            ToolbarButtonRanking[index++] = AnnotUnderlineButton;
            ToolbarButtonRanking[index++] = AnnotMarkerButton;
            ToolbarButtonRanking[index++] = AnnotSquigglyButton;
            ToolbarButtonRanking[index++] = AnnotStamperButton;
            ToolbarButtonRanking[index++] = AnnotSquareButton;
            ToolbarButtonRanking[index++] = AnnotCircleButton;
            ToolbarButtonRanking[index++] = AnnotLineButton;
            ToolbarButtonRanking[index++] = AnnotArrowButton;
            ToolbarButtonRanking[index++] = AnnotMeasureButton;
            ToolbarButtonRanking[index++] = AnnotPolygonButton;
            ToolbarButtonRanking[index++] = AnnotLassoCreateButton;
            ToolbarButtonRanking[index++] = AnnotGroupCreateButton;
            ToolbarButtonRanking[index++] = AnnotZoomButton;
            ToolbarButtonRanking[index++] = AnnotRedactionButton;

            ResizableInkButtons = new List<ButtonBase>();
            ResizableInkButtons.Add(CancelButton);
            ResizableInkButtons.Add(InkButton0);
            ResizableInkButtons.Add(InkButton1);
            ResizableInkButtons.Add(InkButton2);
            ResizableInkButtons.Add(InkEraserButton);
            ResizableInkButtons.Add(InkEraserButtonMultiStroke);
            ResizableInkButtons.Add(InkUndoButton);
            ResizableInkButtons.Add(InkRedoButton);
            ResizableInkButtons.Add(InkSaveButton);

            ResizablePolygonSaveButtons = new List<ButtonBase>();
            ResizablePolygonSaveButtons.Add(ClosePolygonSaveButton);
            ResizablePolygonSaveButtons.Add(AnnotPolylineButton);
            ResizablePolygonSaveButtons.Add(AnnotPolygonButton2);
            ResizablePolygonSaveButtons.Add(AnnotCloudButton);

            bool isPhone = ApiInformation.IsTypePresent("Windows.Phone.UI.Input.HardwareButtons");
            if (isPhone)
            {
                SmallScreenLayout = true;
                InkButtonGrid1.Visibility = Visibility.Collapsed;
                InkButtonGrid2.Visibility = Visibility.Collapsed;
            }
            else
            {
                SmallScreenLayout = false;
            }

            var dataContext = DataContext as AnnotationToolbarViewModel;
            dataContext.ToolSelected += DataContext_ToolSelected;

            if (Resources.ContainsKey("AppBarThemeCompactHeight"))
            {
                dataContext.AppBarCompactHeight = (double)Resources["AppBarThemeCompactHeight"];
            }

            ContentPanel.SizeChanged += ContentPanel_SizeChanged;
            InkToolbar.SizeChanged += InkToolbar_SizeChanged;
            PolygonSaveToolbar.SizeChanged += PolygonSaveToolbar_SizeChanged;
        }

        private void PolygonSaveToolbar_SizeChanged(object sender, SizeChangedEventArgs e)
        {
            if (!double.IsNaN(e.NewSize.Width))
            {
                Size newSize = e.NewSize;
                newSize.Width = Math.Max(0, newSize.Width - (double)Resources["AppBarExpandButtonThemeWidth"]);
                ResizeButtons(newSize);
            }
        }

        private void ContentPanel_SizeChanged(object sender, SizeChangedEventArgs e)
        {
            ResizeButtons(e.NewSize);
        }
        private void InkToolbar_SizeChanged(object sender, SizeChangedEventArgs e)
        {
            if (!double.IsNaN(e.NewSize.Width))
            {
                Size newSize = e.NewSize;

                if (IsExpandButtonVisible)
                    newSize.Width = Math.Max(0, newSize.Width - (double)Resources["AppBarExpandButtonThemeWidth"]);

                ResizeButtons(newSize);
            }
        }

        protected override void AnnotationCommandBar_SizeChanged(object sender, SizeChangedEventArgs e)
        {
            // This disables the base classes SizeChangedEvent
        }

        /// <summary>
        /// ResizeButtons will take care of transfering the tool bar buttons from the primary tool bar
        /// to the secondary tool bar based on the current visibility state of the tool buttons
        /// </summary>
        /// <param name="newSize"></param>
        protected override void ResizeButtons(Size newSize)
        {
            RectangleGeometry clipRect = new RectangleGeometry();
            clipRect.Rect = new Windows.Foundation.Rect(0, 0, newSize.Width, newSize.Height);
            ContentPanel.Clip = clipRect;

            base.ResizeButtons(newSize);

#if SUBSCRIPTIONS
            ViewModel.IsRedactionTipVisible = AnnotRedactionButton.Visibility != Visibility.Collapsed;
#endif

            foreach (var child in ToolbarButtonMainStack.Children)
            {
                ButtonBase childButton = child as ButtonBase;
                if (childButton == null)
                    continue;

                TextBlock text = (childButton.Content as StackPanel)?.Children[1] as TextBlock;
                if (text != null)
                {
                    // NOTE: Magic Number -> Manually changes the font size accordingly to the Screen size
                    text.FontSize = SmallScreenLayout ? 10 : 12;
                }

                if (childButton.Tag == null)
                    return;

                string tag = childButton.Tag.ToString();
                bool isInSecondaryPanel = _SecondaryButtons.ContainsKey(tag) && _SecondaryButtons[tag].Item2;

                // Check if button is collapsed in toolbar, and add it to secondary commands if its not already in there
                if (childButton.Visibility == Visibility.Collapsed)
                {
                    if (!IsButtonHidden(childButton))
                    {
                        if (isInSecondaryPanel)
                            continue;

                        AppBarTabButton secondaryButton = _SecondaryButtons.ContainsKey(tag) ? _SecondaryButtons[tag].Item1 : CreateSecondaryTabButton(childButton as AppBarTabButton);
                        _SecondaryButtons[tag] = new Tuple<AppBarTabButton, bool>(secondaryButton, true);

                        OverflowFlyoutPanel.Children.Add(secondaryButton);
#if SUBSCRIPTIONS
                        if (childButton.Tag.Equals("Redaction"))
                        {
                            if (!ViewModel.Subscribed)
                            {
                                secondaryButton.Flyout = (Flyout)Application.Current.Resources["RedactionUpgradeFlyout"];
                                secondaryButton.Flyout.Placement = FlyoutPlacementMode.LeftEdgeAlignedTop;
                            }
                            else
                            {
                                secondaryButton.Flyout = null;
                            }
                        }
#endif
                    }
                }
                // Check if button is readded to toolbar and remove from secondary commands if its in there
                else if (isInSecondaryPanel)
                {
                    OverflowFlyoutPanel.Children.Remove(_SecondaryButtons[tag].Item1);
                    _SecondaryButtons[tag] = new Tuple<AppBarTabButton, bool>(_SecondaryButtons[tag].Item1, false);
#if SUBSCRIPTIONS
                    if (tag.Equals("Redaction") && !ViewModel.Subscribed)
                    {
                        _SecondaryButtons[tag].Item1.Flyout = (Flyout)Application.Current.Resources["RedactionUpgradeFlyout"];
                        _SecondaryButtons[tag].Item1.Flyout.Placement = FlyoutPlacementMode.Bottom;
                    }
                    else
                    {
                        _SecondaryButtons[tag].Item1.Flyout = null;
                    }
#endif
                }
            }
        }

        /// <summary>
        /// It will copy all data from childButton and create a secondary tab button that will be
        /// used on the OverFlowLyoutPanel
        /// </summary>
        /// <param name="childButton"> Button from the Primary tab button bar </param>
        /// <returns></returns>
        private AppBarTabButton CreateSecondaryTabButton(AppBarTabButton childButton)
        {
            AppBarTabButton secondaryButton = new AppBarTabButton();

            secondaryButton.Tag = childButton.Tag;
            secondaryButton.Command = childButton.Command;
            secondaryButton.CommandParameter = childButton.CommandParameter;
            secondaryButton.SetBinding(StyleProperty, childButton.GetBindingExpression(StyleProperty).ParentBinding);

            // NOTE: Make sure to bind Enabled/Disabled when creating new Button
            var bindIsEnableDP = childButton.GetBindingExpression(ButtonBase.IsEnabledProperty);
            if (bindIsEnableDP != null)
                secondaryButton.SetBinding(AppBarButton.IsEnabledProperty, new Binding { Source = bindIsEnableDP.ParentBinding.Source, Path = bindIsEnableDP.ParentBinding.Path });

            secondaryButton.Width = DEFAULT_SECONDARY_BUTTON_WIDTH;
            secondaryButton.Height = DEFAULT_SECONDARY_BUTTON_HEIGHT;
            secondaryButton.HorizontalContentAlignment = HorizontalAlignment.Stretch;
            secondaryButton.VerticalContentAlignment = VerticalAlignment.Center;

            StackPanel stackPanel = new StackPanel();
            stackPanel.Margin = new Thickness(10, -10, 10, 0);
            stackPanel.Orientation = Orientation.Horizontal;
            stackPanel.HorizontalAlignment = HorizontalAlignment.Stretch;
            stackPanel.VerticalAlignment = VerticalAlignment.Stretch;

            if (childButton.Content is FontIcon)
            {
                FontIcon fontIcon = childButton.Content as FontIcon;
                FontIcon newIcon = new FontIcon();
                newIcon.FontFamily = fontIcon.FontFamily;
                newIcon.Glyph = fontIcon.Glyph;
                newIcon.VerticalAlignment = VerticalAlignment.Stretch;
                newIcon.Height = DEFAULT_SECONDARY_BUTTON_HEIGHT;
                stackPanel.Children.Add(newIcon);
            }
            else if (childButton.Content is Grid)
            {
                /*TODO: Right now it does not copy the ForeGRound binding to display
                 * the selected color when the button is on the secondary menu
                 */

                Grid grid = childButton.Content as Grid;
                Grid newGrid = new Grid();

                foreach (UIElement uIElement in grid.Children)
                {
                    var fontIcon = uIElement as FontIcon;
                    FontIcon newIcon = new FontIcon();
                    newIcon.FontFamily = fontIcon.FontFamily;
                    newIcon.Glyph = fontIcon.Glyph;
                    newIcon.FontSize = fontIcon.FontSize;
                    newIcon.VerticalAlignment = VerticalAlignment.Center;

                    newGrid.Children.Add(newIcon);
                }
                stackPanel.Children.Add(newGrid);
            }

            TextBlock newTextBlock = new TextBlock();
            newTextBlock.Text = childButton.Label;
            newTextBlock.FontSize = childButton.FontSize;
            newTextBlock.FontFamily = childButton.FontFamily;
            newTextBlock.FontSize = 14;
            Thickness margin = newTextBlock.Margin;
            margin.Left = 10;
            newTextBlock.Margin = margin;
            newTextBlock.HorizontalAlignment = HorizontalAlignment.Left;
            newTextBlock.VerticalAlignment = VerticalAlignment.Center;
            stackPanel.Children.Add(newTextBlock);

            secondaryButton.Content = stackPanel;
            return secondaryButton;
        }

        /// <summary>
        /// Will go back if possible
        /// </summary>
        /// <returns>true if it went back, false otherwise</returns>
        public bool GoBack()
        {
            if (DataContext is AnnotationToolbarViewModel)
            {
                return ViewModel.GoBack();
            }
            return false;
        }


        #region Command Bar Expansion
        private void MainAppBar_Opening(object sender, object e)
        {
            IsAppBarOpen = true;
            AppBarChanged?.Invoke(true);

            if (OverflowFlyoutPanel.Children.Count > 0)
            {
                OverflowFlyout.ShowAt(OverflowFlyoutButton);
            }
        }
        private void MainAppBar_Closing(object sender, object e)
        {
            IsAppBarOpen = false;
            AppBarChanged?.Invoke(false);
        }

        #endregion

        private void DataContext_ToolSelected()
        {
            OverflowFlyout?.Hide();
        }

        /// <summary>
        /// Open or close the MainAppBar
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void ExpandButton_Click(object sender, RoutedEventArgs e)
        {
            if (MainAppBar.IsOpen)
                MainAppBar.IsOpen = false;
            else
                MainAppBar.IsOpen = true;
        }

        private void AnnotationToolbarBase_Loaded(object sender, RoutedEventArgs e)
        {
#if SUBSCRIPTIONS
#if IS_SUBSCRIBED
            ViewModel.Subscribed = true;
#else
            ViewModel.Subscribed = ViewModel.GetSubscribed();
#endif
            if (!ViewModel.Subscribed)
            {
                //Lock the redaction tools since no subscription exists
                AnnotRedactionButton.Content = new FontIcon
                {
                    Style = (Style)Resources["ToolsFontIconStyle"],
                    FontSize = (double)Resources["FontIconDefaultFontSize"],
                    Glyph = "\uE781"
                };

                AnnotRedactionButton.Flyout = (Flyout)Application.Current.Resources["RedactionUpgradeFlyout"];
                AnnotRedactionButton.Flyout.Placement = FlyoutPlacementMode.Bottom;
            }
            else
            {
                AnnotRedactionButton.Flyout = null;
            }
#endif
        }
    }
}
