﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;

using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using UIPoint = Windows.Foundation.Point;

using pdftron.PDF.Tools.Controls.ViewModels;


namespace pdftron.PDF.Tools.Controls
{
    public sealed partial class ContextMenuControl : UserControl, IDisposable
    {
        #region Private Properties
        private ContextMenuControlViewModel _ViewModel;
        #endregion

        #region Constructors

        /// <summary>
        /// Creates a new ContextMenuControl with preset buttons
        /// </summary>
        /// <param name="ctrl"></param>
        public ContextMenuControl(FrameworkElement fe)
        {
            this.InitializeComponent();
            _ViewModel = new ContextMenuControlViewModel(fe);
            _ViewModel.ItemClicked += RaiseItemClicked;
            DataContext = _ViewModel;
        }

        /// <summary>
        /// Creates a new ContextMenuControl with user defined buttons
        /// </summary>
        /// <param name="ctrl"></param>
        /// <param name="buttonDictionary">
        /// Keys are used for the Button's Content(converted to an icon), and Tag to handle events
        /// Values are used for tooltips
        /// </param>
        public ContextMenuControl(FrameworkElement fe, Dictionary<string, string> buttonDictionary) //add null check
        {
            this.InitializeComponent();
            _ViewModel = new ContextMenuControlViewModel(fe, buttonDictionary);
            _ViewModel.ItemClicked += RaiseItemClicked;
            DataContext = _ViewModel;
        }
        #endregion

        #region Private Methods
        private void RaiseItemClicked(string tag)
        {
            if (_ItemClicked != null)
                _ItemClicked.Invoke(tag);
        }

        /// <summary>
        /// Context Menu's size changed
        /// </summary>
        private void MainPopupBorder_SizeChanged(object sender, Windows.UI.Xaml.SizeChangedEventArgs e)
        {
            SetDimensions(e.NewSize.Width, e.NewSize.Height);
        }

        /// <summary>
        /// Set the dimensions within the ViewModel to be for positioning.
        /// </summary>
        /// /// <param name="width"></param>
        /// /// <param name="height"></param>
        private void SetDimensions(double width, double height)
        {
            if (_ViewModel != null)
            {
                _ViewModel.SetDimensions(width, height);
            }
        }
        #endregion

        #region Public Properties
        /// <summary>
        /// Visibilty of the ContextMenuControl
        /// </summary>
        public bool IsOpen
        {
            get
            {
                if (_ViewModel != null)
                    return _ViewModel.IsOpen;

                return false;
            }
        }

        /// <summary>
        /// Visibilty of overflow menu
        /// </summary>
        public bool ShowOverflow
        {
            get
            {
                if (_ViewModel != null)
                    return _ViewModel.ShowOverflow;

                return false;
            }
        }

        /// <summary>
        /// Horizontal position
        /// </summary>
        public double HorizontalOffset
        {
            get
            {
                if (_ViewModel != null)
                    return _ViewModel.HorizontalOffset;

                return -1;
            }
        }

        /// <summary>
        /// Vertical position (top of application window to top of control)
        /// </summary>
        public double VerticalOffset
        {
            get
            {
                if (_ViewModel != null)
                    return _ViewModel.VerticalOffset;

                return -1;
            }
        }

        /// <summary>
        /// Maximum amount of buttons allowed per row from the ButtonCollection
        /// </summary>
        public int MaxButtonsPerRow
        {
            get
            {
                if (_ViewModel != null)
                    return _ViewModel.MaxButtonsPerRow;

                return -1;
            }
        }

        /// <summary>
        /// Position of the OverflowCollection, above or below the ButtonCollection
        /// </summary>
        public int OverflowContainerRow
        {
            get
            {
                if (_ViewModel != null)
                    return _ViewModel.OverflowContainerRow;

                return -1;
            }
        }

        /// <summary>
        /// Button collection
        /// </summary>
        public ObservableCollection<Button> ButtonCollection
        {
            get
            {
                if (_ViewModel != null)
                    return _ViewModel.ButtonCollection;

                return null;
            }
        }

        /// <summary>
        /// Overflow item collection
        /// </summary>
        public ObservableCollection<MenuFlyoutItem> OverflowCollection
        {
            get
            {
                if (_ViewModel != null)
                    return _ViewModel.OverflowCollection;

                return null;
            }
        }

        #endregion

        #region Public Methods

        /// <summary>
        /// Show the ContextMenuControl as long the <c>TargetSquare</c> or <c>TargetPoint</c> has been set
        /// </summary>
        /// <remarks>
        /// A base set of tools will be initialized if the ButtonCollection is empty
        /// </remarks>
        public void Show()
        {
            if (_ViewModel != null)
            {
                _ViewModel.Show();
                ShowStoryboard.Begin();
            }
        }

        /// <summary>
        /// Shows the ContextMenuControl
        /// </summary>
        /// <remarks>
        /// A base set of tools will be initialized if the ButtonCollection is empty
        /// </remarks>
        public void Show(UIPoint point)
        {
            if (_ViewModel != null && point != null)
            {
                _ViewModel.Show(point);
                ShowStoryboard.Begin();
            }
        }

        /// <summary>
        /// Hides the ContextMenuControl
        /// </summary>
        public void Hide(bool shouldDispose = false)
        {
            if (_ViewModel != null)
            {
                HideStoryboard.Begin();
                _ViewModel.Hide();
                if (shouldDispose)
                    Dispose();
            }
        }

        /// <summary>
        /// Creates and Adds multiple MenuFlyOutItems to an ObservableCollection.
        /// </summary>
        /// <param name="overflowDictionary">
        /// Keys are used for tags, Values are used for Text
        /// </param>
        public void CreateOverflowFromDictionary(Dictionary<string, string> overflowDictionary)
        {
            if(_ViewModel != null && overflowDictionary != null)
                _ViewModel.CreateOverflowFromDictionary(overflowDictionary);
        }

        /// <summary>
        /// Adds a pre-made MenuFlyoutItem to an ObservableCollection.
        /// </summary>
        /// <param name="flyoutItem"></param>
        public void AddItemToOverflowCollection(MenuFlyoutItem flyoutItem)
        {
            if(_ViewModel != null && flyoutItem != null)
                _ViewModel.AddItemToOverflowCollection(flyoutItem);
        }

        /// <summary>
        /// Add a single Button to the ObservableCollection based on a button dictionary
        /// </summary>
        /// <param name="button">The button to be added to the collection</param>
        public void AddButtonToCollection(Button button)
        {
            if(_ViewModel != null && button != null)
                _ViewModel.AddButtonToCollection(button);
        }

        /// <summary>
        /// Remove a button from the ObservableCollection
        /// </summary>
        /// <param name="button">
        /// The button that should be removed from the ObservableCollection
        /// </param>
        public void RemoveButtonFromCollection(Button button)
        {
            if(_ViewModel != null && button != null)
            {
                _ViewModel.RemoveButtonFromCollection(button);
            }
        }

        /// <summary>
        /// Add buttons to the observable collection based on a button dictionary.
        /// </summary>
        /// <param name="buttonDictionary">
        /// Keys will be used for Button Tags and Icons, 
        /// values will be used for text Content and Tooltips
        /// </param>
        public void CreateButtonsFromDictionary(Dictionary<string, string> buttonDictionary)
        {
            if (_ViewModel != null && buttonDictionary != null)
                _ViewModel.CreateButtonsFromDictionary(buttonDictionary);
        }

        /// <summary>
        /// Clears the Button collection.
        /// </summary>
        public void ClearButtonCollection()
        {
            if(_ViewModel != null)
                _ViewModel.ClearButtonCollection();
        }

        /// <summary>
        /// Clears the overflow collection.
        /// </summary>
        public void ClearOverflowCollection()
        {
            if(_ViewModel != null)
                _ViewModel.ClearOverflowCollection();
        }

        /// <summary>
        /// Initializes a base set tools.
        /// </summary>
        public void InitBaseTools()
        {
            if (_ViewModel != null)
                _ViewModel.InitBaseTools();
        }

        /// <summary>
        /// Gives a target rectangle as 4 points in PDFViewCtrl space
        /// </summary>
        public void TargetSquare(double x1, double y1, double x2, double y2)
        {
            if (_ViewModel == null)
                return;

            _ViewModel.TargetSquare(x1, y1, x2, y2);
        }

        /// <summary>
        /// Provide a point in screen space that the menu should be attached to.
        /// </summary>
        public void TargetPoint(UIPoint pt)
        {
            if (_ViewModel == null)
                return;

            _ViewModel.TargetPoint(pt);
        }

        public void Dispose()
        {
            if (_ViewModel != null)
            {
                _ViewModel.ItemClicked -= RaiseItemClicked;
                _ViewModel.Dispose();
            }
        }
        #endregion

        #region Events
        private ContextMenuControlViewModel.ItemClickedDelegate _ItemClicked;
        public event ContextMenuControlViewModel.ItemClickedDelegate ItemClicked
        {
            add
            {
                if (_ItemClicked?.GetInvocationList().Contains(value) != true)
                    _ItemClicked += value;
            }
            remove => _ItemClicked -= value;
        }
        #endregion
    }
}
