﻿using pdftron.PDF.Tools.Controls.ControlBase;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Runtime.InteropServices.WindowsRuntime;
using System.Threading.Tasks;
using Windows.Foundation;
using Windows.Foundation.Collections;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Controls.Primitives;
using Windows.UI.Xaml.Data;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Navigation;

// The User Control item template is documented at http://go.microsoft.com/fwlink/?LinkId=234236

namespace pdftron.PDF.Tools.Controls
{
    public sealed partial class CropView : UserControl, ICloseableControl
    {
        private ViewModels.CropViewViewModel _ViewModel;
        public ViewModels.CropViewViewModel ViewModel { get { return _ViewModel; } }

        private const int POINT_THRESHOLD = 20;
        private const int MIN_CROP_LENGTH = 50;

        private ControlPoint _ControlPoint;

        private Rect _CropBox;

        // Cropped margins by the user
        private double _LeftCrop;
        private double _TopCrop;
        private double _RightCrop;
        private double _BottomCrop;

        // Center point for calculating translation for middle control point
        private Windows.Foundation.Point _PreviousPointForDrag;
        private Windows.Foundation.Point _CropRectCornerForDrag;

        public CropView(PDFViewCtrl ctrl)
        {
            this.InitializeComponent();
            this.SizeChanged += CropView_SizeChanged;
            _ViewModel = new ViewModels.CropViewViewModel(this, ctrl);
            _ViewModel.PropertyChanged += ViewModel_PropertyChanged;
            this.DataContext = _ViewModel;

            this.PointerPressed += CropRectangle_PointerPressed;
            this.PointerMoved += CropView_PointerMoved;
            this.PointerReleased += CropView_PointerGone;
            this.PointerCanceled += CropView_PointerGone;
            this.PointerCaptureLost += CropView_PointerCaptureLost;

            // Wait until the UI has loaded before rasterizing initial page
            this.Loaded += (s, e) =>
            {
                RasterizeInitialPage();
            };
        }



        private void ViewModel_PropertyChanged(object sender, System.ComponentModel.PropertyChangedEventArgs e)
        {
            if (e.PropertyName.Equals("CurrCropItem"))
            {
                UpdateImageSize();
                UpdateCropRectangle();
                PositionWidgets();
            }
        }

        private void CropView_SizeChanged(object sender, SizeChangedEventArgs e)
        {
            if (e.NewSize.Width == 0)
                return;


            bool wideLayout = false;
            bool squishedTallLayout = false;
            bool centerPageButtons = false;

            if (e.NewSize.Width > e.NewSize.Height + 50 && e.NewSize.Height < 700)
            {
                wideLayout = true;
            }
            else if (e.NewSize.Height < 600)
            {
                squishedTallLayout = true;
            }
            if (wideLayout || e.NewSize.Width > 450)
            {
                centerPageButtons = true;
            }


            // Handle UI for landscape mode
            if (wideLayout)
            {
                CropGrid.SetValue(Grid.RowSpanProperty, 3);
                CropGrid.SetValue(Grid.ColumnSpanProperty, 1);
                PageNavigationGrid.SetValue(Grid.RowProperty, 0);
                PageNavigationGrid.SetValue(Grid.ColumnProperty, 1);
                PageNavigationGrid.SetValue(Grid.RowSpanProperty, 1);
            }
            // Handle UI for portrait mode
            else if (squishedTallLayout)
            {
                CropGrid.SetValue(Grid.RowSpanProperty, 2);
                CropGrid.SetValue(Grid.ColumnSpanProperty, 1);
                PageNavigationGrid.SetValue(Grid.RowProperty, 2);
                PageNavigationGrid.SetValue(Grid.ColumnProperty, 0);
                PageNavigationGrid.SetValue(Grid.RowSpanProperty, 1);
            }
            else
            {
                CropGrid.SetValue(Grid.RowSpanProperty, 1);
                CropGrid.SetValue(Grid.ColumnSpanProperty, 2);
                CropInnerGrid.Margin = new Thickness(0);
                if (e.NewSize.Width > 900)
                {
                    CropGrid.SetValue(Grid.ColumnSpanProperty, 1);
                }
                else if (e.NewSize.Width > 600)
                {
                    if (!double.IsNaN(CropButtonHost.ActualWidth))
                    {
                        double leftMargin = CropButtonHost.ActualWidth * (e.NewSize.Width - 600.0) / 300.0;
                        CropInnerGrid.Margin = new Thickness(0, 0, leftMargin, 0);
                    }
                }
                PageNavigationGrid.SetValue(Grid.RowProperty, 1);
                PageNavigationGrid.SetValue(Grid.ColumnProperty, 0);
                PageNavigationGrid.SetValue(Grid.RowSpanProperty, 2);
            }
            if (centerPageButtons)
            {
                PageNavigationGrid.HorizontalAlignment = HorizontalAlignment.Center;
                PageNavigationGrid.Margin = new Thickness(0, 0, 0, 0);
            }
            else
            {
                PageNavigationGrid.HorizontalAlignment = HorizontalAlignment.Left;
                PageNavigationGrid.Margin = new Thickness(15, 0, 0, 0);
            }
        }

        /// <summary>
        /// Use this to apply the current custom color mode to the crop view.
        /// </summary>
        /// <param name="mode"></param>
        /// <param name="background"></param>
        /// <param name="foreground"></param>
        public void SetColorPostProcessing(PDFRasterizerColorPostProcessMode mode, Windows.UI.Color? background, Windows.UI.Color? foreground)
        {
            if (mode == PDFRasterizerColorPostProcessMode.e_postprocess_night_mode || mode == PDFRasterizerColorPostProcessMode.e_postprocess_invert)
            {
                CropImageBorder.Background = new SolidColorBrush(Windows.UI.Colors.Black);
                TopFill.Fill = this.Resources["NightModeGrayoutBrush"] as SolidColorBrush;
                BottomFill.Fill = this.Resources["NightModeGrayoutBrush"] as SolidColorBrush;
                LeftFill.Fill = this.Resources["NightModeGrayoutBrush"] as SolidColorBrush;
                RightFill.Fill = this.Resources["NightModeGrayoutBrush"] as SolidColorBrush;
            }
            else if (mode == PDFRasterizerColorPostProcessMode.e_postprocess_gradient_map && background != null && foreground != null)
            {
                CropImageBorder.Background = new SolidColorBrush(background.Value);
            }
            _ViewModel.SetColorPostProcessing(mode, background, foreground);
        }

        private void RasterizeInitialPage()
        {
            _ViewModel.RasterizeInitialPage();
        }

        private ControlClosedDelegate _controlClosed;
        public event ControlClosedDelegate ControlClosed
        {
            add
            {
                if (_controlClosed?.GetInvocationList().Contains(value) != true)
                    _controlClosed += value;
            }
            remove => _controlClosed -= value;
        }

        public void CloseControl()
        {
            _controlClosed?.Invoke();
            ViewModel.CleanUp();
        }

        private void _ViewModel_CropBackgroundUpdated(double width, double height)
        {
            UpdateImageSize();
        }

        private void UpdateCropRectangle()
        {
            if (ViewModel.CurrCropItem == null)
            {
                return;
            }

            Rect cropRatioRect = ViewModel.CurrCropRatioRect;

            _LeftCrop = cropRatioRect.x1 * CropImage.Width;
            _TopCrop = cropRatioRect.y1 * CropImage.Height;
            _RightCrop = cropRatioRect.x2 * CropImage.Width;
            _BottomCrop = cropRatioRect.y2 * CropImage.Height;
        }

        /// <summary>
        /// This scales the BackgroundRectangle and ImageBoarder for a PDF page's width and height.
        /// </summary>
        /// <param name="width"></param>
        /// <param name="height"></param>
        private void UpdateImageSize()
        {
            if (ViewModel.CurrCropItem == null)
            {
                return;
            }
            // Perform uniform scaling to UI width/height
            Size size = UtilityFunctions.GetUniformStretch(CropInnerGrid.ActualWidth, CropInnerGrid.ActualHeight, ViewModel.CurrCropItem.Width, ViewModel.CurrCropItem.Height);

            CropImage.Width = size.Width;
            CropImage.Height = size.Height;
            double extraMargin = -CroppingCanvas.Margin.Left * 2;
            CroppingCanvas.Width = CropImage.Width + extraMargin;
            CroppingCanvas.Height = CropImage.Height + extraMargin;
            ImageBorder.Width = CropImage.Width + ImageBorder.BorderThickness.Left + ImageBorder.BorderThickness.Right;
            ImageBorder.Height = CropImage.Height + ImageBorder.BorderThickness.Top + ImageBorder.BorderThickness.Bottom;
        }

        /// <summary>
        ///  Handle re-scaling of crop and UI if user resizes the screen
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void CropInnerGrid_SizeChanged(object sender, SizeChangedEventArgs e)
        {
            if (e.PreviousSize.Height == 0 || e.PreviousSize.Width == 0)
                return;

            UpdateImageSize();
            UpdateCropRectangle();
            PositionWidgets();
        }

        /// <summary>
        ///  Updates the ctrl widgets of a crop when the UI has moved.
        /// </summary>
        private void UpdateCtrlPositions()
        {
            UpperLeftCtrl.SetValue(Canvas.LeftProperty, CropRectangle.GetValue(Canvas.LeftProperty));
            UpperLeftCtrl.SetValue(Canvas.TopProperty, CropRectangle.GetValue(Canvas.TopProperty));

            BottomLeftCtrl.SetValue(Canvas.LeftProperty, CropRectangle.GetValue(Canvas.LeftProperty));
            BottomLeftCtrl.SetValue(Canvas.TopProperty, (double)CropRectangle.GetValue(Canvas.TopProperty) + CropRectangle.ActualHeight);

            BottomRightCtrl.SetValue(Canvas.LeftProperty, (double)CropRectangle.GetValue(Canvas.LeftProperty) + CropRectangle.ActualWidth);
            BottomRightCtrl.SetValue(Canvas.TopProperty, (double)CropRectangle.GetValue(Canvas.TopProperty) + CropRectangle.ActualHeight);

            UpperRightCtrl.SetValue(Canvas.LeftProperty, (double)CropRectangle.GetValue(Canvas.LeftProperty) + CropRectangle.ActualWidth);
            UpperRightCtrl.SetValue(Canvas.TopProperty, CropRectangle.GetValue(Canvas.TopProperty));

            PositionGrayout();
        }

        /// <summary>
        ///  Gets the control point depending on where the crop rectangle was pressed
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void CropRectangle_PointerPressed(object sender, PointerRoutedEventArgs e)
        {
            Windows.Foundation.Point point = e.GetCurrentPoint(CropRectangle).Position;

            if (point.X < -POINT_THRESHOLD || point.X > CropRectangle.Width + POINT_THRESHOLD
                || point.Y < -POINT_THRESHOLD || point.Y > CropRectangle.Height + POINT_THRESHOLD)
            {
                return;
            }

            if (Math.Abs(point.X - CropRectangle.Width) <= POINT_THRESHOLD && Math.Abs(point.Y - CropRectangle.Height) <= POINT_THRESHOLD)
            {
                _ControlPoint = ControlPoint.BottomRight;
            }
            else if (Math.Abs(point.X) <= POINT_THRESHOLD && Math.Abs(point.Y - CropRectangle.Height) <= POINT_THRESHOLD)
            {
                _ControlPoint = ControlPoint.BottomLeft;
            }
            else if (Math.Abs(point.X) <= POINT_THRESHOLD && Math.Abs(point.Y - 0) <= POINT_THRESHOLD)
            {
                _ControlPoint = ControlPoint.TopLeft;
            }
            else if (Math.Abs(point.X - CropRectangle.Width) <= POINT_THRESHOLD && Math.Abs(point.Y - 0) <= POINT_THRESHOLD)
            {
                _ControlPoint = ControlPoint.TopRight;
            }
            else if (Math.Abs(point.X - CropRectangle.Width) <= POINT_THRESHOLD)
            {
                _ControlPoint = ControlPoint.RightMiddle;
            }
            else if (Math.Abs(point.Y - CropRectangle.Height) <= POINT_THRESHOLD)
            {
                _ControlPoint = ControlPoint.BottomMiddle;
            }
            else if (Math.Abs(point.X - 0) <= POINT_THRESHOLD)
            {
                _ControlPoint = ControlPoint.LeftMiddle;
            }
            else if (Math.Abs(point.Y - 0) <= POINT_THRESHOLD)
            {
                _ControlPoint = ControlPoint.TopMiddle;
            }
            else if (point.X >= 0 && point.X <= CropRectangle.Width && point.Y >= 0 && point.Y <= CropRectangle.Height)
            {
                _ControlPoint = ControlPoint.Middle;
                _PreviousPointForDrag = e.GetCurrentPoint(CropImage).Position;
            }
            else
            {
                _ControlPoint = ControlPoint.None;
            }

            if (_ControlPoint != ControlPoint.None)
            {
                this.CapturePointer(e.Pointer);
            }
        }

        /// <summary>
        /// If a Control Point is being used, drag the crop as appropriate to the pointer
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void CropView_PointerMoved(object sender, PointerRoutedEventArgs e)
        {
            if (_ControlPoint == ControlPoint.None)
                return;

            Windows.Foundation.Point point = e.GetCurrentPoint(CropImage).Position;

            double pointX = Math.Max(point.X, 0);
            double pointY = Math.Max(point.Y, 0);

            switch (_ControlPoint)
            {
                case ControlPoint.BottomRight:
                    // bound pointX so that it cannot shrink past a min length, and expand past the image boundaries
                    pointX = Math.Max(pointX, _LeftCrop + MIN_CROP_LENGTH);
                    pointX = Math.Min(pointX, CropImage.Width);
                    //CropRectangle.Width = Math.Max(pointX - _LeftCrop, MIN_CROP_LENGTH);
                    _RightCrop = CropImage.Width - pointX;

                    // bound pointY so that it cannot shrink past a min length, and expand past the image boundaries
                    pointY = Math.Max(pointY, _TopCrop + MIN_CROP_LENGTH);
                    pointY = Math.Min(pointY, CropImage.Height);
                    //CropRectangle.Height = Math.Max(pointY - _TopCrop, MIN_CROP_LENGTH);
                    _BottomCrop = CropImage.Height - pointY;
                    break;
                case ControlPoint.BottomLeft:
                    pointX = Math.Min(pointX, CropImage.Width - MIN_CROP_LENGTH - _RightCrop);
                    //CropRectangle.SetValue(Canvas.LeftProperty, CropImage.Width / -2 + pointX);
                    //CropRectangle.Width = Math.Max(CropImage.Width - pointX - _RightCrop, MIN_CROP_LENGTH);
                    _LeftCrop = pointX;

                    pointY = Math.Max(pointY, _TopCrop + MIN_CROP_LENGTH);
                    pointY = Math.Min(pointY, CropImage.Height);
                    //CropRectangle.Height = Math.Max(pointY - _TopCrop, MIN_CROP_LENGTH);
                    _BottomCrop = CropImage.Height - pointY;
                    break;
                case ControlPoint.TopLeft:
                    pointX = Math.Min(pointX, CropImage.Width - MIN_CROP_LENGTH - _RightCrop);
                    //CropRectangle.SetValue(Canvas.LeftProperty, CropImage.Width / -2 + pointX);
                    //CropRectangle.Width = Math.Max(CropImage.Width - pointX - _RightCrop, MIN_CROP_LENGTH);
                    _LeftCrop = pointX;

                    pointY = Math.Min(pointY, CropImage.Height - _BottomCrop - MIN_CROP_LENGTH);
                    //CropRectangle.SetValue(Canvas.TopProperty, CropImage.Height / -2 + pointY);
                    //CropRectangle.Height = Math.Max(CropImage.Height - pointY -_BottomCrop, MIN_CROP_LENGTH);
                    _TopCrop = pointY;
                    break;
                case ControlPoint.TopRight:
                    pointX = Math.Max(pointX, _LeftCrop + MIN_CROP_LENGTH);
                    pointX = Math.Min(pointX, CropImage.Width);
                    //CropRectangle.Width = Math.Max(pointX - _LeftCrop, MIN_CROP_LENGTH);
                    _RightCrop = CropImage.Width - pointX;

                    pointY = Math.Min(pointY, CropImage.Height - _BottomCrop - MIN_CROP_LENGTH);
                    //CropRectangle.SetValue(Canvas.TopProperty, CropImage.Height / -2 + pointY);
                    //CropRectangle.Height = Math.Max(CropImage.Height - pointY - _BottomCrop, MIN_CROP_LENGTH);
                    _TopCrop = pointY;
                    break;
                case ControlPoint.RightMiddle:
                    pointX = Math.Max(pointX, _LeftCrop + MIN_CROP_LENGTH);
                    pointX = Math.Min(pointX, CropImage.Width);
                    //CropRectangle.Width = Math.Max(pointX - _LeftCrop, MIN_CROP_LENGTH);
                    _RightCrop = CropImage.Width - pointX;
                    break;
                case ControlPoint.BottomMiddle:
                    pointY = Math.Max(pointY, _TopCrop + MIN_CROP_LENGTH);
                    pointY = Math.Min(pointY, CropImage.Height);
                    //CropRectangle.Height = Math.Max(pointY - _TopCrop, MIN_CROP_LENGTH);
                    _BottomCrop = CropImage.Height - pointY;
                    break;
                case ControlPoint.LeftMiddle:
                    pointX = Math.Min(pointX, CropImage.Width - MIN_CROP_LENGTH - _RightCrop);
                    //CropRectangle.SetValue(Canvas.LeftProperty, CropImage.Width / -2 + pointX);
                    //CropRectangle.Width = Math.Max(CropImage.Width - pointX - _RightCrop, MIN_CROP_LENGTH);
                    _LeftCrop = pointX;
                    break;
                case ControlPoint.TopMiddle:
                    pointY = Math.Min(pointY, CropImage.Height - _BottomCrop - MIN_CROP_LENGTH);
                    //CropRectangle.SetValue(Canvas.TopProperty, CropImage.Height / -2 + pointY);
                    //CropRectangle.Height = Math.Max(CropImage.Height - pointY - _BottomCrop, MIN_CROP_LENGTH);
                    _TopCrop = pointY;
                    break;
                case ControlPoint.Middle:
                    // Get difference from previous position
                    double diffX = pointX - _PreviousPointForDrag.X;
                    double diffY = pointY - _PreviousPointForDrag.Y;

                    double cropWidth = CropImage.Width - _RightCrop - _LeftCrop;
                    double cropHeight = CropImage.Height - _BottomCrop - _TopCrop;

                    // Bound new X position to not go past CropImage boundaries
                    double newX = Math.Min(_LeftCrop + diffX, CropImage.Width - cropWidth);
                    newX = Math.Max(newX, 0);

                    // Bound new Y position to not go past CropImage boundaries
                    double newY = Math.Min(_TopCrop + diffY, CropImage.Height - cropHeight);
                    newY = Math.Max(newY, 0);

                    // Update crop values
                    _LeftCrop = newX;
                    _TopCrop = newY;
                    _RightCrop = Math.Max(CropImage.Width - _LeftCrop - cropWidth, 0);
                    _BottomCrop = Math.Max(CropImage.Height - _TopCrop - cropHeight, 0);

                    pointX = Math.Max(0, Math.Min(pointX, CropImage.Width));
                    pointY = Math.Max(0, Math.Min(pointY, CropImage.Height));
                    _PreviousPointForDrag = new Windows.Foundation.Point(pointX, pointY);
                    break;
            }

            _LeftCrop = Math.Max(_LeftCrop, 0);
            _TopCrop = Math.Max(_TopCrop, 0);
            _RightCrop = Math.Max(_RightCrop, 0);
            _BottomCrop = Math.Max(_BottomCrop, 0);

            PositionWidgets();
        }

        /// <summary>
        /// Notify the ViewModel of the updated crop depending on the notify param.
        /// </summary>
        /// <param name="notify"></param>
        private void DoneCropping(bool notify)
        {
            if (_ControlPoint != ControlPoint.None)
            {
                _ControlPoint = ControlPoint.None;

                if (notify)
                {
                    Rect cropRatio = new Rect(_LeftCrop / CropImage.Width, _TopCrop / CropImage.Height, _RightCrop / CropImage.Width, _BottomCrop / CropImage.Height);
                    ViewModel.CurrCropRatioRect = cropRatio;
                }
            }
        }


        private void CropView_PointerGone(object sender, PointerRoutedEventArgs e)
        {
            DoneCropping(true);
            this.ReleasePointerCaptures();
        }

        private void CropView_PointerCaptureLost(object sender, PointerRoutedEventArgs e)
        {
            // Don't notify ViewModel on PointerCaptureLost (since its called on navigating pages)
            DoneCropping(false);
        }

        private void MessageHost_SizeChanged(object sender, SizeChangedEventArgs e)
        {
            MessageHost.SetValue(Canvas.LeftProperty, e.NewSize.Width / -2);
            MessageHost.SetValue(Canvas.TopProperty, e.NewSize.Height / -2);
        }

        private void PositionWidgets()
        {
            if (ViewModel.CurrCropItem == null)
            {
                return;
            }

            double offsetExtra = -CroppingCanvas.Margin.Left - (CropRectangle.StrokeThickness / 2);
            CropRectangle.SetValue(Canvas.LeftProperty, _LeftCrop + offsetExtra);
            CropRectangle.SetValue(Canvas.TopProperty, _TopCrop + offsetExtra);
            CropRectangle.Width = Math.Max(CropImage.Width - _LeftCrop - _RightCrop + CropRectangle.StrokeThickness, MIN_CROP_LENGTH);
            CropRectangle.Height = Math.Max(CropImage.Height - _TopCrop - _BottomCrop + CropRectangle.StrokeThickness, MIN_CROP_LENGTH);

            GrayoutCanvas.Width = CropImage.Width;
            GrayoutCanvas.Height = CropImage.Height;

            PositionGrayout();
            UpdateCtrlPositions();
        }

        private void PositionGrayout()
        {
            Size centerArea = new Size(CropRectangle.Width - CropRectangle.StrokeThickness, CropRectangle.Height - CropRectangle.StrokeThickness);
            TopFill.Height = _TopCrop;
            TopFill.Width = GrayoutCanvas.Width;
            BottomFill.Height = _BottomCrop;
            BottomFill.Width = GrayoutCanvas.Width;
            BottomFill.SetValue(Canvas.TopProperty, TopFill.Height + centerArea.Height);
            LeftFill.Height = centerArea.Height;
            LeftFill.Width = _LeftCrop;
            LeftFill.SetValue(Canvas.TopProperty, TopFill.Height);
            RightFill.Height = centerArea.Height;
            RightFill.Width = _RightCrop;
            RightFill.SetValue(Canvas.LeftProperty, LeftFill.Width + centerArea.Width);
            RightFill.SetValue(Canvas.TopProperty, TopFill.Height);
        }

        /// <summary>
        /// Will perform a back operation if applicable.
        /// Use this when the back button is pressed to give this control a chance to handle it.
        /// </summary>
        /// <returns>true if the control could go back</returns>
        public bool GoBack()
        {
            return _ViewModel.GoBack();
        }
    }
}
