﻿using pdftron.PDF.Tools.Controls.Resources;
using pdftron.PDF.Tools.Controls.ViewModels;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using Windows.ApplicationModel.Core;
using Windows.Foundation;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Media.Animation;


namespace pdftron.PDF.Tools.Controls
{
    public class BookmarkItem
    {
        public Bookmark SourceBookmark { get; set; }
        public string Title { get; set; }
        public bool HasChildren { get; set; }
        public int ChildIndex { get; set; }

        public BookmarkItem(Bookmark bookmark, int index = 0)
        {
            SourceBookmark = bookmark;
            Title = bookmark.GetTitle();
            HasChildren = bookmark.HasChildren();
            ChildIndex = index;
        }
    }

    public class AnimationCollection
    {
        public Storyboard EnterFromLeft { get; set; }
        public Storyboard LeaveToTheLeft { get; set; }
        public Storyboard EnterFromRight { get; set; }
        public Storyboard LeaveToTheRight { get; set; }

        public AnimationCollection()
        { }

        public void Refresh(double newWidth)
        {
            DoubleAnimation da = EnterFromLeft.Children[0] as DoubleAnimation;
            da.From = -newWidth;
            da = LeaveToTheLeft.Children[0] as DoubleAnimation;
            da.To = -newWidth;
            da = EnterFromRight.Children[0] as DoubleAnimation;
            da.From = newWidth;
            da = LeaveToTheRight.Children[0] as DoubleAnimation;
            da.To = newWidth;
        }
    }

    partial class Outline : IDisposable
    {
        private PDFViewCtrl _PDFViewCtrl;
        private NavigationStack _NavigationStack;
        private List<Grid> _BookmarkPages;
        private List<ScrollViewer> _ScrollViewerPages;
        private List<AnimationCollection> _Animations;
        private int _CurrentIndex = -1;
        private TextBlock _RootTextBlock = null;

        private List<int> _CurrentLocation = null;
        private List<int> _RestorePath = null;
        private bool _InitialOutlineLoaded = false;

        private Windows.UI.Xaml.Input.KeyEventHandler _KeyDownHandler = null;
        private RoutedEventHandler _FocusLostHandler = null;
        private string _OriginalText = "";

        private bool _Loaded = false;
        private bool _IsConverting = false;
        private OnConversionEventHandler _OnConversionHandlder = null;

        private Size _CurrentSize;
        private string _DocumentTitle = string.Empty;

        private SemaphoreSlim OutlineReadingSemaphore { get; set; }

        public OutlineViewModel ViewModel { get; private set; }

        /// <summary>
        /// The PDFViewCtrl form which the outline should get the document with PDFs.
        /// </summary>
        public PDFViewCtrl PDFViewCtrl
        {
            get { return _PDFViewCtrl; }
            set
            {
                _PDFViewCtrl = value;
                if (_PDFViewCtrl != null)
                {
                    _BookmarkPages = new List<Grid>();
                    _ScrollViewerPages = new List<ScrollViewer>();
                    _Animations = new List<AnimationCollection>();
                    _CurrentLocation = new List<int>();
                }
            }
        }

        public NavigationStack NavigationStack
        {
            get { return _NavigationStack; }
            set { _NavigationStack = value; }
        }

        /// <summary>
        /// Enable weather the ability to edit the outline is enabled or disabled
        /// </summary>
        public bool IsEditEnabled
        {
            get { return ViewModel.IsEditEnabled; }
            set { ViewModel.IsEditEnabled = value; }
        }

        public string DocumentTitle
        {
            get { return _DocumentTitle; }
            set
            {
                if (!string.Equals(value, _DocumentTitle, StringComparison.Ordinal))
                {
                    _DocumentTitle = value;
                    if (_RootTextBlock != null)
                    {
                        _RootTextBlock.Text = _DocumentTitle;
                    }
                }
            }
        }

        public delegate void PageChangedDelegate(int prevPage, int page);
        public PageChangedDelegate PageChanged;

        /// <summary>
        /// Creates a new outline (bookmark) control without any association with a PDFViewCtrl.
        /// Use the PDFViewCtrl property to associated the outline control with a document and populate it.
        /// </summary>
        public Outline()
        {
            this.InitializeComponent();
            OutlineReadingSemaphore = new SemaphoreSlim(1);
            this.Loaded += Outline_Loaded;
            this.Unloaded += Outline_Unloaded;
        }

        /// <summary>
        /// Creates a new outline (bookmark) control associated with a PDFViewCtrl. The control will show the bookmarks
        /// in the document that is open in the PDFViewCtrl.
        /// </summary>
        public Outline(PDFViewCtrl ctrl)
        {
            this.InitializeComponent();

            ViewModel = new OutlineViewModel(ctrl);
            this.DataContext = ViewModel;
            ViewModel.FocusOnSelectedItemRequested += ViewModelFocusOnSelectedItemRequested;
            ViewModel.DocumentModified += ViewModelDocumentModified;

            if(OutlineTreeView != null)
            {
                OutlineTreeView.ItemInvoked += OutlineTreeView_ItemInvoked;
                OutlineTreeView.Expanding += OutlineTreeView_ExpandingAsync;
                OutlineTreeView.Collapsed += OutlineTreeView_CollapsedAsync;
            }

            this.Loaded += Outline_Loaded;
            this.Unloaded += Outline_Unloaded;
            this.PDFViewCtrl = ctrl;
        }
       
        private void OutlineTreeView_ItemInvoked(Microsoft.UI.Xaml.Controls.TreeView sender, Microsoft.UI.Xaml.Controls.TreeViewItemInvokedEventArgs args)
        {
            BookmarkNode node = args.InvokedItem as BookmarkNode;
            if (node != null)
            {               
                pdftron.PDF.Action action = node.SourceBookmark.GetAction();
                if (action.IsValid())
                {
                    ActionParameter action_param = new ActionParameter(action);
                    ActionUtils.ExecuteAction(action_param, this.PDFViewCtrl, NavigationStack);
                }
            }
        }

        private void OverFlowButton_Click(object sender, RoutedEventArgs e)
        {
            Button button = sender as Button;
            BookmarkNode node = button.DataContext as BookmarkNode;
            if (ViewModel.IsEditing)
            {
                if (ViewModel.LastInvokedNode != null)
                    ViewModel.LastInvokedNode.IsInvoked = false;
                node.IsInvoked = true;
                ViewModel.LastInvokedNode = node;
            }
        }

        private void Outline_Unloaded(object sender, RoutedEventArgs e)
        {
            _Loaded = false;
            if (_OnConversionHandlder != null)
            {
                _PDFViewCtrl.OnConversionChanged -= _OnConversionHandlder;
                _OnConversionHandlder = null;
            }
        }

        private void Outline_Loaded(object sender, RoutedEventArgs e)
        {
            _Loaded = true;
            if (_IsConverting)
            {
                RegisterForConversion();
            }
        }

        private void RegisterForConversion()
        {
            _OnConversionHandlder = new OnConversionEventHandler(PDFViewCtrl_OnConversionChanged);
            _PDFViewCtrl.OnConversionChanged += _OnConversionHandlder;
        }

        private async void PDFViewCtrl_OnConversionChanged(PDFViewCtrlConversionType type, int totalPagesConverted)
        {
            if (type == PDFViewCtrlConversionType.e_conversion_finished)
            {
                await ShowAsync();
            }
        }

        private void SetSize(Size size)
        {
            _CurrentSize = size;
            RectangleGeometry clipRect = new RectangleGeometry();
            clipRect.Rect = new Windows.Foundation.Rect(0, 0, size.Width, size.Height);
            ContentPanel.Clip = clipRect;

            foreach(AnimationCollection collection in _Animations)
            {
                collection.Refresh(size.Width);
            }

            if (_ScrollViewerPages != null)
            {
                foreach (ScrollViewer scroller in _ScrollViewerPages)
                {
                    FrameworkElement scrollContent = scroller.Content as FrameworkElement;
                    if (scrollContent != null)
                    {
                        scrollContent.Width = size.Width;
                    }
                }
            }

            foreach (UIElement child in ContentPanel.Children)
            {
                Grid g = child as Grid;
                g.Width = size.Width;
            }
        }

        public event EventHandler<ItemClickEventArgs> ItemClicked;

        public List<int> GetState()
        {
            return _CurrentLocation;
        }

        public async void RestoreState(List<int> list)
        {
            _RestorePath = list;
            if (_RestorePath != null && _RestorePath.Count > 0 &&
                _InitialOutlineLoaded && _ScrollViewerPages != null && _ScrollViewerPages.Count > 0)
            {
                object child = _ScrollViewerPages[0].Content;
                StackPanel childStack = child as StackPanel;
                if (childStack != null && childStack.Children.Count > _RestorePath[0])
                {
                    UIElement stackChild = childStack.Children[_RestorePath[0]];
                    _RestorePath.RemoveAt(0);
                    Grid childGrid = stackChild as Grid;
                    if (childGrid != null && childGrid.Children.Count > 0)
                    {
                        Button button = childGrid.Children[0] as Button;
                        if (button != null)
                        {
                            BookmarkItem bookmark = button.Tag as BookmarkItem;
                            if (bookmark != null && bookmark.HasChildren)
                            {
                                _BookmarkPages[0].Visibility = Visibility.Collapsed;
                                await NavigateToChildAsync(bookmark);
                            }
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Resets the Bookmark display, so that next time it will create bookmarks from scratch.
        /// </summary>
        public void Reset()
        {
            _BookmarkPages.Clear();
            _ScrollViewerPages.Clear();
            _Animations.Clear();
            _CurrentIndex = -1;
            ContentPanel.Children.Clear();
        }

        public async Task WaitForOutlineToLoadAsync()
        {
            //await OutlineReadingSemaphore.WaitAsync();
            //OutlineReadingSemaphore.Release();
            await ViewModel.WaitForOutlineToLoadAsync();
        }

        /// <summary>
        /// Animates the Bookmarks panel onto the screen.
        /// </summary>
        public async Task ShowAsync()
        {
            if (_PDFViewCtrl == null)
                return;

            if (_PDFViewCtrl.IsConverting)
            {
                _IsConverting = true;
                if (_Loaded)
                {
                    RegisterForConversion();
                }
                return;
            }

            if (ViewModel != null)
            {
                await CoreApplication.MainView.CoreWindow.Dispatcher.RunAsync(Windows.UI.Core.CoreDispatcherPriority.Normal, async () =>
                {
                    await ViewModel.CreateTreeAsync();
                });
            }
        }

        /// <summary>
        /// Animates the Bookmarks panel of the screen.
        /// </summary>
        public void Hide()
        {
            //SlideOutPanel.Begin();
        }

        private void SlideOutCompleted(object sender, object e)
        {
            ContentPanel.Opacity = 0;
        }

        private IAsyncOperation<IList<BookmarkItem>> GetBookmarkListAsync(Bookmark root, PDFDoc doc)
        {
            Task<IList<BookmarkItem>> t = new Task<IList<BookmarkItem>>(() => GetBookmarkList(root, doc));
            t.Start();
            return t.AsAsyncOperation();
        }

        private IList<BookmarkItem> GetBookmarkList(Bookmark root, PDFDoc doc, bool gotLock = false)
        {
            IList<BookmarkItem> items = new List<BookmarkItem>();

            try
            {
                doc?.LockRead();
                if (!gotLock)
                {
                    OutlineReadingSemaphore?.Wait();
                }

                Bookmark current = root ?? doc?.GetFirstBookmark();

                int numbookmarks = 0;
                while (current?.IsValid() == true)
                {
                    BookmarkItem item = new BookmarkItem(current, numbookmarks);
                    items.Add(item);
                    current = current.GetNext();
                    numbookmarks++;
                }

                if (root == null && numbookmarks == 1)
                {
                    IList<BookmarkItem> nextLevel = GetBookmarkList(doc?.GetFirstBookmark().GetFirstChild(), doc, true);
                    if (nextLevel?.Count > 0)
                        return nextLevel;
                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(
                    Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY,
                    Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                doc?.UnlockRead();
                if (!gotLock)
                {
                    OutlineReadingSemaphore?.Release();
                }
            }

            return items;
        }

        private Grid CreateBookmarkGrid(Bookmark root, IList<BookmarkItem> items)
        {
            Grid hostGrid = new Grid();
            if (_CurrentSize != null && !double.IsNaN(_CurrentSize.Width) && _CurrentSize.Width > 0)
            {
                hostGrid.Width = _CurrentSize.Width;
            }

            RowDefinition topRow = new RowDefinition();
            topRow.Height = new GridLength(1, GridUnitType.Auto);
            hostGrid.RowDefinitions.Add(topRow);
            RowDefinition mainRow = new RowDefinition();
            mainRow.Height = new GridLength(1, GridUnitType.Star);
            hostGrid.RowDefinitions.Add(mainRow);
            hostGrid.RenderTransform = new TranslateTransform();

            Grid titleGrid = new Grid();
            titleGrid.Style = this.Resources["TitleGridStyle"] as Style;
            titleGrid.ColumnDefinitions.Add(new ColumnDefinition() { Width = new GridLength(1, GridUnitType.Auto) });
            titleGrid.ColumnDefinitions.Add(new ColumnDefinition() { Width = new GridLength(1, GridUnitType.Star) });
            hostGrid.Children.Add(titleGrid);

            TextBlock title = new TextBlock();
            title.Style = this.Resources["HeaderTextBlockStyle"] as Style;
            title.Tapped += Title_Tapped;
            if (root == null)
            {
                try
                {
                    _RootTextBlock = title;
                    if (string.IsNullOrWhiteSpace(_DocumentTitle))
                    {
                        _PDFViewCtrl.DocLockRead();
                        try
                        {
                            PDFDoc doc = _PDFViewCtrl.GetDoc();
                            if (doc != null)
                            {
                                string fileName = doc.GetFileName();
                                _DocumentTitle = System.IO.Path.GetFileName(doc.GetFileName());
                                if (string.IsNullOrWhiteSpace(_DocumentTitle))
                                {
                                    PDFDocInfo docInfo = doc.GetDocInfo();
                                    if (docInfo != null)
                                    {
                                        _DocumentTitle = docInfo.GetTitle();
                                    }
                                }
                            }
                        }
                        finally
                        {
                            _PDFViewCtrl.DocUnlockRead();
                        }
                    }
                    if (string.IsNullOrWhiteSpace(DocumentTitle))
                    {
                        _DocumentTitle = ResourceHandler.GetString("Outline_DefaultHeader");
                    }
                    title.Text = DocumentTitle;
                }
                catch (Exception ex)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                }
            }
            else
            {
                title.Text = root.GetTitle();
                Button backButton = new Button();
                backButton.Style = this.Resources["BackwardNavigationButtonStyle"] as Style;
                backButton.Click += backButton_Click;
                titleGrid.Children.Add(backButton);
            }
            title.SetValue(Grid.ColumnProperty, 1);
            titleGrid.Children.Add(title);

            ScrollViewer itemScroller = new ScrollViewer();
            itemScroller.Content = CreateBookmarkStack(items);
            itemScroller.HorizontalScrollBarVisibility = ScrollBarVisibility.Hidden;
            itemScroller.HorizontalScrollMode = ScrollMode.Disabled;
            itemScroller.VerticalScrollBarVisibility = ScrollBarVisibility.Auto;
            itemScroller.VerticalScrollMode = ScrollMode.Auto;
            itemScroller.ZoomMode = ZoomMode.Disabled;
            _ScrollViewerPages.Add(itemScroller);

            // Animations
            AnimationCollection animations = new AnimationCollection();
            animations.EnterFromRight = EnterFromRight(hostGrid);
            animations.EnterFromLeft = EnterFromLeft(hostGrid);
            animations.LeaveToTheRight = LeaveToTheRight(hostGrid);
            animations.LeaveToTheLeft = LeaveToTheLeft(hostGrid);
            _Animations.Add(animations);
            itemScroller.SetValue(Grid.RowProperty, 1);
            hostGrid.Children.Add(itemScroller);

            return hostGrid;

        }

        private void Title_Tapped(object sender, Windows.UI.Xaml.Input.TappedRoutedEventArgs e)
        {
            TextBlock tb = e.OriginalSource as TextBlock;
            double aw = tb.ActualWidth;
            System.Diagnostics.Debug.WriteLine("Actual width is " + aw);
        }

        private StackPanel CreateBookmarkStack(IList<BookmarkItem> items)
        {
            StackPanel bookmarkStack = new StackPanel();
            if (_CurrentSize != null && !double.IsNaN(_CurrentSize.Width) && _CurrentSize.Width > 0)
            {
                bookmarkStack.Width = _CurrentSize.Width;
            }
            bookmarkStack.Padding = new Thickness(0, 0, 8, 0);
            foreach (BookmarkItem item in items)
            {
                Grid itemGrid = new Grid();
                itemGrid.ColumnDefinitions.Add(new ColumnDefinition() { Width = new GridLength(1, GridUnitType.Star) });
                itemGrid.ColumnDefinitions.Add(new ColumnDefinition() { Width = new GridLength(1, GridUnitType.Auto) });
                itemGrid.HorizontalAlignment = Windows.UI.Xaml.HorizontalAlignment.Stretch;
                itemGrid.Height = (double)this.Resources["ToolControlsItemHeight"];
                Button itemButton = new Button();
                TextBlock title = new TextBlock();
                title.TextTrimming = TextTrimming.CharacterEllipsis;
                title.Text = item.Title;
                itemButton.Content = title;

                itemButton.Tag = item;
                itemButton.Click += itemButton_Click;
                itemGrid.Children.Add(itemButton);
                if (item.HasChildren)
                {
                    Button childButton = new Button();
                    childButton.Tag = item;
                    childButton.Style = this.Resources["ForwardNavigationButtonStyle"] as Style;
                    childButton.SetValue(Grid.ColumnProperty, 1);
                    itemGrid.Children.Add(childButton);
                    childButton.Click += childButton_Click;
                }
                bookmarkStack.Children.Add(itemGrid);
            }

            return bookmarkStack;
        }

        private void itemButton_Click(object sender, RoutedEventArgs e)
        {
            Utilities.AnalyticsHandlerBase.CURRENT.SendEvent("Bookmark", "Navigated by Outline List");

            Button button = sender as Button;
            BookmarkItem item = button.Tag as BookmarkItem;
            Bookmark bookmark = item.SourceBookmark;
            pdftron.PDF.Action action = bookmark.GetAction();
            if (action.IsValid())
            {
                ActionParameter action_param = new ActionParameter(action);
                ActionUtils.ExecuteAction(action_param, this.PDFViewCtrl, NavigationStack);
            }
            if (ItemClicked != null)
            {
                ItemClicked(this, new ItemClickEventArgs());
            }
        }

        private async void childButton_Click(object sender, RoutedEventArgs e)
        {
            Button button = sender as Button;
            BookmarkItem item = button.Tag as BookmarkItem;
            await NavigateToChildAsync(item);
        }

        private async Task NavigateToChildAsync(BookmarkItem item)
        {
            Bookmark bookmark = item.SourceBookmark;
            _CurrentLocation.Add(item.ChildIndex);
            IList<BookmarkItem> items = await GetBookmarkListAsync(bookmark.GetFirstChild(), PDFViewCtrl.GetDoc());

            Grid ContentGrid = CreateBookmarkGrid(bookmark, items);
            _BookmarkPages.Add(ContentGrid);
            ContentPanel.Children.Add(ContentGrid);
            _Animations[_CurrentIndex].LeaveToTheLeft.Begin();
            _Animations.Last().EnterFromRight.Begin();
            _CurrentIndex++;
            if (_RestorePath != null && _RestorePath.Count > 0)
            {
                if (items.Count > _RestorePath[0])
                {
                    BookmarkItem nextItem = items[_RestorePath[0]];
                    _RestorePath.RemoveAt(0);
                    if (nextItem.HasChildren)
                    {
                        ContentGrid.Visibility = Visibility.Collapsed;
                        await NavigateToChildAsync(nextItem);
                    }
                }
            }
        }

        private void backButton_Click(object sender, RoutedEventArgs e)
        {
            _BookmarkPages.RemoveAt(_CurrentIndex);
            _ScrollViewerPages.RemoveAt(_CurrentIndex);
            _ScrollViewerPages.Last().VerticalScrollBarVisibility = ScrollBarVisibility.Hidden;
            _Animations.Last().LeaveToTheRight.Begin();
            _Animations.RemoveAt(_CurrentIndex);
            _Animations.Last().EnterFromLeft.Begin();
            _CurrentIndex--;
            _BookmarkPages[_CurrentIndex].Visibility = Windows.UI.Xaml.Visibility.Visible;
            if (_CurrentLocation.Count > 0)
            {
                _CurrentLocation.RemoveAt(_CurrentLocation.Count - 1);
            }
        }

        private Storyboard EnterFromRight(FrameworkElement target)
        {
            Storyboard sb = new Storyboard();
            String ID = target.Name.ToString();
            double from = _CurrentSize.Width;
            double to = 0;
            DoubleAnimation translate = new DoubleAnimation
            {
                From = from,
                To = to,
                Duration = new Duration(TimeSpan.FromMilliseconds(300)),
                RepeatBehavior = new RepeatBehavior { Count = 1 },
                EasingFunction = new QuadraticEase() { EasingMode = EasingMode.EaseOut },
            };

            target.Resources.Add(ID + "EnterFromRightAnimation", sb);
            Storyboard.SetTarget(translate, target);
            Storyboard.SetTargetName(translate, target.Name);
            Storyboard.SetTargetProperty(translate, "(UIElement.RenderTransform).(TranslateTransform.X)");
            sb.Completed += EnterLeftCompleted;
            sb.Children.Add(translate);

            return sb;
        }

        private void EnterLeftCompleted(object sender, object e)
        {
        }

        private Storyboard EnterFromLeft(FrameworkElement target)
        {
            Storyboard sb = new Storyboard();
            String ID = target.Name.ToString();
            double from = -_CurrentSize.Width;
            double to = 0;
            DoubleAnimation translate = new DoubleAnimation
            {
                From = from,
                To = to,
                Duration = new Duration(TimeSpan.FromMilliseconds(300)),
                RepeatBehavior = new RepeatBehavior { Count = 1 },
                EasingFunction = new QuadraticEase() { EasingMode = EasingMode.EaseOut },
            };

            target.Resources.Add(ID + "EnterFromLeftAnimation", sb);
            Storyboard.SetTarget(translate, target);
            Storyboard.SetTargetName(translate, target.Name);
            Storyboard.SetTargetProperty(translate, "(UIElement.RenderTransform).(TranslateTransform.X)");
            sb.Completed += EnterRightCompleted;
            sb.Children.Add(translate);

            return sb;
        }

        private void EnterRightCompleted(object sender, object e)
        {
            _ScrollViewerPages.Last().VerticalScrollBarVisibility = ScrollBarVisibility.Auto;
        }

        private Storyboard LeaveToTheRight(FrameworkElement target)
        {
            Storyboard sb = new Storyboard();
            String ID = target.Name.ToString();
            double from = 0;
            double to = _CurrentSize.Width;
            DoubleAnimation translate = new DoubleAnimation
            {
                From = from,
                To = to,
                Duration = new Duration(TimeSpan.FromMilliseconds(300)),
                RepeatBehavior = new RepeatBehavior { Count = 1 },
                EasingFunction = new QuadraticEase() { EasingMode = EasingMode.EaseOut },
            };

            target.Resources.Add(ID + "LeaveToTheRightAnimation", sb);
            Storyboard.SetTarget(translate, target);
            Storyboard.SetTargetName(translate, target.Name);
            Storyboard.SetTargetProperty(translate, "(UIElement.RenderTransform).(TranslateTransform.X)");
            sb.Children.Add(translate);
            sb.Completed += LeaveRightCompleted;

            return sb;
        }

        private void LeaveRightCompleted(object sender, object e)
        {
            if (ContentPanel.Children.Count > 0)
            {
                ContentPanel.Children.RemoveAt(ContentPanel.Children.Count - 1);
            }
        }

        private Storyboard LeaveToTheLeft(FrameworkElement target)
        {
            Storyboard sb = new Storyboard();
            String ID = target.Name.ToString();
            double from = 0;
            double to = -_CurrentSize.Width;
            DoubleAnimation translate = new DoubleAnimation
            {
                From = from,
                To = to,
                Duration = new Duration(TimeSpan.FromMilliseconds(300)),
                RepeatBehavior = new RepeatBehavior { Count = 1 },
                EasingFunction = new QuadraticEase() { EasingMode = EasingMode.EaseOut },
            };

            target.Resources.Add(ID + "LeaveToTheLeft", sb);
            Storyboard.SetTarget(translate, target);
            Storyboard.SetTargetName(translate, target.Name);
            Storyboard.SetTargetProperty(translate, "(UIElement.RenderTransform).(TranslateTransform.X)");
            sb.Children.Add(translate);
            sb.Completed += LeaveLeftCompleted;

            return sb;
        }

        private void LeaveLeftCompleted(object sender, object e)
        {
            if (_BookmarkPages.Count > 1)
            {
                _BookmarkPages[_BookmarkPages.Count - 2].Visibility = Windows.UI.Xaml.Visibility.Collapsed;
            }
        }

        /// <summary>
        /// Will go back if possible.
        /// </summary>
        /// <returns>Return true if it went back, false otherwise</returns>
        public bool GoBack()
        {
            return false;
        }

        private void RemoveItems(object sender, RoutedEventArgs e)
        {
            ViewModel.RemoveBookmarkNodes();
        }

        private DependencyObject FindChildControl<T>(DependencyObject control)
        {
            int childNumber = VisualTreeHelper.GetChildrenCount(control);
            for (int i = 0; i < childNumber; i++)
            {
                DependencyObject child = VisualTreeHelper.GetChild(control, i);
                FrameworkElement fe = child as FrameworkElement;
                // Not a framework element or is null
                if (fe == null) return null;

                if (child is T)
                {
                    // Found the control so return
                    return child;
                }
                else
                {
                    // Not found it - search children
                    DependencyObject nextLevel = FindChildControl<T>(child);
                    if (nextLevel != null)
                        return nextLevel;
                }
            }
            return null;
        }

        private void SetFocus(TextBox box)
        {
            box.LostFocus += TextBox_LostFocus;
            box.KeyDown += TextBox_KeyDown;
            box.Focus(FocusState.Programmatic);
            box.SelectAll();
            _OriginalText = box.Text;
        }

        void TextBox_LostFocus(object sender, RoutedEventArgs e)
        {            
            TextBox box = sender as TextBox;
            ViewModel.FinishEditingText();
            CloseTextBox(box);
        }

        void TextBox_KeyDown(object sender, Windows.UI.Xaml.Input.KeyRoutedEventArgs e)
        {
            if (e.Key == Windows.System.VirtualKey.Enter)
            {
                ViewModel.FinishEditingText();
                e.Handled = true;
            }
            if (e.Key == Windows.System.VirtualKey.Escape)
            {
                ViewModel.CancelEditingText(_OriginalText);
                TextBox box = sender as TextBox;
                box.Text = _OriginalText;
                e.Handled = true;
            }
        }

        private void CloseTextBox(TextBox box)
        {
            if (_KeyDownHandler != null)
            {
                box.KeyDown -= _KeyDownHandler;
            }
            if (_FocusLostHandler != null)
            {
                box.LostFocus -= _FocusLostHandler;
            }
        }

        private TextBox GetChildTextbox(FrameworkElement containerElement)
        {
            DependencyObject box = FindChildControl<TextBox>(containerElement);
            return box as TextBox;
        }

        private void ShowTextBox(BookmarkNode item)
        {
            FrameworkElement containerElement = OutlineTreeView.ContainerFromItem(item) as FrameworkElement;
            if (containerElement != null)
            {               
                TextBox box = GetChildTextbox(containerElement);
                if (box != null)
                {
                    SetFocus(box);
                }
            }
        }

        private void ViewModelFocusOnSelectedItemRequested(ViewModels.BookmarkNode node)
        {
            ShowTextBox(node);
        }

        public delegate void DocumentModifiedEventHandler(PDFDoc doc);
        public DocumentModifiedEventHandler DocumentModified;

        private void RaiseDocumentModified(PDFDoc doc)
        {
            if (DocumentModified != null)
            {
                DocumentModified(doc);
            }
        }

        private void ViewModelDocumentModified(PDFDoc doc)
        {
            RaiseDocumentModified(doc);
        }

        private void MultiSelectCheckBoxCustom_Checked(object sender, RoutedEventArgs e)
        {
            CheckBox box = sender as CheckBox;
            if (box.IsChecked != false)
            {
                ViewModel.CheckedAmount++;
            }
            else
            {
                ViewModel.CheckedAmount--;
            }
        }

        private async void OutlineTreeView_ExpandingAsync(Microsoft.UI.Xaml.Controls.TreeView sender, Microsoft.UI.Xaml.Controls.TreeViewExpandingEventArgs args)
        {
            BookmarkNode bookmarkNode = args.Item as BookmarkNode;

            if (bookmarkNode != null && ViewModel != null)
            {
                if (bookmarkNode.ChildrenLoadedFromDoc == false && !bookmarkNode.HasChildren)
                {
                    await ViewModel.ExpandTreeViewNodeAsync(bookmarkNode);
                }
                if (bookmarkNode.IsExpanded && bookmarkNode.HasChildren)
                {
                    await ViewModel.AdjustIsOpen(bookmarkNode);
                }
            }
        }

        private async void OutlineTreeView_CollapsedAsync(Microsoft.UI.Xaml.Controls.TreeView sender, Microsoft.UI.Xaml.Controls.TreeViewCollapsedEventArgs args)
        {
            BookmarkNode bookmarkNode = args.Item as BookmarkNode;
            if (!bookmarkNode.IsExpanded && bookmarkNode.HasChildren)
            {
                await ViewModel.AdjustIsOpen(bookmarkNode);
            }
        }
 
        /// <summary>
        /// Unsubscribes the ViewModel from events and Disposes the ViewModel 
        /// </summary>
        public void Dispose()
        {

            if (OutlineTreeView != null)
            {
                OutlineTreeView.ItemInvoked -= OutlineTreeView_ItemInvoked;
                OutlineTreeView.Expanding -= OutlineTreeView_ExpandingAsync;
                OutlineTreeView.Collapsed -= OutlineTreeView_CollapsedAsync;
            }

            if (ViewModel != null)
            {
                ViewModel.FocusOnSelectedItemRequested -= ViewModelFocusOnSelectedItemRequested;
                ViewModel.DocumentModified -= ViewModelDocumentModified;
                ViewModel.Dispose();
                ViewModel = null;
            }

            this.Loaded -= Outline_Loaded;
            this.Unloaded -= Outline_Unloaded;           
        }
    }
}
