﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Runtime.InteropServices.WindowsRuntime;
using System.Threading;
using System.Threading.Tasks;

using Windows.Foundation;
using Windows.Foundation.Collections;
using Windows.UI;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Controls.Primitives;
using Windows.UI.Xaml.Data;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Navigation;
using Windows.Data.Json;
using UIRect = Windows.Foundation.Rect;

using pdftron.Common;
using pdftron.PDF.Tools.Controls.ViewModels;
using pdftron.PDF.Tools.Controls.ViewModels.Common;
using pdftron.PDF.Tools.Utilities;


namespace pdftron.PDF.Tools.Controls
{
    public sealed partial class ReflowView : UserControl, IDisposable
    {
        private const string BLANK_PAGE_STRING_PRE_SCRIPT = "<?xml version=\"1.0\" encoding=\"UTF-8\"?> <html><head xmlns = \"http://www.w3.org/1999/xhtml\" > <style>a:link{color:#24a9db}</style>";
        private const string BLANK_PAGE_STRING_POST_SCRIPT = "</head> <body class=\"big\"></body> </html>";

        private string _BlankPageWithKeyEventsString;
        private int _StartingReflowPage = -1;

        private Dictionary<int, string> HtmlPages = new Dictionary<int, string>();
        private ReflowViewModel _ReflowViewModel;
        private string _lastSelectedText = string.Empty;
        private JsonObject _selectedAnnot = null;
        private Windows.Foundation.Point _contextMenuLastPos; // last position of the context menu
        private ContextMenuControl _contextMenu = null;
        private ToolManager _toolManager = null;
        private static Windows.ApplicationModel.Resources.ResourceLoader _ResourceLoader;

        public event TappedEventHandler ReflowTapped;

        public event PropertyChangedEventHandler PropertyChanged;
        public void OnPropertyChanged(string propertyName)
        {
            PropertyChanged?.Invoke(this, new PropertyChangedEventArgs(propertyName));
        }

        public delegate void PageChangedDelegate(int pageNumber, int totalPages);

        private PageChangedDelegate _pageChanged;
        public event PageChangedDelegate PageChanged
        {
            add
            {
                if (_pageChanged?.GetInvocationList().Select(x => x.Method).Contains(value.Method) != true)
                    _pageChanged += value;
            }
            remove
            {
                if (_pageChanged != null)
                {
                    PageChangedDelegate eventMethod = (PageChangedDelegate)Array.Find(
                        _pageChanged.GetInvocationList(),
                        x => x.Method == value.Method);
                    if (eventMethod != null)
                        _pageChanged -= eventMethod;
                }
            }
        }

        public delegate void DocumentChangedDelegate();

        private DocumentChangedDelegate _docChanged;

        /// <summary>
        /// Raised when the current reflowed document has been modified
        /// </summary>
        public event DocumentChangedDelegate DocumentChanged
        {
            add
            {
                if (_docChanged?.GetInvocationList().Select(x => x.Method).Contains(value.Method) != true)
                    _docChanged += value;
            }
            remove
            {
                if (_docChanged != null)
                {
                    DocumentChangedDelegate eventMethod = (DocumentChangedDelegate)Array.Find(
                        _docChanged.GetInvocationList(),
                        x => x.Method == value.Method);
                    if (eventMethod != null)
                        _docChanged -= eventMethod;
                }
            }
        }

        enum ReflowAnnotType
        {
            Unknown = -1,
            Highlight = 8,
            Underline = 9,
            Squiggly = 10,
            Strikeout = 11
        }

        public ReflowView()
        {
            Init(null);
        }

        public ReflowView(PDFDoc doc)
        {
            Init(doc);
        }

        private void Init(PDFDoc doc)
        {
            this.InitializeComponent();

            this.Loaded += ReflowView_Loaded;
            this.PointerWheelChanged += ReflowView_PointerWheelChanged;

            _BlankPageWithKeyEventsString = BLANK_PAGE_STRING_PRE_SCRIPT + ReflowViewModel.KEYDOWN_STRING + ReflowViewModel.TAP_STRING + BLANK_PAGE_STRING_POST_SCRIPT;

            if (doc == null)
                ViewModel = new ReflowViewModel();
            else
                ViewModel = new ReflowViewModel(doc);

            ViewModel.ReflowContentUpdated += ViewModel_ReflowContentUpdated;
            ViewModel.PageChanged += ViewModel_PageChanged;
            this.DataContext = ViewModel;

            ReflowWebView.Loaded += ReflowWebView_Loaded;
            ReflowWebView.ScriptNotify += ReflowWebView_ScriptNotify;
            ReflowWebView.NavigationCompleted += ReflowWebView_NavigationCompleted;

            MainGrid.Tapped += MainGrid_Tapped;
            PreviousButtonHorizontal.Tapped += NavigationButtonHorizontal_Tapped;
            NextButtonHorizontal.Tapped += NavigationButtonHorizontal_Tapped;

            _ResourceLoader = Windows.ApplicationModel.Resources.ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources");
        }

        #region Public Properties

        public ReflowViewModel ViewModel
        {
            get { return _ReflowViewModel; }
            set
            {
                ReflowViewModel oldVM = _ReflowViewModel;
                if (_ReflowViewModel != value)
                {
                    _ReflowViewModel = value;
                    if (oldVM != null)
                    {
                        oldVM.PropertyChanged -= ViewModel_PropertyChanged;
                    }
                    if (_ReflowViewModel != null)
                    {
                        _ReflowViewModel.PropertyChanged += ViewModel_PropertyChanged;
                    }
                }
            }
        }

        /// <summary>
        /// Get or Set the current displayed page
        /// </summary>
        public int CurrentPage
        {
            get
            {
                if (ViewModel != null)
                {
                    return ViewModel.CurrentPageNumber;
                }
                return 0;
            }

            set 
            {
                if (ViewModel != null)
                    ViewModel.SetPageNumber(value);
            }
        }

        /// <summary>
        /// Set weather or not to include images in the reflowed page
        /// </summary>
        public bool IncludeImage
        {
            get 
            {
                return ViewModel.IncludeImages; 
            }
            set 
            { 
                ViewModel.IncludeImages = value; 
            }
        }

        /// <summary>
        /// Set weather or not to display the change page buttons (previous and next page) on both sides
        /// </summary>
        public bool ShowChangePageButtons { get; set; } = true;

        public ToolManager ToolManager
        {
            get { return _toolManager; }
            set
            { 
                _toolManager = value;
            }
        }

        /// <summary>
        /// Set weather or not editing in reflow is enabled or not
        /// </summary>
        public bool IsEditingEnabled
        {
            get { return ViewModel.IsEditingEnabled; }
            set { ViewModel.IsEditingEnabled = value; }
        }

        /// <summary>
        /// Excludes images that are detected as background images
        /// </summary>
        public bool SetHideBackgroundImages
        {
            get { return ViewModel.SetHideBackgroundImages; }
            set { ViewModel.SetHideBackgroundImages = value; }
        }

        /// <summary>
        /// Excludes images that appear under text
        /// </summary>
        public bool SetHideImagesUnderText
        {
            get { return ViewModel.SetHideImagesUnderText; }
            set { ViewModel.SetHideImagesUnderText = value; }
        }

        /// <summary>
        /// Excludes text that is on top of an image
        /// </summary>
        public bool SetDoNotReflowTextOverImages
        {
            get { return ViewModel.SetDoNotReflowTextOverImages; }
            set { ViewModel.SetDoNotReflowTextOverImages = value; }
        }

        #endregion

        #region Events handling

        private async void ReflowView_Loaded(object sender, RoutedEventArgs e)
        {
            if (ViewModel != null || _StartingReflowPage > 0)
                await ViewModel.RequestPageAsync(_StartingReflowPage);
        }

        private void NavigationButtonHorizontal_Tapped(object sender, TappedRoutedEventArgs e)
        {
            e.Handled = true;
        }

        private void MainGrid_Tapped(object sender, TappedRoutedEventArgs e)
        {
            if (e.PointerDeviceType != Windows.Devices.Input.PointerDeviceType.Mouse)
            {
                ReflowTapped?.Invoke(this, new TappedRoutedEventArgs());
            }
        }

        private void ReflowWebView_Loaded(object sender, RoutedEventArgs e)
        {
            ReflowWebView.Focus(FocusState.Programmatic);
        }

        private void ViewModel_ReflowContentUpdated(ReflowItem item)
        {
            DebugLogHelper.CallSarted();

            if (ViewModel == null)
                return;

            if (item.HasReflowedContent)
            {
                ReflowWebView.NavigateToString(ViewModel.ReflowedContent);
            }
            else
            {
                ReflowWebView.NavigateToString(_BlankPageWithKeyEventsString);
            }

            DebugLogHelper.CallEnded();
        }

        private void ViewModel_PageChanged(int pageNumber, int totalPages)
        {
            _pageChanged?.Invoke(pageNumber, totalPages);
        }

        private void ViewModel_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            if (e.PropertyName.Equals("CurrentPageNumber", StringComparison.OrdinalIgnoreCase))
            {
                string progressText = _ResourceLoader.GetString("pdftron_Controls_Reflow_WorkingOnPage");
                ProgressTextBlock.Text = string.Format(progressText, ViewModel.CurrentPageNumber);
            }
            else if (e.PropertyName.Equals("PostProcessingMode", StringComparison.OrdinalIgnoreCase))
            {
                UpdateColors();
            }
        }

        private void ReflowView_PointerWheelChanged(object sender, PointerRoutedEventArgs e)
        {
            int delta = e.GetCurrentPoint(this).Properties.MouseWheelDelta;
            if (delta < 0)
            {
                ViewModel.CMDNextPage.Execute(null);
            }
            else if (delta > 0)
            {
                ViewModel.CMDPreviousPage.Execute(null);
            }
        }

        private async void ReflowWebView_ScriptNotify(object sender, NotifyEventArgs e)
        {
            var cmd = GetReflowCommand(e.Value);
            if (string.IsNullOrWhiteSpace(cmd))
                return;

            HideContextMenu();

            Debug.WriteLine("DEBUG ====> ReflowWebView_ScriptNotify: " + cmd);

            string[] result;

            switch (cmd)
            {
                case "addAnnotRequested":
                    AppCenterHandlerBase.CURRENT.SendEvent("Reflow", new Dictionary<string, string>() {["Annotate"] = "add"});

                    await HandleAddAnnotAsync(e.Value);
                    break;

                case "removeSelectedAnnot":
                    result = e.Value.Split(" ");
                    if (result.Length <= 1 || string.IsNullOrWhiteSpace(result[1]))
                        return;

                    AppCenterHandlerBase.CURRENT.SendEvent("Reflow", new Dictionary<string, string>() { ["Annotate"] = "delete" });

                    await DeleteAnnotationAsync(result[1]);
                    break;

                case "annotSelected":

                    result = e.Value.Split(" ");
                    if (result.Length <= 1 || string.IsNullOrWhiteSpace(result[1]))
                        return;

                    if (IsEditingEnabled == false)
                        return;

                    AppCenterHandlerBase.CURRENT.SendEvent("Reflow", new Dictionary<string, string>() { ["Annotate"] = "edit" });

                    HandleAnnotEdit(result[1]);
                    break;

                case "tap":
                case "rightClick":
                    if (_contextMenu != null)
                        _contextMenu.Hide();

                    ReflowTapped?.Invoke(this, new TappedRoutedEventArgs());
                    return;

                case "mouseup:":

                    result = e.Value.Split(" ");
                    if (result.Length <= 1 || string.IsNullOrWhiteSpace(result[1]))
                        return;

                    if (IsEditingEnabled == false)
                        return;

                    var text = e.Value.Remove(0, 9); // remove "mouseup:" from the message
                    HandleTextSelected(text);
                    break;

                case "key:":
                    HandleKeyDown(e.Value);
                    break;

                default:
                    ViewModel.NavigateInternalLink(e.Value);
                    break;
            }
        }

        private async void ReflowWebView_NavigationCompleted(WebView sender, WebViewNavigationCompletedEventArgs args)
        {
            await LoadAnnotationsAsync();
        }

        private async void _contextMenu_ItemClicked(string title)
        {
            var annotType = ReflowAnnotType.Unknown;

            if (title.Equals("copy text", StringComparison.OrdinalIgnoreCase))
            {
                UtilityFunctions.CopySelectedTextToClipBoard(_lastSelectedText);
            }
            else if (title.Equals("highlight", StringComparison.OrdinalIgnoreCase))
            {
                annotType = ReflowAnnotType.Highlight;
            }
            else if (title.Equals("underline", StringComparison.OrdinalIgnoreCase))
            {
                annotType = ReflowAnnotType.Underline;
            }
            else if (title.Equals("strikeout", StringComparison.OrdinalIgnoreCase))
            {
                annotType = ReflowAnnotType.Strikeout;
            }
            else if (title.Equals("squiggly", StringComparison.OrdinalIgnoreCase))
            {
                annotType = ReflowAnnotType.Squiggly;
            }
            else if (title.Equals("Style", StringComparison.OrdinalIgnoreCase))
            {
                ToolManager.RaiseOnReflowEditRequestedEvent();

                if (!IsValidSub())
                    return;

                HandleStyle();
            }
            else if (title.Equals("delete", StringComparison.OrdinalIgnoreCase))
            {
                ToolManager.RaiseOnReflowEditRequestedEvent();

                if (!IsValidSub())
                    return;

                await HandleDeleteAnnotAsync();
            }

            if (annotType != ReflowAnnotType.Unknown)
            {
                ToolManager.RaiseOnReflowEditRequestedEvent();

                if (!IsValidSub())
                    return;

                await RequestAddAnnotationAsync(annotType);
            }

            HideContextMenu();
        }

        /// <summary>
        /// Check if a valid pro subscription
        /// </summary>
        /// <returns></returns>
        private bool IsValidSub()
        {
#if SUBSCRIPTIONS

            return Viewer.ViewerSettings.IsSubscriptionActivated;
#endif
            return true;
        }

        private async void Display_AnnotationPropertyDialogClosed(AnnotationPropertyDialog dialog)
        {
            var color = UtilityFunctions.ConvertRgbToHex(dialog.ViewModel.MainColor, true);

            await EditAnnotationAsync(color, dialog.ViewModel.Opacity);
        }

        #endregion

        private void HandleAnnotEdit(string data)
        {
            if (string.IsNullOrEmpty(data))
                return;

            JsonObject jsonResult;
            JsonObject.TryParse(data, out jsonResult);

            if (jsonResult != null)
            {
                var editAnnotJson = jsonResult.ToString();
                if (string.IsNullOrWhiteSpace(editAnnotJson))
                    return;

                _selectedAnnot = jsonResult;

                Dictionary<string, string> textSelectOptionsDict = new Dictionary<string, string>();

                // TODO: add copy text when selecting annotation
                //textSelectOptionsDict["copy text"] = _ResourceLoader.GetString("pdftron_Tool_TextSelect_Option_Copy");
                textSelectOptionsDict["Style"] = _ResourceLoader.GetString("pdftron_Tool_EditOptiob_Style");
                textSelectOptionsDict["delete"] = _ResourceLoader.GetString("pdftron_Tool_EditOption_Delete");

                if (_contextMenu == null)
                {
                    _contextMenu = new ContextMenuControl(ReflowWebView, textSelectOptionsDict);
                    _contextMenu.ItemClicked += _contextMenu_ItemClicked;
                }

                ShowContextMenu();
            }
        }            

        /// <summary>
        /// Handle mouse up evenet, generally after selecting text
        /// </summary>
        /// <param name="data">Mouse up event data to be looked at</param>
        private void HandleTextSelected(string data)
        {
            if (string.IsNullOrEmpty(data))
            {
                HideContextMenu();
                return;
            }

            _lastSelectedText = data;

            Dictionary<string, string> textSelectOptionsDict = new Dictionary<string, string>();

            textSelectOptionsDict["copy text"] = _ResourceLoader.GetString("pdftron_Tool_TextSelect_Option_Copy");
            textSelectOptionsDict["highlight"] = _ResourceLoader.GetString("pdftron_Tool_TextSelect_Option_Highlight");
            textSelectOptionsDict["strikeout"] = _ResourceLoader.GetString("pdftron_Tool_TextSelect_Option_Strikeout");
            textSelectOptionsDict["underline"] = _ResourceLoader.GetString("pdftron_Tool_TextSelect_Option_Underline");
            //NOTE: Text decoration not supported by WebView (Edge). Ie.: annotNode.style.textDecorationStyle = "wavy"
            //textSelectOptionsDict["squiggly"] = ResourceHandler.GetString("TextSelect_Option_Squiggly");

            if (_contextMenu == null)
            {
                _contextMenu = new ContextMenuControl(ReflowWebView, textSelectOptionsDict);
                _contextMenu.ItemClicked += _contextMenu_ItemClicked;
            }

            ShowContextMenu();
        }

        private void HandleKeyDown(string data)
        {
            if (string.IsNullOrEmpty(data))
                return;

            string[] keysplit = data.Split(new char[] { ' ' });
            if (keysplit.Length == 2)
            {
                string keyString = keysplit[1];
                int num = 0;

                if (int.TryParse(keyString, out num))
                {
                    if (num == 37)
                    {
                        ViewModel.CMDPreviousPage.Execute(null);
                    }
                    else if (num == 39)
                    {
                        ViewModel.CMDNextPage.Execute(null);
                    }
                }

            }
        }

        private void HandleStyle()
        {
            if (_selectedAnnot == null)
                return;

            try
            {
                var sColor = _selectedAnnot["color"].GetString();
                var dOpacity = _selectedAnnot["opacity"].GetNumber();
                var iType = (int)_selectedAnnot["type"].GetNumber();

                var color = UtilityFunctions.ConvertHexToSettingsColor(sColor);

                AnnotationPropertyDialog dialog = new AnnotationPropertyDialog();
                dialog.HasThickness = false;
                dialog.ViewModel.MainColor = color;
                dialog.ViewModel.Opacity = dOpacity;
                dialog.ViewModel.AssociatedTool = GetToolFromReflowAnnot((ReflowAnnotType)iType);

                AnnotationPropertiesDisplay display = AnnotationPropertiesDisplay.CreateAnnotationPropertiesDisplay(
                    new UIRect(_contextMenuLastPos.X, _contextMenuLastPos.Y, 480, 480),
                    dialog,
                    AnnotationPropertiesDisplay.Placement.PreferBeside);
                display.IsPopupLightDismissed = true;
                display.PlacePopup();

                display.AnnotationPropertyDialogClosed += Display_AnnotationPropertyDialogClosed;
            }
            catch (Exception ignored)
            { }
        }

        private async Task HandleDeleteAnnotAsync()
        {
            if (_selectedAnnot == null)
                return;

            try
            {
                await ReflowWebView.InvokeScriptAsync("eval", new string[] { "javascript:ReflowJS.removeSelectedAnnot();" });
            }
            catch (Exception ex)
            {
                AppCenterHandlerBase.CURRENT.TrackError(ex);
            }
        }

        private void ShowContextMenu()
        {
            if (_contextMenu != null)
            {
                // Get the current mouse position
                var pointerPosition = Windows.UI.Core.CoreWindow.GetForCurrentThread().PointerPosition;
                var x = pointerPosition.X - Window.Current.Bounds.X;
                var y = pointerPosition.Y - Window.Current.Bounds.Y;

                _contextMenuLastPos = new Windows.Foundation.Point(x, y);
                _contextMenu.Show(_contextMenuLastPos);
            }
        }

        private void HideContextMenu()
        {
            if (_contextMenu != null)
            {
                _contextMenu.Hide();
                _contextMenu.ItemClicked -= _contextMenu_ItemClicked;
                _contextMenu = null;
            }
        }

        private void UpdateColors()
        {
            if (ViewModel != null)
            {
                try
                {
                    switch (ViewModel.PostProcessingMode)
                    {
                        case PDFRasterizerColorPostProcessMode.e_postprocess_gradient_map:
                            MainGrid.Background = new SolidColorBrush(ViewModel.PostProcessingWhiteColor);
                            ReflowWebView.DefaultBackgroundColor = ViewModel.PostProcessingWhiteColor;
                            ProgressRing.Foreground = new SolidColorBrush(ViewModel.PostProcessingBlackColor);
                            ProgressTextBlock.Foreground = new SolidColorBrush(ViewModel.PostProcessingBlackColor);
                            break;

                        case PDFRasterizerColorPostProcessMode.e_postprocess_night_mode:
                        case PDFRasterizerColorPostProcessMode.e_postprocess_invert:
                            MainGrid.Background = (SolidColorBrush)Resources["NightModeBackgroundBrush"];
                            ReflowWebView.DefaultBackgroundColor = (Color)Resources["NightModeBackgroundColor"];
                            ProgressRing.Foreground = (SolidColorBrush)Resources["NightModeForegroundBrush"];
                            ProgressTextBlock.Foreground = (SolidColorBrush)Resources["NightModeForegroundBrush"];
                            break;

                        default:
                            MainGrid.Background = (SolidColorBrush)Resources["DayModeBackgroundBrush"];
                            ReflowWebView.DefaultBackgroundColor = (Color)Resources["DayModeBackgroundColor"];
                            ProgressRing.Foreground = (SolidColorBrush)Resources["DayModeForegroundBrush"];
                            ProgressTextBlock.Foreground = (SolidColorBrush)Resources["DayModeForegroundBrush"];
                            break;
                    }
                }
                catch (Exception e)
                {
                    PDFNetException pdfe = new PDFNetException(e.HResult);
                    if (!pdfe.IsPDFNetException)
                    {
                        throw;
                    }
                    else
                    {
                        Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(e));
                    }
                }
            }
        }

        /// <summary>
        /// Returned value from JS events have a command name at the beggining separated by space " "
        /// </summary>
        /// <param name="arg">The argument returned from a script notify event</param>
        /// <returns>The command string if valid</returns>
        private string GetReflowCommand(string arg)
        {
            if (string.IsNullOrEmpty(arg))
                return "";

            var cmd = arg.Split(" ");
            return cmd[0];
        }

        private ToolType GetToolFromReflowAnnot(ReflowAnnotType type)
        {
            switch (type)
            {
                case ReflowAnnotType.Highlight:
                    return ToolType.e_text_highlight;
                case ReflowAnnotType.Squiggly:
                    return ToolType.e_text_squiggly;
                case ReflowAnnotType.Underline:
                    return ToolType.e_text_underline;
                case ReflowAnnotType.Strikeout:
                    return ToolType.e_text_strikeout;
            }

            return ToolType.e_none;
        }

        #region JS call

        private async Task LoadAnnotationsAsync()
        {
            if (ReflowWebView == null)
                return;

            if (ViewModel != null)
                ViewModel.HasReflowedContent = true;

            try
            {
                string annots = ViewModel.GetAnnotations();

                await ReflowWebView.InvokeScriptAsync("eval", new string[] { "javascript:ReflowJS.loadAnnotations(" + annots + ");" });
            }
            catch (Exception ex)
            {
                AppCenterHandlerBase.CURRENT.TrackError(ex);
            }
        }

        private async Task RequestAddAnnotationAsync(ReflowAnnotType annotType)
        {
            if (ReflowWebView == null)
                return;

            SettingsColor settingsColor = new SettingsColor(255, 0, 0, false); // default color
            double opacity = 1; // default opacity

            switch (annotType)
            {
                case ReflowAnnotType.Highlight:
                    settingsColor = Settings.GetShapeColor(ToolType.e_text_highlight, false);
                    opacity = Settings.GetOpacity(ToolType.e_text_highlight);
                    break;

                case ReflowAnnotType.Strikeout:
                    settingsColor = Settings.GetShapeColor(ToolType.e_text_strikeout, false);
                    opacity = Settings.GetOpacity(ToolType.e_text_strikeout);
                    break;

                case ReflowAnnotType.Squiggly:
                    settingsColor = Settings.GetShapeColor(ToolType.e_text_squiggly, false);
                    opacity = Settings.GetOpacity(ToolType.e_text_squiggly);
                    break;

                case ReflowAnnotType.Underline:
                    settingsColor = Settings.GetShapeColor(ToolType.e_text_underline, false);
                    opacity = Settings.GetOpacity(ToolType.e_text_underline);
                    break;
            }

            var color = UtilityFunctions.ConvertRgbToHex(settingsColor);
            int pagePosition = ViewModel.CurrentPage.PageNumber;

            try
            {
                await ReflowWebView.InvokeScriptAsync("eval", new string[]
                {
                    "javascript:ReflowJS.addAnnot("
                    + ((int)annotType).ToString() + "," 
                    + "\"" + color.ToString() + "\"" + ","
                    + ((double)opacity).ToString() + ","
                    + ((int)pagePosition).ToString() + ");"
                }); ;
            }
            catch (Exception ex)
            {
                AppCenterHandlerBase.CURRENT.TrackError(ex);
            }
        }

        /// <summary>
        /// Handle adding a new annotation to the current selected reflowed page
        /// </summary>
        /// <param name="data"></param>
        private async Task HandleAddAnnotAsync(string data)
        {
            if (string.IsNullOrEmpty(data))
                return;

            var annotReq = data.Split(" ");

            JsonObject jsonResult;
            JsonObject.TryParse(annotReq[1], out jsonResult);

            if (jsonResult != null)
            {
                var addAnnotJson = jsonResult.ToString();
                if (string.IsNullOrWhiteSpace(addAnnotJson))
                    return;                               

                try
                {
                    var result = ViewModel.AddAnnot(addAnnotJson);
                    if (string.IsNullOrWhiteSpace(result))
                        return;

                    var page = ViewModel.Doc.GetPage(CurrentPage);
                    var annot = page.GetAnnot(page.GetNumAnnots() - 1);

                    ToolManager.RaiseAnnotationAddedEvent(annot, CurrentPage);
                    _docChanged?.Invoke();

                    await ReflowWebView.InvokeScriptAsync("eval", new string[] { "javascript:ReflowJS.setAnnotFromResult(" + result + ");" });
                }
                catch (Exception ex)
                {
                    AppCenterHandlerBase.CURRENT.TrackError(ex);
                }
            }
        }

        private async Task DeleteAnnotationAsync(string data)
        {
            if (string.IsNullOrEmpty(data))
                return;

            try
            {
                JsonObject jsonRslt;
                JsonObject.TryParse(data, out jsonRslt);

                string reflowId = string.Empty;
                string jsId = string.Empty;

                if (jsonRslt["id"].ValueType == JsonValueType.String)
                {
                    reflowId = jsonRslt["id"].GetString();
                    jsId = "\"" + reflowId + "\"";
                }
                else if (jsonRslt["id"].ValueType == JsonValueType.Number)
                {
                    jsId = reflowId = jsonRslt["id"].GetNumber().ToString();
                }

                var deleteResult = ViewModel.DeleteAnnot(reflowId);
                if (string.IsNullOrEmpty(deleteResult))
                    return;

                _docChanged?.Invoke();

                await ReflowWebView.InvokeScriptAsync("eval", new string[] { "javascript:ReflowJS.removeSelectedAnnotFromResult(" + jsId + ");" });

                /* NOTE: due to an unknown js issue some annotations are not properly removed in runtime so for now
                 * just update the reflow page again */
                await ViewModel.RequestPageAsync(CurrentPage);
            }
            catch (Exception ex)
            {
                AppCenterHandlerBase.CURRENT.TrackError(ex);
            }
        }

        private async Task EditAnnotationAsync(string color, double opacity)
        {
            if (ReflowWebView == null)
                return;

            try
            {
                var jsonId = _selectedAnnot["id"];

                _selectedAnnot["color"] = JsonValue.CreateStringValue(color);
                _selectedAnnot["opacity"] = JsonValue.CreateNumberValue(opacity);
                if (_selectedAnnot["id"].ValueType == JsonValueType.Number)
                    _selectedAnnot["id"] = JsonValue.CreateStringValue(jsonId.ToString());

                var jsonStr = _selectedAnnot.ToString();

                var result = ViewModel.EditAnnot(jsonStr);
                if (string.IsNullOrEmpty(result))
                    return;

                _docChanged?.Invoke();

                // Set the annot style
                await ReflowWebView.InvokeScriptAsync("eval", new string[] 
                {
                    "javascript:ReflowJS.setSelectedAnnotStyle(" + "\"#" + color + "\"" + "," + opacity + ");"
                });

                string scriptResult = string.Empty;

                if (jsonId.ValueType == JsonValueType.Number)
                {
                    scriptResult = "javascript:ReflowJS.setSelectedAnnotStyleFromResult(" + jsonId.GetNumber() + ");";
                }
                else if (jsonId.ValueType == JsonValueType.String)
                {
                    scriptResult = "javascript:ReflowJS.setSelectedAnnotStyleFromResult(\"" + _selectedAnnot["id"].GetString() + "\");";
                }

                await ReflowWebView.InvokeScriptAsync("eval", new string[] { scriptResult });
            }
            catch (Exception ex)
            {
                AppCenterHandlerBase.CURRENT.TrackError(ex);
            }
        }

        #endregion

        #region Public Functions

        public void SetColorPostProcessingMode(PDFRasterizerColorPostProcessMode mode, Windows.UI.Color whiteColor, Windows.UI.Color blackColor)
        {
            if (_ReflowViewModel == null)
                return;

            _ReflowViewModel.SetColorPostProcessingMode(mode, whiteColor, blackColor);
        }

        public void SetFocus(FocusState focusState)
        {
            ReflowWebView.Focus(focusState);
        }

        public void Dispose()
        {
            if (ViewModel != null)
            {
                ViewModel.Dispose();
            }

            HideContextMenu();
        }

        #endregion Public functions
    }
}
