﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Runtime.InteropServices.WindowsRuntime;
using Windows.Foundation;
using Windows.Foundation.Collections;
using Windows.UI.Input;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Controls.Primitives;
using Windows.UI.Xaml.Data;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Navigation;
using Windows.UI.Xaml.Shapes;

using UIPoint = Windows.Foundation.Point;

namespace pdftron.PDF.Tools.Controls.Resources
{
    public sealed partial class SnapResizer : UserControl
    {
        static DependencyProperty OffsetsProperty =
         DependencyProperty.Register("Offsets", typeof(List<double>),
         typeof(SnapResizer), new PropertyMetadata(new List<double> { 0 }, OnOffsetsChanged));

        private static void OnOffsetsChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            SnapResizer resizer = d as SnapResizer;
            if (resizer != null)
            {
                resizer.UpdatePosition();
            }
        }

        public List<double> Offsets
        {
            get { return (List<double>)GetValue(OffsetsProperty); }
            set { SetValue(OffsetsProperty, value); }
        }

        static DependencyProperty PositionProperty =
         DependencyProperty.Register("Position", typeof(int),
         typeof(SnapResizer), new PropertyMetadata(0, OnPositionChanged));

        private static void OnPositionChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            SnapResizer resizer = d as SnapResizer;
            if (resizer != null)
            {
                resizer.UpdatePosition();
            }
        }

        public int Position
        {
            get { return (int)GetValue(PositionProperty); }
            set { SetValue(PositionProperty, value); }
        }

        public event RoutedEventHandler PositionUpdated;

        public SnapResizer()
        {
            this.InitializeComponent();
            //ResizeHandle.Height = HandleHeight;
            this.SizeChanged += SnapResizer_SizeChanged;
            SetupInteraction();
        }

        private void SnapResizer_SizeChanged(object sender, SizeChangedEventArgs e)
        {
            if (e.NewSize.Width > 0 && e.NewSize.Height > 0)
            {
                UpdatePosition(e.NewSize.Width, e.NewSize.Height);
            }
        }

        private void UpdatePosition()
        {
            if (ActualWidth > 0 && ActualHeight > 0)
            {
                UpdatePosition(ActualWidth, ActualHeight);
            }
        }

        private void UpdatePosition(double width, double height)
        {
            if (Offsets == null || Offsets.Count == 0 || Position >= Offsets.Count)
            {
                ResizeHandle.Visibility = Visibility.Collapsed;
                return;
            }

            ResizeHintHost.Children.Clear();
            foreach (double offset in Offsets)
            {
                Rectangle hintRect = new Rectangle();
                hintRect.Style = Resources["HorizontalResieHintRectStyle"] as Style;
                ResizeHintHost.Children.Add(hintRect);
                hintRect.Margin = new Thickness(offset - ((double)Resources["ResizeHintSize"] / 2.0), 0, 0, 0);
            }

            ResetView();
        }


        #region Movement

        private int _ResizerPosition = 0;

        private bool _HasContactPoint = false;
        private bool _ContactPointIsInsideRadius = false;
        private uint _CurrentContactPoint = 0;
        private double _HandleHalfSize = 0;
        private double _ResizerHalfSize = 0;

        private void SetupInteraction()
        {
            ResizeHandle.PointerPressed += ResizeHandle_PointerPressed;
            ResizeHandle.PointerMoved += ResizeHandle_PointerMoved;
            ResizeHandle.PointerReleased += ResizeHandle_PointerReleased;
            ResizeHandle.PointerCaptureLost += ResizeHandle_PointerCaptureLost;
            ResizeHandle.PointerCanceled += ResizeHandle_PointerCanceled;
            ResizeHandle.PointerExited += ResizeHandle_PointerExited;
        }


        private void ResizeHandle_PointerPressed(object sender, PointerRoutedEventArgs e)
        {
            if (Offsets == null || Offsets.Count == 0 || Position >= Offsets.Count || e.Pointer.PointerDeviceType != Windows.Devices.Input.PointerDeviceType.Mouse)
            {
                return;
            }
            else
            {
                UIPoint downpoint = e.GetCurrentPoint(ResizeHandle).Position;
                _ContactPointIsInsideRadius = IsInsideHandleRadius(downpoint);
                if (_ContactPointIsInsideRadius)
                {
                    _ResizerPosition = Position;
                    _CurrentContactPoint = e.Pointer.PointerId;
                    _HasContactPoint = true;
                    HorizontalResizer.Visibility = Visibility.Visible;
                    ResizeHintHost.Visibility = Visibility.Visible;
                    ResizeHandle.CapturePointer(e.Pointer);
                    ResizeHintHost.IsHitTestVisible = true;
                    _HandleHalfSize = (double)Resources["HandleTargetThickness"] / 2.0;
                    _ResizerHalfSize = (double)Resources["ResizerSize"] / 2.0;
                    if (e.Pointer.PointerDeviceType == Windows.Devices.Input.PointerDeviceType.Mouse)
                    {
                        UtilityFunctions.SetCursor(Windows.UI.Core.CoreCursorType.SizeWestEast);
                    }
                }
            }
        }

        private void ResizeHandle_PointerMoved(object sender, PointerRoutedEventArgs e)
        {
            if (_HasContactPoint && e.Pointer.PointerId == _CurrentContactPoint)
            {
                PointerPoint ppt = e.GetCurrentPoint(MainGrid);
                ResizeHandleTranslation.X = ppt.Position.X - Offsets[Position];
                int bestIndex = FindClosestSnapPoint(ppt.Position.X);
                if (bestIndex != _ResizerPosition)
                {
                    _ResizerPosition = bestIndex;
                    VerticalSnapAnimation.To = Offsets[bestIndex] - _ResizerHalfSize;
                    VerticalSnapStoryboard.Begin();
                }
            }
            else if (!_HasContactPoint)
            {
                if (e.Pointer.PointerDeviceType == Windows.Devices.Input.PointerDeviceType.Mouse)
                {
                    if (IsInsideHandleRadius(e.GetCurrentPoint(ResizeHandle).Position))
                    {
                        UtilityFunctions.SetCursor(Windows.UI.Core.CoreCursorType.SizeWestEast);
                    }
                    else
                    {
                        UtilityFunctions.SetCursor(Windows.UI.Core.CoreCursorType.Arrow);
                    }
                }
            }
            else
            {
                UtilityFunctions.SetCursor(Windows.UI.Core.CoreCursorType.Arrow);
            }
        }

        private void ResizeHandle_PointerReleased(object sender, PointerRoutedEventArgs e)
        {
            if (_HasContactPoint && e.Pointer.PointerId == _CurrentContactPoint)
            {
                if (_ResizerPosition != Position)
                {
                    Position = _ResizerPosition;
                    PositionUpdated?.Invoke(this, new RoutedEventArgs());
                }
                ResizeHandle.ReleasePointerCapture(e.Pointer);
                _HasContactPoint = false;
                ResetView();
            }
        }

        private void ResizeHandle_PointerCaptureLost(object sender, PointerRoutedEventArgs e)
        {
            if (_HasContactPoint && e.Pointer.PointerId == _CurrentContactPoint)
            {
                AbortMove();
            }
        }

        private void ResizeHandle_PointerCanceled(object sender, PointerRoutedEventArgs e)
        {
            if (_HasContactPoint && e.Pointer.PointerId == _CurrentContactPoint)
            {
                AbortMove();
            }
        }
        private void ResizeHandle_PointerExited(object sender, PointerRoutedEventArgs e)
        {
            if (!_HasContactPoint || e.Pointer.PointerId != _CurrentContactPoint)
            {
                UtilityFunctions.SetCursor(Windows.UI.Core.CoreCursorType.Arrow);
            }
        }

        private void AbortMove()
        {
            _HasContactPoint = false;
            ResetView();
        }

        private void ResetView()
        {
            _HasContactPoint = false;
            _ContactPointIsInsideRadius = false;
            ResizeHintHost.IsHitTestVisible = false;
            HorizontalResizer.Visibility = Visibility.Collapsed;
            ResizeHintHost.Visibility = Visibility.Collapsed;
            HorizontalResizerTranslation.X = Offsets[Position] - ((double)Resources["ResizerSize"]) / 2.0;
            ResizeHandle.Margin = new Thickness(Offsets[Position] - (double)Resources["HandleTargetThickness"] / 2.0, 0, 0, 0);
            ResizeHandleTranslation.X = 0;
            UtilityFunctions.SetCursor(Windows.UI.Core.CoreCursorType.Arrow);
        }

        private bool IsInsideHandleRadius(UIPoint p)
        {
            return (p.X >= 0 && p.X <= ResizeHandle.ActualWidth && p.Y >= 0 && p.Y <= ResizeHandle.ActualHeight);
        }

        private int FindClosestSnapPoint(double offset)
        {
            int bestIndex = 0;
            double bestDistance = -1;
            for (int i = 0; i < Offsets.Count; ++i)
            {
                double distance = Math.Abs(offset - Offsets[i]);
                if (bestDistance < 0 || distance < bestDistance)
                {
                    bestDistance = distance;
                    bestIndex = i;
                }
            }
            return bestIndex;
        }

        #endregion Movement
    }
}
