﻿using pdftron.PDF.Tools.Controls.ViewModels;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.IO;
using System.Linq;
using System.Runtime.InteropServices.WindowsRuntime;
using System.Threading.Tasks;
using Windows.Foundation;
using Windows.Foundation.Collections;
using Windows.Graphics.Display;
using Windows.UI;
using Windows.UI.Core;
using Windows.UI.ViewManagement;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Controls.Primitives;
using Windows.UI.Xaml.Data;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Media.Animation;
using Windows.UI.Xaml.Navigation;

namespace pdftron.PDF.Tools.Controls
{
    /// <summary>
    /// Control that displays thumbnails in a grid pattern, and allows rearranging, adding, and removing of pages.
    /// </summary>
    /// <remarks>
    /// This control seem to have problems when a debugger is a attached in terms of memory usage
    /// as if garbage collection doesn't work. If you find it running up the memory, try it without a debugger attached.</remarks>
    public sealed partial class ThumbnailViewer : UserControl, ControlBase.ICloseableControl
    {
        private ThumbnailsViewViewModel _viewModel;
        private PDFViewCtrl _pdfViewCtrl;
        private ToolManager _toolManager;
        private string _documentTag;
        private NavigationStack _navigationStack;
        private bool _isActive = true;
        private ThumbnailItem _currentItemWhenGoingOutOfView;
        private ISet<int> _requestedIndexesSinceViewChanged = new HashSet<int>();
        private bool _shouldRefreshWhenComingIntoView;
        private bool _viewIsVisible = true;
        private ItemTemplates _lastItemTemplate;
        private double _lastWidthCalculated = double.NaN;
        private int _columnsCalculcated = 1;
        private int _previousLowest = -1;
        private int _currentLowest = -1;
        private int _previousHighest = -1;
        private int _currentHighest = -1;

        private static DependencyProperty NavigationOnlyProperty = DependencyProperty.Register(
            "NavigationOnly",
            typeof(bool),
            typeof(ThumbnailViewer),
            new PropertyMetadata(false, OnNavigationOnlyChanged));

        private static void OnNavigationOnlyChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            if (d is ThumbnailViewer viewer)
                viewer.UpdateNavigationOnly();
        }

        public bool NavigationOnly
        {
            get => (bool)GetValue(NavigationOnlyProperty);
            set => SetValue(NavigationOnlyProperty, value);
        }

        /// <summary>
        /// Dependency property to set whether Items should be automatically fitted to the width of the ThumbnailViewer,
        /// accoriding to NumberOfColumns
        /// </summary>
        private static DependencyProperty FitItemsToWidthProperty = DependencyProperty.Register(
            "FitItemsToWidth",
            typeof(bool),
            typeof(AnnotationList),
            new PropertyMetadata(false, OnFitItemsToWidthChanged)); // TODO UPDATE

        /// <summary>
        /// Gets or sets whether Items should be automatically fitted to the width of the ThumbnailViewer,
        /// according to NumberOfColumns
        /// </summary>
        public bool FitItemsToWidth
        {
            get => (bool)GetValue(FitItemsToWidthProperty);
            set => SetValue(FitItemsToWidthProperty, value);
        }

        private static void OnFitItemsToWidthChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            if (d is ThumbnailViewer viewer && e.NewValue is bool)
            {
                viewer._lastWidthCalculated = 0;
                if (viewer.ActualWidth > 0)
                    viewer.SetItemTemplateAndSize(new Size(viewer.ActualWidth, viewer.ActualHeight));
            }
        }

        /// <summary>
        /// Dependency property to set the number of columns of thumbnails in the ThumbnailsView
        /// </summary>
        private static DependencyProperty NumberOfColumnsProperty = DependencyProperty.Register(
            "NumberOfColumns",
            typeof(int),
            typeof(ThumbnailViewer),
            new PropertyMetadata(null, OnNumberOfColumnsChanged));

        private static void OnNumberOfColumnsChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            if (d is ThumbnailViewer viewer && !double.IsNaN(viewer.MainListView.ActualWidth))
            {
                viewer._lastWidthCalculated = 0;
                viewer.SetItemTemplateAndSize(new Size(viewer.ActualWidth, viewer.ActualHeight));
            }
        }

        /// <summary>
        /// Gets or sets the number of columns of thumbnails in the ThumbnailsView
        /// Null or any number less than 1 means that the View will calculate an appropriate number of
        /// columns based on its size
        /// </summary>
        public int NumberOfColumns
        {
            get => (int)GetValue(NumberOfColumnsProperty);
            set => SetValue(NumberOfColumnsProperty, value);
        }

        /// <summary>
        /// This property is used only by the control itself, when it dynamically calculates the size of elements
        /// to fit the width of the ListView
        /// </summary>
        private static DependencyProperty ListViewItemWidthProperty = DependencyProperty.Register(
            "ListViewItemWidth",
            typeof(double),
            typeof(AnnotationList),
            new PropertyMetadata(10.0));

        public double ListViewItemWidth
        {
            get => (double)GetValue(ListViewItemWidthProperty);
            set => SetValue(ListViewItemWidthProperty, value);
        }

        /// <summary>
        /// This property is used only by the control itself, when it dynamically calculates the size of elements
        /// to fit the width of the ListView
        /// </summary>
        private static DependencyProperty ListViewItemHeightProperty = DependencyProperty.Register(
            "ListViewItemHeight",
            typeof(double),
            typeof(AnnotationList),
            new PropertyMetadata(10.0));

        public double ListViewItemHeight
        {
            get => (double)GetValue(ListViewItemHeightProperty);
            set => SetValue(ListViewItemHeightProperty, value);
        }

        /// <summary>
        /// This property is used only by the control itself, when it dynamically calculates the size of elements
        /// to fit the width of the ListView
        /// </summary>
        private static DependencyProperty ItemPaddingProperty = DependencyProperty.Register(
            "ItemPadding",
            typeof(Thickness),
            typeof(AnnotationList),
            new PropertyMetadata(new Thickness(8), OnItemPaddingChanged));

        /// <summary>
        /// Sets the padding between items in the ThumbnailsView
        /// </summary>
        public Thickness ItemPadding
        {
            get => (Thickness)GetValue(ItemPaddingProperty);
            set => SetValue(ItemPaddingProperty, value);
        }

        private static void OnItemPaddingChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            if (d is ThumbnailViewer viewer && e.NewValue is Thickness)
            {
                viewer._lastWidthCalculated = 0;
                viewer.SetItemPadding();
                if (!double.IsNaN(viewer.MainListView.ActualWidth))
                {
                    viewer._lastWidthCalculated = 0;
                    viewer.SetItemTemplateAndSize(new Size(viewer.ActualWidth, viewer.ActualHeight));
                }
            }
        }

        private void SetItemPadding()
        {
            Thickness padding = (Thickness)GetValue(ItemPaddingProperty);
            Style itemContainerStyle = Resources["ListViewItemWihtoutExtraPadding"] as Style;
            Setter marginSetter = null;
            foreach (Setter setter in itemContainerStyle.Setters)
            {
                if (setter.Property == MarginProperty)
                {
                    marginSetter = setter;
                    break;
                }
            }

            if (marginSetter != null)
                itemContainerStyle.Setters.Remove(marginSetter);

            marginSetter = new Setter(MarginProperty, padding);
            itemContainerStyle.Setters.Add(marginSetter);
            MainListView.ItemContainerStyle = itemContainerStyle;
        }

        /// <summary>
        /// Dependency property for max width of items when automatically sizing items.
        /// This property only has an effect if NumberOfColumns is NaN or 0
        /// </summary>
        private static DependencyProperty AutoSizingMaxWidthProperty = DependencyProperty.Register(
            "AutoSizingMaxWidth",
            typeof(double),
            typeof(AnnotationList),
            new PropertyMetadata(250.0, OnAutoSizingMaxWidthChanged));

        /// <summary>
        /// The max width of items when automatically sizing items.
        /// This property only has an effect if NumberOfColumns is NaN or 0
        /// </summary>
        public double AutoSizingMaxWidth
        {
            get => (double)GetValue(AutoSizingMaxWidthProperty);
            set => SetValue(AutoSizingMaxWidthProperty, value);
        }

        private static void OnAutoSizingMaxWidthChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            if (d is ThumbnailViewer viewer && e.NewValue is double && !double.IsNaN(viewer.MainListView.ActualWidth))
            {
                viewer._lastWidthCalculated = 0;
                viewer.SetItemTemplateAndSize(new Size(viewer.MainListView.ActualWidth, viewer.MainListView.ActualHeight));
            }
        }

        public ThumbnailsViewViewModel ViewModel => _viewModel;

        public PDFViewCtrl PDFViewCtrl
        {
            private get => _pdfViewCtrl;
            set
            {
                if (!string.IsNullOrWhiteSpace(_documentTag) && _pdfViewCtrl != null)
                    _viewModel.CleanUp();

                _pdfViewCtrl = value;
                if (!string.IsNullOrWhiteSpace(_documentTag) && _pdfViewCtrl != null)
                    SetPDFViewCtrl(_pdfViewCtrl);
            }
        }

        public ToolManager ToolManager
        {
            private get => _toolManager;
            set
            {
                _toolManager = value;
                if (_viewModel != null)
                    _viewModel.ToolManager = _toolManager;
            }
        }

        public string DocumentTag
        {
            get => _documentTag;
            set
            {
                _documentTag = value;
                if (!string.IsNullOrWhiteSpace(_documentTag) && _pdfViewCtrl != null)
                    SetPDFViewCtrl(_pdfViewCtrl);
            }
        }

        public Color BlankPageDefaultColor
        {
            get => _viewModel.PageDefaultColor;
            set
            {
                if (value != _viewModel.PageDefaultColor)
                {
                    _viewModel.PageDefaultColor = value;
                    if (ViewIsVisible)
                    {
                        _viewModel.Refresh();
                    }
                    else
                    {
                        _shouldRefreshWhenComingIntoView = true;
                    }
                }
            }
        }

        public new Brush Background
        {
            get => BackgroundGrid.Background;
            set => BackgroundGrid.Background = value;
        }

        public NavigationStack NavigationStack
        {
            get => _viewModel != null ? _viewModel.NavigationStack : _navigationStack;
            set
            {
                _navigationStack = value;
                if (_viewModel != null)
                    _viewModel.NavigationStack = _navigationStack;
            }
        }

        private bool IsActive
        {
            get => _isActive;
            set
            {
                if (_isActive != value)
                    _isActive = value;
            }
        }

        /// <summary>
        /// This boolean can be set to inform the AnnotationList that it can not scroll.
        /// This happens for example when the AnnotationList is in a Pivot and not in view.
        /// Setting this to true will scroll the currently selected item into view
        /// and also affects restoring state.
        /// </summary>
        public bool ViewIsVisible
        {
            get => _viewIsVisible;
            set
            {
                if (_viewIsVisible != value)
                {
                    _viewIsVisible = value;
                    if (ViewModel != null)
                    {
                        ViewModel.Pause = !_viewIsVisible;
                        if (_viewIsVisible)
                        {
                            if (ViewModel.CurrentItem != null && ViewModel.CurrentItem != _currentItemWhenGoingOutOfView)
                                MainListView.ScrollIntoView(ViewModel.CurrentItem);

                            if (_shouldRefreshWhenComingIntoView)
                            {
                                ViewModel.Refresh();
                                _shouldRefreshWhenComingIntoView = false;
                                _requestedIndexesSinceViewChanged.Clear();
                            }

                            foreach (int modifiedPage in _requestedIndexesSinceViewChanged)
                            {
                                ViewModel.PageModified(modifiedPage);
                            }

                            _requestedIndexesSinceViewChanged.Clear();
                        }
                        else
                        {
                            _currentItemWhenGoingOutOfView = ViewModel.CurrentItem;
                        }
                    }
                }
            }
        }

        public ThumbnailViewer()
        {
            InitializeComponent();
            Init();
        }

        public ThumbnailViewer(PDFViewCtrl ctrl, string docTag)
        {
            InitializeComponent();
            Init();

            PDFViewCtrl = ctrl;
            DocumentTag = docTag;
            _viewModel.PageAdded += ViewModel_PageAdded;
        }

        public ThumbnailViewer(PDFViewCtrl ctrl, string docTag, Color background)
        {
            InitializeComponent();
            Init();

            Background = new SolidColorBrush(background);
            PDFViewCtrl = ctrl;
            DocumentTag = docTag;
        }

        private void DebugButton_Click(object sender, RoutedEventArgs e) => ControlClosed();

        private void Init()
        {
            SetItemPadding();
            SizeChanged += ThumbnailViewer_SizeChanged;
            MainListView.SizeChanged += MainListView_SizeChanged;
            MainListView.Loaded += MainListView_Loaded;
        }

        private void MainListView_Loaded(object sender, RoutedEventArgs e)
        {
            if (ViewIsVisible)
            {
                if (ViewModel?.CurrentItem != null)
                    MainListView.ScrollIntoView(ViewModel.CurrentItem);
            }
            else
            {
                _currentItemWhenGoingOutOfView = null;
            }
        }

        private void ThumbnailViewer_SizeChanged(object sender, SizeChangedEventArgs e)
        {
            // Show/hide the excess Command Bar elements depending on screen width
            if (e.NewSize.Width < 450)
            {
                UtilityFunctions.HideElement(CancelSelectionButton);
                UtilityFunctions.HideElement(CancelSelectionButtonSeparator);
            }
            else
            {
                UtilityFunctions.ShowElement(CancelSelectionButton, CancelSelectionButton.Height);
                UtilityFunctions.ShowElement(CancelSelectionButtonSeparator, CancelSelectionButtonSeparator.Height);
            }
        }

        private async void ViewModel_PageAdded(int pageNumber, ThumbnailItem item)
        {
            //int pageCount = PDFViewCtrl.GetPageCount();
            await Task.Delay(200); // Hack to ensure the ListView's layout is updated.
            if (item != null && ViewModel?.ThumbnailsList?.Contains(item) == true && MainListView != null)
                MainListView.ScrollIntoView(item, ScrollIntoViewAlignment.Default);
        }

        private void MainListView_SizeChanged(object sender, SizeChangedEventArgs e) => SetItemTemplateAndSize(e.NewSize);

        private void SetItemTemplateAndSize(Size listSize)
        {
            if (FitItemsToWidth)
            {
                CalculateItemWidth(listSize.Width);
            }
            else if (listSize.Width < 500 || listSize.Height < 500)
            {
                if (_lastItemTemplate != ItemTemplates.Small)
                {
                    MainListView.ItemTemplate = (DataTemplate)Resources["ThumbnailTemplateSmall"];
                    _lastItemTemplate = ItemTemplates.Small;
                }

                double w = (double)Resources["ThumbnailItemWidthSmall"];
                double h = (double)Resources["ThumbnailItemHeightSmall"];
                Thickness t = (Thickness)Resources["ThumbnailMarginSmall"];
                if (_viewModel != null)
                    _viewModel.ImageMaxDimensions = new Size(w - t.Left - t.Right, h - t.Top - t.Bottom);
            }
            else
            {
                double w = (double)Resources["ThumbnailItemWidthMedium"];
                double h = (double)Resources["ThumbnailItemWidthMedium"];
                double rows = listSize.Height / w;
                double columns = listSize.Width / h;

                if (rows * columns <= 15)
                {
                    if (_lastItemTemplate != ItemTemplates.Medium)
                    {
                        MainListView.ItemTemplate = (DataTemplate)Resources["ThumbnailTemplateMedium"];
                        _lastItemTemplate = ItemTemplates.Medium;
                    }

                    Thickness t = (Thickness)Resources["ThumbnailMarginMedium"];
                    if (_viewModel != null)
                        _viewModel.ImageMaxDimensions = new Size(w - t.Left - t.Right, h - t.Top - t.Bottom);
                }
                else
                {
                    if (_lastItemTemplate != ItemTemplates.Large)
                    {
                        MainListView.ItemTemplate = (DataTemplate)Resources["ThumbnailTemplateLarge"];
                        _lastItemTemplate = ItemTemplates.Large;
                    }

                    w = (double)Resources["ThumbnailItemWidthLarge"];
                    h = (double)Resources["ThumbnailItemHeightLarge"];
                    Thickness t = (Thickness)Resources["ThumbnailMarginLarge"];
                    if (_viewModel != null)
                        _viewModel.ImageMaxDimensions = new Size(w - t.Left - t.Right, h - t.Top - t.Bottom);
                }
            }
        }

        private void SetPDFViewCtrl(PDFViewCtrl ctrl)
        {
            _viewModel = new ThumbnailsViewViewModel(ctrl, _documentTag, this)
            {
                ReadOnly = NavigationOnly
            };

            _viewModel.PropertyChanged += ViewModel_PropertyChanged;
            _viewModel.ThumbnailReceived += _ViewModel_ThumbnailReceived;
            if (_viewModel.ThumbnailsList != null)
                _viewModel.ThumbnailsList.VisibleRangeChanged += ThumbnailsList_VisibleRangeChanged;

            _viewModel.NavigationStack = _navigationStack;
            DataContext = _viewModel;
            MainListView.DataContext = _viewModel;
        }

        private void _ViewModel_ThumbnailReceived(int pageNumber)
        {
            int index = pageNumber - 1;
            if (index >= _viewModel.ThumbnailsList.FirstVisibleIndex && index <= _viewModel.ThumbnailsList.LastVisibleIndex)
            {
                FrameworkElement container = (FrameworkElement)MainListView.ContainerFromIndex(index);
                Windows.UI.Xaml.Controls.Image image = UtilityFunctions.FindVisualChild<Windows.UI.Xaml.Controls.Image>(container);
                if (image != null)
                {
                    if (image.Parent is FrameworkElement parent)
                    {
                        parent.Opacity = 0;
                        parent.Name = "image" + index;
                        Storyboard opacityAnim = OpacityAnimation(parent, index);
                        opacityAnim.Begin();
                    }
                }
            }
        }

        private Storyboard OpacityAnimation(FrameworkElement target, int index)
        {
            Storyboard sb = new Storyboard();
            string ID = "anim" + index + "OpacityAnimation";
            if (target.Resources.ContainsKey(ID))
                return target.Resources[ID] as Storyboard;

            double from = 0;
            double to = 1;
            DoubleAnimation fadeIn = new DoubleAnimation
            {
                From = from,
                To = to,
                Duration = new Duration(TimeSpan.FromMilliseconds(200)),
                RepeatBehavior = new RepeatBehavior { Count = 1 },
                EasingFunction = new QuadraticEase() { EasingMode = EasingMode.EaseIn },
            };

            target.Resources.Add(ID, sb);
            Storyboard.SetTarget(fadeIn, target);
            Storyboard.SetTargetName(fadeIn, target.Name);
            Storyboard.SetTargetProperty(fadeIn, "(UIElement.Opacity)");
            sb.Children.Add(fadeIn);

            return sb;
        }

        private void ViewModel_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            if (e.PropertyName.Equals("CurrentItem", StringComparison.OrdinalIgnoreCase) && ViewModel.CurrentItem != null)
                MainListView.ScrollIntoView(ViewModel.CurrentItem);
        }

        /// <summary>
        /// Display the flyout (defined in resources) when a ThumbnailItem is right tapped
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void ThumbnailItem_RightTapped(object sender, RightTappedRoutedEventArgs e)
        {
            if (NavigationOnly || _viewModel == null ||  ViewModel.IsConverting || ViewModel.EnabledThumbnailOverlay)
                return;

            ThumbnailItem selectedItem = (sender as FrameworkElement).DataContext as ThumbnailItem;

            ListViewItem item = MainListView.ContainerFromItem(selectedItem) as ListViewItem;

            // Align the flyout
            var pos = e.GetPosition(this);
            MarginTextBlock.Margin = new Thickness(pos.X, pos.Y, 0, 0);
            RightClickButtonFlyout.ShowAt(MarginTextBlock);

            IList<MenuFlyoutItemBase> buttons = RightClickButtonFlyout.Items;

            // Since flyout is defined in resources, set the CommandParameter programatically
            (buttons[0] as MenuFlyoutItem).CommandParameter = selectedItem;
            (buttons[1] as MenuFlyoutItem).CommandParameter = selectedItem;
            (buttons[2] as MenuFlyoutItem).CommandParameter = selectedItem;

            e.Handled = true;
        }

        private void MainListView_ItemClick(object sender, ItemClickEventArgs e) => ItemClicked?.Invoke(sender, e);

        public event ControlBase.ControlClosedDelegate ControlClosed = delegate { };
        public event EventHandler<ItemClickEventArgs> ItemClicked;

        public void CloseControl()
        {
            _viewIsVisible = false;
            _viewModel.CleanUp();
            ControlClosed();
        }

        private void UpdateNavigationOnly()
        {
            if (NavigationOnly)
            {
                MainCommandBar.Visibility = Visibility.Collapsed;
                BackgroundGrid.Margin = new Thickness(0);
                FabButton.Visibility = Visibility.Collapsed;
            }
            else
            {
                MainCommandBar.Visibility = Visibility.Visible;
                BackgroundGrid.Margin = new Thickness(0, 48, 0, 0);
                FabButton.Visibility = Visibility.Visible;
            }
            if (_viewModel != null)
                _viewModel.ReadOnly = NavigationOnly;
        }

        /// <summary>
        /// Will perform a back operation if applicable.
        /// Use this when the back button is pressed to give this control a chance to handle it.
        /// </summary>
        /// <returns>true if the control could go back</returns>
        public bool GoBack() => _viewModel.GoBack();

        public void PageModified(int pageNumber)
        {
            if (ViewIsVisible)
            {
                _viewModel.PageModified(pageNumber);
            }
            else if (!_requestedIndexesSinceViewChanged.Contains(pageNumber))
            {
                _requestedIndexesSinceViewChanged.Add(pageNumber);
            }
        }

        public void SelectPages(IList<int> pagesToSelect)
        {
            if (_viewModel != null)
            {
                //_viewModel.EditCommand.Execute(null);
                _viewModel.SelectionMode = ListViewSelectionMode.Multiple;
                _viewModel.IsSelectButtonVisible = false;
                MainListView.SelectedItems?.Clear();

                foreach (int i in pagesToSelect)
                {
                    if (MainListView.Items != null && i >= 1 && i <= MainListView.Items.Count)
                    {
                        MainListView.SelectedItems.Add(MainListView.Items[i - 1]);
                    }
                }
            }
        }

        private int CalculateNumberOfcolumns(double listWidth)
        {
            Thickness itemPadding = (Thickness)GetValue(ItemPaddingProperty);
            double perItemWidth = AutoSizingMaxWidth + itemPadding.Left + itemPadding.Right;
            double numItems = listWidth / perItemWidth;
            int idealColumnNumber = (int)Math.Ceiling(numItems);
            if (idealColumnNumber < 0)
                idealColumnNumber = 1;

            return idealColumnNumber;
        }

        private async void CalculateItemWidth(double listWidth)
        {
            if (double.IsNaN(listWidth) || listWidth <= 0 || listWidth == _lastWidthCalculated)
                return;

            _lastWidthCalculated = listWidth;
            _columnsCalculcated = 1;
            if (NumberOfColumns <= 0)
            {
                _columnsCalculcated = CalculateNumberOfcolumns(listWidth);
            }
            else
            {
                _columnsCalculcated = NumberOfColumns;
            }

            double itemWidth = 10;

            if (_columnsCalculcated > 0)
            {
                Thickness itemPadding = (Thickness)GetValue(ItemPaddingProperty);
                double perGapSize = itemPadding.Left + itemPadding.Right;
                double gapSize = perGapSize * _columnsCalculcated;
                double remainingwidth = listWidth - gapSize - ListViewMarginAndPadding;
                itemWidth = remainingwidth / _columnsCalculcated;
                double rawPixels = DisplayInformation.GetForCurrentView().RawPixelsPerViewPixel;

                int numRawPixels = (int)(itemWidth * rawPixels);
                itemWidth = numRawPixels / rawPixels;
            }

            int previousFirst = _previousLowest;// _ViewModel.ThumbnailsList.FirstVisibleIndex;
            int previousLast = _previousHighest; //_ViewModel.ThumbnailsList.LastVisibleIndex;

            ListViewItemWidth = Math.Max(itemWidth, 10);
            double widthToHeightRatio = (double)Resources["ThumbnailItemWidthToHeightRatio"];
            ListViewItemHeight = ListViewItemWidth / widthToHeightRatio;

            // Need to set this again, so that we re-use the binding.
            MainListView.ItemTemplate = (DataTemplate)Resources["ThumbnailTemplateVariableSize"];

            if (_viewModel != null)
            {
                Thickness t = (Thickness)Resources["TemplateImageMargin"];
                _viewModel.ImageMaxDimensions = new Size(ListViewItemWidth - t.Left - t.Right, ListViewItemHeight - t.Top - t.Bottom);
            }

            await Dispatcher.RunAsync(CoreDispatcherPriority.Normal, () =>
            {
                if (_viewModel.ThumbnailsList.Count >= 1 && previousFirst >= 0 && previousLast >= 0)
                {
                    int newfirst = _viewModel.ThumbnailsList.FirstVisibleIndex;
                    int newlast = _viewModel.ThumbnailsList.LastVisibleIndex;
                    int oldCenter = (previousFirst + previousLast) / 2;
                    int newCenter = (newfirst + newlast) / 2;
                    if (oldCenter != newCenter)
                    {
                        int newSpan = newlast - newfirst + 1;
                        int scrollToIndex = oldCenter;
                        if (newCenter > oldCenter)
                        {
                            scrollToIndex -= (newSpan / 2);
                        }
                        else
                        {
                            scrollToIndex += (newSpan / 2);
                        }

                        scrollToIndex = Math.Max(0, Math.Min(scrollToIndex, _viewModel.ThumbnailsList.Count - 1));
                        MainListView.ScrollIntoView(_viewModel.ThumbnailsList[scrollToIndex]);
                    }
                }
            });
        }

        private void ThumbnailsList_VisibleRangeChanged(object sender, ItemIndexRange e)
        {
            _previousLowest = _currentLowest;
            _currentLowest = ViewModel.ThumbnailsList.FirstVisibleIndex;
            _previousHighest = _currentHighest;
            _currentHighest = ViewModel.ThumbnailsList.LastVisibleIndex;
        }

        private double ListViewMarginAndPadding => MainListView.Margin.Left + MainListView.Margin.Right + MainListView.Padding.Left + MainListView.Padding.Right;
    }
}
