﻿using System;
using Windows.ApplicationModel.Resources;
using Windows.Foundation;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Media.Animation;
using Windows.UI.Xaml.Shapes;

namespace pdftron.PDF.Tools.Controls
{
    /// <summary>
    /// A Content Control which allows to be dragged around the screen
    /// </summary>
    public class FloatingContentControl : ContentControl
    {
        private const string BorderPartName = "PART_Border";
        private const int BOTTOM_PADDING = 70;

        // Using a Border element to be dragged
        private Border _border;
        private ResourceDictionary _ResourceDictControlStyles;
        private bool _isInitialized = false;

        #region Dependency Properties

        public static readonly DependencyProperty DraggableProperty =
            DependencyProperty.Register(
                "Draggable",
                typeof(bool),
                typeof(FloatingContentControl),
                new PropertyMetadata(true));

        public static readonly DependencyProperty BoundaryProperty =
            DependencyProperty.Register(
                "Boundary",
                typeof(FloatingBoundary),
                typeof(FloatingContentControl),
                new PropertyMetadata(FloatingBoundary.None));

        public static readonly DependencyProperty StartPositionProperty =
                DependencyProperty.Register(
                "StartPosition",
                typeof(FloatingStartPosition),
                typeof(FloatingContentControl),
                new PropertyMetadata(FloatingStartPosition.TopLeft));

        #endregion

        /// <summary>
        /// Initializes a new instance of the <see cref="FloatingContentControl"/> class.
        /// </summary>
        public FloatingContentControl()
        {
            _ResourceDictControlStyles = new ResourceDictionary();
            _ResourceDictControlStyles.Source = new Uri("ms-appx:///pdftron.PDF.Tools/Controls/Resources/ControlStyles.xaml");

            this.Style = (Style)_ResourceDictControlStyles["FloatingContentControlStyle"];
        }
        
        /// <summary>
        /// Set wether the control allows it to be dragged around
        /// </summary>
        public bool Draggable
        {
            get { return (bool)GetValue(DraggableProperty); }
            set { SetValue(DraggableProperty, value); }
        }

        /// <summary>
        /// Choose which type of boundary restrictions the control should respect
        /// </summary>
        public FloatingBoundary Boundary
        {
            get { return (FloatingBoundary)GetValue(StartPositionProperty); }
            set { SetValue(StartPositionProperty, value); }
        }

        /// <summary>
        /// Set the starting position for the control
        /// </summary>
        public FloatingStartPosition StartPosition
        {
            get { return (FloatingStartPosition)GetValue(StartPositionProperty); }
            set { SetValue(StartPositionProperty, value); }
        }

        /// <summary>
        /// Invoked whenever application code or internal processes (such as a rebuilding layout pass) call ApplyTemplate.
        /// In simplest terms, this means the method is called just before a UI element displays in your app.
        /// Override this method to influence the default post-template logic of a class.
        /// </summary>
        protected override void OnApplyTemplate()
        {
            // Border
            this._border = this.GetTemplateChild(BorderPartName) as Border;
            if (this._border != null)
            {
                this._border.ManipulationDelta += Border_ManipulationDelta;

                // Move Canvas properties from control to border.
                Canvas.SetLeft(this._border, Canvas.GetLeft(this));
                Canvas.SetLeft(this, 0);
                Canvas.SetTop(this._border, Canvas.GetTop(this));
                Canvas.SetTop(this, 0);

                // Move Margin to border.
                this._border.Padding = this.Margin;
                this.Margin = new Thickness(0);
            }
            else
            {
                // Exception
                throw new Exception("Floating Control Style has no Border.");
            }

            this.Loaded += Floating_Loaded;
        }

        #region Events

        private void Floating_Loaded(object sender, RoutedEventArgs e)
        {
            FrameworkElement el = GetClosestParentWithSize(this);
            if (el == null)
            {
                return;
            }

            SetToStartingPosition();

            el.SizeChanged += Floating_SizeChanged;
        }

        private void Floating_SizeChanged(object sender, SizeChangedEventArgs e)
        {
            SetToStartingPosition();

            var left = Canvas.GetLeft(this._border);
            var top = Canvas.GetTop(this._border);

           var rect = new Windows.Foundation.Rect(left, top, this._border.ActualWidth, this._border.ActualHeight);

            AdjustCanvasPosition(rect);
        }

        private void Border_ManipulationDelta(object sender, ManipulationDeltaRoutedEventArgs e)
        {
            if (Draggable == false)
                return;

            var left = Canvas.GetLeft(this._border) + e.Delta.Translation.X;
            var top = Canvas.GetTop(this._border) + e.Delta.Translation.Y;

            var rect = new Windows.Foundation.Rect(left, top, this._border.ActualWidth, this._border.ActualHeight);
            var moved = AdjustCanvasPosition(rect);
        }

        #endregion

        #region Private methods

        /// <summary>
        /// Adjusts the canvas position according to the IsBoundBy* properties.
        /// </summary>
        private bool AdjustCanvasPosition(Windows.Foundation.Rect rect)
        {
            // Free floating.
            if (this.Boundary == FloatingBoundary.None)
            {
                Canvas.SetLeft(this._border, rect.Left);
                Canvas.SetTop(this._border, rect.Top);

                return true;
            }

            FrameworkElement el = GetClosestParentWithSize(this);

            // No parent
            if (el == null)
            {
                // We probably never get here.
                return false;
            }

            var position = new Windows.Foundation.Point(rect.Left, rect.Top); ;

            if (this.Boundary == FloatingBoundary.Parent)
            {
                var parentRect = new Windows.Foundation.Rect(0, 0, el.ActualWidth, el.ActualHeight);
                position = AdjustedPosition(rect, parentRect);
            }

            if (this.Boundary == FloatingBoundary.Window)
            {
                var ttv = el.TransformToVisual(Window.Current.Content);
                var topLeft = ttv.TransformPoint(new Windows.Foundation.Point(0, 0));

                var parentRect = new Windows.Foundation.Rect(
                    topLeft.X,
                    topLeft.Y, 
                    Window.Current.Bounds.Width - topLeft.X,
                    Window.Current.Bounds.Height - topLeft.Y);

                position = AdjustedPosition(rect, parentRect);
            }

            // Set new position
            Canvas.SetLeft(this._border, position.X);
            Canvas.SetTop(this._border, position.Y);

            return position == new Windows.Foundation.Point(rect.Left, rect.Top);
        }

        /// <summary>
        /// Returns the adjusted the topleft position of a rectangle so that is stays within a parent rectangle.
        /// </summary>
        /// <param name="rect">The rectangle.</param>
        /// <param name="parentRect">The parent rectangle.</param>
        /// <returns></returns>
        private Windows.Foundation.Point AdjustedPosition(Windows.Foundation.Rect rect, Windows.Foundation.Rect parentRect)
        {
            var left = rect.Left;
            var top = rect.Top;

            if (left < -parentRect.Left)
            {
                // Left boundary hit.
                left = -parentRect.Left;
            }
            else if (left + rect.Width > parentRect.Width)
            {
                // Right boundary hit.
                left = parentRect.Width - rect.Width;
            }

            if (top < -parentRect.Top)
            {
                // Top hit.
                top = -parentRect.Top;
            }
            else if (top + rect.Height > parentRect.Height)
            {
                // Bottom hit.
                top = parentRect.Height - rect.Height;
            }

            return new Windows.Foundation.Point(left, top);
        }

        /// <summary>
        /// Gets the closest parent with a real size.
        /// </summary>
        private FrameworkElement GetClosestParentWithSize(FrameworkElement element)
        {
            while (element != null && (element.ActualHeight == 0 || element.ActualWidth == 0))
            {
                // Crawl up the Visual Tree.
                element = element.Parent as FrameworkElement;
            }

            return element;
        }

        private void SetToStartingPosition()
        {
            if (this._border == null)
                return;

            if (_isInitialized && Draggable)
                return;

            FrameworkElement el = GetClosestParentWithSize(this);
            Windows.Foundation.Rect parentRect;

            if (this.Boundary == FloatingBoundary.Parent)
            {
                parentRect = new Windows.Foundation.Rect(0, 0, el.ActualWidth, el.ActualHeight);
            }
            else if (this.Boundary == FloatingBoundary.Window)
            {
                var ttv = el.TransformToVisual(Window.Current.Content);
                var topLeft = ttv.TransformPoint(new Windows.Foundation.Point(0, 0));

                parentRect = new Windows.Foundation.Rect(
                    topLeft.X,
                    topLeft.Y,
                    Window.Current.Bounds.Width - topLeft.X,
                    Window.Current.Bounds.Height - topLeft.Y);
            }

            switch (StartPosition)
            {
                case FloatingStartPosition.TopLeft:
                    break;

                case FloatingStartPosition.BottomCenter:

                    var position = new Windows.Foundation.Point(parentRect.Width / 2, parentRect.Height);

                    // Set new position
                    Canvas.SetLeft(this._border, position.X - (this._border.ActualWidth / 2)); // ensure to center it
                    Canvas.SetTop(this._border, position.Y - BOTTOM_PADDING);
                    break;
            }

            _isInitialized = true;
        }

        #endregion
    }

    public enum FloatingBoundary
    {
        None,
        Parent,
        Window
    }

    public enum FloatingStartPosition
    {
        TopLeft,
        BottomCenter
    }
}
