﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Windows.UI.Xaml.Controls;
using System.IO;
using System.Windows.Input;
using Windows.Storage.Pickers;
using Windows.Storage;
using Windows.UI.Popups;

using pdftron.PDF.Tools.Controls.ViewModels.Common;
using pdftron.PDF.Tools.Commands;
using pdftron.PDF.Tools.Controls.ViewModels.FileOpening;
using Windows.ApplicationModel.Resources;
using pdftron.PDF.Tools.Utilities;
using Windows.Storage.Provider;
using System.Threading;
using pdftron.PDF.Tools.Controls.ViewModels.Viewer.Password;
using Windows.UI.Xaml.Controls.Primitives;
using pdftron.PDF.Tools.Controls.ControlBase;


namespace pdftron.PDF.Tools.Controls.ViewModels.Compare
{
    public class PdfCompareViewModel : ViewModelBase
    {
        double DEFAULT_ZOOM_START_POSITION = 0.80;  // Initial document zoom
        double DEFAULT_ZOOM_FACTOR_STEP = 1.25;     // Amount of zoom when Zooming in or out by button click
        int MAX_PDF_NAME_LENGTH = 70;               // Max amout of chars

        private const int mScrollLoopbackTimeout = 250;
        private object mlastScrollingElement;
        private int mlastScrollChange = Environment.TickCount;

        private const int mZoomLoopbackTimeout = 250;
        private int mlastZoomChange = Environment.TickCount;

        private double mLastVerticalOffsetA;
        private double mLastVerticalOffsetB;

        bool mIsPdfAForCleanup = true; // When init PdfCompare with a storageFile we might not want to delete the file
        bool misAnnotToolbarSetToPdfA = false;
        bool misAnnotToolbarSetToPdfB = false;
        string mTxtDiffCurrentPageTextCached = string.Empty; // Cach current language for future use

        ScrollViewer mScrollViewerA;
        ScrollViewer mScrollViewerB;

        StorageFile mStorageFilePdfA;
        StorageFile mStorageFilePdfATemp;
        StorageFile mStorageFilePdfB;
        StorageFile mStorageFilePdfBTemp;

        ToolManager mToolManagerDocB;
        ToolManager mToolManagerDocA;
        ToolManager mToolManagerTxtDiff;

        CancellationTokenSource mTokenSource;

        ResourceLoader mResourceLoader;

        private readonly ICloseableControl _view;

        public PdfCompareViewModel()
        {
            Init();
        }

        /// <summary>
        /// Initialize the control with a StorageFile that will be used as the left hand side document to compare
        /// </summary>
        /// <param name="storageFile">The PDF document as a StorageFile</param>
        public PdfCompareViewModel(PDFViewCtrl ctrl, ICloseableControl view)
        {
            Init();

            if (ctrl == null)
                return;

            _view = view;

            if (PdfViewCtrlA == null)
            {
                PdfViewCtrlA = new PDFViewCtrl();
                PdfViewCtrlA.OnSetDoc += MPDFViewA_OnSetDoc;
                PdfAName = Path.GetFileNameWithoutExtension(ctrl.GetDoc().GetFileName());
            }

            PdfViewCtrlA.SetDoc(ctrl.GetDoc());
            IsPdfAVisible = true;
            mIsPdfAForCleanup = false;
        }

        /// <summary>
        /// Initialize the control with a StorageFile that will be used as the left hand side document to compare
        /// </summary>
        /// <param name="storageFile">The PDF document as a StorageFile</param>
        public PdfCompareViewModel(IStorageFile storageFile)
        {
            Init();

            if (storageFile == null || storageFile.FileType != ".pdf")
                return;

            PDFDoc doc = new PDFDoc(storageFile);

            if (PdfViewCtrlA == null)
            {
                PdfViewCtrlA = new PDFViewCtrl();
                PdfAName = storageFile.Name + storageFile.FileType;
                PdfViewCtrlA.OnSetDoc += MPDFViewA_OnSetDoc;
            }

            PdfViewCtrlA.SetDoc(doc);
            mIsPdfAForCleanup = false;
        }

        private void Init()
        {
            // Init and setup Annot toolbar
            _AnnotToolbar = new AnnotationCommandBar();
            _AnnotToolbar.IsEnabled = false;

            _AnnotToolbar.IsEraserToolButtonVisible = true;
            _AnnotToolbar.IsUnderlineToolButtonVisible = true;
            _AnnotToolbar.IsHighlightToolButtonVisible = true;

            _AnnotToolbar.IsInkToolButtonVisible = true;
            _AnnotToolbar.InkMultiStroke = false;

            _AnnotToolbar.IsSignatureToolButtonVisible = false;
            _AnnotToolbar.IsStamperToolButtonVisible = false;
            _AnnotToolbar.IsStrikeoutToolButtonVisible = false;
            _AnnotToolbar.IsRedactionToolButtonVisible = false;
            _AnnotToolbar.IsCircleToolButtonVisible = false;
            _AnnotToolbar.IsStickyNoteToolButtonVisible = false;
            _AnnotToolbar.IsLineToolButtonVisible = false;
            _AnnotToolbar.IsArrowToolButtonVisible = false;
            _AnnotToolbar.IsMultiSelectToolButtonVisible = false;
            _AnnotToolbar.IsMarqueeZoomToolButtonVisible = false;
            _AnnotToolbar.IsSquareToolButtonVisible = false;
            _AnnotToolbar.IsSquigglyToolButtonVisible = false;
            _AnnotToolbar.IsMeasureToolsButtonVisible = false;
            _AnnotToolbar.IsShapeToolsButtonVisible = false;
            _AnnotToolbar.IsCloseToolbarButtonVisible = false;
            _AnnotToolbar.IsUndoRedoToolButtonVisible = false;
            _AnnotToolbar.IsExpandButtonVisible = false;

            _AnnotToolbar.ButtonsStayDown = PersistAnnotTool;

            _PageNumberIndicatorPdfA.ShowingBehaviour = ClickablePageNumberIndicator.ShowingBehaviours.AlwaysShow;
            _PageNumberIndicatorPdfB.ShowingBehaviour = ClickablePageNumberIndicator.ShowingBehaviours.AlwaysShow;

            // Setup commands
            CMDOpenPdfA = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.Create(OpenFileAAsync);
                command.Execute(parameter);
            });

            CMDOpenPdfB = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.Create(OpenFileBAsync);
                command.Execute(parameter);
            });

            CMDSavePdfA = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.Create(SaveDocAAsync);
                command.Execute(parameter);
            });

            CMDSavePdfB = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.Create(SaveDocBAsync);
                command.Execute(parameter);
            });

            CMDClosePdfA = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.Create(CloseDocAAsync);
                command.Execute(parameter);
            });

            CMDClosePdfB = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.Create(CloseDocBAsync);
                command.Execute(parameter);
            });

            CMDPageTxtDiff = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.Create(PageTxtDiffAsync);
                command.Execute(parameter);
            });

            CMDAllPagesTxtDiff = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.Create(AllPagesTxtDiffAsync);
                command.Execute(parameter);
            });

            CMDSaveTxtDiff = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.Create(SaveTxtDiffAsync);
                command.Execute(parameter);
            });

            CMDCloseTxtDiff = new RelayCommand(CloseTxtDiff);
            CMDCancelTxtDiff = new RelayCommand(CancelTxtDiff);
            CMDGoBack = new RelayCommand(PageBack);

            CMDZoomIn = new RelayCommand(ZoomIn);
            CMDZoomOut = new RelayCommand(ZoomOut);

            // Init string resources
            mResourceLoader = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources");
            // NOTE: let's cache this string to avoid too many resource loader calls
            TxtDiffCurrentPageText = mTxtDiffCurrentPageTextCached = mResourceLoader.GetString("PDFCompareControl_TxtDiff_CurPage");
        }

        private void PdfViewCtrlB_PointerMoved(object sender, Windows.UI.Xaml.Input.PointerRoutedEventArgs e)
        {
            if (_AnnotToolbar == null || mToolManagerDocB == null)
                return;

            if (misAnnotToolbarSetToPdfB)
                return;

            if (PdfViewCtrlA != null && misAnnotToolbarSetToPdfA == false)
                return;

            if (mToolManagerDocA != null)
                mToolManagerDocB.CreateTool(mToolManagerDocA.CurrentTool.ToolMode, mToolManagerDocB.CurrentTool);

            _AnnotToolbar.ToolManager = mToolManagerDocB;
            _AnnotToolbar.ButtonsStayDown = PersistAnnotTool;
            misAnnotToolbarSetToPdfA = false;
            misAnnotToolbarSetToPdfB = true;
        }

        private void PdfViewCtrlA_PointerMoved(object sender, Windows.UI.Xaml.Input.PointerRoutedEventArgs e)
        {
            if (_AnnotToolbar == null || mToolManagerDocA == null)
                return;

            if (misAnnotToolbarSetToPdfA)
                return;

            if (mToolManagerDocB != null)
                mToolManagerDocA.CreateTool(mToolManagerDocB.CurrentTool.ToolMode, mToolManagerDocB.CurrentTool);

            _AnnotToolbar.ToolManager = mToolManagerDocA;
            _AnnotToolbar.ButtonsStayDown = PersistAnnotTool;

            misAnnotToolbarSetToPdfA = true;
            misAnnotToolbarSetToPdfB = false;
        }

        #region Public Properties
        private AnnotationCommandBar _AnnotToolbar;
        public AnnotationCommandBar AnnotToolbar
        {
            get { return _AnnotToolbar; }
            set
            {
                _AnnotToolbar = value;
                RaisePropertyChanged();
            }
        }

        private PDFViewCtrl _pdfViewCtrlA;
        public PDFViewCtrl PdfViewCtrlA 
        {
            get { return _pdfViewCtrlA; }
            set
            {
                _pdfViewCtrlA = value;
                RaisePropertyChanged();
            }
        }

        private PDFViewCtrl _pdfViewCtrlB;
        public PDFViewCtrl PdfViewCtrlB
        {
            get { return _pdfViewCtrlB; }
            set
            {
                _pdfViewCtrlB = value;
                RaisePropertyChanged();
            }
        }

        private PDFViewCtrl _PdfViewCtrlTxtDiff;
        public PDFViewCtrl PdfViewCtrlTxtDiff
        {
            get { return _PdfViewCtrlTxtDiff; }
            set
            {
                _PdfViewCtrlTxtDiff = value;
                
                RaisePropertyChanged();
            }
        }

        private ClickablePageNumberIndicator _PageNumberIndicatorPdfA = new ClickablePageNumberIndicator();
        public ClickablePageNumberIndicator PageNumbIndicatorPdfA
        {
            get 
            {
                return _PageNumberIndicatorPdfA;
            }
            set 
            {
                _PageNumberIndicatorPdfA = value;
                RaisePropertyChanged();
            }
        }

        private ClickablePageNumberIndicator _PageNumberIndicatorPdfB = new ClickablePageNumberIndicator();
        public ClickablePageNumberIndicator PageNumbIndicatorPdfB
        {
            get
            {
                return _PageNumberIndicatorPdfB;
            }
            set
            {
                _PageNumberIndicatorPdfB = value;
                RaisePropertyChanged();
            }
        }

        private bool _syncScrolling = true;
        public bool SyncScrolling
        {
            get { return _syncScrolling; }
            set
            {
                _syncScrolling = value;
                RaisePropertyChanged();
            }
        }

        private bool _syncPage = true;
        public bool SyncPage
        {
            get { return _syncPage; }
            set
            {
                _syncPage = value;
                RaisePropertyChanged();
            }
        }

        private bool _syncZoom = false;
        public bool SyncZoom
        {
            get { return _syncZoom; }
            set
            {
                _syncZoom = value;
                RaisePropertyChanged();
            }
        }

        private string _pdfAName = string.Empty;
        public string PdfAName
        {
            get
            {
                if (!string.IsNullOrEmpty(_pdfAName))
                    if (_pdfAName.Length > MAX_PDF_NAME_LENGTH)
                        return _pdfAName.Substring(0, MAX_PDF_NAME_LENGTH);

                return _pdfAName;
            }
            set
            {
                _pdfAName = value;
                RaisePropertyChanged();
            }
        }

        private string _pdfBName = string.Empty;
        public string PdfBName
        {
            get
            {
                if (!string.IsNullOrEmpty(_pdfBName))
                    if (_pdfBName.Length > MAX_PDF_NAME_LENGTH)
                        return _pdfBName.Substring(0, MAX_PDF_NAME_LENGTH);

                return _pdfBName;
            }
            set
            {
                _pdfBName = value;
                RaisePropertyChanged();
            }
        }

        private string _TxtDiffName = string.Empty;
        public string TxtDiffName
        {
            get { return _TxtDiffName; }
            set
            {
                _TxtDiffName = value;
                RaisePropertyChanged();
            }
        }

        private string _TxtDiffCurrentPageText = string.Empty;
        public string TxtDiffCurrentPageText
        {
            get { return _TxtDiffCurrentPageText; }
            set
            {
                _TxtDiffCurrentPageText = value;
                RaisePropertyChanged();
            }
        }

        private bool _isPdfAVisible = false;
        public bool IsPdfAVisible
        {
            get { return _isPdfAVisible; }
            set
            {
                _isPdfAVisible = value;
                RaisePropertyChanged();
            }
        }

        private bool _isPdfBVisible = false;
        public bool IsPdfBVisible
        {
            get { return _isPdfBVisible; }
            set 
            {
                _isPdfBVisible = value;
                RaisePropertyChanged();
            }
        }

        private bool _IsPdfAOpenPageVisible = true;
        public bool IsPdfAOpenPageVisible
        {
            get { return _IsPdfAOpenPageVisible; }
            set
            {
                _IsPdfAOpenPageVisible = value;
                RaisePropertyChanged();
            }
        }

        private bool _IsPdfBOpenPageVisible = true;
        public bool IsPdfBOpenPageVisible
        {
            get { return _IsPdfBOpenPageVisible; }
            set
            {
                _IsPdfBOpenPageVisible = value;
                RaisePropertyChanged();
            }
        }        

        private bool _IsTextDiffResultVisible = false;
        public bool IsTextDiffResultVisible
        {
            get { return _IsTextDiffResultVisible; }
            set
            {
                _IsTextDiffResultVisible = value;
                RaisePropertyChanged();
            }
        }

        private bool _IsTextDiffEnabled = false;
        public bool IsTextDiffEnabled
        {
            get { return _IsTextDiffEnabled; }
            set
            {
                _IsTextDiffEnabled = value;
                RaisePropertyChanged();
            }
        }

        private bool _IsToolbarVisible = false;
        public bool IsToolbarVisible
        {
            get { return _IsToolbarVisible; }
            set
            {
                _IsToolbarVisible = value;
                RaisePropertyChanged();
            }
        }

        private bool _persistAnnotTool = true;
        /// <summary>
        /// When enabled it will make sure to not switch back to Pan tool after an annotation
        /// tool is used
        /// </summary>
        public bool PersistAnnotTool 
        {
            get { return _persistAnnotTool; }
            set { _persistAnnotTool = value; }
        }

        private bool _IsProgressRingVisible = false;
        public bool IsProgressRingVisible
        {
            get { return _IsProgressRingVisible; }
            set 
            {
                _IsProgressRingVisible = value;
                RaisePropertyChanged();
            }
        }
        #endregion

        #region Public Methods
        public bool IsAnyDocsModified()
        {
            if (PdfViewCtrlA != null)
            {
                if (PdfViewCtrlA.GetDoc().IsModified())
                    return true;
            }

            if (PdfViewCtrlB != null)
            {
                if (PdfViewCtrlB.GetDoc().IsModified())
                    return true;
            }

            if (PdfViewCtrlTxtDiff != null)
            {
                if (PdfViewCtrlTxtDiff.GetDoc().IsModified())
                    return true;
            }

            return false;
        }

        public async Task OpenAndCompareFileAAsync(StorageFile file)
        {
            await OpenFileAAsyncImpl(file);
        }

        public async Task OpenAndCompareFileBAsync(StorageFile file)
        {
            await OpenFileBAsyncImpl(file);
        }

        public void CleanUp()
        {
            if (PdfViewCtrlA != null)
            {
                PdfViewCtrlA.OnPageNumberChanged -= PdfViewCtrl_OnPageNumberChanged;
                PdfViewCtrlA.PointerMoved -= PdfViewCtrlA_PointerMoved;

                if (mScrollViewerA != null)
                    mScrollViewerA.ViewChanged -= scrollViewer_ViewChanged;

                mToolManagerDocA?.Dispose();

                PdfViewCtrlA = null;

                IsPdfAVisible = false;
                IsPdfAOpenPageVisible = true;
            }

            if (PdfViewCtrlB != null)
            {
                PdfViewCtrlB.OnPageNumberChanged -= PdfViewCtrl_OnPageNumberChanged;
                PdfViewCtrlB.PointerMoved -= PdfViewCtrlB_PointerMoved;

                if (mScrollViewerB != null)
                    mScrollViewerB.ViewChanged -= scrollViewer_ViewChanged;

                mToolManagerDocB?.Dispose();

                PdfViewCtrlB = null;

                IsPdfBVisible = false;
                IsPdfBOpenPageVisible = true;
            }

            IsTextDiffEnabled = false;

            if (AnnotToolbar != null)
            {
                AnnotToolbar.CloseControl();
                AnnotToolbar.IsEnabled = false;
                IsToolbarVisible = false;
            }

            if (PdfViewCtrlTxtDiff != null)
            {
                mToolManagerTxtDiff?.Dispose();

                PdfViewCtrlTxtDiff = null;
            }
        }
        #endregion

        #region Commands
        public ICommand CMDOpenPdfA { get; set; }

        public ICommand CMDOpenPdfB { get; set; }

        public ICommand CMDSavePdfA { get; set; }

        public ICommand CMDSavePdfB { get; set; }

        public ICommand CMDClosePdfA { get; set; }

        public ICommand CMDClosePdfB { get; set; }

        public ICommand CMDGoBack { get; set; }

        public ICommand CMDPageTxtDiff { get; set; }

        public ICommand CMDAllPagesTxtDiff { get; set; }

        public ICommand CMDSaveTxtDiff { get; set; }

        public ICommand CMDCloseTxtDiff { get; set; }

        public ICommand CMDCancelTxtDiff { get; set; }

        public ICommand CMDZoomIn { get; set; }

        public ICommand CMDZoomOut { get; set; }

        #endregion

        #region Operations

        /// <summary>
        /// Open document A to compare with document B.
        /// The original file is never used to open in the viewer, instead a copy is created in the temporary folder
        /// so all editing can be done.
        /// </summary>
        private async Task OpenFileAAsync() 
        {
            FileOpenPicker filePicker = new FileOpenPicker();
            filePicker.ViewMode = PickerViewMode.List;
            filePicker.FileTypeFilter.Add(".pdf");

            var file = await filePicker.PickSingleFileAsync();
            if (file == null)
                return;

            await OpenFileAAsyncImpl(file);
        }

        private async Task OpenFileAAsyncImpl(StorageFile file)
        {
            // Ensure to cleanup current loaded document
            if (PdfViewCtrlA != null)
            {
                await CleanupPdfAAsync();
            }

            mStorageFilePdfA = file;
            PdfAName = mStorageFilePdfA.DisplayName + mStorageFilePdfA.FileType;

            mStorageFilePdfATemp = await UtilityFunctions.CreateTemporaryAccessFileAsync(mStorageFilePdfA.DisplayName);
            if (mStorageFilePdfATemp == null)
            {
                ResourceLoader loader = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources");

                ContentDialog dialog = new ContentDialog 
                {
                    Title = loader.GetString("DocumentsPage_FileOpeningError_Title"),
                    Content = loader.GetString("DocumentsPage_FileOpeningError_FailedToCreateTempFile"),
                    CloseButtonText = loader.GetString("pdftron_Controls_Content_Dialog_OK")
                };

                await DialogHelper.ShowContentDialogAsync(dialog);
                return;
            }

            await mStorageFilePdfA.CopyAndReplaceAsync(mStorageFilePdfATemp);
            await UtilityFunctions.RemoveReadOnlyAndTemporaryFlagsAsync(mStorageFilePdfA);

            var doc = await OpenDocPDFViewerAsync(mStorageFilePdfATemp, FileAccessMode.ReadWrite);
            if (doc == null)
                return;

            if (PdfViewCtrlA == null)
            {
                PdfViewCtrlA = new PDFViewCtrl();
                PdfViewCtrlA.OnSetDoc += MPDFViewA_OnSetDoc;
            }

            PdfViewCtrlA.SetDoc(doc);
            IsPdfAVisible = true;
            IsPdfAOpenPageVisible = false;
        }

        /// <summary>
        /// Open document B to compare with document A.
        /// The original file is never used to open in the viewer, instead a copy is created in the temporary folder
        /// so all editing can be done.
        /// </summary>
        private async Task OpenFileBAsync()
        {
            FileOpenPicker filePicker = new FileOpenPicker();
            filePicker.ViewMode = PickerViewMode.List;
            filePicker.FileTypeFilter.Add(".pdf");

            var file = await filePicker.PickSingleFileAsync();
            if (file == null)
                return;

            await OpenFileBAsyncImpl(file);
        }

        private async Task OpenFileBAsyncImpl(StorageFile file)
        {
            // Ensure to cleanup current loaded document
            if (PdfViewCtrlB != null)
            {
                await CleanupPdfBAsync();
            }

            mStorageFilePdfB = file;
            PdfBName = mStorageFilePdfB.DisplayName + mStorageFilePdfB.FileType;

            mStorageFilePdfBTemp = await UtilityFunctions.CreateTemporaryAccessFileAsync(mStorageFilePdfB.DisplayName);
            if (mStorageFilePdfBTemp == null)
            {
                ResourceLoader loader = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources");

                ContentDialog dialog = new ContentDialog 
                {
                    Title = loader.GetString("DocumentsPage_FileOpeningError_Title"),
                    Content = loader.GetString("DocumentsPage_FileOpeningError_FailedToCreateTempFile"),
                    CloseButtonText = loader.GetString("Generic_Cancel_Text")
                };

                await DialogHelper.ShowContentDialogAsync(dialog);
                return;
            }

            await mStorageFilePdfB.CopyAndReplaceAsync(mStorageFilePdfBTemp);
            await UtilityFunctions.RemoveReadOnlyAndTemporaryFlagsAsync(mStorageFilePdfB);

            var doc = await OpenDocPDFViewerAsync(mStorageFilePdfBTemp, FileAccessMode.ReadWrite);
            if (doc == null)
                return;

            if (PdfViewCtrlB == null)
            {
                PdfViewCtrlB = new PDFViewCtrl();
                PdfViewCtrlB.OnSetDoc += MPDFViewB_OnSetDoc;
            }

            PdfViewCtrlB.SetDoc(doc);
            IsPdfBVisible = true;
            IsPdfBOpenPageVisible = false;
        }

        private async Task<PDFDoc> OpenDocPDFViewerAsync(IStorageFile file, FileAccessMode mode)
        {
            if (file == null)
                return null;

            try
            {
                var stream = await file.OpenAsync(mode);
                var doc = new PDFDoc(stream);

                var securityH = doc.GetSecurityHandler();

                // In case init fails we ask for password
                if (doc.InitSecurityHandler() == false)
                {
                    ContentDialog contentDialog = new ContentDialog();
                    var pwdBox = new PasswordBox();
                    contentDialog.Title = mResourceLoader.GetString("PDFCompareControl_EnterPassword");
                    contentDialog.PrimaryButtonText = mResourceLoader.GetString("pdftron_Controls_ThumbnailsView_Delete_Message_Confirm");
                    contentDialog.CloseButtonText = mResourceLoader.GetString("pdftron_Controls_ThumbnailsView_Delete_Message_Cancel");
                    contentDialog.Content = pwdBox;

                    var result = await DialogHelper.ShowContentDialogAsync(contentDialog);

                    if (result != ContentDialogResult.Primary)
                        return null;

                    if (string.IsNullOrEmpty(pwdBox.Password))
                        return null;

                    if (doc.InitStdSecurityHandler(pwdBox.Password) == false)
                    {
                        ContentDialog dialog = new ContentDialog
                        {
                            Content = mResourceLoader.GetString("PasswordDialog_ManyFails_Title"),
                            CloseButtonText = mResourceLoader.GetString("pdftron_Controls_Content_Dialog_OK")
                        };

                        await DialogHelper.ShowContentDialogAsync(dialog);
                        return null;
                    }
                }

                return doc;
            }
            catch (Exception e)
            {
                // NOTE: If file already opened it will cause an exception
                ContentDialog contentDialog = new ContentDialog
                {
                    Content = e.Message,
                    CloseButtonText = mResourceLoader.GetString("pdftron_Controls_Content_Dialog_OK")
                };

                await DialogHelper.ShowContentDialogAsync(contentDialog);
                return null;
            }
        }

        private async Task SaveDocAAsync()
        {
            if (PdfViewCtrlA == null)
                return;

            if (PdfViewCtrlA.HasDocument == false)
                return;

            await SaveAsAsync(PdfViewCtrlA.GetDoc());
        }

        private async Task SaveDocBAsync()
        {
            if (PdfViewCtrlB == null)
                return;

            if (PdfViewCtrlB.HasDocument == false)
                return;

            await SaveAsAsync(PdfViewCtrlB.GetDoc());
        }

        /// <summary>
        /// Close document A and properly delete the temporary document created
        /// </summary>
        /// <returns></returns>
        private async Task CloseDocAAsync()
        {
            await CleanupPdfAAsync();

            IsPdfAVisible = false;
            IsPdfAOpenPageVisible = true;
            IsTextDiffEnabled = false;
        }

        /// <summary>
        /// Close document B and properly delete the temporary document created
        /// </summary>
        /// <returns></returns>
        private async Task CloseDocBAsync()
        {
            await CleanupPdfBAsync();

            IsPdfBVisible = false;
            IsPdfBOpenPageVisible = true;
            IsTextDiffEnabled = false;
        }

        /// <summary>
        /// Helper method to perform a cleanup on the loaded PDF A
        /// </summary>
        private async Task CleanupPdfAAsync()
        {
            if (PdfViewCtrlA.HasDocument && mIsPdfAForCleanup)
            {
                PdfViewCtrlA.GetDoc().Destroy();
            }

            if (mStorageFilePdfA != null && mIsPdfAForCleanup)
                mStorageFilePdfA = null;

            if (mStorageFilePdfATemp != null && mIsPdfAForCleanup)
            {

                await mStorageFilePdfATemp.DeleteAsync(StorageDeleteOption.Default);
                mStorageFilePdfATemp = null;
            }

            mToolManagerDocA.Dispose();
            misAnnotToolbarSetToPdfA = false;

            PdfViewCtrlA.PointerMoved -= PdfViewCtrlB_PointerMoved;
            PdfViewCtrlA.OnSetDoc -= MPDFViewB_OnSetDoc;
            PdfViewCtrlA = null;

            CloseToolbar();
        }

        /// <summary>
        /// Helper method to perform a cleanup on the loaded PDF B
        /// </summary>
        private async Task CleanupPdfBAsync()
        {
            if (PdfViewCtrlB.HasDocument)
            {
                PdfViewCtrlB.GetDoc().Destroy();
            }

            if (mStorageFilePdfB != null)
                mStorageFilePdfB = null;

            if (mStorageFilePdfBTemp != null)
            {
                await mStorageFilePdfBTemp.DeleteAsync(StorageDeleteOption.Default);
                mStorageFilePdfBTemp = null;
            }

            mToolManagerDocB.Dispose();
            misAnnotToolbarSetToPdfB = false;

            PdfViewCtrlB.PointerMoved -= PdfViewCtrlB_PointerMoved;
            PdfViewCtrlB.OnSetDoc -= MPDFViewB_OnSetDoc;
            PdfViewCtrlB = null;

            CloseToolbar();
        }

        private void CloseToolbar()
        {
            if (PdfViewCtrlA == null && PdfViewCtrlB == null)
            {
                AnnotToolbar.CloseControl();
                AnnotToolbar.IsEnabled = false;
                IsToolbarVisible = false;
            }
        }

        /// <summary>
        /// Compare text difference betweeen all pages
        /// </summary>
        private async Task AllPagesTxtDiffAsync()
        {
            if (PdfViewCtrlA == null || PdfViewCtrlB == null)
                return;

            AppCenterHandlerBase.CURRENT.SendEvent("PDF Compare", new Dictionary<string, string>() { ["Compare Text"] = "Compare ALL pages" });

            mTokenSource = new CancellationTokenSource();
            IsProgressRingVisible = true;

            try
            {
                var rstDoc = await Task.Run(() => TextDiffAllPagesImpl(mTokenSource.Token), mTokenSource.Token);

                IsProgressRingVisible = false;

                IsTextDiffResultVisible = true;
                IsPdfAVisible = false;
                IsPdfBVisible = false;
                IsPdfAOpenPageVisible = false;
                IsPdfBOpenPageVisible = false;
                IsTextDiffEnabled = false;

                PdfViewCtrlTxtDiff = new PDFViewCtrl();
                PdfViewCtrlTxtDiff.SetPagePresentationMode(PDFViewCtrlPagePresentationMode.e_facing_continuous);
                PdfViewCtrlTxtDiff.SetDoc(rstDoc);

                mToolManagerTxtDiff = new ToolManager(PdfViewCtrlTxtDiff);
                _AnnotToolbar.ToolManager = mToolManagerTxtDiff;
            }
            catch (Exception)
            {
                // swallow
            }
            finally
            {
                mTokenSource.Dispose();
            }
        }

        private PDFDoc TextDiffAllPagesImpl(CancellationToken ct)
        {            
            PDFDoc doc = new PDFDoc();

            var docA = PdfViewCtrlA.GetDoc();
            var docB = PdfViewCtrlB.GetDoc();

            int totalPagesA = docA.GetPageCount();
            int totalPagesB = docB.GetPageCount();

            for (int i = 1; totalPagesA >= i; i++)
            {
                if (ct.IsCancellationRequested)
                    break;

                if (i > totalPagesB)
                    break;

                var pageA = docA.GetPage(i);
                var pageB = docB.GetPage(i);

                doc.AppendTextDiff(pageA, pageB);
            }

            return doc;
        }

        private void CancelTxtDiff(object param)
        {
            AppCenterHandlerBase.CURRENT.SendEvent("PDF Compare", new Dictionary<string, string>() { ["Compare Text"] = "CANCEL compare ALL pages" });

            mTokenSource.Cancel();
        }

        /// <summary>
        /// Compare the text difference between the currenct selected pages
        /// </summary>
        private async Task PageTxtDiffAsync()
        {
            if (PdfViewCtrlA == null || PdfViewCtrlB == null)
                return;

            AppCenterHandlerBase.CURRENT.SendEvent("PDF Compare", new Dictionary<string, string>() { ["Compare Text"] = "Compare CURRENT pages" });

            PDFDoc rstDoc = new PDFDoc(); ;

            await Task.Run(() =>
            {
                var currentPageA = PdfViewCtrlA.GetCurrentPage();
                var currentPageB = PdfViewCtrlB.GetCurrentPage();

                var docA = PdfViewCtrlA.GetDoc();
                var docB = PdfViewCtrlB.GetDoc();

                rstDoc.AppendTextDiff(docA.GetPage(currentPageA), docB.GetPage(currentPageB));
            });

            IsTextDiffResultVisible = true;
            IsPdfAVisible = false;
            IsPdfBVisible = false;
            IsPdfAOpenPageVisible = false;
            IsPdfBOpenPageVisible = false;
            IsTextDiffEnabled = false;

            PdfViewCtrlTxtDiff = new PDFViewCtrl();
            PdfViewCtrlTxtDiff.SetPagePresentationMode(PDFViewCtrlPagePresentationMode.e_facing_continuous);
            PdfViewCtrlTxtDiff.SetDoc(rstDoc);

            mToolManagerTxtDiff = new ToolManager(PdfViewCtrlTxtDiff);
            _AnnotToolbar.ToolManager = mToolManagerTxtDiff;            
        }

        private async Task SaveTxtDiffAsync()
        {
            if (PdfViewCtrlTxtDiff == null)
                return;

            if (PdfViewCtrlTxtDiff.HasDocument == false)
                return;

            await SaveAsAsync(PdfViewCtrlTxtDiff.GetDoc());
        }

        private async Task SaveAsAsync(PDFDoc doc)
        {
            if (doc == null)
                return;

            FileSavePicker savePicker = new FileSavePicker();
            savePicker.SuggestedStartLocation = PickerLocationId.DocumentsLibrary;

            savePicker.FileTypeChoices.Add("PDF Document", new List<string>() { ".pdf" });
            savePicker.SuggestedFileName = "New Document"; // TODO: string from resource

            StorageFile file = await savePicker.PickSaveFileAsync();
            if (file != null)
            {
                CachedFileManager.DeferUpdates(file);

                await doc.SaveToNewLocationAsync(file, SDF.SDFDocSaveOptions.e_incremental);

                FileUpdateStatus status = await CachedFileManager.CompleteUpdatesAsync(file);
                if (status == FileUpdateStatus.Complete)
                {
                    // TOOD: file saved status
                }
                else
                {
                    // TODO: file not saved status
                }
            }
        }

        private void CloseTxtDiff(object param)
        {
            IsTextDiffResultVisible = false;
            IsPdfAVisible = true;
            IsPdfBVisible = true;

            IsTextDiffEnabled = true;

            _PdfViewCtrlTxtDiff = null;
        }

        private void PageBack(object param)
        {
            if (IsAnyDocsModified())
            {
                // TODO: Implement logic to confirm exit
            }

            CleanUp();

            if (_view != null)
                _view.CloseControl();
        }

        private void ZoomIn(object param)
        {
            mlastZoomChange = Environment.TickCount;

            if (param.ToString() == "pdfA")
            {
                PerformZoom(PdfViewCtrlA, true);

                if (SyncZoom)
                    PdfViewCtrlB.SetZoom(PdfViewCtrlA.GetZoom());
            }
            else if (param.ToString() == "pdfB")
            {
                PerformZoom(PdfViewCtrlB, true);

                if (SyncZoom)
                    PdfViewCtrlA.SetZoom(PdfViewCtrlB.GetZoom());
            }
            else if (param.ToString() == "txtDiff")
            {
                PerformZoom(PdfViewCtrlTxtDiff, true);
            }
        }

        private void ZoomOut(object param)
        {
            mlastZoomChange = Environment.TickCount;

            if (param.ToString() == "pdfA")
            {
                PerformZoom(PdfViewCtrlA, false);

                if (SyncZoom)
                    PdfViewCtrlB.SetZoom(PdfViewCtrlA.GetZoom());
            }
            else if (param.ToString() == "pdfB")
            {
                PerformZoom(PdfViewCtrlB, false);

                if (SyncZoom)
                    PdfViewCtrlA.SetZoom(PdfViewCtrlB.GetZoom());
            }
            else if (param.ToString() == "txtDiff")
            {
                PerformZoom(PdfViewCtrlTxtDiff, false);
            }
        }

        private void PerformZoom(PDFViewCtrl pDFViewCtrl, bool isZoomIn)
        {
            if (!pDFViewCtrl.HasDocument)
                return;

            var zoomFactor = DEFAULT_ZOOM_FACTOR_STEP; // Amount of zoom on each call

            double currentZoom = pDFViewCtrl.GetZoom();
            var width = (int)(pDFViewCtrl.ActualWidth / 2); // Ensure to keep the zooming centered
            var height = (int)(pDFViewCtrl.ActualHeight / 2); // Ensure to keep the zooming centered

            if (isZoomIn)
                pDFViewCtrl.SetZoom(width, height, currentZoom * zoomFactor, true);
            else
                pDFViewCtrl.SetZoom(width, height, currentZoom / zoomFactor, true);
        }

        #endregion

        #region Events
        float mLastPdfViewAZoom = 0;
        float mLastPdfViewBZoom = 0;

        private void scrollViewer_ViewChanged(object sender, ScrollViewerViewChangedEventArgs e)
        {
            if (Environment.TickCount - mlastZoomChange < mZoomLoopbackTimeout)
                return;
                      
            if (PdfViewCtrlA == null || PdfViewCtrlB == null)
                return;

            if (PdfViewCtrlA.HasDocument == false || PdfViewCtrlB.HasDocument == false)
                return;

            if (mScrollViewerA == null || mScrollViewerB == null)
                return;

            // NOTE: prevent mouse wheel zooming from triggering scrolling which is unstable
            {
                if (sender == mScrollViewerA)
                {
                    if (mLastPdfViewAZoom == 0)
                    {
                        mLastPdfViewAZoom = mScrollViewerA.ZoomFactor;
                    }
                    else
                    {
                       if (mLastPdfViewAZoom != mScrollViewerA.ZoomFactor)
                        {
                            mLastPdfViewAZoom = mScrollViewerA.ZoomFactor;
                            mlastZoomChange = Environment.TickCount;
                            return;
                        }

                    }
                }
                else
                {
                    if (mLastPdfViewBZoom == 0)
                    {
                        mLastPdfViewBZoom = mScrollViewerB.ZoomFactor;
                    }
                    else
                    {
                        if (mLastPdfViewBZoom != mScrollViewerB.ZoomFactor)
                        {
                            mLastPdfViewBZoom = mScrollViewerB.ZoomFactor;
                            mlastZoomChange = Environment.TickCount;
                            return;
                        }

                    }

                }
            }

            if (SyncScrolling)
            {
                double sourceLastVerticalOffset = 0;

                // Note: Using the scroll wheel of a mouse causes the built-in ScrollViewer animation to trigger a few times, so the 
                // timeout helps to make it smooth
                if (mlastScrollingElement != sender && Environment.TickCount - mlastScrollChange < mScrollLoopbackTimeout)
                    return;

                mlastScrollingElement = sender;
                mlastScrollChange = Environment.TickCount;

                ScrollViewer sourceScrollViewer;
                ScrollViewer targetScrollViewer;
                double sourcePDFViewCtrlZoom;
                double targetPDFViewCtrlZoom;

                if (sender == mScrollViewerA)
                {
                    sourceScrollViewer = mScrollViewerA;
                    targetScrollViewer = mScrollViewerB;

                    sourcePDFViewCtrlZoom = PdfViewCtrlA.GetZoom();
                    targetPDFViewCtrlZoom = PdfViewCtrlB.GetZoom();

                    sourceLastVerticalOffset = mLastVerticalOffsetA;
                }
                else
                {
                    sourceScrollViewer = mScrollViewerB;
                    targetScrollViewer = mScrollViewerA;

                    sourcePDFViewCtrlZoom = PdfViewCtrlB.GetZoom();
                    targetPDFViewCtrlZoom = PdfViewCtrlA.GetZoom();

                    sourceLastVerticalOffset = mLastVerticalOffsetB;
                }

                if (_syncPage)
                {
                    // NOTE: Here we handle different zoom levels for each document to guarantee a proper sync of pages
                    var factorAdjust = targetPDFViewCtrlZoom / sourcePDFViewCtrlZoom;

                    targetScrollViewer.ChangeView(null, sourceScrollViewer.VerticalOffset * factorAdjust, null, false);
                }
                else
                {
                    var verticalOffset = targetScrollViewer.VerticalOffset;

                    if (sourceScrollViewer.VerticalOffset > sourceLastVerticalOffset)
                    {
                        verticalOffset += sourceScrollViewer.VerticalOffset - sourceLastVerticalOffset;
                    }
                    else
                    {
                        verticalOffset -= Math.Abs(sourceScrollViewer.VerticalOffset - sourceLastVerticalOffset);
                    }

                    targetScrollViewer.ChangeView(null, verticalOffset, null, true);
                }
            }

            mLastVerticalOffsetA = mScrollViewerA.VerticalOffset;
            mLastVerticalOffsetB = mScrollViewerB.VerticalOffset;
        }

        private void MPDFViewA_OnSetDoc()
        {
            mScrollViewerA = PdfViewCtrlA.GetScrollViewer();
            mScrollViewerA.ViewChanged += scrollViewer_ViewChanged;

            mLastVerticalOffsetA = mScrollViewerA.VerticalOffset;

            PdfViewCtrlA.SetZoom(DEFAULT_ZOOM_START_POSITION);
            PdfViewCtrlA.SetZoomEnabled(false);

            if (PdfViewCtrlB != null && PdfViewCtrlB.HasDocument)
                IsTextDiffEnabled = true;

            IsPdfAOpenPageVisible = false;

            AnnotToolbar.IsEnabled = true;
            IsToolbarVisible = true;

            mToolManagerDocA = new ToolManager(PdfViewCtrlA);

            PageNumbIndicatorPdfA.PDFViewCtrl = PdfViewCtrlA;
            PdfViewCtrlA.PointerMoved += PdfViewCtrlA_PointerMoved;
            PdfViewCtrlA.OnPageNumberChanged += PdfViewCtrl_OnPageNumberChanged;

            UpdateTextDiffCurrentPage();
        }

        private void MPDFViewB_OnSetDoc()
        {
            mScrollViewerB = PdfViewCtrlB.GetScrollViewer();
            mScrollViewerB.ViewChanged += scrollViewer_ViewChanged;

            mLastVerticalOffsetB = mScrollViewerB.VerticalOffset;

            PdfViewCtrlB.SetZoom(DEFAULT_ZOOM_START_POSITION);
            PdfViewCtrlB.SetZoomEnabled(false);

            if (PdfViewCtrlA != null && PdfViewCtrlA.HasDocument)
                IsTextDiffEnabled = true;

            AnnotToolbar.IsEnabled = true;
            IsToolbarVisible = true;

            mToolManagerDocB = new ToolManager(PdfViewCtrlB);

            PageNumbIndicatorPdfB.PDFViewCtrl = PdfViewCtrlB;
            PdfViewCtrlB.PointerMoved += PdfViewCtrlB_PointerMoved;
            PdfViewCtrlB.OnPageNumberChanged += PdfViewCtrl_OnPageNumberChanged;

            UpdateTextDiffCurrentPage();
        }

        private void UpdateTextDiffCurrentPage()
        {
            if ((PdfViewCtrlA == null) || (PdfViewCtrlB == null))
                return;

            TxtDiffCurrentPageText = string.Format(mTxtDiffCurrentPageTextCached, PdfViewCtrlA.GetCurrentPage().ToString(), PdfViewCtrlB.GetCurrentPage().ToString());
        }

        private void PdfViewCtrl_OnPageNumberChanged(int current_page, int num_pages)
        {
            UpdateTextDiffCurrentPage();
        }
        #endregion

    }
}
