﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Collections.ObjectModel;

using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Media;
using Windows.Foundation;

using UIRect = Windows.Foundation.Rect;
using UIPoint = Windows.Foundation.Point;

using pdftron.PDF.Tools.Controls.Viewer;
using pdftron.PDF.Tools.Controls.ViewModels.Common;
using pdftron.PDF.Tools.Utilities;


namespace pdftron.PDF.Tools.Controls.ViewModels
{
    public class ContextMenuControlViewModel : ViewModelBase, IDisposable
    {
        #region Defaults
        //Default ideal values
        private const double _IDEAL_SPACE = 20;
        private const double _SIDE_MARGIN = 50;
        private const double _SIDE_MARGIN_THRESHOLD = 10;
        private const int _IDEAL_BUTTON_WIDTH = 40;

        //Max number of items for the collections
        private const int _MAX_OVERFLOW_ITEMS = 10;
        private const int _MAX_BUTTON_ITEMS = 12;
        #endregion

        #region Private Properties
        private bool _IsOpen;
        private bool _ShowOverflow;

        // Popup position
        private double _HorizontalOffset;
        private double _VerticalOffset;

        private int _MaxButtonsPerRow = 6;
        private int _OverflowContainerRow = 2;

        private UIRect _CurrentPosition;

        private Thickness _OverflowAbove = new Thickness(0, 0, 0, 1);
        private Thickness _OverflowBelow = new Thickness(0, 1, 0, 0);
        private Thickness _OverflowBorderThickness;

        // ContextMenuControl dimensions
        private double _ActualWidth;
        private double _ActualHeight;
        private double _OldHeight;

        private FrameworkElement _frameworkElement; // used for positioning reference
        private Button _OverflowButton = new Button();

        // Button styling for icons
        private FontFamily _IconFontFamily = new FontFamily("ms-appx:///pdftron.PDF.Tools/Resources/glyph-icon.ttf#PDFTron Glyphicons");
        private double _IconFontSize = 20;
        #endregion

        #region Constructors

        public ContextMenuControlViewModel(FrameworkElement fe)
        {
            _frameworkElement = fe;
            InitCollections();
        }

        public ContextMenuControlViewModel(FrameworkElement fe, Dictionary<string, string> buttonDictionary)
        {
            _frameworkElement = fe;
            InitCollections();
            CreateButtonsFromDictionary(buttonDictionary);
        }
        #endregion

        #region Init
        private void InitCollections()
        {
            ButtonCollection = new ObservableCollection<Button>();
            OverflowCollection = new ObservableCollection<MenuFlyoutItem>();
            InitOverflowButton();
            OverflowCollection.CollectionChanged += OverflowCollection_CollectionChanged;
            ButtonCollection.CollectionChanged += ButtonCollection_CollectionChanged;
        }

        private void InitOverflowButton()
        {
            _OverflowButton.Tag = "overflow";
            _OverflowButton.Content = GetIconFromName("overflow");
            _OverflowButton.Click += HandleClickOverflow;
            _OverflowButton.FontFamily = _IconFontFamily;
            _OverflowButton.FontSize = _IconFontSize;
        }
        #endregion

        #region Public Properties
        /// <summary>
        /// The collection of buttons that are used as an ItemSource within the view.
        /// </summary>
        public ObservableCollection<Button> ButtonCollection { get; private set; }

        /// <summary>
        /// The collection of MenuFlyoutItemBase that are used as the overflow ItemSource within the view.
        /// </summary>
        public ObservableCollection<MenuFlyoutItem> OverflowCollection { get; private set; }

        /// <summary>
        /// Visibilty of underlying control
        /// </summary>
        public bool IsOpen
        {
            get { return _IsOpen; }
            set
            {
                if (Set(ref _IsOpen, value))
                {
                    RaisePropertyChanged();
                }
            }
        }

        /// <summary>
        /// Visibilty of overflow menu
        /// </summary>
        public bool ShowOverflow
        {
            get { return _ShowOverflow; }
            private set
            {
                if (Set(ref _ShowOverflow, value))
                {
                    RaisePropertyChanged();

                    ViewerSettings.IsContextMenuFlyoutOpened = _ShowOverflow;
                }
            }
        }

        /// <summary>
        /// Horizontal position
        /// </summary>
        public double HorizontalOffset
        {
            get { return _HorizontalOffset; }
            private set
            {
                if (Set(ref _HorizontalOffset, value))
                {
                    RaisePropertyChanged();
                }
            }
        }

        /// <summary>
        /// Vertical position (top of application window to top of control)
        /// </summary>
        public double VerticalOffset
        {
            get { return _VerticalOffset; }
            private set
            {
                if (Set(ref _VerticalOffset, value))
                {
                    RaisePropertyChanged();
                }
            }
        }

        /// <summary>
        /// Maximum amount of buttons allowed per row from the ButtonCollection
        /// </summary>
        public int MaxButtonsPerRow
        {
            get { return _MaxButtonsPerRow; }
            private set
            {
                if (Set(ref _MaxButtonsPerRow, value))
                {
                    RaisePropertyChanged();
                }
            }
        }

        /// <summary>
        /// Above or below the ButtonCollection
        /// </summary>
        public int OverflowContainerRow
        {
            get { return _OverflowContainerRow; }
            private set
            {
                if (Set(ref _OverflowContainerRow, value))
                {
                    RaisePropertyChanged();
                }
            }
        }

        /// <summary>
        /// Helper to position the horizontal border separating
        /// the ButtonCollection from the OverflowCollection
        /// </summary>
        public Thickness OverflowBorderThickness
        {
            get { return _OverflowBorderThickness; }
            private set
            {
                if (Set(ref _OverflowBorderThickness, value))
                {
                    RaisePropertyChanged();
                }
            }
        }
        #endregion

        #region Private Methods
        /// <summary>
        /// Automatically adds or removes the overflow button
        /// based on if the OverflowCollection has items
        /// <summary>
        private void OverflowCollection_CollectionChanged(object sender, System.Collections.Specialized.NotifyCollectionChangedEventArgs e)
        {
            if (OverflowCollection == null || ButtonCollection == null)
                return;

            if (OverflowCollection.Count > 0)
            {
                if (!ButtonCollection.Contains(_OverflowButton))
                {
                    if (ButtonCollection.Count == _MAX_BUTTON_ITEMS)
                    {
                        Button button = ButtonCollection[ButtonCollection.Count - 1];
                        if(button != null)
                        {
                            MenuFlyoutItem item = new MenuFlyoutItem();
                            item.Text = ToolTipService.GetToolTip(button).ToString();
                            item.Tag = button.Tag;

                            OverflowCollection.CollectionChanged -= OverflowCollection_CollectionChanged;
                            AddItemToOverflowCollection(item);
                            OverflowCollection.CollectionChanged -= OverflowCollection_CollectionChanged;
                        }

                        RemoveButtonFromCollection(ButtonCollection[ButtonCollection.Count - 1]);
                    }
                    if(ButtonCollection.Count > 1)
                        ButtonCollection.Insert(ButtonCollection.Count, _OverflowButton);
                }
            }
            else
            {
                if (ButtonCollection.Contains(_OverflowButton))
                {
                    ButtonCollection.Remove(_OverflowButton);
                }
            }

            if(ButtonCollection.Count == 0 && OverflowCollection.Count > 0)
            {
                ShowOverflow = true;
            }
        }

        /// <summary>
        /// Automatically adjusts how many items are in the top and bottom rows
        /// and adjusts widths in the case of button content being filled with text instead of icons
        /// <summary>
        private void ButtonCollection_CollectionChanged(object sender, System.Collections.Specialized.NotifyCollectionChangedEventArgs e)
        {
            if (ButtonCollection == null)
                return;

            int count = ButtonCollection.Count;

            //Ensures we have as little empty space on the second row as possible
            if (count > _MAX_BUTTON_ITEMS / 2 && count < _MAX_BUTTON_ITEMS - 2)
            {
                MaxButtonsPerRow = (int)Math.Round(count / 2.0);
            }
            else
            {
                MaxButtonsPerRow = _MAX_BUTTON_ITEMS / 2;
            }

            if (count <= 0)
                return;

            //Ensures that all elements within the variableSizedWrapGrid occupy the same space
            double maxWidth = ButtonCollection.Max(x => x.Width);
            if (maxWidth <= _IDEAL_BUTTON_WIDTH)
                return;

            foreach (Button btn in ButtonCollection)
            {
                if (btn.Width != maxWidth)
                {
                    btn.Width = maxWidth;
                }
            }
        }

        /// <summary>
        /// Get the icon unicode from name for the Font Icon: glyph-icon.ttf
        /// </summary>
        /// <param name="name">The name of the action</param>
        /// <returns>Icon unicode</returns>
        private string GetIconFromName(string name)
        {
            switch (name.ToLower())
            {
                case "ink":
                    return "\uE719";
                case "signature":
                    return "\uE703";
                case "marker":
                    return "\uE722";
                case "text":
                    return "\uE725";
                case "arrow":
                    return "\uE755";
                case "line":
                    return "\uE756";
                case "rectangle":
                    return "\uE749";
                case "ellipse":
                    return "\uE747";
                case "add note":
                case "note":
                case "view note":
                    return "\uE744";
                case "style":
                    return "\uE596";
                case "delete":
                    return "\uE591";
                case "flatten":
                    return "\uE789";
                case "rotate":
                    return "\uE524";
                case "calibrate":
                    return "\uE787";
                case "done":
                    return "\uE504";
                case "stamper filepicker":
                    return "\uE647";
                case "image":
                case "stamper camera":
                    return "\uE649";
                case "edit text":
                case "type":
                    return "\uE790";
                case "link":
                    return "\uE556";
                case "share":
                    return "\uE560";
                case "read":
                case "play":
                case "speech":
                    return "\uE634";
                case "squiggly":
                    return "\uE729";
                case "highlight":
                    return "\uE726";
                case "strikeout":
                    return "\uE732";
                case "underline":
                    return "\uE735";
                case "copy text":
                case "copy annot text":
                    return "\uE788";
                case "copy":
                case "copy to clipboard":
                case "copy annot to clipboard":
                    return "\uE785";
                case "paste":
                case "paste annot from clipboard":
                    return "\uE671";
                case "duplicate annot":
                    return "\uE827";
                case "cut annot":
                    return "\uE832";
                case "group":
                    return "\uE791";
                case "ungroup":
                    return "\uE792";
                case "opacity":
                    return "\uE793";
                case "fill color":
                    return "\uE794";
                case "color":
                    return "\uE795";
                case "overflow":
                    return "\uE610";
                case "mic":
                    return "\uE632";

                default:
                    return string.Empty;
            }
        }

        /// <summary>
        /// Gets the position of the element in PDFview screen coordinates.
        /// /// <summary>
        private UIRect GetElementRect(FrameworkElement element)
        {
            GeneralTransform elementtransform = element.TransformToVisual(null);
            UIRect rect = elementtransform.TransformBounds(new UIRect(new UIPoint(0, 0), new Size(element.ActualWidth, element.ActualHeight)));
            return rect;
        }

        /// <summary>
        /// Gets the current window's bounds which will be used later to position
        /// the flyout within these bounds
        /// <summary>
        private UIRect GetWindowBounds()
        {
            UIRect pdfrect = GetElementRect(_frameworkElement);
            UIRect windrect = Window.Current.Bounds;

            // windrect might not start at 0 (if you have multiple monitors)
            // So shift it to 0, since all calculations are relative to it.
            windrect.X = 0;
            windrect.Y = 0;

            double left = Math.Max(windrect.Left, pdfrect.Left);
            double top = Math.Max(windrect.Top, pdfrect.Top);
            double right = Math.Min(windrect.Right, pdfrect.Right);
            double bottom = Math.Min(windrect.Bottom, pdfrect.Bottom);
            return new UIRect(left, top, right - left, bottom - top);
        }

        /// <summary>
        /// Positions the popup within the bounds of the window
        /// <summary>
        private void PositionPopup()
        {
            UIRect window = GetWindowBounds();

            //Horizontal Offsetting
            double left = Math.Max(_CurrentPosition.Left, window.Left);
            double right = Math.Min(_CurrentPosition.Right, window.Right);

            if (_ActualWidth < window.Width - (2 * _SIDE_MARGIN))
            {
                //Center on top of the element
                HorizontalOffset = ((right + left) / 2) - (_ActualWidth / 2);

                //Check left side
                if (HorizontalOffset < _SIDE_MARGIN)
                {
                    HorizontalOffset = _SIDE_MARGIN;
                }

                if (HorizontalOffset + _ActualWidth > window.Right - _SIDE_MARGIN)
                {
                    HorizontalOffset = window.Right - _ActualWidth - _SIDE_MARGIN;
                }
            }
            else if (_ActualWidth < window.Width - (2 * _SIDE_MARGIN_THRESHOLD))
            {
                HorizontalOffset = (window.Width - _ActualWidth) / 2;
            }

            //Vertical Offsetting
            double topHeight = _CurrentPosition.Top - window.Top;
            double bottomHeight = window.Bottom - _CurrentPosition.Bottom;

            // Base case to adjust the _VerticalOffset if ShowOverflow changes while IsOpen is true          
            if (_VerticalOffset > 0 && (_VerticalOffset + _OldHeight > topHeight || _VerticalOffset + _OldHeight < bottomHeight))
            {
                if (!_ShowOverflow && _OldHeight != _ActualHeight && _OverflowContainerRow == 0)
                {
                    VerticalOffset += _OldHeight - _ActualHeight;
                }
            }
            else if (topHeight >= _ActualHeight + _IDEAL_SPACE)
            {
                VerticalOffset = window.Top + topHeight - (_ActualHeight + _IDEAL_SPACE);
            }
            else if (bottomHeight >= _ActualHeight + _IDEAL_SPACE)
            {
                VerticalOffset = window.Bottom - bottomHeight + _IDEAL_SPACE;
            }
            else if (topHeight >= _ActualHeight)
            {
                VerticalOffset = window.Top + ((topHeight - _ActualHeight) / 2);
            }
            else if (bottomHeight >= _ActualHeight)
            {
                VerticalOffset = window.Bottom - bottomHeight + ((bottomHeight - _ActualHeight) / 2);
            }
            else
            {
                VerticalOffset = window.Top + (window.Height / 2) - _ActualHeight;
            }

            //Adjust the Overflow menu's position based on the Overflow menu's visibility
            if (_ShowOverflow)
            {
                if ((VerticalOffset + _ActualHeight) > window.Bottom)
                {
                    VerticalOffset -= _ActualHeight - _OldHeight;
                    OverflowContainerRow = 0;
                    OverflowBorderThickness = _OverflowAbove;
                }
                else
                {
                    OverflowContainerRow = 2;
                    OverflowBorderThickness = _OverflowBelow;
                }
                if (ButtonCollection.Count == 0)
                {
                    OverflowBorderThickness = new Thickness(0, 0, 0, 0);
                }
            }
        }

        /// <summary>
        /// Handles click and tapped events from the ButtonCollection and OverflowCollection
        /// <summary>
        private void HandleItemClick(object sender, RoutedEventArgs e)
        {
            var element = sender as FrameworkElement;
            if (element != null)
            {
                string itemTag = element.Tag as string;
                RaiseItemClicked(itemTag);
            }
        }

        /// <summary>
        /// Toggle the visibility of the overflow menu
        /// <summary>
        private void HandleClickOverflow(object sender, RoutedEventArgs e)
        {
            ShowOverflow = !ShowOverflow;
        }

        /// <summary>
        /// Raises an event for the user control to be able to subscribe 
        /// to ButtonCollection and OverflowCollection item clicks
        /// <summary>
        private void RaiseItemClicked(string tag)
        {
            _itemClicked?.Invoke(tag);
        }

        /// <summary>
        /// Provide a point in screen space that the menu should be attached to
        /// </summary>
        /// <param name="p">Note: Not a pdftron point</param>
        public void TargetPoint(UIPoint p)
        {
            UIRect r = GetElementRect(_frameworkElement);
            _CurrentPosition = new UIRect(p.X + r.Left, p.Y + r.Top, 1, 1);
        }

        /// <summary>
        /// Gives a target rectangle as 4 points in PDFViewCtrl space
        /// </summary>
        public void TargetSquare(double x1, double y1, double x2, double y2)
        {
            UIRect r = GetElementRect(_frameworkElement);

            double minx = Math.Min(x1, x2) + 5; // Add 5 to compensate for calculation error
            double miny = Math.Min(y1, y2);
            double w = Math.Abs(x2 - x1);
            double h = Math.Abs(y2 - y1);
            _CurrentPosition = new UIRect(new UIPoint(r.Left + minx, r.Top + miny), new Size(w, h));
        }

        /// <summary>
        /// Returns true if the ButtonCollection is null or empty
        /// <summary>
        private bool IsButtonCollectionNullEmpty()
        {
            if (ButtonCollection == null || ButtonCollection.Count == 0)
                return true;
            else
                return false;
        }        

        /// <summary>
        /// Attempts to find localized Item content
        /// </summary>
        /// <param name="item">Resource's name that should be searched for</param>
        private string GetStringFromResource(string item)
        {
            string text = Tools.ResourceHandler.GetString(item);
            if (text == string.Empty)
            {
                return item;
            }
            else
            {
                return text;
            }
        }
        #endregion

        #region Public Methods
        /// <summary>
        /// Create and add Buttons to the ButtonCollection from a Dictionary.
        /// </summary>
        /// <param name="buttonDictionary">
        /// Keys will be used for Button Tags and Icons, 
        /// values will be used for text Content and Tooltips
        /// </param>
        public void CreateButtonsFromDictionary(Dictionary<string, string> buttonDictionary)
        {
            if (ButtonCollection == null || buttonDictionary == null)
                return;

            if (ButtonCollection.Count > 0)
            {
                ClearButtonCollection();
            }

            IEnumerable<string> buttanKeys = null;
            if (buttonDictionary != null)
            {
                buttanKeys = buttonDictionary.Keys;
            }
            else
            {
                return;
            }

            foreach (string key in buttanKeys)
            {
                if (ButtonCollection.Count >= _MAX_BUTTON_ITEMS)
                {
                    if (OverflowCollection == null || OverflowCollection.Count >= _MAX_OVERFLOW_ITEMS)
                        return;

                    MenuFlyoutItem item = new MenuFlyoutItem();
                    item.Text = buttonDictionary[key];
                    item.Tag = key;
                    AddItemToOverflowCollection(item);
                }

                else
                {
                    Button btn = new Button();
                    TextBlock tb = new TextBlock();
                    tb.Text = buttonDictionary[key];
                    tb.FontSize = 14;
                    btn.Content = tb;
                    btn.Tag = key;
                    ToolTipService.SetToolTip(btn, buttonDictionary[key]);
                    AddButtonToCollection(btn);
                }
            }

        }

        /// <summary>
        /// Add a single Button to the ObservableCollection
        /// </summary>
        /// <param name="button">The button to be added to the collection</param>
        public void AddButtonToCollection(Button button)
        {
            if (ButtonCollection == null || button == null)
                return;

            if (ButtonCollection.Count < _MAX_BUTTON_ITEMS)
            {
                button.Click += HandleItemClick;
                if (!string.IsNullOrEmpty(button.Tag.ToString()))
                {
                    string icon = GetIconFromName(button.Tag.ToString());
                    if (icon != string.Empty)
                    {
                        button.FontFamily = _IconFontFamily;
                        button.FontSize = _IconFontSize;
                        button.Content = icon;
                    }
                    else
                    {
                        //Will use text as button's content, adjust the button's width to fit the text
                        button.Measure(new Size(Double.PositiveInfinity, Double.PositiveInfinity));
                        button.Width = button.DesiredSize.Width + 4;
                        button.Padding = new Thickness(0, 0, 0, 0);
                    }
                }
                ButtonCollection.Add(button);
            }
        }

        /// <summary>
        /// Adds a single MenuFlyoutItemBase to the ObservableCollection.
        /// </summary>
        /// <param name="item">The MenuFlyoutItem to be added to the collection</param>
        public void AddItemToOverflowCollection(MenuFlyoutItem item)
        {
            if (OverflowCollection == null || item == null)
                return;

            if (OverflowCollection.Count < _MAX_OVERFLOW_ITEMS)
            {
                item.Tapped += HandleItemClick;
                OverflowCollection.Add(item);
            }
        }

        /// <summary>
        /// Creates MenuFlyoutItems and adds them to the collection.
        /// </summary>
        /// <param name="overflowDictionary">
        /// Keys are used MenuFlyoutItem Tags, Values are used for content
        /// </param>
        public void CreateOverflowFromDictionary(Dictionary<string, string> overflowDictionary)
        {
            if (overflowDictionary == null || OverflowCollection == null)
                return;

            if(OverflowCollection.Count > 0)
            {
                ClearOverflowCollection();
            }

            if(OverflowCollection != null && overflowDictionary != null)
            {
                IEnumerable<string> menuKeys = overflowDictionary.Keys;
                foreach (string key in menuKeys)
                {
                    if (OverflowCollection.Count == _MAX_OVERFLOW_ITEMS)
                        return;

                    MenuFlyoutItem menuItem = new MenuFlyoutItem();
                    menuItem.Text = overflowDictionary[key];
                    menuItem.Tag = key.ToLower();
                    AddItemToOverflowCollection(menuItem);
                }
            }
        }

        /// <summary>
        /// Initializes a base set of tools within the ButtonCollection and a base set
        /// of tools within the OverflowCollection.
        /// </summary>
        public void InitBaseTools()
        {
            if (OverflowCollection != null)
            {
                if (OverflowCollection.Count > 1)
                    return;
            }
            Dictionary<string, string> creationItemsDict = new Dictionary<string, string>();
            creationItemsDict["note"] = GetStringFromResource("CreateOption_Note");
            creationItemsDict["signature"] = GetStringFromResource("CreateOption_Signature");
            creationItemsDict["ink"] = GetStringFromResource("CreateOption_Ink");
            creationItemsDict["marker"] = GetStringFromResource("CreateOption_Marker");
            creationItemsDict["text"] = GetStringFromResource("CreateOption_Text");
            creationItemsDict["arrow"] = GetStringFromResource("CreateOption_Arrow");
            creationItemsDict["line"] = GetStringFromResource("CreateOption_Line");
            creationItemsDict["rectangle"] = GetStringFromResource("CreateOption_Rectangle");
            creationItemsDict["ellipse"] = GetStringFromResource("CreateOption_Ellipse");
            creationItemsDict["image"] = GetStringFromResource("CreateOption_Image");
            CreateButtonsFromDictionary(creationItemsDict);

            Dictionary<string, string> creationOverflowDict = new Dictionary<string, string>();
            creationOverflowDict["paste"] = GetStringFromResource("CreateOption_PasteAnnot");
            CreateOverflowFromDictionary(creationOverflowDict);
        }

        /// <summary>
        /// Clears the Button collection.
        /// </summary>
        public void ClearButtonCollection()
        {
            if (ButtonCollection == null)
                return;

            foreach (Button item in ButtonCollection)
            {
                if (item != null)
                    item.Click -= HandleItemClick;
            }
            ButtonCollection.Clear();
        }

        /// <summary>
        /// Removes a button from the collection.
        /// </summary>
        public void RemoveButtonFromCollection(Button button)
        {
            if (ButtonCollection == null)
                return;

            int index = ButtonCollection.IndexOf(button);
            if (index != -1 && ButtonCollection[index] != null)
            {
                ButtonCollection[index].Click -= HandleItemClick;
                ButtonCollection.RemoveAt(index);
            }
        }

        /// <summary>
        /// Clears the overflow collection.
        /// </summary>
        public void ClearOverflowCollection()
        {
            if (OverflowCollection == null)
                return;

            foreach (MenuFlyoutItemBase item in OverflowCollection)
            {
                if(item != null)
                    item.Tapped -= HandleItemClick;
            }
            OverflowCollection.Clear();
            if (ButtonCollection != null && _OverflowButton != null)
            {
                if (ButtonCollection.Contains(_OverflowButton))
                {
                    ButtonCollection.Remove(_OverflowButton);
                }
            }

        }

        /// <summary>
        /// Makes the control visible
        /// </summary>
        /// <remarks>
        /// A base set of tools will be initialized if the ButtonCollection is empty
        /// </remarks>
        public void Show()
        {
            if (_CurrentPosition == null || IsButtonCollectionNullEmpty())
                return;

            if (OverflowCollection != null && OverflowCollection.Count > 0)
            {
                ShowOverflow = ViewerSettings.IsContextMenuFlyoutOpened;
            }

            IsOpen = true;
        }

        /// <summary>
        /// Makes the control visible
        /// </summary>
        /// <remarks>
        /// A base set of tools will be initialized if the ButtonCollection is empty
        /// </remarks>
        public void Show(UIPoint point)
        {
            if (point == null || IsButtonCollectionNullEmpty())
                return;

            if (OverflowCollection != null && OverflowCollection.Count > 0)
            {
                ShowOverflow = ViewerSettings.IsContextMenuFlyoutOpened;
            }

            TargetPoint(point);
            IsOpen = true;
        }

        /// <summary>
        /// Hides the control.
        /// </summary>
        public void Hide()
        {
            IsOpen = false;
        }

        /// <summary>
        /// Set the dimensions of the context menu.
        /// </summary>
        public void SetDimensions(double width, double height)
        {
            _OldHeight = _ActualHeight;
            _ActualWidth = width;
            _ActualHeight = height;
            PositionPopup();
        }

        public void Dispose()
        {
            if(_OverflowButton != null)
            {
                _OverflowButton.Click -= HandleClickOverflow;
            }

            if(ButtonCollection != null)
            {
                ButtonCollection.CollectionChanged -= ButtonCollection_CollectionChanged;
                ClearButtonCollection();
            }

            if(OverflowCollection != null)
            {
                OverflowCollection.CollectionChanged -= OverflowCollection_CollectionChanged;
                ClearOverflowCollection();
            }
        }
        #endregion

        #region Events
        public delegate void ItemClickedDelegate(string tag);
        private ItemClickedDelegate _itemClicked;
        public event ItemClickedDelegate ItemClicked
        {
            add
            {
                if (_itemClicked?.GetInvocationList().Contains(value) != true)
                    _itemClicked += value;
            }
            remove => _itemClicked -= value;
        }
        #endregion
    }
}
