﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

using Windows.ApplicationModel.Resources;
using Windows.Graphics.Imaging;
using Windows.Media.Capture;
using Windows.Storage;
using Windows.Storage.Pickers;
using Windows.Storage.Streams;
using Windows.System.UserProfile;
using Windows.UI.Xaml.Media.Imaging;

using pdftron.Common;
using pdftron.PDF.Tools.Commands;
using pdftron.PDF.Tools.Controls.Viewer;
using pdftron.PDF.Tools.Controls.ViewModels.Common;
using pdftron.PDF.Tools.Utilities;
using pdftron.SDF;


namespace pdftron.PDF.Tools.Controls.ViewModels
{
    public class CreateDocControlViewModel : ViewModelBase
    {
        #region Enums
        public enum CreateDocFrom
        {
            Blank,
            Image
        }

        public enum ImageCreationType
        {
            File,
            Camera,
            None
        }
        #endregion

        #region Constants
        protected const double _inConversionRate = 1 / 72.0;
        protected const double _mmConversionRate = 0.3528;

        private const int MAX_PAGE_COUNT = 50;
        #endregion

        #region Private Properties
        private string _DocumentTitle;

        private int _TotalPages;

        private CreateDocFrom _DocumentType;

        private bool _IsPortrait = true;

        private PaperSizeHolder _SelectedPaperSize;
        private PaperTypeHolder _SelectedPaperType;

        private Dictionary<PaperSizes, Tuple<double, double>> _PaperSizeMap;

        private BitmapImage _previewImage;

        private readonly List<string> letterSizeCultures = new List<string> { "US", "CA", "MX", "CU", "DO", "GT", "CR", "SV", "HN", "CO", "VE", "PH", "CL" };
        private readonly IDictionary<ImageCreationType, StorageFile> _filesPicked = new Dictionary<ImageCreationType, StorageFile>();
        #endregion

        #region Constructors
        /// <summary>
        /// Creates a new CreateDocControlViewModel
        /// </summary>
        public CreateDocControlViewModel()
        {
            InitCollections();

            CMDPickImageFromFile = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.CreateWithParameter(PickImageFromFileAsync);
                command.Execute(parameter);
            });

            CMDPickImageFromCamera = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.CreateWithParameter(PickImageFromCameraAsync);
                command.Execute(parameter);
            });

            CMDCreateDocument = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.CreateWithParameter(CreateDocumentAsync);
                command.Execute(parameter);
            });
        }
        #endregion

        #region Public Properties
        public ObservableCollection<PaperTypeHolder> AvailablePaperTypes { get; private set; }
        public ObservableCollection<PaperSizeHolder> AvailablePaperSizes { get; private set; }

        public string DocumentTitle
        {
            get { return _DocumentTitle; }
            set
            {
                if (Set(ref _DocumentTitle, value))
                {
                    RaisePropertyChanged();
                }
            }
        }

        public CreateDocFrom DocumentType
        {
            get { return _DocumentType; }
            set
            {
                if (Set(ref _DocumentType, value))
                {
                    RaisePropertyChanged();
                    ResetViewModel();
                    RaisePropertyChanged(nameof(IsCreateButtonEnabled));
                }
            }
        }

        public bool IsPortrait
        {
            get { return _IsPortrait; }
            set
            {
                if (Set(ref _IsPortrait, value))
                {
                    RaisePropertyChanged();
                }
            }
        }

        public PaperSizeHolder SelectedPaperSize
        {
            get { return _SelectedPaperSize; }
            set
            {
                if (Set(ref _SelectedPaperSize, value))
                {
                    RaisePropertyChanged();
                }
            }
        }

        public PaperTypeHolder SelectedPaperType
        {
            get { return _SelectedPaperType; }
            set
            {
                if (Set(ref _SelectedPaperType, value))
                {
                    RaisePropertyChanged();
                }
            }
        }

        public int TotalPages
        {
            get { return _TotalPages; }
            set
            {
                if (Set(ref _TotalPages, value))
                {
                    RaisePropertyChanged();
                }
            }
        }

        public string DefaultDocumentName
        {
            get
            {
                ResourceLoader loader = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources");
                return loader.GetString("DocumentCreation_AllDocuments_DefaultTitle");
            }
        }

        public BitmapImage PreviewImage
        {
            get => _previewImage;
            set
            {
                _previewImage = value;
                RaisePropertyChanged();
                RaisePropertyChanged(nameof(IsImageSelected));
                RaisePropertyChanged(nameof(IsCreateButtonEnabled));
            }
        }

        public bool IsImageSelected => _previewImage != null;

        public bool IsCreateButtonEnabled => IsImageSelected || _DocumentType == CreateDocFrom.Blank;
        #endregion

        #region Commands
        public RelayCommand CMDPickImageFromFile { get; }

        public RelayCommand CMDPickImageFromCamera { get; }

        public RelayCommand CMDCreateDocument { get; set; }
        #endregion

        #region Private Methods
        private string GetFileName()
        {
            if (string.IsNullOrEmpty(_DocumentTitle))
                return DefaultDocumentName;

            return _DocumentTitle;
        }

        private Task<PDFDoc> GetBlankPDFDocAsync()
        {
            AppCenterHandlerBase.CURRENT.SendEvent("PDF Creation",
                new Dictionary<string, string>()
                {
                    ["Type"] = "Blank Document"
                });

            AppCenterHandlerBase.CURRENT.SendEvent("PDF Creation",
                new Dictionary<string, string>()
                {
                    ["Paper Type"] = Enum.GetName(typeof(PaperType), SelectedPaperType.PaperType)
                });

            return Task.Run(() =>
            {
                PDFDocCreator.GetPaperDimensionsFromSize(SelectedPaperSize.PaperSize, out double width, out double height);
                if (!IsPortrait)    //landscape
                {
                    double temp = width;
                    width = height;
                    height = temp;
                }

                return PDFDocCreator.Create(SelectedPaperType.PaperType, width, height, _TotalPages);
            });
        }

        private async Task<PDFDoc> GetImagePDFDOC()
        {

            if (_filesPicked.Count == 0)
            {
                return null;
            }

            if (_filesPicked.Count > 0)
            {
                if (_filesPicked.Keys.Contains(ImageCreationType.Camera))
                {
                    AppCenterHandlerBase.CURRENT.SendEvent("PDF Creation",
                        new Dictionary<string, string>()
                        {
                            ["Type"] = "From Camera"
                        });
                }
                else
                {
                    AppCenterHandlerBase.CURRENT.SendEvent("PDF Creation",
                        new Dictionary<string, string>()
                        {
                            ["Type"] = "From Image File"
                        });
                }
            }

            PDFDoc doc = new PDFDoc();

            foreach (StorageFile file in _filesPicked.Values)
            {
                doc = await PDFDocCreator.CreateFromImageAsync(file);
            }

            return doc;
        }

        private string GetFileNameWithExtension()
        {
            string fileName = GetFileName();
            if (fileName.Length < 4)
                fileName += ".pdf";

            if (!fileName.Substring(fileName.Length - 4).Equals(".pdf", StringComparison.OrdinalIgnoreCase))
                fileName += ".pdf";

            return fileName;
        }

        private async Task<IReadOnlyList<StorageFile>> GetOpenFileFromPicker(IList<string> fileTypes)
        {
            ResourceLoader loader = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources");

            FileOpenPicker fileOpenPicker = new FileOpenPicker
            {
                ViewMode = PickerViewMode.Thumbnail,
                CommitButtonText = loader.GetString("DocumentCreation_ImageDocument_PickFileConform")
            };

            foreach (string filetype in fileTypes)
            {
                fileOpenPicker.FileTypeFilter.Add(filetype);
            }

            StorageFile theFile = await fileOpenPicker.PickSingleFileAsync();
            return new List<StorageFile>
            {
                theFile
            };
        }

        private async Task PickImageFromFileSystemAsync()
        {
            try
            {

                IReadOnlyList<StorageFile> pickedFiles = await GetOpenFileFromPicker(ViewerSettings.ImagesTypesForConversion);
                if (pickedFiles[0] == null)
                    return;

                _filesPicked.Clear();
                _filesPicked.Add(ImageCreationType.File, pickedFiles[0]);

                await AddFileToBeShownAsync(_filesPicked[0]);
                RaisePropertyChanged("IsInputValid");
            }
            finally
            {
            }
        }

        private async Task AddFileToBeShownAsync(StorageFile file)
        {
            try
            {
                if (file.FileType == ".xps" || file.FileType == ".oxps")
                {
                    PreviewImage = new BitmapImage
                    {
                        UriSource = new Uri("ms-appx:///Assets/DocumentPage/FilePlaceHolder_pdf.png", UriKind.Absolute)
                    };
                }
                else
                {
                    using (IRandomAccessStream fileStream = await file.OpenReadAsync())
                    {
                        BitmapImage bitmapImage = new BitmapImage();
                        await bitmapImage.SetSourceAsync(fileStream);
                        PreviewImage = bitmapImage;
                    }
                }

                _imageLoaded?.Invoke(file);
            }
            catch
            {
                _imageLoadingFailed?.Invoke();
            }
        }

        private async Task GetImageFromCameraAsync()
        {
            try
            {
                CameraCaptureUI cameraUI = new CameraCaptureUI();
                cameraUI.PhotoSettings.AllowCropping = false;
                cameraUI.PhotoSettings.MaxResolution = CameraCaptureUIMaxPhotoResolution.HighestAvailable;
                cameraUI.PhotoSettings.AllowCropping = true;

                StorageFile capturedMedia = await cameraUI.CaptureFileAsync(CameraCaptureUIMode.Photo);
                if (capturedMedia != null)
                {
                    _filesPicked.Clear();
                    _filesPicked.Add(ImageCreationType.Camera, capturedMedia);

                    await AddFileToBeShownAsync(capturedMedia);
                    RaisePropertyChanged("IsInputValid");
                }
            }
            finally
            {
            }
        }

        private async Task PickImageFromFileAsync(object parameter) => await PickImageFromFileSystemAsync();

        private async Task PickImageFromCameraAsync(object sender) => await GetImageFromCameraAsync();

        #endregion

        #region Public Methods
        /// <summary>
        /// Create a PDF Document and invokes 
        /// NewDocumentCreated which passes the StorageFile document as the parameter
        /// </summary>
        private async Task CreateDocumentAsync(object sender)
        {
            bool success = false;
            StorageFile savedFile = null;
            PDFDoc doc = null;
            try
            {
                if (_DocumentType == CreateDocFrom.Blank)
                    doc = await GetBlankPDFDocAsync();

                else if (_DocumentType == CreateDocFrom.Image)
                {
                    doc = await GetImagePDFDOC();
                }

                if (doc != null)
                {
                    savedFile = await DocumentManager.Instance.SaveToTemporaryFileAsync(doc, GetFileNameWithExtension());
                    doc.GetDocInfo().SetTitle(GetFileName());

                    if (doc != null && savedFile != null)
                        success = true;
                }
            }
            catch (Exception ex)
            {
                AppCenterHandlerBase.CURRENT.TrackError(ex);
            }

            if (savedFile != null)
            {
                if (success)
                {
                    _NewDocumentCreated?.Invoke(savedFile);
                }
                else
                {
                    _documentCreationFailed?.Invoke();
                }
            }
            else
            {
                _documentCreationFailed?.Invoke();
            }
        }
        #endregion

        #region Init
        private void InitCollections()
        {
            AvailablePaperSizes = new ObservableCollection<PaperSizeHolder>();
            AvailablePaperTypes = new ObservableCollection<PaperTypeHolder>();

            InitPapers();
        }

        private void InitPapers()
        {
            ResourceLoader loader = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources");
            PaperSizes regionSize = GetRegionDefaultPageSize();
            double convRate = _inConversionRate;
            string units = loader.GetString("DocumentCreation_AllDocuments_UnitsInch");
            if (regionSize == PaperSizes.A4)
            {
                convRate = _mmConversionRate;
                units = loader.GetString("DocumentCreation_AllDocuments_UnitsMilliMeter");
            }

            foreach (PaperSizes size in (PaperSizes[])Enum.GetValues(typeof(PaperSizes)))
            {
                PDFDocCreator.GetPaperDimensionsFromSize(size, out double height, out double width);
                PaperSizeHolder psh = new PaperSizeHolder(size, width * convRate, height * convRate, units);
                AvailablePaperSizes.Add(psh);
                if (size == regionSize)
                    SelectedPaperSize = psh;
            }

            foreach (PaperType type in (PaperType[])Enum.GetValues(typeof(PaperType)))
            {
                AvailablePaperTypes.Add(new PaperTypeHolder(type));
            }

            SelectedPaperType = AvailablePaperTypes[0]; // Default is Blank Page
        }

        private PaperSizes GetRegionDefaultPageSize()
        {
            PaperSizes paperSize = PaperSizes.A4;

            string language = string.Empty;
            if (GlobalizationPreferences.Languages.Count > 0)
                language = GlobalizationPreferences.Languages[0];

            if (!string.IsNullOrWhiteSpace(language) && language.Length > 2 && letterSizeCultures.Contains(language.Substring(language.Length - 2).ToUpper()))
                paperSize = PaperSizes.Letter;

            return paperSize;
        }

        private void ResetViewModel()
        {
            _DocumentTitle = "";
            RaisePropertyChanged("DocumentTitle");
            if (_DocumentType == CreateDocFrom.Image)
            {
                PreviewImage = null;
                IsPortrait = true;
                TotalPages = 1;
            }
            if (_DocumentType == CreateDocFrom.Blank)
            {
                SelectedPaperType = AvailablePaperTypes[0];
                SelectedPaperSize = AvailablePaperSizes[0];
            }
        }
        #endregion

        #region Events
        public delegate void DocumentCreationFailedDelegate();
        private DocumentCreationFailedDelegate _documentCreationFailed;

        /// <summary>
        /// Raised when a document creation attempt fails
        /// </summary>
        public event DocumentCreationFailedDelegate DocumentCreationFailed
        {
            add
            {
                if (_documentCreationFailed?.GetInvocationList().Contains(value) != true)
                    _documentCreationFailed += value;
            }
            remove => _documentCreationFailed -= value;
        }

        private NewDocumentCreatedDelegate _NewDocumentCreated;
        public delegate void NewDocumentCreatedDelegate(StorageFile file);

        /// <summary>
        /// Raised when a document is successfully created
        /// </summary>
        public event NewDocumentCreatedDelegate NewDocumentCreated
        {
            add
            {
                if (_NewDocumentCreated?.GetInvocationList().Select(x => x.Method).Contains(value.Method) != true)
                    _NewDocumentCreated += value;
            }
            remove => _NewDocumentCreated -= value;
        }

        /// <summary>
        /// Raised when document creation is cancelled.
        /// </summary>
        public delegate void DocumentCreationCancelledDelegate();
        private DocumentCreationCancelledDelegate _DocumentCreationCancelled;
        public event DocumentCreationCancelledDelegate DocumentCreationCancelled
        {
            add
            {
                if (_DocumentCreationCancelled?.GetInvocationList().Select(x => x.Method).Contains(value.Method) != true)
                    _DocumentCreationCancelled += value;
            }
            remove => _DocumentCreationCancelled -= value;
        }

        private ImageLoadedDelegate _imageLoaded;
        public delegate void ImageLoadedDelegate(StorageFile file);

        /// <summary>
        /// Raised when an image has been loaded in the document creation dialog
        /// </summary>
        public event ImageLoadedDelegate ImageLoaded
        {
            add
            {
                if (_imageLoaded?.GetInvocationList().Select(x => x.Method).Contains(value.Method) != true)
                    _imageLoaded += value;
            }
            remove => _imageLoaded -= value;
        }

        private ImageLoadingFailedDelegate _imageLoadingFailed;
        public delegate void ImageLoadingFailedDelegate();

        /// <summary>
        /// Raised when an image failed to during loading in the document creation dialog
        /// </summary>
        public event ImageLoadingFailedDelegate ImageLoadingFailed
        {
            add
            {
                if (_imageLoadingFailed?.GetInvocationList().Select(x => x.Method).Contains(value.Method) != true)
                    _imageLoadingFailed += value;
            }
            remove => _imageLoadingFailed -= value;
        }

        #endregion
    }
}