﻿using pdftron.PDF.Tools.Commands;
using pdftron.PDF.Tools.Controls.ViewModels.Common;
using pdftron.PDF.Tools.Utilities;
using System.Linq;
using System.Threading.Tasks;
using Windows.ApplicationModel.Resources;
using Windows.System;
using Windows.UI.Popups;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Input;

namespace pdftron.PDF.Tools.Controls.ViewModels.FileOpening
{
    public class PasswordViewModel : ViewModelBase
    {
        public PasswordViewModel(PDFDoc doc)
        {
            _Doc = doc;
            InitCommands();
            HandlePassword();
        }

        private PasswordHandledDelegate _passwordHandled;
        public delegate void PasswordHandledDelegate(bool success, string password);
        public event PasswordHandledDelegate PasswordHandled
        {
            add
            {
                if (_passwordHandled?.GetInvocationList().Contains(value) != true)
                    _passwordHandled += value;
            }
            remove => _passwordHandled -= value;
        }

        #region Commands

        private void InitCommands()
        {
            PasswordKeyDownCommand = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.CreateWithParameter(PasswordKeyDownAsync);
                command.Execute(parameter);
            });
            PasswordKeyUpCommand = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.CreateWithParameter(PasswordKeyUpAsync);
                command.Execute(parameter);
            });
            PasswordChangedCommand = new RelayCommand(PasswordChangedCommandImpl);
            PasswordOkPressedCommand = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.CreateWithParameter(PasswordOkPressedAsync);
                command.Execute(parameter);
            });
            PasswordCancelPressedCommand = new RelayCommand(PasswordCancelPressedCommandImpl);
        }

        public RelayCommand PasswordKeyDownCommand { get; private set; }
        public RelayCommand PasswordKeyUpCommand { get; private set; }
        public RelayCommand PasswordChangedCommand { get; private set; }
        public RelayCommand PasswordOkPressedCommand { get; private set; }
        public RelayCommand PasswordCancelPressedCommand { get; private set; }

        private async Task PasswordKeyUpAsync(object parameter)
        {
            KeyRoutedEventArgs args = parameter as KeyRoutedEventArgs;
            if (args?.Key == VirtualKey.Enter)
                await PasswordKeyDownAsync(args.Key);
        }

        private void PasswordChangedCommandImpl(object newPassword)
        {
            if (newPassword is string password)
                PasswordChanged(password);
        }

        private async Task PasswordOkPressedAsync(object parameter) => await VerifyPasswordAsync();

        private void PasswordCancelPressedCommandImpl(object sender)
        {
            CancelPassword();
        }

        #endregion Commands


        private const int DEFAULT_ATTEMPTS_AT_PASSWORD = 3;
        private int _AttemptsAtPassword = 0;
        private PDFDoc _Doc;

        private string _CurrentPassword = string.Empty;
        public string CurrentPassword
        {
            get { return _CurrentPassword; }
            set
            {
                if (Set(ref _CurrentPassword, value))
                {
                    RaisePropertyChanged(nameof(HasPasswordBoxGotContent));
                    IsIncorrectPasswordNotificationVisible = false;
                }
            }
        }

        private bool _IsIncorrectPasswordNotificationVisible = false;
        public bool IsIncorrectPasswordNotificationVisible
        {
            get { return _IsIncorrectPasswordNotificationVisible; }
            set
            {
                Set(ref _IsIncorrectPasswordNotificationVisible, value);
                RaisePropertyChanged(nameof(HasPasswordBoxGotContent));
            }
        }

        public bool HasPasswordBoxGotContent
        {
            get { return !string.IsNullOrEmpty(CurrentPassword) && CurrentPassword.Length > 0 && IsIncorrectPasswordNotificationVisible == false; }
        }

        private void HandlePassword()
        {
            CurrentPassword = string.Empty;
            IsIncorrectPasswordNotificationVisible = false;
            _AttemptsAtPassword = DEFAULT_ATTEMPTS_AT_PASSWORD;
        }

        private void PasswordChanged(string newPassword)
        {
            if (newPassword != null)
            {
                CurrentPassword = newPassword;
            }
        }

        private async Task PasswordKeyDownAsync(object parameter)
        {
            if (parameter is KeyRoutedEventArgs args)
                await PasswordKeyDownAsync(args.Key);
        }

        private async Task PasswordKeyDownAsync(VirtualKey key)
        {
            if (key == VirtualKey.Enter)
            {
                await VerifyPasswordAsync();
            }
            else if (key == VirtualKey.Escape)
            {
                CancelPassword();
            }
        }

        private async Task VerifyPasswordAsync()
        {
            if (string.IsNullOrEmpty(CurrentPassword))
                return;

            if (_Doc.InitStdSecurityHandler(CurrentPassword))
            {
                _passwordHandled?.Invoke(true, CurrentPassword);
            }
            else
            {
                --_AttemptsAtPassword;
                IsIncorrectPasswordNotificationVisible = true;
                if (_AttemptsAtPassword == 0)
                {
                    _passwordHandled?.Invoke(false, null);

                    AnalyticsHandlerBase.CURRENT.SendEvent(AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, "Document Opened XFA");
                    //await System.Threading.Tasks.Task.Delay(200);
                    ResourceLoader loader = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources");

                    ContentDialog dialog = new ContentDialog
                    {
                        Title = loader.GetString("PasswordDialog_ManyFails_Title"),
                        Content = string.Format(loader.GetString("PasswordDialog_ManyFails_Info"), DEFAULT_ATTEMPTS_AT_PASSWORD),
                        CloseButtonText = loader.GetString("pdftron_Controls_Content_Dialog_OK")
                    };

                    await DialogHelper.ShowContentDialogAsync(dialog);
                }
            }
        }

        private void CancelPassword() => _passwordHandled?.Invoke(false, null);
    }
}
