﻿using pdftron.Filters;
using pdftron.PDF.Tools.Commands;
using pdftron.PDF.Tools.Controls.Viewer;
using pdftron.PDF.Tools.Controls.ViewModels.Common;
using pdftron.PDF.Tools.Utilities;
using pdftron.SDF;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Diagnostics.CodeAnalysis;
using System.IO;
using System.Linq;
using System.Text.RegularExpressions;
using System.Threading.Tasks;
using Windows.ApplicationModel.Resources;
using Windows.Storage;
using Windows.System;
using Windows.UI.Popups;
using Windows.UI.Xaml.Controls;

namespace pdftron.PDF.Tools.Controls.ViewModels.FileOpening
{
    public class PdfPackageViewModel : ViewModelBase
    {
        private ObservableCollection<PdfPackageItem> _packageItems;
        private string _displayFileName = string.Empty;
        private readonly List<PDFDoc> _documentStack;
        private bool _errorHappened;
        private string _errorMessage;

        private StorageFile _selectedFile;

        private OnFileSelectedHandler _fileSelcted;
        public delegate void OnFileSelectedHandler(PDFDoc doc, StorageFile file, bool err, string errorMessage);
        public event OnFileSelectedHandler FileSelcted
        {
            add
            {
                if (_fileSelcted?.GetInvocationList().Contains(value) != true)
                    _fileSelcted += value;
            }
            remove => _fileSelcted -= value;
        }

        public RelayCommand ItemClickCommand { get; private set; }

        public RelayCommand PackageDialogBackCommand { get; private set; }

        public PdfPackageViewModel(PDFDoc doc, string fileName)
        {
            InitCommands();
            PackageItems = new ObservableCollection<PdfPackageItem>();
            _documentStack = new List<PDFDoc>
            {
                doc
            };

            DisplayFileName = fileName;
            PopulatePackageList();
        }

        public ObservableCollection<PdfPackageItem> PackageItems
        {
            get => _packageItems;
            set => Set(ref _packageItems, value);
        }

        public string DisplayFileName
        {
            get => _displayFileName;
            set => Set(ref _displayFileName, value);
        }

        private void InitCommands()
        {
            ItemClickCommand = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.CreateWithParameter(ItemClickAsync);
                command.Execute(parameter);
            });
            PackageDialogBackCommand = new RelayCommand(PackageDialogBack);
        }

        private async Task ItemClickAsync(object parameter)
        {
            if (parameter is PdfPackageItem item)
            {
                if (item.IsPDF)
                {
                    await OpenPdfAsync(item);
                }
                else
                {
                    await OpenNonPdfAsync(item);
                }
            }
        }

        private void PackageDialogBack(object sender) => HandleBackButton();

        private void PopulatePackageList()
        {
            PackageItems.Clear();
            foreach (PdfPackageItem item in GetPackageContent(_documentStack[_documentStack.Count - 1]))
                PackageItems.Add(item);
        }

        private IList<PdfPackageItem> GetPackageContent(PDFDoc doc)
        {
            List<PdfPackageItem> items = new List<PdfPackageItem>();

            NameTree files = NameTree.Find(doc.GetSDFDoc(), "EmbeddedFiles");
            if (files.IsValid())
            {
                NameTreeIterator nameTreeIterator = files.GetIterator();
                while (nameTreeIterator.HasNext())
                {
                    string entryName = nameTreeIterator.Key().GetAsPDFText();
                    bool isPDF = false;
                    if (HasExtension(entryName))
                    {
                        string extension = GetExtension(entryName);
                        if (extension.Equals("pdf", StringComparison.OrdinalIgnoreCase))
                            isPDF = true;
                    }

                    items.Add(new PdfPackageItem(entryName, isPDF));
                    nameTreeIterator.Next();
                }
            }

            return items;
        }

        private async Task<StorageFile> ExtractFileFromPortFolio(PDFDoc doc, string fileName)
        {
            try
            {
                NameTree files = NameTree.Find(doc.GetSDFDoc(), "EmbeddedFiles");
                if (files.IsValid())
                {
                    // Traverse the list of embedded files.
                    NameTreeIterator nameTreeIterator = files.GetIterator();
                    while (nameTreeIterator.HasNext())
                    {
                        string entryName = nameTreeIterator.Key().GetAsPDFText();
                        if (entryName.Equals(fileName, StringComparison.OrdinalIgnoreCase))
                        {
                            FileSpec fileSpec = new FileSpec(nameTreeIterator.Value());
                            IFilter filter = fileSpec.GetFileData();
                            if (filter != null)
                            {
                                Regex regex = new Regex("<\\d*>");

                                string tagFreeFileName = regex.Replace(fileName, string.Empty);
                                string cleanFileName = UtilityFunctions.SanitizeFileName(tagFreeFileName);
                                _selectedFile = await UtilityFunctions.CreateTemporaryAccessFileAsync(cleanFileName);
                                string newFileName = _selectedFile.Path;
                                filter.WriteToFile(newFileName, false);

                                return _selectedFile;
                            }
                        }
                        nameTreeIterator.Next();
                    }
                }
            }
            catch (Exception)
            {
                _errorHappened = true;

                ResourceLoader loader = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources");
                _errorMessage = string.Format(loader.GetString("PackageDialog_FileExtractionError_Info"), ViewerSettings.DisplayName);
            }
            return null;
        }

        [SuppressMessage("Major Code Smell", "S1144:Unused private types or members should be removed", Justification = "<Pending>")]
        [SuppressMessage("CodeQuality", "IDE0051:Remove unused private members", Justification = "<Pending>")]
        private async Task<StorageFile> ExtractNonPdfDocFromPortFolio(PDFDoc doc, string fileName)
        {
            try
            {
                StorageFolder folder = await GetPdfPackageFolderAsync();
                NameTree files = NameTree.Find(doc.GetSDFDoc(), "EmbeddedFiles");
                if (files.IsValid())
                {
                    // Traverse the list of embedded files.
                    NameTreeIterator nameTreeIterator = files.GetIterator();
                    while (nameTreeIterator.HasNext())
                    {
                        string entryName = nameTreeIterator.Key().GetAsPDFText();
                        if (entryName.Equals(fileName, StringComparison.OrdinalIgnoreCase))
                        {
                            FileSpec fileSpec = new FileSpec(nameTreeIterator.Value());
                            IFilter filter = fileSpec.GetFileData();
                            if (filter != null)
                            {
                                string newFileName = "tempfile." + GetExtension(fileName);
                                string newFullName = Path.Combine(folder.Path, newFileName);
                                filter.WriteToFile(newFullName, false);
                                return await folder.GetFileAsync(newFileName);
                            }
                        }

                        nameTreeIterator.Next();
                    }
                }
            }
            catch (Exception)
            {
                _errorHappened = true;
                _errorMessage = "Sorry, " + ViewerSettings.DisplayName + " failed to extract the selected file.";
            }

            return null;
        }

        private async Task OpenPdfAsync(PdfPackageItem item)
        {
            StorageFile file = await ExtractFileFromPortFolio(_documentStack[_documentStack.Count - 1], item.FileName);
            PDFDoc newDoc = null;
            try
            {
                newDoc = new PDFDoc(file);
            }
            catch
            {
                //Swallow
            }

            Obj collectionObj = newDoc.GetRoot().FindObj("Collection");
            if (collectionObj != null)
            {
                _documentStack.Add(newDoc);
                PopulatePackageList();
            }
            else
            {
                _fileSelcted?.Invoke(newDoc, _selectedFile, _errorHappened, _errorMessage);
            }
        }

        private async Task OpenNonPdfAsync(PdfPackageItem item)
        {
            bool success = false;
            StorageFile file = await ExtractFileFromPortFolio(_documentStack[_documentStack.Count - 1], item.FileName);
            if (file != null)
            {
                const string xpsExtension = ".oxps";
                string fileExtension = Path.GetExtension(file.DisplayName);
                if (xpsExtension.Equals(fileExtension, StringComparison.OrdinalIgnoreCase))
                {
                    _fileSelcted?.Invoke(null, file, _errorHappened, _errorMessage);
                }
                else
                {
                    try
                    {
                        success = await Launcher.LaunchFileAsync(file);
                    }
                    catch
                    {
                        //Swallow
                    }
                }
            }

            if (!success)
            {
                ResourceLoader loader = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources");

                ContentDialog dialog = new ContentDialog
                {
                    Title = item.FileName,
                    Content = string.Format(loader.GetString("PackageDialog_FailedToLaunchNonPDF_Info")),
                    CloseButtonText = loader.GetString("pdftron_Controls_Content_Dialog_OK")
                };

                await DialogHelper.ShowContentDialogAsync(dialog);
            }
        }

        private void HandleBackButton() => _fileSelcted?.Invoke(null, null, false, null);

        /// <summary>
        /// Because the path might not be valid, we need to implement our own check for this
        /// </summary>
        /// <param name="fileName"></param>
        /// <returns></returns>
        private bool HasExtension(string fileName) => fileName.Contains(".");

        private string GetExtension(string fileName)
        {
            if (HasExtension(fileName))
            {
                string[] pathParts = fileName.Split('.');
                if (pathParts.Length > 0)
                    return pathParts[pathParts.Length - 1];
            }

            return null;
        }

        private async Task<StorageFolder> GetPdfPackageFolderAsync()
        {
            IStorageItem item = await ApplicationData.Current.TemporaryFolder.TryGetItemAsync("PDFPackageFolder");
            return item == null
                ? await ApplicationData.Current.TemporaryFolder.CreateFolderAsync("PDFPackageFolder", CreationCollisionOption.ReplaceExisting)
                : item as StorageFolder;
        }

        public override bool GoBack()
        {
            HandleBackButton();
            return true;
        }
    }
}
