﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Input;

using Windows.Foundation;

using pdftron.Common;
using pdftron.PDF.Tools.Commands;
using pdftron.PDF.Tools.Controls.ViewModels.Common;
using pdftron.PDF.Tools.Utilities;
using Windows.UI.Xaml;
using Windows.Data.Json;

namespace pdftron.PDF.Tools.Controls.ViewModels
{
    public class ReflowItem : ViewModelBase
    {
        private string _ReflowedContent = "";
        public string ReflowedContent
        {
            get { return _ReflowedContent; }
            internal set
            {
                Set(ref _ReflowedContent, value);
            }
        }

        private int _PageNumber = 0;
        public int PageNumber
        {
            get { return _PageNumber; }
            internal set { Set(ref _PageNumber, value); }
        }

        private Reflow _reflow = null;
        public Reflow Reflow { get { return _reflow; } set { _reflow = value; } }

        public bool HasReflowedContent { get { return !string.IsNullOrEmpty(_ReflowedContent); } }

        public ReflowItem(int pageNumber)
        {
            PageNumber = pageNumber;
        }

        public void Reset()
        {
            ReflowedContent = string.Empty;
            Reflow = null;
        }
    }

    public class ReflowViewModel : ViewModelBase, IDisposable
    {
        public const string KEYDOWN_STRING = "<script> document.onkeydown = function (e) {window.external.notify(\"key: \" + e.keyCode); } </script>";
        public const string MOUSE_UP_STRING = "<script> document.onmouseup = function (e) {window.external.notify(\"mouseup: \" + window.getSelection()); } </script>";

        // Ensures paragraphs don't span to the right edge of the display, to allow for tapping to bring up command bar.
        public const string PARAGRAPH_STYLE_STRING = "<style>p{display:inline;} p:not(:last-of-type):after{content:\".\";display:block;height:0;clear:both;visibility:hidden;margin-bottom:1em;} p:not(:first-of-type):before{content:\".\";display:block;height:0;clear:both;visibility:hidden;margin-top:1em;}</style>";
        public const string TAP_STRING = "<script> document.onclick = function (e) { if (e.pointerType === 'touch' && (e.target === document.body || e.target.tagName == 'DIV')) window.external.notify(\"tap\"); }</script>";
        private const string REFLOW_EDITING = "<script type=\"text/javascript\" src=\"ms-appx-web:///pdftron.PDF.Tools/Resources/Reflow.js\"></script>";
        private const int PAGE_FETCHING_RANGE = 3;
        private const string LINK_PREFIX = "LaunchLink:";

        private PostProcessBundle _PostProcessBundle;
        private PDFViewCtrl _PostProcessPDFView;
        private Dictionary<int, ReflowItem> _ReflowedPages;
        private PDFDoc _Doc = null;
        private int _currentPageNumber = 0;
        private int _totalPageNumbers = 0;
        private bool _showChangePageButtons = true;
        private bool _isReflowing = false;
        private ReflowItem _CurrentPage = new ReflowItem(0);
        private SemaphoreSlim _ReflowSemaphore = null;

        #region Events

        private delegate void PageChangeRequestedDelegate(int pageNumber);
        private event PageChangeRequestedDelegate PageChangeRequested;

        public delegate void PageChangedDelegate(int pageNumber, int totalPages);
        public event PageChangedDelegate PageChanged;

        public delegate bool PageChangeIntendedDelegate(int pageNumber);
        /// <summary>
        /// Gets raised when the ReflowViewModel is about to change pages through a link being clicked
        /// Handle this and return true to make the ReflowViewModel abort.
        /// </summary>
        public event PageChangeIntendedDelegate PageChangeIntended;

        public delegate void ReflowContentUpdatedDelegate(ReflowItem item);
        public event ReflowContentUpdatedDelegate ReflowContentUpdated;

        #endregion

        public ReflowViewModel()
        {
            Init();
        }

        public ReflowViewModel(PDFDoc doc)
        {
            Init();
            Doc = doc;
        }

        private void Init()
        {
            _ReflowedPages = new Dictionary<int, ReflowItem>();
            _ReflowSemaphore = new SemaphoreSlim(1);
            _PostProcessBundle = new PostProcessBundle(PDFRasterizerColorPostProcessMode.e_postprocess_none);

            this.PageChangeRequested += ReflowViewModel_PageChangeRequested;

            InitCommands();
        }

        #region Properties

        /// <summary>
        /// Set weather or not to include images in the reflowed pages
        /// </summary>
        public bool IncludeImages { get; set; } = false;

        /// <summary>
        /// Set wheather or not annotate in reflow is enabled
        /// </summary>
        public bool IsEditingEnabled { get; set; } = false;

        /// <summary>
        /// Show/hide the side navigation buttons
        /// </summary>
        public bool ShowChangePageButtons 
        {
            get { return _showChangePageButtons; }
            set { Set(ref _showChangePageButtons, value); }
        }

        /// <summary>
        /// Excludes images that are detected as background images
        /// </summary>
        public bool SetHideBackgroundImages { get; set; } = false;

        /// <summary>
        /// Excludes images that appear under text
        /// </summary>
        public bool SetHideImagesUnderText { get; set; } = false;

        /// <summary>
        /// Excludes text that is on top of an image
        /// </summary>
        public bool SetDoNotReflowTextOverImages { get; set; } = false;

        public Dictionary<int, ReflowItem> ReflowedPages
        {
            get { return _ReflowedPages; }
        }

        public PDFDoc Doc
        {
            get { return _Doc; }
            set
            {
                Set(ref _Doc, value);

                if (_Doc != null)
                    _totalPageNumbers = Doc.GetPageCount();
            }
        }

        public int TotalPageNumbers
        {
            get
            {
                if (Doc != null)
                {
                    return _totalPageNumbers;
                }
                return 0;
            }
        }

        public int CurrentPageNumber
        {
            get { return _currentPageNumber;  }
            private set
            {
                if (Doc != null && Set(ref _currentPageNumber, value))
                {
                    PageChanged?.Invoke(CurrentPageNumber, TotalPageNumbers);

                    RaisePropertyChanged(nameof(IsPrevButtonEnabled));
                    RaisePropertyChanged(nameof(IsNextButtonEnabled));
                }
            }
        }

        public ReflowItem CurrentPage
        {
            get { return _CurrentPage; }
            private set
            {
                Set(ref _CurrentPage, value);
            }
        }

        private bool _hasReflowedContent = false;
        public bool HasReflowedContent
        {
            get
            {
                return _hasReflowedContent;
            }
            set { Set(ref _hasReflowedContent, value); }
        }

        public string ReflowedContent
        {
            get
            {
                if (CurrentPage == null)
                {
                    return "";
                }
                return CurrentPage.ReflowedContent;
            }
        }

        public string GetReflowProgressText
        {
            get
            {
                int pageNumber = 1;
                if (CurrentPage != null)
                {
                    pageNumber = CurrentPage.PageNumber;
                }

                return string.Format(ResourceHandler.GetString("Reflow_WorkingOnPage"), pageNumber);
            }
        }

        /// <summary>
        /// Weather or not previous button is enabled
        /// </summary>
        public bool IsPrevButtonEnabled { get { return Doc != null && CurrentPageNumber > 1; } }

        /// <summary>
        /// Weather or not next button is enabled
        /// </summary>
        public bool IsNextButtonEnabled { get { return Doc != null && CurrentPageNumber < TotalPageNumbers; } }

        public PDFRasterizerColorPostProcessMode PostProcessingMode { get { return _PostProcessBundle.PostProcessMode; } }

        public Windows.UI.Color PostProcessingWhiteColor { get { return _PostProcessBundle.PostProcessWhiteColor; } }

        public Windows.UI.Color PostProcessingBlackColor { get { return _PostProcessBundle.PostProcessBlackColor; } }

        public PDFViewCtrl PostProcessPDFView
        {
            get
            {
                if (_PostProcessPDFView == null)
                {
                    _PostProcessPDFView = new PDFViewCtrl();
                }
                return _PostProcessPDFView;
            }
        }

        private void UpdateReflowColors()
        {
            RaisePropertyChanged(nameof(PostProcessingMode));
            RaisePropertyChanged(nameof(PostProcessingWhiteColor));
            RaisePropertyChanged(nameof(PostProcessingBlackColor));
        }

        #endregion Properties

        #region Public Functions

        /// <summary>
        /// Sets the current page to pageNumber
        /// </summary>
        /// <param name="pageNumber">the page number to navigate to</param>
        /// <returns>True if the pageNumber is a valid page within range</returns>
        public bool SetPageNumber(int pageNumber)
        {
            if (Doc != null && pageNumber > 0 && pageNumber <= TotalPageNumbers)
            {
                PageChangeRequested?.Invoke(pageNumber);

                return true;
            }
            return false;
        }

        public void Dispose()
        {
            if (ReflowedPages != null)
            {
                ReflowedPages.Clear();
            }

            this.PageChangeRequested -= ReflowViewModel_PageChangeRequested;

            Doc = null;
        }

        public async void SetColorPostProcessingMode(PDFRasterizerColorPostProcessMode mode, Windows.UI.Color whiteColor, Windows.UI.Color blackColor)
        {
            try
            {
                PostProcessBundle newPostProcess = new PostProcessBundle(mode, whiteColor, blackColor);
                if (!newPostProcess.Equals(_PostProcessBundle))
                {
                    _PostProcessBundle = newPostProcess;
                    if (mode == PDFRasterizerColorPostProcessMode.e_postprocess_none ||
                        mode == PDFRasterizerColorPostProcessMode.e_postprocess_invert ||
                        mode == PDFRasterizerColorPostProcessMode.e_postprocess_night_mode)
                    {
                        PostProcessPDFView.SetColorPostProcessMode(mode);
                    }
                    else
                    {
                        PostProcessPDFView.SetColorPostProcessColors(whiteColor, blackColor);
                    }

                    foreach (ReflowItem item in ReflowedPages.Values)
                    {
                        item.Reset();

                    }

                    await RequestPageAsync(CurrentPageNumber);

                    UpdateReflowColors();
                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
        }

        /// <summary>
        /// Will try to navigate to an internal link. Returns true if successful
        /// </summary>
        /// <param name="link"></param>
        /// <returns>true if link was a link and it could sucessfully navigate to it</returns>
        public bool NavigateInternalLink(string link)
        {
            int linkLength = LINK_PREFIX.Length;
            if (link.Length < linkLength || Doc == null)
            {
                return false;
            }
            if (link.Substring(0, linkLength).Equals(LINK_PREFIX))
            {
                string htmlLink = link.Substring(linkLength);
                if (htmlLink.StartsWith("file:///") && htmlLink.EndsWith(".html"))
                {
                    try
                    {
                        int slashPos = htmlLink.LastIndexOf('/');
                        string objNumString = htmlLink.Substring(slashPos + 1, htmlLink.Length - 5 - (slashPos + 1));
                        int objNum = int.Parse(objNumString);
                        int pageNum = 0;
                        Doc.LockRead();
                        try
                        {
                            for (PageIterator itr = Doc.GetPageIterator(); itr.HasNext(); itr.Next())
                            {
                                if (itr.Current().GetSDFObj().GetObjNum() == objNum)
                                {
                                    pageNum = itr.GetPageNumber();
                                    break;
                                }
                            }
                        }
                        catch (Exception e)
                        {
                            PDFNetException pdfe = new PDFNetException(e.HResult);
                            if (!pdfe.IsPDFNetException)
                            {
                                throw;
                            }
                            else
                            {
                                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(e));
                            }
                        }
                        finally
                        {
                            Doc.UnlockRead();
                        }

                        if (pageNum > 0)
                        {
                            if (PageChangeIntended == null || !PageChangeIntended(pageNum))
                            {
                                CurrentPageNumber = pageNum;
                            }
                        }
                    }
                    catch (Exception e)
                    {
                        PDFNetException pdfe = new PDFNetException(e.HResult);
                        if (!pdfe.IsPDFNetException)
                        {
                            throw;
                        }
                        else
                        {
                            Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(e));
                        }
                    }
                }
                else
                {
                    ActionUtils.LaunchBrowser(htmlLink);
                }
            }

            return false;
        }

        // TODO: test/check with ViewerControl
        public void UpdatePageNumberProperties()
        {
            RaisePropertyChanged(nameof(TotalPageNumbers));
            RaisePropertyChanged(nameof(IsPrevButtonEnabled));
            RaisePropertyChanged(nameof(IsNextButtonEnabled));
        }

        #endregion Public Functions

        #region Commands

        private void InitCommands()
        {
            CMDNextPage = new RelayCommand(GoToNextPage);
            CMDPreviousPage = new RelayCommand(GoToPreviousPage);
        }

        public ICommand CMDNextPage { get; set; }

        public ICommand CMDPreviousPage { get; set; }

        private async void GoToNextPage(object param)
        {
            if (CurrentPage == null || _isReflowing)
                return;

            if (CurrentPage.PageNumber > TotalPageNumbers)
                return;

            await RequestPageAsync(CurrentPage.PageNumber + 1, true);
        }

        private async void GoToPreviousPage(object param)
        {
            if (CurrentPage == null || _isReflowing)
                return;

            if (CurrentPage.PageNumber <= 1)
                return;

            await RequestPageAsync(CurrentPage.PageNumber - 1, true);
        }

        #endregion Commands

        #region Impl

        private async void ReflowViewModel_PageChangeRequested(int pageNumber)
        {
            await RequestPageAsync(pageNumber);
        }

        private class PostProcessBundle
        {
            public PDFRasterizerColorPostProcessMode PostProcessMode { get; private set; }
            public Windows.UI.Color PostProcessWhiteColor { get; private set; }
            public Windows.UI.Color PostProcessBlackColor { get; private set; }

            public PostProcessBundle(PDFRasterizerColorPostProcessMode mode)
            {
                PostProcessMode = mode;
                PostProcessWhiteColor = Windows.UI.Colors.White;
                PostProcessBlackColor = Windows.UI.Colors.Black;
        }

            public PostProcessBundle(PDFRasterizerColorPostProcessMode mode, Windows.UI.Color whiteColor, Windows.UI.Color blackColor)
            {
                PostProcessMode = mode;
                PostProcessWhiteColor = whiteColor;
                PostProcessBlackColor = blackColor;
            }

            public override bool Equals(object obj)
            {
                PostProcessBundle other = obj as PostProcessBundle;
                if (other != null)
                {
                    if (PostProcessMode != other.PostProcessMode)
                    {
                        return false;
                    }

                    if (PostProcessMode == PDFRasterizerColorPostProcessMode.e_postprocess_none ||
                        PostProcessMode == PDFRasterizerColorPostProcessMode.e_postprocess_invert ||
                        PostProcessMode == PDFRasterizerColorPostProcessMode.e_postprocess_night_mode)
                    {
                        return true;
                    }

                    return PostProcessWhiteColor == other.PostProcessWhiteColor &&
                        PostProcessBlackColor == other.PostProcessBlackColor;
                }

                return base.Equals(obj);
            }

            public override int GetHashCode()
            {
                if (PostProcessMode == PDFRasterizerColorPostProcessMode.e_postprocess_none ||
                       PostProcessMode == PDFRasterizerColorPostProcessMode.e_postprocess_invert ||
                       PostProcessMode == PDFRasterizerColorPostProcessMode.e_postprocess_night_mode)
                {
                    return (int)PostProcessMode;
                }

                return (int)PostProcessMode * PostProcessBlackColor.GetHashCode() * PostProcessWhiteColor.GetHashCode();
            }
        }

        /// <summary>
        /// Injects JavaScript events to the HTML page
        /// </summary>
        /// <param name="htmlPage">The HTML page</param>
        /// <returns>The HTML page with JavaScript events</returns>
        private string InjectJavaScriptEvents(string htmlPage)
        {
            var str = htmlPage;

            // WebView steals all input gesture commands. Need to manually send it back out.
            string js = "<script> document.onmousedown = function (e) {if (e.which == 3) window.external.notify(\"rightClick\"); } </script>";
            int index = str.IndexOf("</head>");
            str = str.Insert(index, js);
                        
            str = str.Insert(index, TAP_STRING);
            str = str.Insert(index, PARAGRAPH_STYLE_STRING);
            str = str.Insert(index, KEYDOWN_STRING);
            str = str.Insert(index, MOUSE_UP_STRING);
            str = str.Insert(index, REFLOW_EDITING);

            index = str.IndexOf("</html>");
            js = string.Format("{0}{1}{2}",
                "<script>for (var i = 0; i < document.links.length; i++) { document.links[i].onclick = function() { window.external.notify('",
                LINK_PREFIX, "' + this.href); return false; } }</script>");

            str = str.Insert(index, js);

            return str;
        }

        private string HandlePostProcessColor(string reflowContent)
        {
            string postProcessColorResult = string.Empty;

            bool postProcessed = false;
            if (PostProcessingMode != PDFRasterizerColorPostProcessMode.e_postprocess_none)
            {
                try
                {
                    postProcessColorResult = PostProcessReflowContent(reflowContent);
                    postProcessed = true;
                }
                catch (Exception e)
                {
                    PDFNetException pdfe = new PDFNetException(e.HResult);
                    if (!pdfe.IsPDFNetException)
                    {
                        throw;
                    }
                    else
                    {
                        Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(e));
                    }
                }
            }
            if (!postProcessed)
            {
                postProcessColorResult = reflowContent;
            }

            return postProcessColorResult;
        }

        private string PostProcessReflowContent(string input)
        {
            StringBuilder sb = new StringBuilder();

            Regex tagStr = new Regex("<.*?>");
            Regex colorStr = new Regex("color:#[0-9|a-f|A-F]{6}");
            Regex backgroundColorStr = new Regex("background-color:#[0-9|a-f|A-F]{6}");
            Regex paragraphStr = new Regex("<p.*?>");
            Regex spanStr = new Regex("<span.*?>");

            MatchCollection matches = tagStr.Matches(input);
            int currStringIndex = 0;
            foreach (Match match in matches)
            {
                sb.Append(input.Substring(currStringIndex, match.Index - currStringIndex));
                currStringIndex = match.Index + match.Length;
                string matchedString = match.Value;
                int matchedStringIndex = 0;
                MatchCollection colorMatches = colorStr.Matches(matchedString);
                MatchCollection paraMatches = paragraphStr.Matches(matchedString);
                MatchCollection spanMatches = spanStr.Matches(matchedString);
                MatchCollection backgroundMatches = backgroundColorStr.Matches(matchedString);

                foreach (Match colorMatch in colorMatches)
                {
                    bool isInsideParahraph = IsInsideMatch(colorMatch.Index, paraMatches);
                    bool isInsideSpan = IsInsideMatch(colorMatch.Index, spanMatches);
                    bool isInsideBackgroundColor = IsInsideMatch(colorMatch.Index, backgroundMatches);

                    sb.Append(matchedString.Substring(matchedStringIndex, colorMatch.Index - matchedStringIndex));
                    matchedStringIndex = colorMatch.Index + colorMatch.Length;
                    string colorString = colorMatch.Value;
                    if ((isInsideParahraph || isInsideSpan) && !isInsideBackgroundColor)
                    {
                        sb.Append("color:#" + PostProcessColor(colorString.Substring(7, 6)));
                    }
                    else
                    {
                        sb.Append(colorString);
                    }
                }
                sb.Append(matchedString.Substring(matchedStringIndex));
            }
            sb.Append(input.Substring(currStringIndex));

            return sb.ToString();
        }

        /// <summary>
        /// Convers colorString into a color that can be PostProcessed by the PDFViewCtrl
        /// </summary>
        /// <param name="colorString">string of the form RRGGBB</param>
        /// <returns></returns>
        private string PostProcessColor(string colorString)
        {
            if (colorString.Length == 6)
            {
                double r = System.Convert.ToInt32(colorString.Substring(0, 2), 16);
                double g = System.Convert.ToInt32(colorString.Substring(2, 2), 16);
                double b = System.Convert.ToInt32(colorString.Substring(4, 2), 16);
                pdftron.PDF.ColorPt cp = new ColorPt(r / 255.0, g / 255.0, b / 255.0);
                cp = PostProcessPDFView.GetPostProcessedColor(cp);
                int ppr = (int)(cp.Get(0) * 255);
                int ppg = (int)(cp.Get(1) * 255);
                int ppb = (int)(cp.Get(2) * 255);
                string retString = "";
                if (ppr < 16)
                {
                    retString += "0";
                }
                retString += ppr.ToString("X");
                if (ppg < 16)
                {
                    retString += "0";
                }
                retString += ppg.ToString("X");
                if (ppb < 16)
                {
                    retString += "0";
                }
                retString += ppb.ToString("X");
                return retString;
            }
            return colorString;
        }

        private bool IsInsideMatch(int index, MatchCollection matches)
        {
            foreach (Match match in matches)
            {
                if (index > match.Index && index < match.Index + match.Length)
                    return true;
            }

            return false;
        }

        #endregion Impl

        #region New Implementation        

        /// <summary>
        /// Request a document page to be refloed
        /// </summary>
        /// <param name="pageNumber">The page number (1-base) to be reflowed</param>
        /// <param name="prevNext">True to request previous and next pages as well</param>
        public async Task RequestPageAsync(int pageNumber, bool prevNext = false)
        {
            if (Doc == null)
                return;

            if (pageNumber <= 0 || pageNumber > TotalPageNumbers)
                return;

            await _ReflowSemaphore.WaitAsync();
            try
            {
                _isReflowing = true;
                HasReflowedContent = false;

                if (ReflowedPages.ContainsKey(pageNumber))
                {
                    ReflowedPages[pageNumber].Reset();
                }

                var result = await RequestPageImplAsync(pageNumber);
                if (result == null)
                    return;

                var reflowedPage = new ReflowItem(pageNumber);
                reflowedPage.Reflow = result.Item1;
                reflowedPage.ReflowedContent = result.Item2;

                ReflowedPages.TryAdd(pageNumber, reflowedPage);
                CurrentPage = reflowedPage;
                CurrentPageNumber = CurrentPage.PageNumber;

                ReflowContentUpdated?.Invoke(CurrentPage);

                if (prevNext)
                    await RequestNextPrevPageAsync();
            }
            catch (Exception ex)
            {
                AppCenterHandlerBase.CURRENT.TrackError(ex);
            }
            finally 
            {
                _ReflowSemaphore.Release();
                _isReflowing = false;
            }
        }

        public string AddAnnot(string data)
        {
            if (CurrentPage == null || CurrentPage.HasReflowedContent == false)
                return string.Empty;

            return CurrentPage.Reflow.SetAnnot(data);
        }

        /// <summary>
        /// Perform the editing/update of a specific annotation considering it's ID is valid.
        /// </summary>
        /// <remarks>
        /// Example of JSON data: {"id":"1","type":11,"color":"FF0000","opacity":1,"ranges":[28,48],"startOffset":20,"endOffset":41}
        /// </remarks>
        /// <param name="data">The annotation JSON data to update</param>
        /// <returns>Result in JSON format</returns>
        public string EditAnnot(string data)
        {
            return CurrentPage.Reflow.SetAnnot(data);
        }

        /// <summary>
        /// Delete a specific annotation from the current page
        /// </summary>
        /// <param name="id">Annotation id</param>
        /// <returns>Result in JSON format</returns>
        public string DeleteAnnot(string id)
        {
            return CurrentPage.Reflow.SetAnnot("{\"id\": \"" + id + "\"}");
        }

        /// <summary>
        /// Get all annotations for the current selected page
        /// </summary>
        /// <returns>Return a JSON string with annotations</returns>
        public string GetAnnotations()
        {
            if (CurrentPage == null || CurrentPage.HasReflowedContent == false)
                return string.Empty;

            var annots = CurrentPage.Reflow.GetAnnot("");
            return annots;
        }

        /// <summary>
        /// Load 
        /// </summary>
        /// <returns></returns>
        private async Task RequestNextPrevPageAsync()
        {
            var nextPage = CurrentPageNumber + 1;

            if (nextPage >= TotalPageNumbers)
                return;

            if (ReflowedPages.ContainsKey(nextPage) == false)
            {
                var reflowedPage = new ReflowItem(nextPage);
                var result = await RequestPageImplAsync(nextPage);
                if (result != null)
                    reflowedPage.ReflowedContent = result.Item2;

                ReflowedPages.Add(nextPage, reflowedPage);
            }

            var prevPage = CurrentPageNumber - 1;

            if (prevPage <= 0)
                return;

            if (ReflowedPages.ContainsKey(prevPage) == false)
            {
                var reflowedPage = new ReflowItem(prevPage);
                var result = await RequestPageImplAsync(prevPage);
                if (result != null)
                    reflowedPage.ReflowedContent = result.Item2;

                ReflowedPages.Add(prevPage, reflowedPage);
            }
        }

        /// <summary>
        /// Process the request to reflow the page
        /// </summary>
        /// <param name="pageNumber">The page number to be reflowed</param>
        /// <returns>True if success, false otherwise</returns>
        private IAsyncOperation<Tuple<Reflow, string>> RequestPageImplAsync(int pageNumber)
        {
           var task = new Task<Tuple<Reflow, string>>
                    (() =>
                    {
                        if (Doc == null)
                            return null;

                        bool shouldUnlock = false;

                        try
                        {
                            Doc.LockRead();
                            shouldUnlock = true;

                            var page = Doc.GetPage(pageNumber);
                            var reflow = Convert.CreateReflow(page, "");

                            reflow.SetHTMLOutputTextMarkup(!IsEditingEnabled);
                            reflow.SetIncludeImages(IncludeImages);
                            reflow.SetHideBackgroundImages(SetHideBackgroundImages);
                            reflow.SetHideImagesUnderText(SetHideImagesUnderText);
                            reflow.SetDoNotReflowTextOverImages(SetDoNotReflowTextOverImages);                            
                            reflow.SetMessageWhenNoReflowContent("Sorry, nothing to reflow");
                            reflow.SetMessageWhenReflowFailed("Reflow failed");

                            string reflowHtml = reflow.GetHtml();

                            if (string.IsNullOrEmpty(reflowHtml))
                                return null;

                            var reflowHtmlResult = InjectJavaScriptEvents(reflowHtml);

                            reflowHtmlResult = HandlePostProcessColor(reflowHtmlResult);

                            return new Tuple<Reflow, string>(reflow, reflowHtmlResult);
                        }
                        catch (Exception ex)
                        {
                            AppCenterHandlerBase.CURRENT.TrackError(ex);
                            return null;
                        }
                        finally
                        {
                            if (shouldUnlock)
                                Doc.UnlockRead();
                        }
                    });

            task.Start();
            return task.AsAsyncOperation();
        }

        #endregion
    }
}
