﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Input;
using Windows.ApplicationModel.Resources;
using Windows.Storage;
using Windows.Storage.Pickers;
using Windows.UI.Popups;
using Windows.UI.Xaml.Controls;

using pdftron.PDF.Tools.Commands;
using pdftron.PDF.Tools.Controls.ViewModels.Common;
using pdftron.PDF.Tools.Utilities;
using Windows.Storage.Provider;

namespace pdftron.PDF.Tools.Controls.ViewModels.View
{
    public class ViewPdfViewModel : ViewModelBase
    {
        double DEFAULT_ZOOM_START_POSITION = 0.80;  // Initial document zoom
        double DEFAULT_ZOOM_FACTOR_STEP = 1.25;     // Amount of zoom when Zooming in or out by button click
        int MAX_PDF_NAME_LENGTH = 70;               // Max amout of chars

        StorageFile mStorageFilePdf;
        StorageFile mStorageFilePdfTemp;

        ToolManager mToolManagerDoc;

        ResourceLoader mResourceLoader;

        public ViewPdfViewModel()
        {
            Init();
        }

        private void Init()
        {
            // Init and setup Annot toolbar
            _AnnotToolbar = new AnnotationCommandBar();
            _AnnotToolbar.IsEnabled = false;
            _AnnotToolbar.IsCloseToolbarButtonVisible = false;
            _AnnotToolbar.ButtonsStayDown = true;
            _AnnotToolbar.IsRedactionToolButtonVisible = false;

            _PageNumberIndicator.ShowingBehaviour = ClickablePageNumberIndicator.ShowingBehaviours.AlwaysShow;

            // Setup commands
            CMDOpenPdf = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.Create(OpenFileAsync);
                command.Execute(parameter);
            });

            CMDCloseDoc = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.Create(CloseDocAsync);
                command.Execute(parameter);
            });

            CMDSaveAsDoc = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.Create(SaveAsDocAsync);
                command.Execute(parameter);
            });

            CMDSaveDoc = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.Create(SaveDocAsync);
                command.Execute(parameter);
            });

            CMDZoomIn = new RelayCommand(ZoomIn);
            CMDZoomOut = new RelayCommand(ZoomOut);

            // Init string resources
            mResourceLoader = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources");
        }

        #region Public Properties

        private string _pdfName = string.Empty;
        public string PdfName
        {
            get
            {
                if (!string.IsNullOrEmpty(_pdfName))
                    if (_pdfName.Length > MAX_PDF_NAME_LENGTH)
                        return _pdfName.Substring(0, MAX_PDF_NAME_LENGTH);

                return _pdfName;
            }
            set
            {
                _pdfName = value;
                RaisePropertyChanged();
            }
        }

        private bool _IsPdfOpenPageVisible = true;
        public bool IsPdfOpenPageVisible
        {
            get { return _IsPdfOpenPageVisible; }
            set
            {
                _IsPdfOpenPageVisible = value;
                RaisePropertyChanged();
            }
        }

        private ClickablePageNumberIndicator _PageNumberIndicator = new ClickablePageNumberIndicator();
        public ClickablePageNumberIndicator PageNumbIndicator
        {
            get
            {
                return _PageNumberIndicator;
            }
            set
            {
                _PageNumberIndicator = value;
                RaisePropertyChanged();
            }
        }

        private PDFViewCtrl _pdfViewCtrl;
        public PDFViewCtrl PdfViewCtrl
        {
            get { return _pdfViewCtrl; }
            set
            {
                _pdfViewCtrl = value;
                RaisePropertyChanged();
            }
        }

        private AnnotationCommandBar _AnnotToolbar;
        public AnnotationCommandBar AnnotToolbar
        {
            get { return _AnnotToolbar; }
            set
            {
                _AnnotToolbar = value;
                RaisePropertyChanged();
            }
        }

        #endregion

        #region Public Methods
        public async Task OpenFileAsync(StorageFile file)
        {
            await OpenFileAAsyncImpl(file);
        }

        public void CleanUp()
        {
            if (PdfViewCtrl != null)
            {
                mToolManagerDoc?.Dispose();

                PdfViewCtrl = null;

                IsPdfOpenPageVisible = true;
            }

            if (AnnotToolbar != null)
            {
                AnnotToolbar.CloseControl();
                AnnotToolbar.IsEnabled = false;
            }
        }
        #endregion

            #region Commands
        public ICommand CMDOpenPdf { get; set; }

        public ICommand CMDCloseDoc { get; set; }

        public ICommand CMDSaveDoc { get; set; }

        public ICommand CMDSaveAsDoc { get; set; }

        public ICommand CMDZoomIn { get; set; }

        public ICommand CMDZoomOut { get; set; }
        #endregion

        #region Operations

        /// <summary>
        /// Open document A to compare with document B.
        /// The original file is never used to open in the viewer, instead a copy is created in the temporary folder
        /// so all editing can be done.
        /// </summary>
        private async Task OpenFileAsync()
        {
            FileOpenPicker filePicker = new FileOpenPicker();
            filePicker.ViewMode = PickerViewMode.List;
            filePicker.FileTypeFilter.Add(".pdf");

            var file = await filePicker.PickSingleFileAsync();
            if (file == null)
                return;

            await OpenFileAAsyncImpl(file);
        }

        private async Task OpenFileAAsyncImpl(StorageFile file)
        {
            // Ensure to cleanup current loaded document
            if (PdfViewCtrl != null)
            {
                await CleanupPdfAsync();
            }

            mStorageFilePdf = file;
            PdfName = mStorageFilePdf.DisplayName + mStorageFilePdf.FileType;

            mStorageFilePdfTemp = await UtilityFunctions.CreateTemporaryAccessFileAsync(mStorageFilePdf.DisplayName);
            if (mStorageFilePdfTemp == null)
            {
                ResourceLoader loader = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources");

                ContentDialog dialog = new ContentDialog
                {
                    Title = loader.GetString("DocumentsPage_FileOpeningError_Title"),
                    Content = loader.GetString("DocumentsPage_FileOpeningError_FailedToCreateTempFile"),
                    CloseButtonText = loader.GetString("pdftron_Controls_Content_Dialog_OK")
                };

                await DialogHelper.ShowContentDialogAsync(dialog);
                return;
            }

            await mStorageFilePdf.CopyAndReplaceAsync(mStorageFilePdfTemp);
            await UtilityFunctions.RemoveReadOnlyAndTemporaryFlagsAsync(mStorageFilePdf);

            var doc = await OpenDocPDFViewerAsync(mStorageFilePdfTemp, FileAccessMode.ReadWrite);
            if (doc == null)
                return;

            if (PdfViewCtrl == null)
            {
                PdfViewCtrl = new PDFViewCtrl();
                PdfViewCtrl.OnSetDoc += PDFView_OnSetDoc;
            }

            PdfViewCtrl.SetDoc(doc);
            IsPdfOpenPageVisible = false;
        }

        private async Task<PDFDoc> OpenDocPDFViewerAsync(IStorageFile file, FileAccessMode mode)
        {
            if (file == null)
                return null;

            try
            {
                var stream = await file.OpenAsync(mode);
                var doc = new PDFDoc(stream);

                var securityH = doc.GetSecurityHandler();

                // In case init fails we ask for password
                if (doc.InitSecurityHandler() == false)
                {
                    ContentDialog contentDialog = new ContentDialog();
                    var pwdBox = new PasswordBox();
                    contentDialog.Title = mResourceLoader.GetString("PDFCompareControl_EnterPassword");
                    contentDialog.PrimaryButtonText = mResourceLoader.GetString("pdftron_Controls_ThumbnailsView_Delete_Message_Confirm");
                    contentDialog.CloseButtonText = mResourceLoader.GetString("pdftron_Controls_ThumbnailsView_Delete_Message_Cancel");
                    contentDialog.Content = pwdBox;

                    var result = await DialogHelper.ShowContentDialogAsync(contentDialog);

                    if (result != ContentDialogResult.Primary)
                        return null;

                    if (string.IsNullOrEmpty(pwdBox.Password))
                        return null;

                    if (doc.InitStdSecurityHandler(pwdBox.Password) == false)
                    {
                        ContentDialog cnttDialog = new ContentDialog
                        {
                            Title = mResourceLoader.GetString("PasswordDialog_ManyFails_Title"),
                            PrimaryButtonText = mResourceLoader.GetString("pdftron_Controls_Content_Dialog_OK")
                        };

                        await DialogHelper.ShowContentDialogAsync(cnttDialog);
                        return null;
                    }
                }

                return doc;
            }
            catch (Exception e)
            {
                // NOTE: If file already opened it will cause an exception
                var msg = new MessageDialog(e.Message);
                await msg.ShowAsync();
                return null;
            }
        }

        /// <summary>
        /// Close document A and properly delete the temporary document created
        /// </summary>
        /// <returns></returns>
        private async Task CloseDocAsync()
        {
            await CleanupPdfAsync();

            IsPdfOpenPageVisible = true;
        }

        /// <summary>
        /// Helper method to perform a cleanup on the loaded PDF A
        /// </summary>
        private async Task CleanupPdfAsync()
        {
            if (PdfViewCtrl.HasDocument)
            {
                PdfViewCtrl.GetDoc().Destroy();
            }

            if (mStorageFilePdf != null)
                mStorageFilePdf = null;

            if (mStorageFilePdfTemp != null)
            {

                await mStorageFilePdfTemp.DeleteAsync(StorageDeleteOption.Default);
                mStorageFilePdfTemp = null;
            }

            mToolManagerDoc.Dispose();

            PdfViewCtrl.OnSetDoc -= PDFView_OnSetDoc;
            PdfViewCtrl = null;

            CloseToolbar();
        }

        private void CloseToolbar()
        {
            if (PdfViewCtrl == null)
            {
                AnnotToolbar.CloseControl();
                AnnotToolbar.IsEnabled = false;
            }
        }

        private void ZoomIn(object param)
        {
            PerformZoom(PdfViewCtrl, true);
        }

        private void ZoomOut(object param)
        {
             PerformZoom(PdfViewCtrl, false);            
        }

        private void PerformZoom(PDFViewCtrl pDFViewCtrl, bool isZoomIn)
        {
            if (!pDFViewCtrl.HasDocument)
                return;

            var zoomFactor = DEFAULT_ZOOM_FACTOR_STEP; // Amount of zoom on each call

            double currentZoom = pDFViewCtrl.GetZoom();
            var width = (int)(pDFViewCtrl.ActualWidth / 2); // Ensure to keep the zooming centered
            var height = (int)(pDFViewCtrl.ActualHeight / 2); // Ensure to keep the zooming centered

            if (isZoomIn)
                pDFViewCtrl.SetZoom(width, height, currentZoom * zoomFactor, true);
            else
                pDFViewCtrl.SetZoom(width, height, currentZoom / zoomFactor, true);
        }

        private async Task SaveAsDocAsync()
        {
            if (PdfViewCtrl == null)
                return;

            if (PdfViewCtrl.HasDocument == false)
                return;

            await SaveAsAsync(PdfViewCtrl.GetDoc());
        }

        /// <summary>
        /// Save changes made in the Temporary file to the original one
        /// </summary>
        private async Task SaveDocAsync()
        {
            if (PdfViewCtrl == null)
                return;

            if (PdfViewCtrl.HasDocument)
            {
                var doc = PdfViewCtrl.GetDoc();

                bool locked = doc.TryLock();
                if (doc.IsModified())
                {
                    try
                    {
                        doc.Unlock();
                        locked = false;

                        await doc.SaveAsync(SDF.SDFDocSaveOptions.e_incremental);
                    }
                    catch
                    {
                        // Save failed
                        return;
                    }
                    finally
                    {
                        if (locked)
                            PdfViewCtrl.DocUnlock();
                    }
                }
                else
                {
                    return;
                }

            }

            if (mStorageFilePdf == null)
                return;

            if (mStorageFilePdfTemp == null)
                return;

            try
            {
                await mStorageFilePdfTemp.CopyAndReplaceAsync(mStorageFilePdf);
            } 
            catch (Exception)
            {
                // Swallow exception
            }
        }

        private async Task SaveAsAsync(PDFDoc doc)
        {
            if (doc == null)
                return;

            FileSavePicker savePicker = new FileSavePicker();
            savePicker.SuggestedStartLocation = PickerLocationId.DocumentsLibrary;

            savePicker.FileTypeChoices.Add("PDF Document", new List<string>() { ".pdf" });
            savePicker.SuggestedFileName = "New Document"; // TODO: string from resource

            StorageFile file = await savePicker.PickSaveFileAsync();
            if (file != null)
            {
                CachedFileManager.DeferUpdates(file);

                await doc.SaveToNewLocationAsync(file, SDF.SDFDocSaveOptions.e_incremental);

                FileUpdateStatus status = await CachedFileManager.CompleteUpdatesAsync(file);
                if (status == FileUpdateStatus.Complete)
                {
                    // TOOD: file saved status
                }
                else
                {
                    // TODO: file not saved status
                }
            }
        }

        #endregion

        #region Events

        private void PDFView_OnSetDoc()
        {
            PdfViewCtrl.SetZoom(DEFAULT_ZOOM_START_POSITION);
            PdfViewCtrl.SetZoomEnabled(false);

            IsPdfOpenPageVisible = false;

            mToolManagerDoc = new ToolManager(PdfViewCtrl);

            // Set Undo and Redo Manager
            UndoRedoManager undoRedoManager = new UndoRedoManager();
            mToolManagerDoc.SetUndoRedoManager(undoRedoManager);

            AnnotToolbar.ToolManager = mToolManagerDoc;
            AnnotToolbar.IsEnabled = true;

            PageNumbIndicator.PDFViewCtrl = PdfViewCtrl;
        }
        #endregion

    }
}
