﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Diagnostics;
using System.Diagnostics.CodeAnalysis;
using System.Linq;
using System.Runtime.CompilerServices;
using System.Threading;
using System.Threading.Tasks;
using System.Xml.Serialization;

using Windows.ApplicationModel.Core;
using Windows.Foundation;
using Windows.Storage;
using Windows.Storage.AccessCache;
using Windows.Storage.FileProperties;
using Windows.System;
using Windows.UI;
using Windows.UI.Core;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Microsoft.UI.Xaml.Controls;

using pdftron.PDF.Tools.Collections;
using pdftron.PDF.Tools.Commands;
using pdftron.PDF.Tools.Controls.Viewer;
using pdftron.PDF.Tools.Controls.ViewModels.Common;
using pdftron.PDF.Tools.Controls.ViewModels.FileOpening;
using pdftron.PDF.Tools.Utilities;


namespace pdftron.PDF.Tools.Controls.ViewModels.Viewer.Tab
{
    /// <summary>
    /// <para>
    /// This class lets you put multiple PDFViewCtrl's into a tab like control.
    /// You can use the TabControlViewModel.TabControlButtonViewModel property to get a
    /// complimentary view model for Tab Selection.
    /// </para>
    /// <para>Note: Always dispose the TabControlViewModel explicitly when done with it, as this will free up the ToolManagers.</para>
    /// </summary>
    public class TabControlViewModel : ViewModelBase, IDisposable
    {
        internal const bool _showSemaphoreLogging = false; // NOTE: change value here to be used to debugg current semaphore calls
        private const int MAX_ACCESS_LIST_ENTRIES = 60;

        private Windows.UI.Core.CoreWindow _coreWindow;

        private static readonly TabControlViewModel _instance = new TabControlViewModel();

        [SuppressMessage("Simplification", "RCS1085:Use auto-implemented property.", Justification = "<Pending>")]
        public static TabControlViewModel Instance => _instance;

        private VisibleTabChangedDelegate _visibleTabChanged;
        public delegate void VisibleTabChangedDelegate(TabControlViewModel tabControlViewModel);
        public event VisibleTabChangedDelegate VisibleTabChanged
        {
            add
            {
                if (_visibleTabChanged?.GetInvocationList().Select(x => x.Method).Contains(value.Method) != true)
                    _visibleTabChanged += value;
            }
            remove
            {
                if (_visibleTabChanged != null)
                {
                    VisibleTabChangedDelegate eventMethod = (VisibleTabChangedDelegate)Array.Find(
                        _visibleTabChanged.GetInvocationList(),
                        x => x.Method == value.Method);
                    if (eventMethod != null)
                        _visibleTabChanged -= eventMethod;
                }
            }
        }

        private BrowseDocumentRequestedDelegate _browseDocumentRequested;
        public delegate void BrowseDocumentRequestedDelegate(TabControlViewModel tabControlViewModel);

        /// <summary>
        /// Event is raised when user requests brose a document
        /// </summary>
        public event BrowseDocumentRequestedDelegate BrowseDocumentRequested
        {
            add
            {
                if (_browseDocumentRequested?.GetInvocationList().Select(x => x.Method).Contains(value.Method) != true)
                    _browseDocumentRequested += value;
            }
            remove
            {
                if (_browseDocumentRequested != null)
                {
                    BrowseDocumentRequestedDelegate eventMethod = (BrowseDocumentRequestedDelegate)Array.Find(
                        _browseDocumentRequested.GetInvocationList(),
                        x => x.Method == value.Method);
                    if (eventMethod != null)
                        _browseDocumentRequested -= eventMethod;
                }
            }
        }

        private CreateDocumentRequestedDelegate _createDocumentRequested;
        public delegate void CreateDocumentRequestedDelegate(TabControlViewModel tabControlViewModel);

        /// <summary>
        /// Event is raised when user requests create a new document
        /// </summary>
        public event CreateDocumentRequestedDelegate CreateDocumentRequested
        {
            add
            {
                if (_createDocumentRequested?.GetInvocationList().Select(x => x.Method).Contains(value.Method) != true)
                    _createDocumentRequested += value;
            }
            remove
            {
                if (_createDocumentRequested != null)
                {
                    CreateDocumentRequestedDelegate eventMethod = (CreateDocumentRequestedDelegate)Array.Find(
                        _createDocumentRequested.GetInvocationList(),
                        x => x.Method == value.Method);
                    if (eventMethod != null)
                        _createDocumentRequested -= eventMethod;
                }
            }
        }

        private TabCloseRequestDelegate _TabCloseRequest;
        public delegate void TabCloseRequestDelegate(PdfViewCtrlTabInfo item);
        /// <summary>
        /// Event is raised when user requests to close a tab by clicking the tab's close button
        /// </summary>
        public event TabCloseRequestDelegate TabCloseRequest
        {
            add
            {
                if (_TabCloseRequest?.GetInvocationList().Select(x => x.Method).Contains(value.Method) != true)
                    _TabCloseRequest += value;
            }
            remove
            {
                if (_TabCloseRequest != null)
                {
                    TabCloseRequestDelegate eventMethod = (TabCloseRequestDelegate)Array.Find(
                        _TabCloseRequest.GetInvocationList(),
                        x => x.Method == value.Method);
                    if (eventMethod != null)
                        _TabCloseRequest -= eventMethod;
                }
            }
        }

        private ToolManager.ShowingPopupMenuHandler _showToolPopupMenu;
        public event ToolManager.ShowingPopupMenuHandler ShowToolPopupMenu
        {
            add
            {
                if (_showToolPopupMenu?.GetInvocationList().Select(x => x.Method).Contains(value.Method) != true)
                    _showToolPopupMenu += value;
            }
            remove
            {
                if (_showToolPopupMenu != null)
                {
                    ToolManager.ShowingPopupMenuHandler eventMethod = (ToolManager.ShowingPopupMenuHandler)Array.Find(
                        _showToolPopupMenu.GetInvocationList(),
                        x => x.Method == value.Method);
                    if (eventMethod != null)
                        _showToolPopupMenu -= eventMethod;
                }
            }
        }

        private ToolManager.ShowingContextMenuHandler _showToolContextMenu;
        public event ToolManager.ShowingContextMenuHandler ShowToolContextMenu
        {
            add
            {
                if (_showToolContextMenu?.GetInvocationList().Select(x => x.Method).Contains(value.Method) != true)
                    _showToolContextMenu += value;
            }
            remove
            {
                if (_showToolContextMenu != null)
                {
                    ToolManager.ShowingContextMenuHandler eventMethod = (ToolManager.ShowingContextMenuHandler)Array.Find(
                        _showToolContextMenu.GetInvocationList(),
                        x => x.Method == value.Method);
                    if (eventMethod != null)
                        _showToolContextMenu -= eventMethod;
                }
            }
        }


        private SemaphoreSlim _loadingSemaphore;
        private readonly SemaphoreSlim _accessSemaphore;
        private readonly SemaphoreSlim _futureAccessListCleanupSemaphore = new SemaphoreSlim(1);

        private bool _isEnabled = true;
        private ObservableCollection<PdfViewCtrlTabInfo> _tabs;
        private bool _autoSaveState;
        private ShowIfModifiedStates _showIfDocumentModified;
        private DateTime _lastFilePropertyUpdatedTime = DateTime.MinValue;
        private PdfViewCtrlTabInfo _selectedTab;
        private int _visibleTabIndex = -1;
        private bool _showImagePreviews;
        private bool _hasFixedItemAtEnd;
        private int _maximumItems = 20;
        private static int _vmNumberPool;
        private int _vmNumber = ++_vmNumberPool;
        private bool _showFixedItem;
        private DispatcherTimer _autoSaveTimer;
        private bool _blockWhileSaving;
        private TabViewWidthMode _tabWidthMode = TabViewWidthMode.SizeToContent;
        private bool _isAddTabButtonVisible = true;
        private bool _isAddTabButtonEnabled = true;
        private bool _isTabReorderEnabled = true;
        private bool _showTabFileInfo = true;

        protected bool _savingInProgress;
        protected List<string> _futureAccessListTokensToDelete;

        static TabControlViewModel()
        {
        }

        private TabControlViewModel()
        {
            CMDBrowseDocument = new RelayCommand(BrowseDocument);
            CMDCreateDocument = new RelayCommand(CreateDocument);

            _tabs = new ObservableCollection<PdfViewCtrlTabInfo>();
            _futureAccessListTokensToDelete = new List<string>();
            _accessSemaphore = new SemaphoreSlim(1);
        }

        #region Commands
        public RelayCommand CMDBrowseDocument { get; }

        public RelayCommand CMDCreateDocument { get; }
        #endregion

        /// <summary>
        /// Flag to indicate when TabControlViewModel has been initialized
        /// </summary>
        public static bool IsReady { get; private set; }

        public bool IsEnabled
        {
            get => _isEnabled;
            set => Set(ref _isEnabled, value);
        }

        public async Task InitializeInstanceAsync(bool enableTracking = true)
        {
            await LoadingSemaphore.WaitAsync().ConfigureAwait(false);
            try
            {
                if (!IsReady)
                {
                    List<PdfViewCtrlTabInfo> tabList = await _instance.LoadTabStateAsync();
                    if (tabList.Count > 0)
                    {
                        await UtilityFunctions.RunAsyncSaferAsync(GetCoreWindow(), () => 
                        {
                            foreach (PdfViewCtrlTabInfo tab in tabList)
                                _instance.Tabs.Add(tab);
                        });
                    }

                    _instance._lastFilePropertyUpdatedTime = DateTime.Now;
                    _instance.ShowIfDocumentModified = ShowIfModifiedStates.DoNotShow;

                    IsReady = true;
                }

                if (!enableTracking)
                {
                    foreach (var tab in new ObservableCollection<PdfViewCtrlTabInfo>(_instance.Tabs))
                    {
                        try
                        {
                            await CloseTabImplAsync(tab, false);
                        }
                        catch(Exception ex)
                        {
                            AppCenterHandlerBase.CURRENT.TrackError(ex);
                        }
                    }
                }
            }
            catch (Exception e)
            {
                Debug.WriteLine("Error: " + e.ToString());
            }
            finally
            {
                LoadingSemaphore.Release();
            }
        }

        private SemaphoreSlim LoadingSemaphore => _loadingSemaphore ?? (_loadingSemaphore = new SemaphoreSlim(1));

        private void SemWait(
            bool supress = false,
            [CallerMemberName] string member = "",
            [CallerLineNumber] int line = 0)
        {
#if DEBUG
            if (_showSemaphoreLogging && !supress)
            {
                Debug.WriteLine("Function {0}, line {1} waiting for Semaphore", member, line);
            }
#endif
            SemaphoreSlim tmpSema = _accessSemaphore;
            tmpSema.Wait();
#if DEBUG
            if (_showSemaphoreLogging && !supress)
            {
                Debug.WriteLine("Function {0}, line {1} got Semaphore", member, line);
            }
#endif
        }

        private async Task SemWaitAsync(
            bool supress = false,
            [CallerMemberName] string member = "",
            [CallerLineNumber] int line = 0)
        {
#if DEBUG
            if (_showSemaphoreLogging && !supress)
            {
                Debug.WriteLine("Function {0}, line {1} waiting async for Semaphore", member, line);
            }
#endif
            SemaphoreSlim tmpSema = _accessSemaphore;
            await tmpSema.WaitAsync().ConfigureAwait(false);
#if DEBUG
            if (_showSemaphoreLogging && !supress)
            {
                Debug.WriteLine("Function {0}, line {1} got async Semaphore", member, line);
            }
#endif
        }

        private void SemRelease(
            bool supress = false,
            [CallerMemberName] string member = "",
            [CallerLineNumber] int line = 0)
        {
#if DEBUG
            if (_showSemaphoreLogging && !supress)
            {
                Debug.WriteLine("Function {0}, line {1} released Semaphore", member, line);
            }
#endif
            SemaphoreSlim tmpSema = _accessSemaphore;
            tmpSema.Release();
        }

        public void Dispose()
        {
            StopAutoSavingOnTimer();

            foreach (PdfViewCtrlTabInfo tab in Tabs)
                tab.CleanUp();
        }

        public ObservableCollection<PdfViewCtrlTabInfo> Tabs
        {
            get => _tabs;
            set => Set(ref _tabs, value);
        }

        public string DefaultImageURL { get; set; } = "ms-appx:///pdftron.PDF.Tools/Controls/Resources/DefaultFileIcon.png";

        /// <summary>
        /// Sets whether or not to keep the opened tabs saved when ViewModel
        /// gets deactivated
        /// </summary>
        public bool KeepOpenedTabs { get; set; } = false;

        /// <summary>
        /// Get the number of current opened tabs
        /// </summary>
        public int OpenTabs
        {
            get
            {
                if (Tabs == null)
                    return 0;

                return Tabs.Count;
            }
        }

        public PdfViewCtrlTabInfo OldestViewedTab
        {
            get
            {
                //if (Tabs.Count == 0 || (Tabs.Count == 1 && Tabs[0].IsFixedItem))
                //    return null;

                if (Tabs.Count == 0)
                    return null;

                PdfViewCtrlTabInfo oldest = Tabs[0];
                foreach (PdfViewCtrlTabInfo info in Tabs)
                {
                    //if (!info.IsFixedItem && info.MetaData.TabLastViewedTimeStamp < oldest.MetaData.TabLastViewedTimeStamp)
                    if (info.MetaData.TabLastViewedTimeStamp < oldest.MetaData.TabLastViewedTimeStamp)
                        oldest = info;
                }

                return oldest;
            }
        }

        /// <summary>
        /// Enable tab Auto Save to ensure the temporary PDF file is saved over time
        /// NOTE: It does not save back to the original file, only to the temporary one
        /// </summary>
        public bool AutoSaveTabState
        {
            get => _autoSaveState;
            set
            {
                if (value != _autoSaveState)
                {
                    _autoSaveState = value;
                }

                if (_autoSaveState)
                {
                    if (ViewerSettings.AutoSaveOn && !ViewerSettings.AutoSaveOnEveryChange)
                        StartAutoSavingOnTimer();
                }
                else
                {
                    StopAutoSavingOnTimer();
                }
            }
        }

        [XmlIgnore]
        public ShowIfModifiedStates ShowIfDocumentModified
        {
            get => _showIfDocumentModified;
            set
            {
                if (value != _showIfDocumentModified)
                {
                    _showIfDocumentModified = value;
                    foreach (PdfViewCtrlTabInfo tab in Tabs)
                        tab.ShowIfDocumentModified = _showIfDocumentModified;
                }
            }
        }

        private void BrowseDocument(object parameter)
        {
            AnalyticsHandlerBase.CURRENT.SendEvent("Tabs", "Tab Plus Button clicked");
            AppCenterHandlerBase.CURRENT.SendEvent("Open File", new Dictionary<string, string>() { ["Command"] = "Viewer - Tabs Plus Button" });

            _browseDocumentRequested?.Invoke(this);
        }

        private void CreateDocument(object parameter)
        {


            _createDocumentRequested?.Invoke(this);
        }

        private async void Info_TabSelected(PdfViewCtrlTabInfo item)
        {
            System.Diagnostics.Debug.WriteLine("====> DEBUG: " + nameof(Info_TabSelected) + " called");

            await SelectTabAsync(item);
        }

        /// <summary>
        /// Tab is about to be closed
        /// </summary>
        /// <param name="item">Tab to be closed</param>
        private void Info_TabClosed(PdfViewCtrlTabInfo item)
        {
            if(_TabCloseRequest != null)
                _TabCloseRequest.Invoke(item);
        }

        public async Task<PdfViewCtrlTabInfo> AddTabAsync(StorageFile orig, StorageFile tmp, PDFDoc doc, string title)
        {
            PdfViewCtrlTabInfo newInfo = new PdfViewCtrlTabInfo(orig, tmp, doc, title);
            await AddTabAsync(newInfo);

            return newInfo;
        }

        public async Task AddTabAsync(PdfViewCtrlTabInfo newInfo)
        {
            newInfo.SetUpPDFViewCtrlAndTools += SetUpPDFViewCtrlAndTools;

            newInfo.ShowPreview = ShowImagePreviews;
            newInfo.ThumbnailFallback = DefaultImageURL;

            newInfo.TabSelected += Info_TabSelected;
            newInfo.TabClosed += Info_TabClosed;
            newInfo.PdfUpdated += TabItem_PdfUpdated;
            newInfo.MetaData.TabAddedTimeStamp = DateTime.Now;
            newInfo.ShowIfDocumentModified = ShowIfDocumentModified;

            Tabs.Add(newInfo);

            ResolveFirstAndLast();

            SaveFilesForFutureUse(newInfo);
            await CleanUpFutureAccessListAsync();
        }

        [SuppressMessage("Major Code Smell", "S112:General exceptions should never be thrown", Justification = "<Pending>")]
        public PdfViewCtrlTabInfo ReplaceTab(int index, StorageFile orig, StorageFile tmp, PDFDoc doc, string title)
        {
            if (index < 0 || index >= OpenTabs)
                throw new IndexOutOfRangeException(string.Format("Tabs only has {0} items, index {1} is out of range.", Tabs.Count, index));

            PdfViewCtrlTabInfo newInfo = new PdfViewCtrlTabInfo(orig, tmp, doc, title);
            ReplaceTab(index, newInfo);
            
            return newInfo;
        }

        [SuppressMessage("Major Code Smell", "S112:General exceptions should never be thrown", Justification = "<Pending>")]
        protected void ReplaceTab(int index, PdfViewCtrlTabInfo newInfo)
        {
            if (index < 0 || index >= OpenTabs)
                throw new IndexOutOfRangeException(string.Format("Tabs only has {0} items, index {1} is out of range.", OpenTabs, index));

            PdfViewCtrlTabInfo oldTab = Tabs[index];
            oldTab.CleanUp();

            newInfo.SetUpPDFViewCtrlAndTools += SetUpPDFViewCtrlAndTools;
            Tabs[index] = newInfo;
            newInfo.ShowPreview = ShowImagePreviews;
            newInfo.ThumbnailFallback = DefaultImageURL;
            //newInfo.ItemClicked += TabItem_ItemClicked;
            newInfo.TabSelected += Info_TabSelected;
            newInfo.MetaData.TabAddedTimeStamp = DateTime.Now;

            ResolveFirstAndLast();
            SaveFilesForFutureUse(newInfo);
            RemoveTabFromFutureAccessList(oldTab);

            SemWait();
            try
            {
                SaveTabStateNoLock(Tabs.IndexOf(newInfo));
            }
            finally
            {
                SemRelease();
            }
        }

        /// <summary>
        /// Close tab
        /// </summary>
        /// <param name="item">Tab to be closed</param>
        public async Task CloseTabAsync(PdfViewCtrlTabInfo item) 
        {
            if (item == null || Tabs == null)
                return;

            if (Tabs.Contains(item) == false)
                return;

            int index = Tabs.IndexOf(item);
            if (index == -1)
                return;

            await CloseTabAsync(index);
        }

        /// <summary>
        /// Close the tab by index position
        /// </summary>
        /// <param name="index">The tab position</param>
        public async Task CloseTabAsync(int index, bool activateNewTab = true)
        {
            if (Tabs == null || (index >= Tabs.Count))
                return;

            if (index < 0 || index >= OpenTabs)
                return;

            if (Tabs != null)
            {
                var tabToClose = Tabs[index];
                await CloseTabImplAsync(tabToClose, activateNewTab);
            }
        }

        public async Task CloseTabImplAsync(PdfViewCtrlTabInfo tab, bool activateNewTab = true)
        {
            AnalyticsHandlerBase.CURRENT.SendEvent("Tabs", "Tab Closed");

            if (tab != null)
            {
                RemoveTabFromFutureAccessList(tab);

                tab.TabClosed -= Info_TabClosed;
                tab.TabSelected -= Info_TabSelected;
                tab.PdfUpdated -= TabItem_PdfUpdated;
                tab.CleanUp();
                Tabs.Remove(tab);
                tab = null;
            }

            if (activateNewTab)
                await UpdateSelectionIfNecessaryAsync();

            ResolveFirstAndLast();
            SaveAllTabsState();

            // All tabs closed, so go to init state
            if (Tabs.Count <= 0)
            {
                _visibleTabIndex = -1;
            }
        }

        public async Task CloseOldestAddedTabAsync()
        {
            if (Tabs.Count == 0)
                return;

            PdfViewCtrlTabInfo oldest = Tabs[0];
            foreach (PdfViewCtrlTabInfo info in Tabs)
            {
                if (info.MetaData.TabAddedTimeStamp < oldest.MetaData.TabAddedTimeStamp)
                    oldest = info;
            }

            await CloseTabAsync(Tabs.IndexOf(oldest));
        }

        public async Task CloseOldestViewedTabAsync()
        {
            if (Tabs.Count == 0)
                return;

            PdfViewCtrlTabInfo oldest = Tabs[0];
            foreach (PdfViewCtrlTabInfo info in Tabs)
            {
                if (info.MetaData.TabLastViewedTimeStamp < oldest.MetaData.TabLastViewedTimeStamp)
                    oldest = info;
            }

            await CloseTabAsync(Tabs.IndexOf(oldest));
        }

        /// <summary>
        /// Select tab implementation
        /// </summary>
        /// <param name="index">Tab index</param>
        public async Task SelectTabImplAsync(int index)
        {
            if (Tabs != null)
            {
                for (int i = 0; i < Tabs.Count; ++i)
                {
                    if (i != index)
                    {
                        if (Tabs[i].IsSelected && Tabs[i].PDFViewCtrl != null)
                        {
                            if (Tabs[i].IsReflow)
                                Tabs[i].PDFViewCtrl.SetCurrentPage(Tabs[i].ReflowCurrentPage);

                            Tabs[i].PDFViewCtrl.Deactivate();
                            SemWait();
                            try
                            {
                                SaveTabStateNoLock(i);
                            }
                            finally
                            {
                                SemRelease();
                            }

                            await Tabs[i].SaveDocumentAsync();
                        }

                        if (i >= 0 && i < Tabs.Count)
                        {
                            Tabs[i].IsSelected = false;
                            Tabs[i].Deactivate();
                        }
                    }
                }

                if (index >= 0 && index < Tabs.Count)
                {
                    if (!Tabs[index].IsSelected)
                        AnalyticsHandlerBase.CURRENT.SendEvent("Tabs", "Tab clicked");

                    Tabs[index].IsSelected = true;
                    await Tabs[index].ActivateAsync();
                    SelectedTab = Tabs[index];
                    VisibleTabIndex = Tabs.IndexOf(SelectedTab);

                    if (SelectedTab.ToolManager != null) // NOTE: Attempt to avoid Null exception
                        SelectedTab.ToolManager.OnShowContextMenu += ToolManager_OnShowContextMenu;

                    SelectedTab.PDFViewCtrl?.Activate();
                }

                if (SelectedTab != null)
                    AutoSaveTabState = SelectedTab.CanBeSaved;
            }
        }

        private void ToolManager_OnShowContextMenu(ContextMenuControl menu, ShowingContextMenuEventArgs eventArgs)
        {
            _showToolContextMenu?.Invoke(menu, eventArgs);
        }

        public async Task SelectTabAsync(PdfViewCtrlTabInfo tab)
        {
            if (Tabs == null || Tabs.Count <= 0)
                return;

            var index = Tabs.IndexOf(tab);
            if (index == -1)
                return;

            if (Tabs[index].IsSelected)
                return;

            await SelectTabImplAsync(index);
        }

        public async Task ShowLastViewedTabAsync() => await UpdateSelectionIfNecessaryAsync();

        /// <summary>
        /// Check if any of the opened tabs/documents already contain the Storagefile
        /// </summary>
        /// <param name="file">File to check</param>
        /// <returns>True if tab control already contains the file, false otherwise</returns>
        public bool ContainsFile(StorageFile file)
        {
            SemWait();
            try
            {
                foreach (PdfViewCtrlTabInfo info in Tabs)
                {
                    if ((info.TempFile != null && UtilityFunctions.AreFilesEqual(info.TempFile, file))
                            || (info.OriginalFile != null && UtilityFunctions.AreFilesEqual(info.OriginalFile, file))
                            || (info.FileSourceIfNotSaveable != null && UtilityFunctions.AreFilesEqual(info.FileSourceIfNotSaveable, file)))
                    {
                        return true;
                    }
                }
            }
            finally
            {
                SemRelease();
            }

            return false;
        }

        public bool ContainsFilePath(string filePath)
        {
            SemWait();
            try
            {
                foreach (PdfViewCtrlTabInfo info in Tabs)
                {
                    if ((info.TempFile?.Path.Equals(filePath) == true)
                            || (info.OriginalFile?.Path.Equals(filePath) == true))
                    {
                        return true;
                    }
                }
            }
            finally
            {
                SemRelease();
            }

            return false;
        }

        /// <summary>
        /// If the file is already in a tab, activates that tab and returns true. Otherwise, returns false.
        /// This function is virtual as it may not behave as expected with files from all sources, such as remote storage which may create a new temp file each time.
        /// </summary>
        /// <param name="file"></param>
        /// <returns></returns>
        public virtual async Task<bool> CheckIfFileIsOpenAndActivateAsync(NewDocumentProperties documentProperties)
        {
            SemWait();
            PdfViewCtrlTabInfo tabtoActivate = null;
            try
            {
                foreach (PdfViewCtrlTabInfo info in Tabs)
                {
                    if (info.TempFile != null && UtilityFunctions.AreFilesEqual(info.TempFile, documentProperties.TemporaryFile))
                    {
                        tabtoActivate = info;
                        break;
                    }
                }
            }
            finally
            {
                SemRelease();
            }

            if (tabtoActivate != null)
            {
                if (documentProperties.Doc != null)
                {
                    tabtoActivate.Doc = documentProperties.Doc;
                }

                await SelectTabImplAsync(Tabs.IndexOf(tabtoActivate));
                return true;
            }

            return false;
        }

        public async Task<bool> CheckIfOriginalIsOpenAndActivateAsync(StorageFile file)
        {
            SemWait();
            PdfViewCtrlTabInfo tabtoActivate = null;
            try
            {
                foreach (PdfViewCtrlTabInfo info in Tabs)
                {
                    if (info.FileSourceIfNotSaveable != null && UtilityFunctions.AreFilesEqual(info.FileSourceIfNotSaveable, file))
                    {
                        tabtoActivate = info;
                        break;
                    }
                }
            }
            finally
            {
                SemRelease();
            }

            if (tabtoActivate != null)
            {
                await SelectTabImplAsync(Tabs.IndexOf(tabtoActivate));
                return true;
            }

            return false;
        }

        /// <summary>
        /// Save the state of all opened tabs
        /// </summary>
        public void SaveAllTabsState()
        {
            if (OpenTabs == 0)
                return;

            SemWait();
            try
            {             
                TabSetting.OpenTabs = OpenTabs;
                for (int i = 0; i < OpenTabs; i++)
                    SaveTabStateNoLock(i);
            }
            finally
            {
                SemRelease();
            }
        }

        /// <summary>
        /// Save the state of the current opened tab
        /// </summary>
        public void SaveCurrentTabState()
        {
            if (OpenTabs == 0)
                return;

            SemWait();
            try
            {
                TabSetting.OpenTabs = OpenTabs;

                for (int i = 0; i < OpenTabs; i++)
                {
                    if (Tabs[i] == SelectedTab)
                        SaveTabStateNoLock(i);
                }

            }
            finally
            {
                SemRelease();
            }
        }

        /// <summary>
        /// Makes no tab selected. Reactivate by selecting a tab.
        /// It is up to you to save the document in the current tab before Deactivating it.
        /// </summary>
        public void Deactivate()
        {
            if (SelectedTab != null)
            {
                try
                {
                    SemWait();
                    SaveTabStateNoLock(VisibleTabIndex);
                }
                finally
                {
                    SemRelease();
                }

                SelectedTab.IsSelected = false;
                SelectedTab.Deactivate();
            }

            foreach (PdfViewCtrlTabInfo tab in Tabs)
            {
                tab.CleanUp();
            }

            if (KeepOpenedTabs == false && Tabs.Count > 0)
                Tabs.Clear();

            SelectedTab = null;
            AutoSaveTabState = false;
        }

        /// <summary>
        /// Update the properties of every opened tab/document like File Size and Last Modified Date
        /// </summary>
        /// <remarks>
        /// NOTE: this call must be called with await to guarantee it will finish, otherwise there is a risk of deadlock
        /// </remarks>
        public async Task UpdateFilePropertiesAsync()
        {
            try
            {
                /* NOTE:
                 * 
                 * below the call FileStorage.GetBasicPropertiesAsync() needs to be finished to ensure
                 * the UI Thread can continue, so await is necessary otherwise the Semaphore will never be released
                 * and cause a deadlock
                 */
                await SemWaitAsync().ConfigureAwait(false);

                TimeSpan timeSinceUpdate = DateTime.Now - _lastFilePropertyUpdatedTime;
                if (timeSinceUpdate > TimeSpan.FromSeconds(3))
                {
                    foreach (PdfViewCtrlTabInfo tab in Tabs)
                    {
                        if (tab.OriginalFile != null)
                        {
                            BasicProperties props = await tab.OriginalFile.GetBasicPropertiesAsync();
                            DateTimeOffset origOffset = tab.SystemLastModifiedDate;
                            tab.SystemFileSize = props.Size;
                            tab.SystemLastModifiedDate = props.DateModified;
                        }
                    }

                    _lastFilePropertyUpdatedTime = DateTime.Now;
                }
            }
            catch (Exception ex)
            {
                AppCenterHandlerBase.CURRENT.TrackError(ex);
            }
            finally
            {
                SemRelease();
            }
        }

        public PdfViewCtrlTabInfo SelectedTab
        {
            get => _selectedTab;
            set
            {
                if (_selectedTab != value)
                {
                    _selectedTab = value;
                    RaisePropertyChanged();
                    RaiseVisibleTabChanged();
                }
            }
        }

        /// <summary>
        /// Current selected tab index position
        /// </summary>
        public int VisibleTabIndex
        {
            get => _visibleTabIndex;
            private set
            {
                if (value != _visibleTabIndex)
                {
                    _visibleTabIndex = value;
                    if (value >= 0)
                        SelectedTab = Tabs[_visibleTabIndex];

                    RaisePropertyChanged();
                }
            }
        }

        public bool ShowImagePreviews
        {
            get => _showImagePreviews;
            set
            {
                if (value != _showImagePreviews)
                {
                    _showImagePreviews = value;
                    foreach (PdfViewCtrlTabInfo item in Tabs)
                        item.ShowPreview = _showImagePreviews;
                }
            }
        }

        public int MaximumItems
        {
            get => _maximumItems;
            set
            {
                if (value != _maximumItems)
                {
                    _maximumItems = value;
                }
            }
        }

        public bool ShowFixedItem
        {
            get => _showFixedItem;
            set
            {
                if (value != _showFixedItem)
                {
                    _showFixedItem = value;
                    RaisePropertyChanged();
                }
            }
        }

        public TabViewWidthMode TabWidthMode
        {
            get => _tabWidthMode;
            set => Set(ref _tabWidthMode, value);
        }

        /// <summary>
        /// Choose whether or not the add tab button is Visible or Collapsed
        /// </summary>
        public bool IsAddTabButtonVisible
        {
            get 
            {
                if (!_isAddTabButtonEnabled)
                    return false;

                return OpenTabs < MaximumItems;
            }
            set => Set(ref _isAddTabButtonVisible, value);
        }

        /// <summary>
        /// Choose whether or not the add tab button is enabled or disabled
        /// </summary>
        public bool IsAddTabButtonEnabled
        {
            get { return _isAddTabButtonEnabled; }
            set
            {
                if (Set(ref _isAddTabButtonEnabled, value))
                {
                    RaisePropertyChanged(nameof(IsAddTabButtonVisible));
                }
            }
        }

        /// <summary>
        /// Choose whether or not Tab Reordering is enabled
        /// </summary>
        public bool IsTabReorderEnabled
        {
            get { return _isTabReorderEnabled; }
            set => Set(ref _isTabReorderEnabled, value);
        }

        /// <summary>
        /// Enable or Disable the file path popup when right clicking on a tab
        /// </summary>
        public bool ShowTabFileInfo
        {
            get { return _showTabFileInfo; }
            set { Set(ref _showTabFileInfo, value); }
        }

        public void RaisePropertyChangedImpl(string property) => RaisePropertyChanged(property);

        private void RaiseVisibleTabChanged() => _visibleTabChanged?.Invoke(this);

        private async void TabItem_PdfUpdated(object sender, RoutedEventArgs e)
        {
            await SemWaitAsync();
            try
            {
                if (sender is PdfViewCtrlTabInfo tab)
                {
                    SaveFilesForFutureUse(tab);
                    int index = Tabs.IndexOf(tab);
                    if (index >= 0 && index < OpenTabs)
                        TabSetting.SaveSettings(index, tab.MetaData);
                }
            }
            finally
            {
                SemRelease();
            }
        }

        private void ResolveFirstAndLast()
        {
            if (Tabs.Count > 1)
            {
                Tabs[0].First = true;
                Tabs[1].First = false;
                Tabs[Tabs.Count - 1].Last = true;
                Tabs[Tabs.Count - 2].Last = false;
            }
            else if (Tabs.Count == 1)
            {
                Tabs[0].First = true;
                Tabs[0].Last = true;
            }
        }

        private async Task UpdateSelectionIfNecessaryAsync()
        {
            if (Tabs.Count == 0)
            {
                SelectedTab = null;
                return;
            }

            PdfViewCtrlTabInfo lastViewed = Tabs[0];
            foreach (PdfViewCtrlTabInfo info in Tabs)
            {
                if (info.IsSelected)
                {
                    VisibleTabIndex = Tabs.IndexOf(info);
                    return;
                }

                if (info.MetaData.TabLastViewedTimeStamp > lastViewed.MetaData.TabLastViewedTimeStamp)
                    lastViewed = info;
            }

            await SelectTabAsync(lastViewed);
        }

        /// <summary>
        /// This call will save the state of the tab id passed with the following steps:
        /// 1. Refresh/Save metadata like: zoom level, last page seen, page rotation, etc
        /// 2. 
        /// </summary>
        /// <remarks>
        /// Always lock around this, as this will not lock
        /// </remarks>
        /// <param name="tabIndex"></param>
        private void SaveTabStateNoLock(int tabIndex)
        {
            TabSetting.OpenTabs = OpenTabs;

            if (Tabs == null || tabIndex >= Tabs.Count)
                return;

            if (tabIndex >= OpenTabs)
                return;

            Tabs[tabIndex].RefreshState();
            SaveFilesForFutureUse(Tabs[tabIndex]);
            TabSetting.SaveSettings(tabIndex, Tabs[tabIndex].MetaData);
        }

        private async Task<List<PdfViewCtrlTabInfo>> LoadTabStateAsync()
        {
            List<PdfViewCtrlTabInfo> tabs = new List<PdfViewCtrlTabInfo>();
            int numTabs = TabSetting.OpenTabs;
            for (int i = 0; i < numTabs; i++)
            {
                try
                {
                    PdfViewCtrlTabInfo info = null;
                    PdfViewCtrlTabMetaData metaData = TabSetting.GetFromSettings(i);
                    if (!string.IsNullOrWhiteSpace(metaData.FutureAccessListToken))
                    {
                        StorageFile tmpFile = await GetSavedFileAsync(metaData.FutureAccessListToken).ConfigureAwait(false);
                        if (tmpFile != null)
                        {
                            StorageFile origFile = await GetSavedFileAsync(metaData.OriginalFileToken).ConfigureAwait(false);
                            info = new PdfViewCtrlTabInfo(origFile, tmpFile, metaData);
                            if (!string.IsNullOrWhiteSpace(metaData.FileSourceTokenIfNotSaveable))
                                info.FileSourceIfNotSaveable = await GetSavedFileAsync(metaData.FileSourceTokenIfNotSaveable).ConfigureAwait(false);

                            if (origFile != null)
                            {
                                BasicProperties props = await origFile.GetBasicPropertiesAsync().AsTask().ConfigureAwait(false);
                                info.SystemFileSize = props.Size;
                                info.SystemLastModifiedDate = props.DateModified;
                            }
                        }
                    }

                    if (!string.IsNullOrEmpty(metaData.FileSourceTokenIfNotSaveable))
                    {
                        StorageFile sourceFile = await GetSavedFileAsync(metaData.FileSourceTokenIfNotSaveable).ConfigureAwait(false);
                        if (sourceFile != null)
                        {
                            if (info == null)
                            {
                                info = PdfViewCtrlTabInfo.CreateUniversalTabInfo(metaData, sourceFile);
                            }
                            else
                            {
                                info.FileSourceIfNotSaveable = sourceFile;
                            }
                        }
                    }

                    if (info != null)
                    {
                        tabs.Add(info);
                        info.TabSelected += Info_TabSelected;
                        info.TabClosed += Info_TabClosed;
                        info.SetUpPDFViewCtrlAndTools += SetUpPDFViewCtrlAndTools;
                    }
                }
                catch (Exception ex)
                {
                    AnalyticsHandlerBase.CURRENT.SendEvent(AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, AnalyticsHandlerBase.GetExceptionMessage(ex));
                }
            }
            
            return tabs;
        }

        /// <summary>
        /// Save current PDF docs (temp and original) for future access (using StorageFile token) if these docs still exist in the current run
        /// </summary>
        /// <param name="info">The doc tab information</param>
        protected virtual void SaveFilesForFutureUse(PdfViewCtrlTabInfo info)
        {
            StorageItemAccessList accessList = StorageApplicationPermissions.FutureAccessList;

            // Ideally, this should never happen. We're supposed to have clean up code running before we get here.
            if (accessList.Entries.Count + 2 >= accessList.MaximumItemsAllowed)
            {
                accessList.Remove(accessList.Entries[accessList.Entries.Count - 1].Token);
                accessList.Remove(accessList.Entries[accessList.Entries.Count - 1].Token);
                accessList.Remove(accessList.Entries[accessList.Entries.Count - 1].Token);
            }

            if (info.TempFile != null)
            {
                info.MetaData.FutureAccessListToken = accessList.Add(info.TempFile);
            }
            else
            {
                info.MetaData.FutureAccessListToken = string.Empty;
            }

            if (info.OriginalFile != null)
            {
                info.MetaData.OriginalFileToken = accessList.Add(info.OriginalFile);
            }
            else
            {
                info.MetaData.OriginalFileToken = string.Empty;
            }

            if (info.FileSourceIfNotSaveable != null)
            {
                info.MetaData.FileSourceTokenIfNotSaveable = accessList.Add(info.FileSourceIfNotSaveable);
            }
            else
            {
                info.MetaData.FileSourceTokenIfNotSaveable = string.Empty;
            }
        }

        protected virtual async Task<StorageFile> GetSavedFileAsync(string token)
        {
            StorageItemAccessList accessList = StorageApplicationPermissions.FutureAccessList;
            if (accessList.ContainsItem(token))
                return await accessList.GetFileAsync(token);

            return null;
        }

        protected virtual void SetUpPDFViewCtrlAndTools(PDFViewCtrl ctrl, ToolManager toolManager, Deferral deferral)
        {
            try
            {
                // NOTE: control the amount of Zoom In/Out allowed
                ctrl.SetRelativeZoomLimits(PDFViewCtrlPageViewMode.e_fit_page, ViewerSettings.MinZoomLevel, ViewerSettings.MaxZoomLevel);
                ctrl.SetProgressiveRendering(true);
                bool isPhone = UtilityFunctions.GetDeviceFormFactorType() == DeviceFormFactorType.Phone;
                PDFViewCtrlPageViewMode preferredMode = PDFViewCtrlPageViewMode.e_fit_page;
                if (isPhone)
                    preferredMode = PDFViewCtrlPageViewMode.e_fit_width;

                if (ViewerSettings.MaintainZoom)
                {
                    ctrl.SetPageRefViewMode(PDFViewCtrlPageViewMode.e_zoom);
                }
                else
                {
                    ctrl.SetPageRefViewMode(preferredMode);
                }

                if (ctrl.GetPagePresentationMode() == PDFViewCtrlPagePresentationMode.e_single_continuous)
                {
                    ctrl.SetPageViewMode(PDFViewCtrlPageViewMode.e_fit_width);
                }
                else
                {
                    ctrl.SetPageViewMode(preferredMode);
                }

                ctrl.SetUrlExtraction(true);
                ctrl.SetupThumbnails(false, true, true, 500, 100 * 1024 * 1024, 0.1);

                if (ResourceHandler.ViewerResources != null)
                {
                    Color viewerBackground = (Color)ResourceHandler.ViewerResources["MainViewerBackgroundColor"];
                    ctrl.SetBackgroundColor(viewerBackground);
                }

                ctrl.SetHighlightFields(true);
                ctrl.SetPageBox(PageBox.e_user_crop);

                if (isPhone)
                {
                    // Phones have much less memory than tablets, so we set the memory limit accordingly
                    ulong mem = MemoryManager.AppMemoryUsageLimit / (5000000 * 2);
                    ctrl.SetRenderedContentCacheSize(mem);
                }
                else
                {
                    ulong mem = MemoryManager.AppMemoryUsageLimit / (5000000 * 2);
                    mem = Math.Min(mem, 196);
                    mem = Math.Max(mem, 96);
                    ctrl.SetRenderedContentCacheSize(mem);
                }

                toolManager.EnablePopupMenuOnLongPress = true;
                toolManager.UseSmallPageNumberIndicator = false;
                toolManager.PopupAuthorDialogFirstTime = false;
            }
            finally
            {
                deferral.Complete();
            }
        }

        /// <summary>
        /// Will remove every FutureAccessList item except those that are currently in the tabs.
        /// Override or delete in case you handle this list yourself.
        /// </summary>
        public virtual async Task CleanUpFutureAccessListAsync()
        {
            await _futureAccessListCleanupSemaphore.WaitAsync();
            try
            {
                PinnedRootFolders rootFolders = await PinnedRootFolders.GetItemSourceAsync();
                StorageItemAccessList accessList = StorageApplicationPermissions.FutureAccessList;

                if (accessList.Entries.Count < MAX_ACCESS_LIST_ENTRIES + rootFolders.PinnedItems.Count)
                    return;

                List<string> tokensToRemove = new List<string>();
                foreach (AccessListEntry entry in accessList.Entries)
                {
                    bool removeEntry = true;
                    foreach (PdfViewCtrlTabInfo tab in Tabs)
                    {
                        if (entry.Token.Equals(tab.MetaData.FutureAccessListToken)
                                || entry.Token.Equals(tab.MetaData.OriginalFileToken)
                                || (tab.MetaData.FileSourceTokenIfNotSaveable != null && entry.Token.Equals(tab.MetaData.FileSourceTokenIfNotSaveable)))
                        {
                            removeEntry = false;
                            break;
                        }
                    }

                    if (removeEntry && !rootFolders.ContainsToken(entry.Token))
                        tokensToRemove.Add(entry.Token);
                }

                foreach (string token in tokensToRemove)
                {
                    if (accessList.ContainsItem(token))
                        accessList.Remove(token);
                }
            }
            finally
            {
                _futureAccessListCleanupSemaphore.Release();
            }
        }

        protected virtual void RemoveTabFromFutureAccessList(PdfViewCtrlTabInfo tab)
        {
            if (tab?.MetaData != null)
            {
                if (!string.IsNullOrEmpty(tab.MetaData.FutureAccessListToken))
                    _futureAccessListTokensToDelete.Add(tab.MetaData.FutureAccessListToken);

                if (!string.IsNullOrEmpty(tab.MetaData.OriginalFileToken))
                    _futureAccessListTokensToDelete.Add(tab.MetaData.OriginalFileToken);

                UpdateFutureAccessListWithRemovedTabs();
            }
        }

        protected virtual void UpdateFutureAccessListWithRemovedTabs()
        {
            if (_futureAccessListTokensToDelete?.Count > 0)
            {
                StorageItemAccessList accessList = StorageApplicationPermissions.FutureAccessList;

                foreach (string token in _futureAccessListTokensToDelete)
                {
                    if (!string.IsNullOrEmpty(token) && accessList.ContainsItem(token))
                        accessList.Remove(token);
                }

                _futureAccessListTokensToDelete.Clear();
            }
        }

        private void StartAutoSavingOnTimer()
        {
            if (_blockWhileSaving || _autoSaveTimer != null)
                return;

            _autoSaveTimer = new DispatcherTimer
            {
                Interval = TimeSpan.FromSeconds(ViewerSettings.AutoSaveInterval)
            };

            _autoSaveTimer.Tick += AutoSaveTimer_Tick;
            _autoSaveTimer.Start();
        }

        private void StopAutoSavingOnTimer()
        {
            if (_autoSaveTimer != null)
            {
                _autoSaveTimer.Stop();
                _autoSaveTimer.Tick -= AutoSaveTimer_Tick;
                _autoSaveTimer = null;
            }
        }

        private void RefreshAutoSaver()
        {
            if (_blockWhileSaving)
                return;

            if (AutoSaveTabState && _autoSaveTimer != null)
            {
                _autoSaveTimer.Stop();
                _autoSaveTimer.Start();
            }
        }

        /// <summary>
        /// Process the auto save of the opened temporary document
        /// </summary>
        protected virtual async void AutoSaveTimer_Tick(object sender, object e)
        {
            DebugLogHelper.CallSarted();

            // Just for safety if the event was triggered and not time to disabled it properly
            if (AutoSaveTabState == false)
            {
                StopAutoSavingOnTimer();
                return;
            }

            await UtilityFunctions.RunAsyncSaferAsync(GetCoreWindow(), () =>
            {
                if (_autoSaveTimer != null)
                    _autoSaveTimer.Stop();
            });


            _blockWhileSaving = true;

            await SaveCurrentDocumentChangesAsync();

            _blockWhileSaving = false;

            await UtilityFunctions.RunAsyncSaferAsync(GetCoreWindow(), () =>
            {
                if (AutoSaveTabState && _autoSaveTimer != null)
                    _autoSaveTimer.Start();
            });

            DebugLogHelper.CallEnded();
        }

        /// <summary>
        /// Perform a save operation to the temporary document (not original)
        /// </summary>
        internal async Task SaveCurrentDocumentChangesAsync()
        {
            DebugLogHelper.CallSarted();

            if (SelectedTab != null)
            {
                var doc = SelectedTab.Doc;

                // Save changes if document is modified on the selected tab
                if (doc != null && doc.IsModified())
                {
                    await SelectedTab.DocSemWaitAsync().ConfigureAwait(false);
                    bool locked = false;
                    try
                    {
                        // so we know we can write to the document, if exception is thrown just handle it
                        doc.LockRead();
                        locked = true;

                        doc.UnlockRead();
                        locked = false;

                        // check if modified again since semaphoro could have taken long enoug for something to change
                        if (doc.IsModified())
                        {
                            if ((SelectedTab.MetaData.DocumentState == OpenedDocumentStates.Corrupted
                                || SelectedTab.MetaData.DocumentState == OpenedDocumentStates.CorruptedAndModified)
                                && !SelectedTab.MetaData.IsBrokenDocumentRestored)
                            {
                                await doc.SaveAsync(pdftron.SDF.SDFDocSaveOptions.e_remove_unused).AsTask().ConfigureAwait(false);
                            }
                            else
                            {
                                await doc.SaveAsync().AsTask().ConfigureAwait(false);
                            }
                        }
                    }
                    catch (Exception ex)
                    {
                        AppCenterHandlerBase.CURRENT.TrackError(ex);
                    }
                    finally
                    {
                        if (locked)
                            doc.UnlockRead();

                        SelectedTab.DocSemRelease();
                    }
                }

                // Update info (metadata) regarding selected tab
                try
                {
                    await SemWaitAsync();

                    if (Tabs.Contains(SelectedTab))
                    {
                        int tabIndex = Tabs.IndexOf(SelectedTab);
                        SelectedTab.RefreshState();
                        TabSetting.SaveSettings(tabIndex, SelectedTab.MetaData);
                    }
                }
                catch (Exception ex)
                {
                    AppCenterHandlerBase.CURRENT.TrackError(ex);
                }
                finally
                {
                    SemRelease();
                }
            }

            DebugLogHelper.CallEnded();
        }

        /// <summary>
        /// Get cached CoreWindow
        /// </summary>
        private Windows.UI.Core.CoreWindow GetCoreWindow()
        {
            if (_coreWindow == null)
                _coreWindow = CoreApplication.MainView.CoreWindow;

            return _coreWindow;
        }
    }
}
