﻿using pdftron.PDF.Tools.Controls.Viewer;
using pdftron.PDF.Tools.Controls.ViewModels.FileOpening;
using System;

namespace pdftron.PDF.Tools.Controls.ViewModels.Viewer.Tab
{
    public static class TabSetting
    {
        private static Windows.Storage.ApplicationDataContainer _LocalSettings;
        private static Windows.Storage.ApplicationDataContainer LocalSettings
        {
            get
            {
                if (null == _LocalSettings)
                {
                    _LocalSettings = Windows.Storage.ApplicationData.Current.LocalSettings;
                }
                return _LocalSettings;
            }
        }

        private const string OPEN_TABS_SETTING_NAME = "CompleteReader_Tabs_NumOpen";
        public static int OpenTabs
        {
            get
            {
                if (LocalSettings.Values.ContainsKey(OPEN_TABS_SETTING_NAME))
                {
                    return (int)LocalSettings.Values[OPEN_TABS_SETTING_NAME];
                }
                return 0;
            }
            set
            {
                LocalSettings.Values[OPEN_TABS_SETTING_NAME] = value;
            }
        }

        /// <summary>
        /// Save the tab doc metadata to the app`s LocalSettings container (disk in local app's folder) 
        /// </summary>
        /// <param name="tabIndex">The doc tab index/position</param>
        /// <param name="metaData">The doc tab metadata</param>
        public static void SaveSettings(int tabIndex, PdfViewCtrlTabMetaData metaData)
        {
            string prefix = "CompleteReader_Tabs_" + tabIndex + "_";
            Windows.Foundation.Collections.IPropertySet locals = LocalSettings.Values;
            locals[prefix + "Title"] = metaData.TabTitle;
            locals[prefix + "HS"] = metaData.HScrollPos;
            locals[prefix + "VS"] = metaData.VScrollPos;
            locals[prefix + "Z"] = metaData.Zoom;
            locals[prefix + "Pg"] = metaData.LastPage;
            locals[prefix + "PgR"] = (int)metaData.PageRotation;
            locals[prefix + "PPM"] = (int)metaData.PagePresentationMode;
            locals[prefix + "R"] = (bool)metaData.IsReflow;
            locals[prefix + "TAdded"] = metaData.TabAddedTimeStamp.ToString("yyyy-MM-ddTHH:mm:ss.fffffffzzz");
            locals[prefix + "TLView"] = metaData.TabLastViewedTimeStamp.ToString("yyyy-MM-ddTHH:mm:ss.fffffffzzz");
            locals[prefix + "PW"] = metaData.Password;
            locals[prefix + "TmpToken"] = metaData.FutureAccessListToken;
            locals[prefix + "OriToken"] = metaData.OriginalFileToken;
            locals[prefix + "SourceToken"] = metaData.FileSourceTokenIfNotSaveable;
            locals[prefix + "TmbLoc"] = metaData.ThumbnailLocation;
            locals[prefix + "TmbToDate"] = metaData.IsThumbnailUpToDate;

            locals[prefix + "DocSt"] = (int)metaData.DocumentState;
            locals[prefix + "SvWarn"] = metaData.HasUserBeenWarnedAboutSaving;
            locals[prefix + "MdSinceO"] = metaData.HasDocumentBeenModifiedSinceOpening;
            locals[prefix + "MRU"] = metaData.MostRecentlyUsedToken;
            locals[prefix + "Drag"] = metaData.IsOpenedThroughDrop;

            locals[prefix + "MdSinceS"] = metaData.IsDocumentModifiedSinceLastSave;
            locals[prefix + "BrkRes"] = metaData.IsBrokenDocumentRestored;

            locals[prefix + "ModDate"] = metaData.LastModifiedDate.ToString("yyyy-MM-ddTHH:mm:ss.fffffffzzz");
            locals[prefix + "FlSize"] = metaData.FileSize;
        }

        public static PdfViewCtrlTabMetaData GetFromSettings(int tabIndex)
        {
            PdfViewCtrlTabMetaData data = new PdfViewCtrlTabMetaData();
            string prefix = "CompleteReader_Tabs_" + tabIndex + "_";
            Windows.Foundation.Collections.IPropertySet locals = LocalSettings.Values;

            data.TabTitle = GetSetting<string>(prefix + "Title", string.Empty);
            data.HScrollPos = GetSetting<double>(prefix + "HS", 0);
            data.VScrollPos = GetSetting<double>(prefix + "VS", 0);
            data.Zoom = GetSetting<double>(prefix + "Z", 1);
            data.LastPage = GetSetting<int>(prefix + "Pg", 1);
            data.PageRotation = (PageRotate)GetSetting<int>(prefix + "PgR", (int)PageRotate.e_0);
            data.PagePresentationMode = (PDFViewCtrlPagePresentationMode)GetSetting<int>(prefix + "PPM", (int)ViewerSettings.PagePresentationMode);
            data.IsReflow = (bool)GetSetting<bool>(prefix + "R", false);
            data.IsOpenedThroughDrop = GetSetting<bool>(prefix + "Drag", false);
            bool hasDate = false;
            string dateTimeString = GetSetting<string>(prefix + "TAdded", string.Empty);
            if (!string.IsNullOrWhiteSpace(dateTimeString))
            {
                try
                {
                    data.TabAddedTimeStamp = DateTime.Parse(dateTimeString);
                    hasDate = true;
                }
                catch (Exception) { }
            }
            if (!hasDate)
            {
                data.TabAddedTimeStamp = DateTime.MinValue;
            }

            hasDate = false;
            dateTimeString = GetSetting<string>(prefix + "TLView", string.Empty);
            if (!string.IsNullOrWhiteSpace(dateTimeString))
            {
                try
                {
                    data.TabLastViewedTimeStamp = DateTime.Parse(dateTimeString);
                    hasDate = true;
                }
                catch (Exception) { }
            }
            if (!hasDate)
            {
                data.TabLastViewedTimeStamp = DateTime.MinValue;
            }

            data.Password = GetSetting<string>(prefix + "PW", string.Empty);
            data.FutureAccessListToken = GetSetting<string>(prefix + "TmpToken", string.Empty);
            data.OriginalFileToken = GetSetting<string>(prefix + "OriToken", string.Empty);
            data.FileSourceTokenIfNotSaveable = GetSetting<string>(prefix + "SourceToken", string.Empty);
            data.ThumbnailLocation = GetSetting<string>(prefix + "TmbLoc", string.Empty);
            data.IsThumbnailUpToDate = GetSetting<bool>(prefix + "TmbToDate", false);

            data.DocumentState = (OpenedDocumentStates)GetSetting<int>(prefix + "DocSt", (int)OpenedDocumentStates.Normal);
            data.HasUserBeenWarnedAboutSaving = GetSetting<bool>(prefix + "SvWarn", false);
            data.HasDocumentBeenModifiedSinceOpening = GetSetting<bool>(prefix + "MdSinceO", false);
            data.MostRecentlyUsedToken = GetSetting<string>(prefix + "MRU", string.Empty);

            data.IsDocumentModifiedSinceLastSave = GetSetting<bool>(prefix + "MdSinceS", false);
            data.IsBrokenDocumentRestored = GetSetting<bool>(prefix + "BrkRes", false);

            dateTimeString = GetSetting<string>(prefix + "ModDate", string.Empty);
            hasDate = false;
            if (!string.IsNullOrWhiteSpace(dateTimeString))
            {
                try
                {
                    data.LastModifiedDate = DateTimeOffset.Parse(dateTimeString);
                    hasDate = true;
                }
                catch (Exception) { }
            }
            if (!hasDate)
            {
                data.LastModifiedDate = DateTimeOffset.MinValue;
            }
            data.FileSize = GetSetting<ulong>(prefix + "FlSize", 0);


            return data;
        }

        private static T GetSetting<T>(string settingName, T def)
        {
            if (LocalSettings.Values.ContainsKey(settingName))
            {
                object val = LocalSettings.Values[settingName];
                if (val is T)
                {
                    return (T)val;
                }
            }
            return def;
        }
    }
}
