﻿using pdftron.PDF.Tools.Controls.ViewModels;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Runtime.InteropServices.WindowsRuntime;
using System.Threading.Tasks;
using Windows.Foundation;
using Windows.Foundation.Collections;
using Windows.UI.Core;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Controls.Primitives;
using Windows.UI.Xaml.Data;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Media.Animation;
using Windows.UI.Xaml.Navigation;
using Windows.UI.Xaml.Shapes;

using PDFDouble = pdftron.Common.DoubleRef;

// The User Control item template is documented at http://go.microsoft.com/fwlink/?LinkId=234236

namespace pdftron.PDF.Tools.Controls
{
    public sealed partial class AnnotationList : UserControl
    {
        private NavigationStack _navigationStack;
        private SolidColorBrush _flashBrush;
        private bool _viewIsVisible = true;
        private RestoreProgressStates _currentRestoreProgress;
        private double _restoreScrollPosition = -1; // negative number means don't restore
        private ScrollViewer _listViewScroller;

        public event EventHandler<ItemClickEventArgs> ItemClicked;

        public AnnotationListViewModel ViewModel { get; private set; }

        public AnnotationList() => InitializeComponent();

        public AnnotationList(PDFViewCtrl ctrl)
        {
            InitializeComponent();
            PDFViewCtrl = ctrl;
        }

        public PDFViewCtrl PDFViewCtrl
        {
            get => ViewModel?.PDFViewCtrl;
            set => SetPDFViewCtrl(value);
        }

        public NavigationStack NavigationStack
        {
            get => ViewModel != null ? ViewModel.NavigationStack : _navigationStack;
            set
            {
                _navigationStack = value;
                if (ViewModel != null)
                    ViewModel.NavigationStack = _navigationStack;
            }
        }

        private void SetPDFViewCtrl(PDFViewCtrl ctrl)
        {
            if (ctrl != null)
            {
                ViewModel = new AnnotationListViewModel(ctrl)
                {
                    NavigationStack = _navigationStack
                };

                ViewModel.AnnotationGroups.CollectionChanged += AnnotationGroups_CollectionChanged;
                ViewModel.PropertyChanged += ViewModel_PropertyChanged;
                AnnotationItemsListView.Loaded += AnnotationItemsListView_Loaded;
                AnnotationItemsListView.SelectionChanged += AnnotationItemsListView_SelectionChanged;
                DataContext = ViewModel;
                InitAnimations();
            }
        }

        private void AnnotationItemsListView_SelectionChanged(object sender, SelectionChangedEventArgs e)
        {
            if (AnnotationItemsListView.SelectedItem != null)
                AnnotationItemsListView.ScrollIntoView(AnnotationItemsListView.SelectedItem);
        }

        private void AnnotationList_ItemClick(object sender, ItemClickEventArgs e) => ItemClicked?.Invoke(this, e);

        public SolidColorBrush FlashBrush
        {
            get => _flashBrush;
            set
            {
                _flashBrush = value;
                if (ViewModel != null)
                    ViewModel.FlashBrush = value;
            }
        }

        private void InitAnimations()
        {
            ViewModel.FlashAnimation = FlashTheRectangle;
            ViewModel.FlashBrush = FlashBrush ?? Resources["PrimaryHighlightBrighBrush"] as SolidColorBrush;
        }

        public void RestoreState(double scrollPosition)
        {
            _currentRestoreProgress = RestoreProgressStates.Started;
            if (ViewModel?.IsFinishedLoading == true)
            {
                ScrollRestoreOffsetIntoView(scrollPosition);
            }
            else
            {
                SetLoading();
                _restoreScrollPosition = scrollPosition;
            }
        }

        public double GetState()
        {
            if (_restoreScrollPosition > 0)
                return _restoreScrollPosition;

            ScrollViewer sv = ListViewScroller;
            if (sv != null)
                return sv.VerticalOffset;

            return 0;
        }

        /// <summary>
        /// This boolean can be set to inform the AnnotationList that it can not scroll.
        /// This happens for example when the AnnotationList is in a Pivot and not in view.
        /// Setting this to true will scroll the currently selected item into view
        /// and also affects restoring state.
        /// </summary>
        public bool ViewIsVisible
        {
            get => _viewIsVisible;
            set
            {
                if (_viewIsVisible != value)
                {
                    _viewIsVisible = value;
                    if (_viewIsVisible)
                    {
                        if (ViewModel.SelectedItem != null)
                        {
                            AnnotationItemsListView.ScrollIntoView(ViewModel.SelectedItem);
                            _currentRestoreProgress = RestoreProgressStates.Finished;
                        }
                        else if (ViewModel.ToolManager != null && ViewModel.ToolManager.SelectedAnnotation != null)
                        {
                            ViewModel.ToolManager_SelectionChanged(ViewModel.ToolManager.SelectedAnnotation, ViewModel.ToolManager.SelectedAnnotationPageNumber);
                            _currentRestoreProgress = RestoreProgressStates.Finished;
                        }
                        else if (_currentRestoreProgress == RestoreProgressStates.LoadedButUnableToScroll)
                        {
                            ScrollRestoreOffsetIntoView(_restoreScrollPosition);
                        }
                    }
                }
            }
        }

        private ScrollViewer ListViewScroller => _listViewScroller ?? (_listViewScroller = UtilityFunctions.FindVisualChild<ScrollViewer>(AnnotationItemsListView));

        private async void ScrollRestoreOffsetIntoView(double scrollPosition)
        {
            AnnotationItemsListView.UpdateLayout();
            await Dispatcher.RunAsync(CoreDispatcherPriority.Normal, () =>
            {
                ScrollViewer sv = ListViewScroller;
                if (sv != null)
                {
                    if (sv.ScrollableHeight >= scrollPosition - 1 || (ViewModel?.IsFinishedLoading == true))
                    {
                        if (ViewIsVisible)
                        {
                            _currentRestoreProgress = RestoreProgressStates.Finished;
                            _restoreScrollPosition = -1;
                            SetLoaded();
                        }
                        else
                        {
                            _currentRestoreProgress = RestoreProgressStates.LoadedButUnableToScroll;
                        }

                    }

                    sv.ChangeView(null, scrollPosition, null, true);
                }
            });
        }

        private void SetLoading()
        {
            AnnotationItemsListView.Opacity = 0;
            AnnotationItemsListView.IsHitTestVisible = false;
            LoadingProgressRing.Visibility = Visibility.Visible;
            LoadingProgressRing.IsActive = true;
        }

        private void SetLoaded()
        {
            AnnotationItemsListView.Opacity = 1;
            AnnotationItemsListView.IsHitTestVisible = true;
            LoadingProgressRing.Visibility = Visibility.Collapsed;
            LoadingProgressRing.IsActive = false;
        }


        private void ViewModel_PropertyChanged(object sender, System.ComponentModel.PropertyChangedEventArgs e)
        {
            if (e.PropertyName.Equals("IsFinishedLoading", StringComparison.OrdinalIgnoreCase) && ViewModel.IsFinishedLoading)
                DelaySetLoaded();
        }

        private async void DelaySetLoaded()
        {
            int attempts = 5;
            while (attempts-- > 0 && _restoreScrollPosition > 0)
            {
                if (_restoreScrollPosition > 0)
                    ScrollRestoreOffsetIntoView(_restoreScrollPosition);

                if (attempts > 0)
                    await Task.Delay(100);
            }

            SetLoaded();
        }

        private void AnnotationItemsListView_Loaded(object sender, RoutedEventArgs e)
        {
            if (_restoreScrollPosition > 0)
                ScrollRestoreOffsetIntoView(_restoreScrollPosition);
        }

        private void AnnotationGroups_CollectionChanged(object sender, System.Collections.Specialized.NotifyCollectionChangedEventArgs e)
        {
            if (_restoreScrollPosition > 0)
                ScrollRestoreOffsetIntoView(_restoreScrollPosition);
        }

        public async Task WaitForAnnotationListToFinish()
        {
            if (ViewModel != null)
            {
                ViewModel.ClearViewModel();
                await ViewModel.WaitForAnnotationListToFinish();
            }
        }

        /// <summary>
        /// Will go back if possible.
        /// </summary>
        /// <returns>Return true if it went back, false otherwise</returns>
        public bool GoBack()
        {
            if (ViewModel != null)
                return ViewModel.GoBack();

            return false;
        }
    }
}
