﻿using Microsoft.UI.Xaml.Controls;
using pdftron.PDF.Tools.Controls.ViewModels.Viewer.Tab;
using System;
using System.Collections.Generic;
using Windows.ApplicationModel.Core;
using Windows.ApplicationModel.DataTransfer;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Media.Animation;
using ProgressRing = Microsoft.UI.Xaml.Controls.ProgressRing;

// The User Control item template is documented at http://go.microsoft.com/fwlink/?LinkId=234236

namespace pdftron.PDF.Tools.Controls.Viewer.Dialogs
{
    public sealed partial class PDFViewCtrlTabButtonControl : UserControl
    {
        TabControlViewModel ViewModel { get; set; }

        public PDFViewCtrlTabButtonControl()
        {
            this.InitializeComponent();

            this.DataContextChanged += PDFViewCtrlTabButtonControl_DataContextChanged;
            CopyPathButton.Click += CopyPathButton_Click;
        }

        private void PDFViewCtrlTabButtonControl_DataContextChanged(FrameworkElement sender, DataContextChangedEventArgs args)
        {
            ViewModel = TabViewControl.DataContext as TabControlViewModel;
        }

        private void CopyPathButton_Click(object sender, RoutedEventArgs e)
        {
            DataPackage dataPackage = new DataPackage
            {
                RequestedOperation = DataPackageOperation.Copy
            };

            dataPackage.SetText(FlyoutPathTextBlock.Text);
            Clipboard.SetContent(dataPackage);
            FadeInAndOutCopyToClipboard.Begin();
        }

        private Storyboard GetFadeInAnimation(FrameworkElement target)
        {
            Storyboard sb = new Storyboard();
            string ID = target.Name;
            TimeSpan delayTime = TimeSpan.FromSeconds(0.5);
            TimeSpan fadeTime = TimeSpan.FromSeconds(0.2);
            DoubleAnimationUsingKeyFrames fade = new DoubleAnimationUsingKeyFrames
            {
                Duration = new Duration(delayTime + fadeTime),
                RepeatBehavior = new RepeatBehavior { Count = 1 },
                //EasingFunction = new QuadraticEase() { EasingMode = EasingMode.EaseOut },
            };

            DiscreteDoubleKeyFrame startFrame = new DiscreteDoubleKeyFrame
            {
                KeyTime = KeyTime.FromTimeSpan(TimeSpan.FromSeconds(0)),
                Value = 0.0
            };

            fade.KeyFrames.Add(startFrame);

            DiscreteDoubleKeyFrame fadeStartFrame = new DiscreteDoubleKeyFrame
            {
                KeyTime = KeyTime.FromTimeSpan(delayTime),
                Value = 0.0
            };

            fade.KeyFrames.Add(fadeStartFrame);

            EasingDoubleKeyFrame fadeFrame = new EasingDoubleKeyFrame
            {
                KeyTime = KeyTime.FromTimeSpan(delayTime + fadeTime),
                Value = 1.0
            };

            fadeFrame.EasingFunction = new QuadraticEase
            {
                EasingMode = EasingMode.EaseIn
            };

            fade.KeyFrames.Add(fadeFrame);

            target.Resources.Add(ID + "FadInAnimation", sb);
            Storyboard.SetTarget(fade, target);
            Storyboard.SetTargetName(fade, target.Name);
            Storyboard.SetTargetProperty(fade, "(UIElement.Opacity)");
            sb.Children.Add(fade);

            return sb;
        }

        private Dictionary<ProgressRing, Storyboard> _AnimationsPlaying = new Dictionary<ProgressRing, Storyboard>();

        private void ProgressRing_Loaded(object sender, RoutedEventArgs e)
        {
            ProgressRing ringy = sender as ProgressRing;
            if (ringy.Visibility == Visibility.Visible && !_AnimationsPlaying.ContainsKey(ringy))
            {
                Storyboard sb = GetFadeInAnimation(ringy);
                _AnimationsPlaying[ringy] = sb;
                sb.Begin();
            }
        }

        private void TabViewItemRightTapped(object sender, RightTappedRoutedEventArgs e)
        {
            if(ViewModel != null)
            {
                if(ViewModel.ShowTabFileInfo == false)
                {
                    e.Handled = true;
                    return;             //Ignore showing the flyout on right clicking
                }
            }

            if (sender is FrameworkElement fe && fe.DataContext is PdfViewCtrlTabInfo selectedTab)
            {
                Windows.Foundation.Point pos = e.GetPosition(this);
                MarginTextBlock.Margin = new Thickness(pos.X, pos.Y, 0, 0);
                FlyoutFileInfo.ShowAt(MarginTextBlock);

                FlyoutTitleTextBlock.Text = selectedTab.MetaData.TabTitle;
                FlyoutPathTextBlock.Text = selectedTab.OriginalFile.Path;
                e.Handled = true;
            }
        }

        /// <summary>
        /// Helper event to ensure the selected tab gets updated and selected after a modal event, which disables
        /// the control
        /// </summary>
        private void TabViewControl_IsEnabledChanged(object sender, DependencyPropertyChangedEventArgs e)
        {            
            if (ViewModel == null)
                return;

            var val_set = (bool)e.NewValue;

            if (val_set)
            {
                TabViewControl.SelectedIndex = ViewModel.VisibleTabIndex;
            }
            else
            {
                TabViewControl.SelectedIndex = -1; // Set to -1 when disabled, so once enabled it can trigger a property change
            }
        }

        /// <summary>
        /// Once the drag of a tab has finished we will ensure the dragged tab is selected as the activated one
        /// </summary>
        private async void TabViewControl_TabDragCompleted(TabView sender, TabViewTabDragCompletedEventArgs args)
        {
            var tabInfo = args.Item as PdfViewCtrlTabInfo;

            if (tabInfo != null)
            {
                TabViewControl.SelectedIndex = -1; // Note: Due some kind of sync issue we are setting it to -1 and then InfoTabSelected will update thorugh binding to ensure property change value
                await ViewModel.SelectTabAsync(tabInfo);
            }

            // Trying to ensure the TabView gets updating when opening new tabs after reordering
            sender.UpdateLayout();
        }

        /// <summary>
        /// Once the tab is selected and starts getting dragged we will ensure it is not selected anymore until the drag and drop process finishes
        /// </summary>
        private void TabViewControl_TabDragStarting(TabView sender, TabViewTabDragStartingEventArgs args)
        {
            var tabInfo = args.Item as PdfViewCtrlTabInfo;
            if (tabInfo != null)
                tabInfo.IsSelected = false;
        }

        private async void DocumentTabViewItem_Tapped(object sender, TappedRoutedEventArgs e)
        {
            var tabViewItem = sender as TabViewItem;
            if (tabViewItem == null)
                return;

            var tabInfo = tabViewItem.DataContext as PdfViewCtrlTabInfo;
            if (tabInfo == null)
                return;

            await ViewModel.SelectTabAsync(tabInfo);
        }

        private void DocumentTabViewItem_CloseRequested(TabViewItem sender, TabViewTabCloseRequestedEventArgs args)
        {
            var tabInfo = sender.DataContext as PdfViewCtrlTabInfo;
            if (tabInfo == null)
                return;

            tabInfo.RaiseTabClosedEvent();
        }
    }
}
