﻿using System;
using Windows.UI.Input;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Shapes;
using Windows.UI.Xaml.Media;
using Windows.UI;

using UIPoint = Windows.Foundation.Point;

using PDFDouble = pdftron.Common.DoubleRef;
using PDFRect = pdftron.PDF.Rect;
using PDFPoint = pdftron.PDF.Point;
using Windows.UI.Xaml.Input;

namespace pdftron.PDF.Tools
{
    /*
     * Marquee Zoom
     * 
     * This tool allows to Zoom into specifc regions on the PDF file by
     * drawing a rectangle on the floating canvas of the PDFViewerCtrl
     */

    public class MarqueeZoom : Tool
    {
        Rectangle mRectangle; // REctangle to be drawn on the PDFViewerCtrl floating canvas

        protected bool mIsDragging = false;
        protected bool mStartMoving = false;
        protected bool mStartDrawing = false;

        protected Canvas mViewerCanvas = null;
        protected int mPointerID;
        protected Point m_rectPt1 = null;
        protected Point m_rectPt2 = null;
        protected Point m_viewerPt1 = null;
        protected Point m_viewerPt2 = null;
        protected Rect mPageCropOnClient;
        protected CompositeTransform mRenderTransform;
        protected bool mMirrorX = false;
        protected bool mMirrorY = false;

        protected const int START_DRAWING_THRESHOLD = 50;
              
        public MarqueeZoom(PDFViewCtrl ctrl, ToolManager tMan) : base(ctrl, tMan)
        {
            mToolMode = ToolType.e_zoom_marquee;
            mNextToolMode = ToolType.e_zoom_marquee;

            // Initialize the rectangle to be drawn
            mRectangle = new Rectangle() 
            {
                Fill = new SolidColorBrush(Colors.Red),
                Opacity = 0.25,
                Stroke = new SolidColorBrush(Colors.Red),
                StrokeDashOffset = 1.5,
                StrokeDashCap = PenLineCap.Triangle,
                StrokeThickness = 3,
                StrokeDashArray = new DoubleCollection { 4, 2}
            };

            // Get the PDFView temporary canvas so we can use it for drawing our rectangle
            mViewerCanvas = ctrl.GetAnnotationCanvas();
        }

        #region Interaction Events

        internal override bool PointerPressedHandler(object sender, Windows.UI.Xaml.Input.PointerRoutedEventArgs e)
        {
            mIsDragging = true;
            mStartMoving = false; // reset whether we're moving or not.

            DisableScrolling();

            mPointerID = (int)e.Pointer.PointerId;
            
            // Get the current point in referece to the PDFViewer - This will be used for Zooming
            UIPoint pp_viewer = e.GetCurrentPoint(mPDFView).Position;
            m_viewerPt1 = new Point(pp_viewer.X, pp_viewer.Y);
            m_viewerPt2 = new Point(m_viewerPt1.x, m_viewerPt1.y);

            // Ensure PDFView has a document loaded
            PDFDoc doc = mPDFView.GetDoc();
            if (doc == null)
                return false;

            // Update the Page number so it can be used for manipulation utilities in Tools
            var PageNumber = mPDFView.GetPageNumberFromScreenPoint(pp_viewer.X, pp_viewer.Y);
            mPage = doc.GetPage(PageNumber);

            // Update current displayed canvas
            mViewerCanvas = mPDFView.GetAnnotationCanvas();

            // Get the origin clicked point so we can use it to start drawing the rectangle
            UIPoint pp = e.GetCurrentPoint(mViewerCanvas).Position;
            m_rectPt1 = new Point(pp.X, pp.Y);
            m_rectPt2 = new Point(m_rectPt1.x, m_rectPt1.y);

            // Get the page's bounding box in canvas space so we can limit it to the bounding box
            var mDownPageNumber = mPDFView.GetCurrentPage();
            mPageCropOnClient = BuildPageBoundBoxOnClient(mDownPageNumber);
            if (mPageCropOnClient != null)
            {
                if (m_rectPt1.x < mPageCropOnClient.x1)
                {
                    m_rectPt1.x = mPageCropOnClient.x1;
                }
                if (m_rectPt1.x > mPageCropOnClient.x2)
                {
                    m_rectPt1.x = mPageCropOnClient.x2;
                }
                if (m_rectPt1.y < mPageCropOnClient.y1)
                {
                    m_rectPt1.y = mPageCropOnClient.y1;
                }
                if (m_rectPt1.y > mPageCropOnClient.y2)
                {
                    m_rectPt1.y = mPageCropOnClient.y2;
                }
            }

            return base.PointerReleasedHandler(sender, e);
        }

        internal override bool PointerMovedHandler(object sender, Windows.UI.Xaml.Input.PointerRoutedEventArgs e)
        {
            if (mIsDragging)
            {
                // Check if this is the same click
                if (e.Pointer.PointerId != mPointerID)                
                    return false;
                
                // Update current click position on the PDFView canvas - This will be used to draw the rectangle guide
                UIPoint pp = e.GetCurrentPoint(mViewerCanvas).Position;
                m_rectPt2.x = pp.X;
                m_rectPt2.y = pp.Y;

                CheckAndLimitBounds(ref m_rectPt2);

                // We don't want to start drawing until we have a bit of a distance
                if (!mStartDrawing)
                {
                    // wait until you're at least START_DRAWING_THRESHOLD distance from the start
                    if ((Math.Abs(m_rectPt2.x - m_rectPt1.x) > START_DRAWING_THRESHOLD)
                        || (Math.Abs(m_rectPt2.y - m_rectPt1.y) > START_DRAWING_THRESHOLD))
                    {
                        mStartDrawing = true;

                        mViewerCanvas.Children.Add(mRectangle);
                        mRectangle.SetValue(Canvas.LeftProperty, m_rectPt1.x);
                        mRectangle.SetValue(Canvas.TopProperty, m_rectPt1.y);
                        mRenderTransform = new CompositeTransform();
                        mRectangle.RenderTransform = mRenderTransform;
                        mMirrorX = false;
                        mMirrorY = false;
                    }
                }

                // Size the canvas according to the two points
                if (mStartDrawing)
                {
                    mRectangle.Width = Math.Abs(m_rectPt2.x - m_rectPt1.x);
                    mRectangle.Height = Math.Abs(m_rectPt2.y - m_rectPt1.y);
                    if (m_rectPt2.x < m_rectPt1.x)
                    {
                        mMirrorX = true;
                        mRenderTransform.ScaleX = -1;
                    }
                    else
                    {
                        mMirrorX = false;
                        mRenderTransform.ScaleX = 1;
                    }
                    if (m_rectPt2.y < m_rectPt1.y)
                    {
                        mMirrorY = true;
                        mRenderTransform.ScaleY = -1;
                    }
                    else
                    {
                        mMirrorY = false;
                        mRenderTransform.ScaleY = 1;
                    }
                }
            }

            return base.PointerReleasedHandler(sender, e);
        }

        internal override bool PointerReleasedHandler(object sender, PointerRoutedEventArgs e)
        {
            // Clear any drawing on the canvas
            mViewerCanvas.Children.Clear();
            mIsDragging = false;

            // Ensure we actually started drawing
            if (mStartDrawing)
            {                
                mStartDrawing = false;

                // Get the final PDFView point for calculating zooming
                UIPoint pp_viewer = e.GetCurrentPoint(mPDFView).Position;
                m_viewerPt2.x = pp_viewer.X;
                m_viewerPt2.y = pp_viewer.Y;

                PDFPoint tempPoint1 = m_viewerPt1;
                PDFPoint tempPoint2 = m_viewerPt2;

                double zoom_factor_w;
                double zoom_factor_h;
                double zoom_x;
                double zoom_y;

                // Check X direction for the drawing rectangle and swap point if necessary
                if (m_viewerPt1.x > m_viewerPt2.x)
                {
                    tempPoint1 = m_viewerPt2;
                    tempPoint2 = m_viewerPt1;
                }

                // Check if dragging in Y is top to bottom else bottom to top
                if (tempPoint2.y > tempPoint1.y)
                {
                    var zoom_w = tempPoint2.x - tempPoint1.x;
                    var zoom_h = tempPoint2.y - tempPoint1.y;

                    zoom_x = tempPoint1.x + ((tempPoint2.x - tempPoint1.x) / 2);
                    zoom_y = tempPoint1.y + ((tempPoint2.y - tempPoint1.y) / 2);

                    zoom_factor_w = mPDFView.ActualWidth / zoom_w;
                    zoom_factor_h = mPDFView.ActualHeight / zoom_h;
                }
                else
                {
                    var zoom_w = tempPoint2.x - tempPoint1.x;
                    var zoom_h = tempPoint1.y - tempPoint2.y;

                    zoom_x = tempPoint1.x + ((tempPoint2.x - tempPoint1.x) / 2);
                    zoom_y = tempPoint2.y + ((tempPoint1.y - tempPoint2.y) / 2);

                    zoom_factor_w = mPDFView.ActualWidth / zoom_w;
                    zoom_factor_h = mPDFView.ActualHeight / zoom_h;
                }

                var zoom_factor = zoom_factor_h > zoom_factor_w ? zoom_factor_w : zoom_factor_h; // We will use the smaller zoom to ensure to fit on the page
                double currentZoom = mPDFView.GetZoom(); // Get current zoom

                mPDFView.SetZoom((int)zoom_x, (int)zoom_y, currentZoom * zoom_factor);
            }
                        
            EnableScrolling();
            EndCurrentTool(ToolType.e_pan);

            return base.PointerReleasedHandler(sender, e);
        }

        #endregion

        #region Utilities

        /// <summary>
        /// Ensure we're within our document bounding box
        /// </summary>
        /// <param name="point"></param>
        private bool CheckAndLimitBounds(ref PDFPoint point)
        {
            bool outBounds = false;

            // want to make sure we're within our bounding box.
            if (mPageCropOnClient != null)
            {
                if (point.x < mPageCropOnClient.x1)
                {
                    point.x = mPageCropOnClient.x1;
                    outBounds = true;
                }
                if (point.x > mPageCropOnClient.x2)
                {
                    point.x = mPageCropOnClient.x2;
                    outBounds = true;
                }
                if (point.y < mPageCropOnClient.y1)
                {
                    point.y = mPageCropOnClient.y1;
                    outBounds = true;
                }
                if (point.y > mPageCropOnClient.y2)
                {
                    point.y = mPageCropOnClient.y2;
                    outBounds = true;
                }
            }

            return outBounds;
        }

        #endregion
    }
}