﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

using Windows.UI.Xaml;
using Windows.UI.Xaml.Input;

using pdftron.PDF;
using pdftron.PDF.Tools.Utilities;
using Windows.UI.Xaml.Controls;
using pdftron.PDF.Tools.Controls;
using static pdftron.PDF.Tools.UtilityFunctions;


namespace pdftron.PDF.Tools
{
    /// <summary>
    /// The various types of tools available for use.
    /// </summary>
    public enum ToolType
    {
        e_none = 0,
        e_pan = 1,
        e_annot_edit = 2,
        e_line_create = 3,
        e_arrow_create = 4,
        e_rect_create = 5,
        e_oval_create = 6,
        e_ink_create = 7,
        e_text_annot_create = 8,
        e_link_action = 9,
        e_text_select = 10,
        e_form_fill = 11,
        e_sticky_note_create = 12,
        e_text_highlight = 13,
        e_text_underline = 14,
        e_text_strikeout = 15,
        e_text_squiggly = 16,
        e_line_edit = 17,
        e_rich_media = 18,
        e_advanced_shape_edit = 19,
        e_signature = 20,
        e_ink_eraser = 21,
        e_annot_edit_text_markup = 22,
        e_polygon_create = 23,
        e_polyline_create = 24,
        e_stamper = 25,
        e_ink_marker = 26,
        e_cloud_create = 27,
        e_annot_edit_rect_group = 28,
        e_annot_edit_group_create = 29,
        e_ruler_create = 30,
        e_perimeter_create = 31,
        e_area_create = 32,
        e_rect_redaction_create = 33,
        e_text_redaction = 34,
        e_lasso_create = 35,
        e_zoom_marquee = 99,
        e_magnifying_rectangle = 100,
    };

    public delegate void OnFileChanged(Windows.Storage.StorageFile file);
    public delegate void NewToolCreatedDelegate(ToolType toolType);
    public delegate void UndoRedoStatusChangedDelegate(UndoRedoStatusEventArgs eventArgs);

    /// <summary>
    /// Describes in what context the ToolManager's SingleTap event is fired.
    /// </summary>
    public enum SingleTapEvents
    {
        BeforeTool = 0, /// <summary>This happens before the tools handle the event</summary>
        SingleTapConfirmed, /// <summary>We know it's not a double tap</summary>
    }

    public delegate void SingleTapDelegate(SingleTapEvents eventType, TappedRoutedEventArgs args);

    public class InterceptAnnotationEventArgs : EventArgs
    {
        private bool handled;
        private IAnnot annotation;
        private int pageNumber;

        public InterceptAnnotationEventArgs(IAnnot annotation, int pageNumber)
        {
            this.annotation = annotation;
            this.pageNumber = pageNumber;

            this.handled = false;
        }

        public bool Handled
        {
            get { return this.handled; }
            set { this.handled = value; }
        }

        public IAnnot Annotation
        {
            get { return this.annotation; }
        }

        public int PageNumber
        {
            get { return this.pageNumber; }
        }
    }

    /// <summary>
    /// The ToolManager class will attach itself to a PDFViewCtrl and provides a lot of interactive features,
    /// such as text selection, annotation creation and editing, and form filling.
    ///
    /// It is important to Dispose the ToolManager when it is no longer needed. This is because it subscribes
    /// to certain events in the PDFViewCtrl that will prevent both the PDFViewCtrl and the ToolManager from
    /// being reclaimed by garbage collection.
    /// </summary>
    public sealed class ToolManager : IDisposable
    {
        private bool _IsDisposed = false;
        private PointerEventHandler _PointerReleasedHandler;
        private KeyEventHandler _KeyDownHandler;

        internal static string SignatureAnnotationIdentifyingString = "pdftronSignatureStamp";
        internal static string StamperAnnotationIdentifyingString = "pdftronImageStamp";
        internal static string StamperRotationAnnotationIdentifyingString = "pdftronImageStampRotation";

        internal static string AnnotationAuthorHasBeenAskedSettingsString = "PDFViewCtrlToolsAnnotationAuthorHasBeenAsked";
        public static string DebugAnnotationAuthorHasBeenAskedSettingsString { get { return AnnotationAuthorHasBeenAskedSettingsString; } }

        private static string _AnnotationAuthorNameSettingsString = "PDFViewCtrlToolsAnnotationAuthorName";
        /// <summary>
        /// The name under which the tools store the author name in the ApplicationData.Current.RoamingSettings container.
        /// </summary>
        public static string AnnotationAuthorNameSettingsString { get {return _AnnotationAuthorNameSettingsString; } }

        private Tool _CurrentTool;
        /// <summary>
        /// Gets the currently active tool.
        /// </summary>
        public Tool CurrentTool
        {
            get
            {
                return _CurrentTool;
            }
        }

        /// <summary>
        /// Get the current selected tool type
        /// </summary>
        public ToolType CurrentToolType
        {
            get 
            {
                if (CurrentTool != null)
                    return CurrentTool.ToolMode;

                return ToolType.e_none;
            }            
        }

        private FrameworkElement _PopupLimitingTarget = null;
        /// <summary>
        /// Gets or sets a FrameworkElement that will be used by the PDFViewCtrl tools as the acceptable area on which to target
        /// </summary>
        public FrameworkElement PopupLimitingTarget
        {
            get { return _PopupLimitingTarget; }
            set { _PopupLimitingTarget = value; }
        }

        internal FrameworkElement PopupTargetInternal
        {
            get
            {
                if (PopupLimitingTarget != null)
                {
                    return PopupLimitingTarget;
                }
                return mPDFView;
            }
        }

        private bool _EnablePopupMenuOnLongPress = false;
        /// <summary>
        /// Gets or sets whether or not the tools will show a popup menu
        /// when the user long presses with their finger. This will only happen while using the pan tool.
        /// </summary>
        public bool EnablePopupMenuOnLongPress
        {
            get
            {
                return _EnablePopupMenuOnLongPress;
            }
            set
            {
                _EnablePopupMenuOnLongPress = value;
            }
        }


        private bool _EnablePopupMenu = true;
        /// <summary>
        /// Gets or sets whether or not the tools will use the popup menu.
        /// If set to false, it is up to users to modify the tools so that they can
        /// be used with, for example, the app bar. Defaults to true.
        /// </summary>
        public bool IsPopupMenuEnabled
        {
            get
            {
                return _EnablePopupMenu;
            }
            set
            {
                _EnablePopupMenu = value;
            }
        }

        private bool _UseSmallPageNumberIndicator = true;
        /// <summary>
        /// This tells the Tools to use a smaller page number indicator.
        /// If true, the page number indicator will be smaller and will only contain the current page number,
        /// but not the total number of pages.
        /// </summary>
        public bool UseSmallPageNumberIndicator
        {
            get
            {
                return _UseSmallPageNumberIndicator;
            }
            set
            {
                if (value)
                {
                    mPageNumberIndicator.Shrink();
                }
                else
                {
                    mPageNumberIndicator.Grow();
                }
                _UseSmallPageNumberIndicator = value;
            }
        }

        internal PageNumberIndicator mPageNumberIndicator;
        /// <summary>
        /// This is a small UI widget that displays a the current page number.
        /// </summary>
        public PageNumberIndicator PageNumberIndicator
        {
            get
            {
                return mPageNumberIndicator;
            }
        }


        private bool _TextMarkupAdobeHack = true;
        /// <summary>
        /// Gets or sets whether the TextMarkup annotations are compatible with Adobe
        /// (Adobe's quads don't follow the specification, but they don't handle quads that do).
        /// </summary>
        public bool TextMarkupAdobeHack {
            get { return _TextMarkupAdobeHack; }
            set { _TextMarkupAdobeHack = value; }
        }


        /// <summary>
        /// Determines the text selection behaviour of the pan tool
        /// </summary>
        public enum TextSelectionBehaviour
        {
            /// <summary>
            /// Clicking and dragging always pans the control
            /// </summary>
            AlwaysPan,
            /// <summary>
            /// Clicking and dragging always selects text
            /// </summary>
            AlwaysSelect,
            /// <summary>
            /// Will see if there is any text at the click point, if there is, it will select, otherwise it will pan
            /// </summary>
            Mixed,
        }

        private TextSelectionBehaviour _PanToolTextSelectionMode = TextSelectionBehaviour.Mixed;
        /// <summary>
        /// Set whether click and drag actions should select text or pan the view.
        /// </summary>
        public TextSelectionBehaviour PanToolTextSelectionMode
        {
            get { return _PanToolTextSelectionMode; }
            set { _PanToolTextSelectionMode = value; }
        }

        public enum AllowedLinks
        {
            /// <summary>
            /// We will activate on both GoTo and URL type links.
            /// </summary>
            All,
            /// <summary>
            /// Only activate links that navigate within the document.
            /// </summary>
            GoTo,
            /// <summary>
            /// Only activate links that navigate to URLs
            /// </summary>
            URL,
            /// <summary>
            /// Don't allow any links to be used.
            /// </summary>
            None,
        }

        private AllowedLinks _AllowedLinkActions = AllowedLinks.All;
        /// <summary>
        /// Gets or sets what types of actions the tool will activate. Defaults to All
        /// </summary>
        public AllowedLinks AllowedLinkActions
        {
            get { return _AllowedLinkActions; }
            set { _AllowedLinkActions = value; }
        }

        private bool _CopyAnnotatedTextToNote = false;
        /// <summary>
        /// Gets or sets whether annotating text will autimatically copy the contents into the
        /// text markup's note.
        /// </summary>
        public bool CopyAnnotatedTextToNote
        {
            get { return _CopyAnnotatedTextToNote; }
            set { _CopyAnnotatedTextToNote = value; }
        }

        /// <summary>
        /// Determines when ink annotations should be smoothed
        /// </summary>
        public enum InkSmoothingOptions
        {
            AllButStylus = 0,
            NoSmoothing,
            SmoothAll,
        }

        private InkSmoothingOptions _InkSmoothingBehaviour = InkSmoothingOptions.AllButStylus;
        /// <summary>
        /// Gets or sets when ink should be smoothed when finished.
        /// </summary>
        public InkSmoothingOptions InkSmoothingBehaviour
        {
            get { return _InkSmoothingBehaviour; }
            set { _InkSmoothingBehaviour = value; }
        }

        private bool _StylusAsPen = true;
        /// <summary>
        /// Gets or sets whether the stylus should pan or ink as the default state
        /// </summary>
        public bool StylusAsPen
        {
            get { return _StylusAsPen; }
            set { _StylusAsPen = value; }
        }

        /// <summary>
        /// True if there currently is text selected as a result of the Tools.
        /// </summary>
        public bool HasSelectedText
        {
            get
            {
                return (CurrentTool.ToolMode == ToolType.e_text_select && mPDFView.HasSelection());
            }
        }

        public string GetSelectedText()
        {
            string text = "";

            // Extract selected text
            int selectionStartPage = mPDFView.GetSelectionBeginPage();
            int selectionEndPage = mPDFView.GetSelectionEndPage();
            for (int pgnm = selectionStartPage; pgnm <= selectionEndPage; pgnm++)
            {
                if (mPDFView.HasSelectionOnPage(pgnm))
                {
                    text += mPDFView.GetSelection(pgnm).GetAsUnicode();
                }
            }

            return text;
        }

        /// <summary>
        /// Makes the ToolManager unselect all annotations.
        /// </summary>
        public void UnselectAllAnnotations()
        {
            if (CurrentTool != null &&
                (CurrentTool.ToolMode == ToolType.e_annot_edit || CurrentTool.ToolMode == ToolType.e_annot_edit_text_markup))
            {
                CreateDefaultTool();
            }
        }

        /// <summary>
        /// Gets the page number of the currently selected annotation.
        /// If no annotation is selected, this number is 0.
        /// </summary>
        public int SelectedAnnotationPageNumber
        {
            get
            {
                if (CurrentTool != null)
                {
                    if (CurrentTool.ToolMode == ToolType.e_annot_edit || CurrentTool.ToolMode == ToolType.e_annot_edit_text_markup)
                    {
                        return CurrentTool.CurrentAnnotPageNumber;
                    }
                }
                return 0;
            }
        }

        /// <summary>
        /// Gets the currently selected annotation.
        /// If no annotation is selected, this is null.
        /// </summary>
        public IAnnot SelectedAnnotation
        {
            get
            {
                if (CurrentTool != null)
                {
                    if (CurrentTool.ToolMode == ToolType.e_annot_edit 
                        || CurrentTool.ToolMode == ToolType.e_annot_edit_text_markup
                        || CurrentTool.ToolMode == ToolType.e_line_edit
                        || CurrentTool.ToolMode == ToolType.e_advanced_shape_edit)
                    {
                        return CurrentTool.CurrentAnnot;
                    }
                }
                return null;
            }
        }

        private bool _ReadOnlyMode = false;
        /// <summary>
        /// Gets or sets whether read-only mode should be enabled.
        /// In read only mode, annotations cannot be added, edited, or deleted.
        /// </summary>
        public bool ReadOnlyMode
        {
            get { return _ReadOnlyMode; }
            set { _ReadOnlyMode = value; }
        }


        /// <summary>
        /// This event is raised when the tools perform a save-as option. This happens when the users signs a
        /// form. This event will provide the file the user chose to save the doc in, and should in the future
        /// be used by the app when the user wants to save the document.
        /// </summary>
        public event OnFileChanged FileHasChanged;

        internal void RaiseFileChangedEvent(Windows.Storage.StorageFile file)
        {
            if (FileHasChanged != null)
            {
                FileHasChanged(file);
            }
        }

        public delegate void TextToSpeechDelegate(MediaElement media);

        public event TextToSpeechDelegate TextToSpeechActivated;
        internal void RaiseTextToSpeechActivatedEvent(MediaElement media)
        {
            if (TextToSpeechActivated != null)
            {
                TextToSpeechActivated(media);
            }
        }

        /// <summary>
        /// This event is raised when the ToolManager receives tap events.
        /// Before the tools handle the event, BeforeTool is raised.
        /// Once a single tap is confirmed (i.e. not part of a double tap)
        /// SingleTapConfirmed is raised
        /// </summary>
        public event SingleTapDelegate SingleTap;

        /// <summary>
        /// This even is raised every time a new tool is created.
        /// The type of the new tools is passed as a parameter.
        /// </summary>
        public event NewToolCreatedDelegate NewToolCreated;

        /// <summary>
        /// This event is raised every time undo and redo manager has
        /// made a change
        /// </summary>
        public event UndoRedoStatusChangedDelegate UndoRedoStatusChanged;

        /// <summary>
        /// Listen to Undo and Redo Manager for status change and trigger ToolManager event change
        /// </summary>
        private void UndoRedoAction_OnUndoRedoStatusChanged(object sender, UndoRedoStatusEventArgs e)
        {
            // Just bubble up the event args
            UndoRedoStatusChanged?.Invoke(e);
        }

        /// <summary>
        /// Create an instance of IAuthorDialog and set it here to overwrite the default the tools are using.
        /// </summary>
        public Utilities.IAuthorDialog AuthorDialog { get; set; }

        /// <summary>
        /// Gets or sets whether an author should be added. Will open the author dialog if none is entered yet.
        /// </summary>
        public bool AddAuthorToAnnotations { get; set; }

        /// <summary>
        /// Gets or sets whether or not the author dialog should pop up the first time the user adds an annotations.
        /// </summary>
        public bool PopupAuthorDialogFirstTime { get; set; }

        internal List<DelayRemoveTimer> mDelayRemoveTimers;
        internal NoteDialogBase mCurrentNoteDialog;
        internal FixedSizedBoxPopup mCurrentBoxPopup;
        internal Controls.ViewModels.AuthorDialogViewModel mOpenAuthorDialogViewModel;

        private bool _IsEnabled = true;
        /// <summary>
        /// Gets or sets whether or not the ToolManager will react to events from the PDFViewCtrl.
        /// </summary>
        public bool IsEnabled
        {
            get { return _IsEnabled; }
            set { _IsEnabled = value; }
        }

        private bool _IsSignatureDialogFullScreen = true;
        /// <summary>
        /// Gets or Sets whether the Signature Dialog should be displayed on Full Screen
        /// or reduced size
        /// </summary>
        public bool IsSignatureDialogFullScreen
        {
            get { return _IsSignatureDialogFullScreen; }
            set { _IsSignatureDialogFullScreen = value; }
        }

        private bool _IsSoundAnnotEnabled = true;

        /// <summary>
        /// Gets or Sets whether the Sound Annotation creation is enabled or not
        /// </summary>
        public bool IsSoundAnnotEnabled
        {
            get { return _IsSoundAnnotEnabled; }
            set { _IsSoundAnnotEnabled = value; }
        }

        private bool _showDefaultSignature = true;
        /// <summary>
        /// Gets or Sets Whether to allow displaying a default (previously saved) signature
        /// </summary>
        public bool ShowDefaultSignature
        {
            get { return _showDefaultSignature; }
            set { _showDefaultSignature = value; }
        }

        // text selection scroll speeds when dragging to edge.
        // How far should you have to move the mouse before we can start scrolling
        public double MOVE_THRESHOLD_TO_START_SCROLL = 25;
        // the speed with which we scroll
        public double TEXT_SELECT_SCROLL_SPEED_X = 20;
        public double TEXT_SELECT_SCROLL_SPEED_Y = 50;
        // speed increases linearly if within this margin
        public double TEXT_SELECT_SCROLL_SOFT_MARGIN_X = 35;
        public double TEXT_SELECT_SCROLL_SOFT_MARGIN_Y = 35;
        // once we're outside this margin, we go full speed
        public double TEXT_SELECT_SCROLL_HARD_MARGIN_X = 10;
        public double TEXT_SELECT_SCROLL_HARD_MARGIN_Y = 10;
        // once margin is passed, increase speed by factor
        public double TEXT_SELECT_SCROLL_SPEED_MULTIPLIER_IF_POST_MARGIN_X = 4;
        public double TEXT_SELECT_SCROLL_SPEED_MULTIPLIER_IF_POST_MARGIN_Y = 4;
        // the amount text quads are inflated for hit testing purposes with fingers
        public static double TEXT_SELECT_FINGER_HIT_TESTING_MARGIN = 5;

        /// <summary>
        /// Stylus inking
        /// If user uses stylus in pan mode, we switch to ink mode. This is the time without additional strokes we will wait before saving this as an annotation.
        /// set to 0 or < 0 to disable this timer completely. Ink will then save as soon as the stroke is finished.
        /// </summary>
        public double INK_TIME_BEFORE_INK_SAVES_ANNOTATION_IN_MILLISECONDS = 2000;

        /// <summary>
        /// When in above ink mode, we will also save and start a new annotation if the user draws far enough from the current ink.
        /// set to 0 or < 0 in order to disable this. If you disable it, every stroke will count as the same annotation.
        /// </summary>
        public double INK_MARGIN_FOR_STARTING_NEW_ANNOTATION = 100;

        // Manipulation Events variables. They need to be here since the tool can change mid manipulation
        private Dictionary<uint, Pointer> mContactPoints = new Dictionary<uint, Pointer>();

        internal Dictionary<uint, Pointer> ContactPoints { get { return mContactPoints; } }
        internal bool EnableScrollByManipulation { get; set; }
        internal bool EnableOneFingerScroll { get; set; }
        internal bool ManipulationTouch { get; set; }
        internal bool ManipulationInertiaStarted { get; set; }
        internal int TouchPoints { get; set; }

        public delegate void AnnotationModificationHandler(IAnnot annotation, int pageNumber);

        /// <summary>
        /// This event is fired when an annotation has been added to a document
        /// </summary>
        public event AnnotationModificationHandler AnnotationAdded = delegate { };

        /// <summary>
        /// This event is raised whenever an annotation on the current document has been edited
        /// </summary>
        public event AnnotationModificationHandler AnnotationEdited = delegate { };

        /// <summary>
        /// This event is raised whenever an annotation has been deleted from the document
        /// </summary>
        public event AnnotationModificationHandler AnnotationRemoved = delegate { };

        /// <summary>
        /// This event is raised just before an annotation is deleted from the document.
        /// At this point, the annotation is still part of the document.
        /// </summary>
        public event AnnotationModificationHandler AnnotationPreRemoved = delegate { };

        /// <summary>
        /// This event is raised when annotations are selected or deselected by the editing tools
        /// </summary>
        public event AnnotationModificationHandler SelectionChanged = delegate { };

        //Todo: Refactor later; in Android AnnotationModificationListener takes a Map of annoations and page numbers. In UWP only one annotation is considered.
        public delegate void AnnotationGroupModificationHandler(Dictionary<IAnnot, int> annotationGroup); //Dictionary values represent page numbers

        /// <summary>
        /// This event is fired when an annotation group has been added to a document
        /// </summary>
        public event AnnotationGroupModificationHandler AnnotationGroupAdded = delegate { };

        /// <summary>
        /// This event is raised whenever an annotation group on the current document has been edited
        /// </summary>
        public event AnnotationGroupModificationHandler AnnotationGroupEdited = delegate { };

        /// <summary>
        /// This event is raised whenever an annotation group on the current document is about to be edited
        /// </summary>
        public event AnnotationGroupModificationHandler AnnotationGroupPreEdited = delegate { };

        /// <summary>
        /// This event is raised whenever an annotation group has been deleted from the document
        /// </summary>
        public event AnnotationGroupModificationHandler AnnotationGroupRemoved = delegate { };

        /// <summary>
        /// This event is raised just before an annotation group is deleted from the document.
        /// At this point, the annotation group is still part of the document.
        /// </summary>
        public event AnnotationGroupModificationHandler AnnotationGroupPreRemoved = delegate { };

        /// <summary>
        /// This event is raised when annotation group is selected or deselected by the editing tools
        /// </summary>
        public event AnnotationGroupModificationHandler SelectionGroupChanged = delegate { };

        public delegate void InterceptAnnotationHandler(object sender, InterceptAnnotationEventArgs e);

        public event InterceptAnnotationHandler ShouldHandleLinkAnnotation = delegate { };

        [Obsolete("ShowingPopupMenuHandler is deprecated, please use ShowingContextMenuHandler instead.")]
        public delegate void ShowingPopupMenuHandler(PopupCommandMenu menu, ShowingPopupMenuEventArgs eventArgs);

        [Obsolete("OnShowPopupMenu is deprecated, please use OnShowContextMenu instead.")]
        /// <summary>
        /// This event is raised before displaying the annotation popup menu.
        /// </summary>
        public event ShowingPopupMenuHandler OnShowPopupMenu;

        public delegate void ShowingContextMenuHandler(ContextMenuControl menu, ShowingContextMenuEventArgs eventArgs);

        /// <summary>
        /// This event is raised before displaying the context menu.
        /// </summary>
        public event ShowingContextMenuHandler OnShowContextMenu;

        public delegate void SoundAnnotCreateHandler(Windows.Foundation.Point target, int pageNumber);
        private SoundAnnotCreateHandler _onSoundAnnotRequested;

        /// <summary>
        /// Event is raised when user selects sound annotation (Mic) on the Context Menu 
        /// </summary>
        public event SoundAnnotCreateHandler OnSoundAnnotRequested
        {
            add
            {
                if (_onSoundAnnotRequested?.GetInvocationList().Contains(value) != true)
                    _onSoundAnnotRequested += value;
            }
            remove => _onSoundAnnotRequested -= value;
        }

        public delegate void ReflowEditRequestedHandler();
        private ReflowEditRequestedHandler _reflowEditRequested;

        /// <summary>
        /// Event is raised when user request to annotate during reflow
        /// </summary>
        public event ReflowEditRequestedHandler ReflowEditRequested
        {
            add
            {
                if (_reflowEditRequested?.GetInvocationList().Contains(value) != true)
                    _reflowEditRequested += value;
            }
            remove => _reflowEditRequested -= value;
        }

        private PDFViewCtrl mPDFView;
        private AnnotManager mAnnotManager;

        public bool IsOnPhone { get; private set; }

        private NavigationStack _NavigationStack = null;
        /// <summary>
        /// Gets or sets the ToolManager navigation stack. If not null, the Tools will use the
        /// NavigationStack for navigation, like following links.
        /// </summary>
        public NavigationStack NavigationStack
        {
            get { return _NavigationStack; }
            set { _NavigationStack = value; }
        }

        private string _DocumentName = null;
        /// <summary>
        /// Gets or sets the DocumentName as the user of the ToolManager wants it displayed.
        /// If null, the tools will use PDFDoc.GetFileName()
        /// Useful when a temp file with a generic name is used.
        /// </summary>
        public string DocumentName
        {
            get { return _DocumentName; }
            set { _DocumentName = value; }
        }

        /// <summary>
        /// Gets the UndoRedoManger. UndoRedoManger contains API for Undo / Redo operation.
        /// </summary>
        public IUndoRedoManager UndoRedoAction { get; private set; }

        private bool _isRichTextEnabled = false;
        /// <summary>
        /// This is for toggling the feature for creating the FreeTextAnnotation to support Rich Text data in the annotation.
        /// </summary>
        public bool IsRichTextEnabled { get { return _isRichTextEnabled; } set { _isRichTextEnabled = value; } }

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="ctrl"></param>
        public ToolManager(PDFViewCtrl ctrl)
        {

            IsOnPhone = false;
#if WINDOWS_PHONE_APP
            IsOnPhone = true;
#elif !WINDOWS_APP
             IsOnPhone =
               Windows.Foundation.Metadata.ApiInformation.IsTypePresent("Windows.Phone.UI.Input.HardwareButtons");
#endif

            mPDFView = ctrl;
            mPageNumberIndicator = new PageNumberIndicator();
            mDelayRemoveTimers = new List<DelayRemoveTimer>();

            EnableScrollByManipulation = false;
            EnableOneFingerScroll = false;
            ManipulationTouch = false;
            ManipulationInertiaStarted = false;
            TouchPoints = 0;

            // tie into PDFViewCtrl events
            ctrl.OnScale += OnScaleHandler;
            ctrl.OnPageNumberChanged += PageNumberChangedHandler;
            ctrl.OnSetDoc += SetDocHandler;
            ctrl.OnSize += OnSizeHandler;
            ctrl.OnViewChanged += ViewChangedHandler;
            ctrl.OnRenderFinished += OnFinishedRenderingHandler;

            ctrl.PointerPressed += PointerPressedHandler;

            // Note: For pointer released, we have to use the AddHandler Syntax in order to catch events whose handled flag is set to true.
            // This is because, internally, PDFViewCtrl will set the focus to itself on a pointer released event. This means that the Handled
            // flag has to be set to true in order to prevent other elements in the visual tree from taking the focus once this event bubbles
            // to them. This means that you can not expect a PointerReleasedEvent to always be preceded by a pointer pressed event.
            _PointerReleasedHandler = new PointerEventHandler(PointerReleasedHandler);
            ctrl.AddHandler(Windows.UI.Xaml.UIElement.PointerReleasedEvent, _PointerReleasedHandler, true);

            ctrl.PointerMoved += PointerMovedHandler;
            ctrl.PointerCanceled += PointerCanceledHandler;
            ctrl.PointerWheelChanged += PointerWheelChangedHandler;
            ctrl.PointerCaptureLost += PointerCaptureLostHandler;
            ctrl.PointerEntered += PointerEnteredHandler;
            ctrl.PointerExited += PointerExitedHandler;

            ctrl.Tapped += TappedHandler;
            ctrl.Holding += HoldingHandler;
            ctrl.RightTapped += RightTappedHandler;
            ctrl.DoubleTapped += DoubleTappedHandler;

            ctrl.ManipulationStarted += ManipulationStartedHandler;
            ctrl.ManipulationCompleted += ManipulationCompletedHandler;
            ctrl.ManipulationDelta += ManipulationDeltaHandler;
            ctrl.ManipulationInertiaStarting += ManipulationInertiaStartingHandler;
            ctrl.ManipulationStarting += ManipulationStartingHandler;

            // This handler needs to fire at all times in order to be able to, for example, change pages when the user presses an arrow key
            // while the page is small enough so that it can't scroll.
            _KeyDownHandler = new KeyEventHandler(KeyDownHandler);
            ctrl.AddHandler(Windows.UI.Xaml.UIElement.KeyDownEvent, _KeyDownHandler, false);
            ctrl.KeyUp += KeyUpHandler;

            AddAuthorToAnnotations = true;
            CreateDefaultTool();
        }

        ~ToolManager()
        {
            Dispose();
        }

        /// <summary>
        /// Disposes the ToolManager. Always perform this when the ToolManager is not needed anymore.
        /// The ToolManager is subscribed to events on the PDFViewCtrl that prevent it from being reclaimed.
        /// Therefore it is very important to dispose the ToolManager, so that it can unsubscribe to these events.
        /// </summary>
        public void Dispose()
        {
            if (_IsDisposed)
            {
                return;
            }
            _IsDisposed = true;

            if (UndoRedoAction != null)
            {
                ((UndoRedoManager)UndoRedoAction).Dispose();
                UndoRedoAction = null;
            }

            if (mPDFView != null)
            {
                Unsubscribe();
            }
        }

        /// <summary>
        /// If this is not done when the ToolManager is no longer needed, the PDFViewCtrl can not be reclaimed
        /// by the garbage collector.
        /// </summary>
        private void Unsubscribe()
        {
            if (_CurrentTool != null)
            {
                _CurrentTool.OnClose();
            }

            mPDFView.OnScale -= OnScaleHandler;
            mPDFView.OnPageNumberChanged -= PageNumberChangedHandler;
            mPDFView.OnSetDoc -= SetDocHandler;
            mPDFView.OnSize -= OnSizeHandler;
            mPDFView.OnViewChanged -= ViewChangedHandler;
            mPDFView.OnRenderFinished -= OnFinishedRenderingHandler;

            mPDFView.PointerPressed -= PointerPressedHandler;
            mPDFView.RemoveHandler(Windows.UI.Xaml.UIElement.PointerReleasedEvent, _PointerReleasedHandler);

            mPDFView.PointerMoved -= PointerMovedHandler;
            mPDFView.PointerCanceled -= PointerCanceledHandler;
            mPDFView.PointerWheelChanged -= PointerWheelChangedHandler;
            mPDFView.PointerCaptureLost -= PointerCaptureLostHandler;
            mPDFView.PointerEntered -= PointerEnteredHandler;
            mPDFView.PointerExited -= PointerExitedHandler;

            mPDFView.Tapped -= TappedHandler;
            mPDFView.Holding -= HoldingHandler;
            mPDFView.RightTapped -= RightTappedHandler;
            mPDFView.DoubleTapped -= DoubleTappedHandler;

            mPDFView.ManipulationStarted -= ManipulationStartedHandler;
            mPDFView.ManipulationCompleted -= ManipulationCompletedHandler;
            mPDFView.ManipulationDelta -= ManipulationDeltaHandler;
            mPDFView.ManipulationInertiaStarting -= ManipulationInertiaStartingHandler;
            mPDFView.ManipulationStarting -= ManipulationStartingHandler;

            // This handler needs to fire at all times in order to be able to, for example, change pages when the user presses an arrow key
            // while the page is small enough so that it can't scroll.
            mPDFView.RemoveHandler(Windows.UI.Xaml.UIElement.KeyDownEvent, _KeyDownHandler);
            mPDFView.KeyUp -= KeyUpHandler;

            foreach (DelayRemoveTimer timer in mDelayRemoveTimers)
            {
                timer.Destroy();
            }
        }

        /// <summary>
        /// Get the PDFViewCtrl which this ToolManager is attached to
        /// </summary>
        /// <returns> PDFViewCtrl instance </returns>
        public PDFViewCtrl GetPDFViewCtrl()
        {
            return mPDFView;
        }

        /// <summary>
        /// Attach UndoRedoManager to the ToolManager
        /// </summary>
        /// <param name="undoRedoManager"> IUndoRedoManager instance </param>
        public void SetUndoRedoManager(IUndoRedoManager undoRedoManager = null)
        {
            if (UndoRedoAction != null)
            {
                ((UndoRedoManager)UndoRedoAction).Dispose();
                UndoRedoAction.OnUndoRedoStatusChanged -= UndoRedoAction_OnUndoRedoStatusChanged;
            }

            if (undoRedoManager == null)
                UndoRedoAction = new UndoRedoManager();
            else
                UndoRedoAction = undoRedoManager;

            UndoRedoAction.OnUndoRedoStatusChanged += UndoRedoAction_OnUndoRedoStatusChanged;
            UndoRedoAction.InitializeUndoRedoActionOnDocumentLoad(mPDFView, this);

            // Just raise a initialization notification
            UndoRedoStatusChanged?.Invoke(new UndoRedoStatusEventArgs
            {
                CanRedo = false,
                CanUndo = false,
                ModifiedPage = -1,
                RedoInfo = "",
                UndoInfo = ""
            });
        }

        /// <summary>
        /// Enables annotation manager for annotation syncing
        /// </summary>
        /// <param name="author"> he unique identifier of the current user </param>
        /// <returns></returns>
        public void EnableAnnotManager(string userId)
        {
            if (string.IsNullOrEmpty(userId))
                mAnnotManager = null;
            else
                mAnnotManager = new AnnotManager(this, userId);
        }

        /// <summary>
        /// Check if AnnotManager is enabled
        /// </summary>
        public bool IsAnnotManagerEnabled
        {
            get
            {
                return mAnnotManager != null ? true : false;
            }
        }

        /// <summary>
        /// Get AnnotManager instance. Make sure to check if it is already enabled
        /// </summary>
        /// <returns> AnnotManager instance </returns>
        public AnnotManager GetAnnotManager()
        {
            return mAnnotManager;
        }

        public MagnifyingTool CreateMagnifyingTool()
        {
            if (_CurrentTool != null)
            {
                _CurrentTool.OnClose();
            }
            CreateTool(ToolType.e_magnifying_rectangle, null);
            return _CurrentTool as MagnifyingTool;
        }

        /// <summary>
        /// Creates the default tool (the Pan tool).
        /// </summary>
        /// <param name="ctrl"></param>
        public void CreateDefaultTool()
        {
            _CurrentTool?.OnClose();
            CreateTool(ToolType.e_pan, null);
        }

        [Windows.Foundation.Metadata.DefaultOverload]
        public Tool CreateTool(ToolType mode, Tool current_tool)
        {
            return CreateTool(mode, current_tool, false);
        }


        public Tool CreateTool(ToolType mode, Tool current_tool, bool toolIsPersistent)
        {
            Tool t = null;
            if (!_ToolChangedInternally)
            {
                ToolModeAfterEditing = ToolType.e_none;
            }

            bool returningToOldTool = false;
            if (ToolModeAfterEditing != ToolType.e_none && mode != ToolType.e_annot_edit && mode != ToolType.e_annot_edit_text_markup && mode != ToolType.e_line_edit)
            {
                mode = ToolModeAfterEditing;
                returningToOldTool = true;
            }

            if (ReadOnlyMode)
            {
                if (mode == ToolType.e_arrow_create || mode == ToolType.e_form_fill || mode == ToolType.e_ink_create || mode == ToolType.e_ink_eraser ||
                    mode == ToolType.e_line_create || mode == ToolType.e_oval_create || mode == ToolType.e_rect_create || mode == ToolType.e_signature ||
                    mode == ToolType.e_sticky_note_create || mode == ToolType.e_text_annot_create || mode == ToolType.e_stamper ||
                    mode == ToolType.e_text_highlight || mode == ToolType.e_text_squiggly || mode == ToolType.e_text_strikeout || mode == ToolType.e_text_underline ||
                    mode == ToolType.e_cloud_create || mode == ToolType.e_polygon_create || mode == ToolType.e_polyline_create || mode == ToolType.e_annot_edit_rect_group ||
                    mode == ToolType.e_annot_edit_group_create || mode == ToolType.e_ruler_create || mode == ToolType.e_perimeter_create || mode == ToolType.e_area_create ||
                    mode == ToolType.e_lasso_create || mode == ToolType.e_annot_edit_text_markup)
                {
                    mode = ToolType.e_pan;
                }
            }
            switch (mode)
            {
                case ToolType.e_pan:
                    t = new Pan(mPDFView, this);
                    break;
                case ToolType.e_annot_edit:
                    t = new AnnotEdit(mPDFView, this);
                    break;
                case ToolType.e_advanced_shape_edit:
                    t = new AnnotEditAdvancedShape(mPDFView, this);
                    break;
                case ToolType.e_annot_edit_rect_group:
                    t = new AnnotEditRectGroup(mPDFView, this);
                    break;
                case ToolType.e_annot_edit_group_create:
                    t = new AnnotEditGroupCreate(mPDFView, this);
                    break;
                case ToolType.e_lasso_create:
                    t = new LassoCreate(mPDFView, this);
                    break;
                case ToolType.e_line_create:
                    t = new LineCreate(mPDFView, this);
                    break;
                case ToolType.e_arrow_create:
                    t = new ArrowCreate(mPDFView, this);
                    break;
                case ToolType.e_rect_create:
                    t = new RectCreate(mPDFView, this);
                    break;
                case ToolType.e_oval_create:
                    t = new OvalCreate(mPDFView, this);
                    break;
                case ToolType.e_ink_create:
                    t = new FreehandCreate(mPDFView, this);
                    break;
                case ToolType.e_ink_marker:
                    t = new FreehandCreate(mPDFView, this, true);
                    break;
                case ToolType.e_text_annot_create:
                    t = new FreeTextCreate(mPDFView, this);
                    break;
                case ToolType.e_link_action:
                    t = new LinkAction(mPDFView, this);
                    break;
                case ToolType.e_text_select:
                    t = new TextSelect(mPDFView, this);
                    break;
                case ToolType.e_form_fill:
                    t = new FormFill(mPDFView, this);
                    break;
                case ToolType.e_sticky_note_create:
                    t = new StickyNoteCreate(mPDFView, this);
                    break;
                case ToolType.e_text_highlight:
                    t = new TextHightlightCreate(mPDFView, this);
                    break;
                case ToolType.e_text_underline:
                    t = new TextUnderlineCreate(mPDFView, this);
                    break;
                case ToolType.e_text_strikeout:
                    t = new TextStrikeoutCreate(mPDFView, this);
                    break;
                case ToolType.e_text_squiggly:
                    t = new TextSquigglyCreate(mPDFView, this);
                    break;
                case ToolType.e_line_edit:
                    t = new LineEdit(mPDFView, this);
                    break;
                case ToolType.e_rich_media:
                    t = new RichMedia(mPDFView, this);
                    break;
                case ToolType.e_signature:
                    t = new Signature(mPDFView, this);
                    break;
                case ToolType.e_stamper:
                    t = new ImageStamper(mPDFView, this);
                    break;
                case ToolType.e_ink_eraser:
                    t = new Eraser(mPDFView, this);
                    break;
                case ToolType.e_annot_edit_text_markup:
                    t = new TextMarkupEdit(mPDFView, this);
                    break;
                case ToolType.e_magnifying_rectangle:
                    t = new MagnifyingTool(mPDFView, this);
                    break;
                case ToolType.e_zoom_marquee:
                    t = new MarqueeZoom(mPDFView, this);
                    break;
                case ToolType.e_polyline_create:
                    t = new PolylineCreate(mPDFView, this);
                    break;
                case ToolType.e_polygon_create:
                    t = new PolygonCreate(mPDFView, this);
                    break;
                case ToolType.e_cloud_create:
                    t = new CloudCreate(mPDFView, this);
                    break;
                case ToolType.e_ruler_create:
                    t = new RulerCreate(mPDFView, this);
                    break;
                case ToolType.e_perimeter_create:
                    t = new PerimeterMeasureCreate(mPDFView, this);
                    break;
                case ToolType.e_area_create:
                    t = new AreaMeasureCreate(mPDFView, this);
                    break;
                case ToolType.e_rect_redaction_create:
                    t = new RectRedactionCreate(mPDFView, this);
                    break;
                case ToolType.e_text_redaction:
                    t = new TextRedactionCreate(mPDFView, this);
                    break;
                default:
                    t = new Pan(mPDFView, this);
                    break;
            }

            if (_CurrentTool != null && _CurrentTool != current_tool)
            {
                _CurrentTool.OnClose();
            }

            if (current_tool != null)
            {
                // Transfer some useful info
                t.Transfer((Tool)current_tool);
                current_tool.OnClose();		//close the old tool; old tool can use this to clean up things.
            }

            //call a tool's onCreate() function in which the tool can initialize things that require the transferred properties.
            t.OnCreate();
            t.ReturnToPanModeWhenFinished = !toolIsPersistent && ToolModeAfterEditing == ToolType.e_none;
            _CurrentTool = t;

            if (returningToOldTool)
            {
                ToolModeAfterEditing = ToolType.e_none;
            }

            if (NewToolCreated != null)
            {
                NewToolCreated(mode);
            }

            return t;
        }

        /// <summary>
        /// Gets the list of options that are normally displayed when the user taps on blank space
        /// in the PDFViewCtrl (i.e. no annotation or text).
        /// </summary>
        /// <returns></returns>
        public IList<string> GetAnnotationCreationOptions()
        {
            return _CurrentTool.ToolCreationOptions();
        }

        /// <summary>
        /// Creates a new tool based on one of the strings from GetAnnotationCreationOptions()
        /// </summary>
        /// <param name="title"></param>
        public void CreateTool(string title)
        {
            CreateTool(title, false);
        }

        public void CreateTool(string title, bool toolIsPersistent)
        {
            _CurrentTool.HandleToolCreationOption(title, toolIsPersistent);
        }

        /// <summary>
        /// Closes any open dialog the Tools have open.
        /// User this to handle things like back buttons.
        /// </summary>
        /// <param name="all">Closes all open dialogs if true, otherwise only 1 (in case there are nested ones).</param>
        /// <returns></returns>
        public bool CloseOpenDialog(bool all = false)
        {
            bool closed = false;
            if (mCurrentNoteDialog != null)
            {
                mCurrentNoteDialog.Close();
                closed = true;
            }

            if (CurrentTool is Signature && (CurrentTool as Signature).GoBack())
            {
                closed = true;
            }

            if (mOpenAuthorDialogViewModel != null)
            {
                mOpenAuthorDialogViewModel.GoBack();
                mOpenAuthorDialogViewModel = null;
                closed = true;
            }

            if (mCurrentBoxPopup != null)
            {
                mCurrentBoxPopup.Hide();
                mCurrentBoxPopup = null;
                closed = true;
            }

            if (all || !closed)
            {
                bool cls = _CurrentTool.CloseOpenDialog();
                closed |= cls;
                if (all)
                {
                    while (cls)
                    {
                        cls = _CurrentTool.CloseOpenDialog();
                        closed |= cls;
                    }
                }
            }

            return closed;
        }

        public bool ClosePopup()
        {
            return CurrentTool.ClosePopup();
        }


        internal void ClearDelayRemoveList()
        {
            foreach (DelayRemoveTimer timer in mDelayRemoveTimers)
            {
                timer.Destroy();
            }
            mDelayRemoveTimers.Clear();
        }

        private ToolType _ToolModeAfterEditing = ToolType.e_none;
        internal ToolType ToolModeAfterEditing
        {
            get { return _ToolModeAfterEditing; }
            set { _ToolModeAfterEditing = value; }
        }

        private bool _ToolChangedInternally = false;
        internal bool ToolChangedInternally
        {
            get { return _ToolChangedInternally; }
            set { _ToolChangedInternally = value; }
        }

        private int _LastPageNumberRaised = 0;
        internal int LastPageNumberRaised
        {
            get { return _LastPageNumberRaised; }
            private set { _LastPageNumberRaised = value; }
        }

        /// <summary>
        /// Raise the request for Sound Annot to be initiated, usually from Context Menu
        /// </summary>
        /// <param name="annot"></param>
        internal void RaiseOnSoundAnnotRequestedEvent(Windows.Foundation.Point target, int pageNumber)
        {
            _onSoundAnnotRequested?.Invoke(target, pageNumber);
        }

        /// <summary>
        /// Raise the event to notify user is requesting editing during reflow
        /// </summary>
        /// <param name="type">The type of annotation</param>
        internal void RaiseOnReflowEditRequestedEvent()
        {
            _reflowEditRequested?.Invoke();
        }

        #region Annotation Modification

        /// <summary>
        /// Lets the various tools raise the AnnotationAdded event from a unified location.
        /// </summary>
        /// <param name="annot"></param>
        internal void RaiseAnnotationAddedEvent(IAnnot annot, int pageNumber)
        {
            AnnotationAdded?.Invoke(annot, pageNumber);
        }

        /// <summary>
        /// Lets the various tools raise the AnnotationEdited event from a unified location.
        /// </summary>
        /// <param name="annot"></param>
        internal void RaiseAnnotationEditedEvent(IAnnot annot, int pageNumber)
        {
            AnnotationEdited?.Invoke(annot, pageNumber);
        }

        /// <summary>
        /// This method should be called before removing annotations
        /// </summary>
        /// <param name="annot"></param>
        /// <param name="pageNumber"></param>
        internal void RaiseAnnotationPreRemovedEvent(IAnnot annot, int pageNumber)
        {
            AnnotationPreRemoved?.Invoke(annot, pageNumber);
        }

        /// <summary>
        /// This method should be called whenever annotations are removed from the document.
        /// </summary>
        /// <param name="annot"></param>
        internal void RaiseAnnotationRemovedEvent(IAnnot annot, int pageNumber)
        {
            AnnotationRemoved?.Invoke(annot, pageNumber);
        }

        internal void RaiseShouldHandleLinkAnnotationEvent(object sender, InterceptAnnotationEventArgs e)
        {
            ShouldHandleLinkAnnotation?.Invoke(sender, e);
        }

        internal void RaiseOnShowMenu(PopupCommandMenu menu, ShowingPopupMenuEventArgs eventArgs)
        {
            OnShowPopupMenu?.Invoke(menu, eventArgs);
        }

        internal void RaiseOnShowContextMenu(ContextMenuControl context, ShowingContextMenuEventArgs eventArgs)
        {
            OnShowContextMenu?.Invoke(context, eventArgs);
        }

        /// <summary>
        /// Lets the various tools raise the SelectionChanged event from a unified location.
        /// </summary>
        /// <param name="annot"></param>
        /// <param name="pageNumber"></param>
        internal void RaiseSelectionChangedEvent(IAnnot annot, int pageNumber)
        {
            SelectionChanged?.Invoke(annot, pageNumber);
        }

        #endregion Annotation Modification

        #region Annotation Group Modification

        /// <summary>
        /// Lets the various tools raise the AnnotationGroupAdded event from a unified location.
        /// </summary>
        /// <param name="annotationGroup"></param>
        internal void RaiseAnnotationGroupAddedEvent(Dictionary<IAnnot, int> annotationGroup)
        {
            if (AnnotationGroupAdded != null)
            {
                AnnotationGroupAdded(annotationGroup);
            }
        }

        /// <summary>
        /// Lets the various tools raise the AnnotationGroupEdited event from a unified location.
        /// </summary>
        /// <param name="annotationGroup"></param>
        internal void RaiseAnnotationGroupEditedEvent(Dictionary<IAnnot, int> annotationGroup)
        {
            if (AnnotationGroupEdited != null)
            {
                AnnotationGroupEdited(annotationGroup);
            }
        }

        internal void RaiseAnnotationGroupPreEditedEvent(Dictionary<IAnnot, int> annotationGroup)
        {
            AnnotationGroupPreEdited?.Invoke(annotationGroup);
        }

        internal void RaiseAnnotationGroupPreRemovedEvent(Dictionary<IAnnot, int> annotationGroup)
        {
            AnnotationGroupPreRemoved?.Invoke(annotationGroup);
        }

        /// <summary>
        /// Lets the various tools raise the AnnotationGroupRemoved event from a unified location.
        /// </summary>
        /// <param name="annotationGroup"></param>
        internal void RaiseAnnotationGroupRemovedEvent(Dictionary<IAnnot, int> annotationGroup)
        {
            if (AnnotationGroupRemoved != null)
            {
                AnnotationGroupRemoved(annotationGroup);
            }
        }

        /// <summary>
        /// Lets the various tools raise the SelectionGroupChanged event from a unified location.
        /// </summary>
        /// <param name="annotationGroup"></param>
        internal void RaiseSelectionGroupChangedEvent(Dictionary<IAnnot, int> annotationGroup)
        {
            SelectionGroupChanged?.Invoke(annotationGroup);
        }

        /// <summary>
        /// Select all annotations part of the same group of annotation
        /// Note: the annotation must be part of a annotation group in order to select it
        /// </summary>
        /// <param name="annotation">The annotation part of a group</param>
        /// <param name="pageNumber">The page number where the annotation is located</param>
        public void SelectAnnotationGroup(IAnnot annotation, int pageNumber)
        {
            if (annotation == null || annotation.IsValid() == false)
                return;

            var result = Tool.CanSelectGroupAnnotation(mPDFView, annotation, pageNumber);
            if (result.Key == ToolType.e_none)
                return;

            AnnotEditRectGroup t = CreateTool(ToolType.e_annot_edit_rect_group, null) as AnnotEditRectGroup;
            t.SelectAnnotationGroup(annotation, pageNumber);
        }

        /// <summary>
        /// Select the annotation passed if it exists in the page number and displays the
        /// context menu
        /// NOTE: only works with IMarkup style of annotations
        /// </summary>
        /// <param name="annotation">the annotation to be selected</param>
        /// <param name="pageNumber">the page number where the annotation lives</param>
        public void SelectAnnotation(IAnnot annotation, int pageNumber)
        {
            if (annotation == null || annotation.IsValid() == false)
                return;

            if (annotation is Annots.IMarkup)
            {
                AnnotEdit t = CreateTool(ToolType.e_annot_edit, null) as AnnotEdit;
                t.SelectAnnotation(annotation, pageNumber);
            }
        }

        /// <summary>
        /// Delete the annotation passed if it exists
        /// </summary>
        /// <param name="pdfViewCtrl">the active PDF View Control</param>
        /// <param name="selectedAnnotation">the currently selected annotation to be deleted</param>
        /// <param name="pageNumber">the page number that contains the selected annotation</param>
        public void DeleteAnnotation(PDFViewCtrl pdfViewCtrl, IAnnot selectedAnnotation, int pageNumber)
        {
            try
            {
                if (pdfViewCtrl == null || selectedAnnotation == null || selectedAnnotation.IsValid() == false)
                    return;

                pdfViewCtrl.DocLock(true);

                this.RaiseAnnotationPreRemovedEvent(selectedAnnotation, pageNumber);

                // Perform the Annotation removal and update PDFViewer
                var page = pdfViewCtrl.GetDoc().GetPage(pageNumber);
                selectedAnnotation = AnnotUtils.SafeDeleteAnnotAndUpdate(pdfViewCtrl, page, selectedAnnotation, pageNumber);

                this.RaiseAnnotationRemovedEvent(selectedAnnotation, pageNumber);
            }
            catch (Exception ex)
            {
                AppCenterHandlerBase.CURRENT.TrackError(ex);
            }
            finally
            {
                pdfViewCtrl.DocUnlock();
            }
        }

        #endregion Annotation Group Modification

        #region Single Tap Handling
        private class SingleTapCanceller
        {
            public bool Cancel = false;
        }

        SingleTapCanceller _SingleTapCanceller = null;
        internal static int _SingleTapDelayMS = 300;

        private void CancelCurrentSingleTap()
        {
            if (_SingleTapCanceller != null)
            {
                _SingleTapCanceller.Cancel = true;
            }
        }

        private async void DelaySingleTap(TappedRoutedEventArgs e, SingleTapCanceller canceller)
        {
            await Task.Delay(_SingleTapDelayMS);
            if (!canceller.Cancel)
            {
                SingleTap?.Invoke(SingleTapEvents.SingleTapConfirmed, e);
            }
        }

        #endregion Single Tap Handling

        #region Event Handlers
        //////////////////////////////////////////////////////////////////////////
        // All events that the ToolManager gets follows the same pattern.
        // First, the event is forwarded to the current tool.
        // Once the current tool is finished, we look to see if the next tool mode
        // is the same as the current tool mode. This gives the current tool a
        // chance to process the event, and if it detects that another tool should
        // handle this event, it can set the next tool mode to the appropriate
        // tool.
        // This is done for example when the Pan tool handles a MouseLeftButtonDown
        // event and notices a form field at the current cursor location.
        //////////////////////////////////////////////////////////////////////////


        internal void OnScaleHandler()
        {
            PDFViewCtrlEventLoop(0, 0, (s, args) =>
            {
                return _CurrentTool.OnScale();
            });
        }

        /// <summary>
        /// Set this variable to true during any of PointerPressed or PointerReleased in the tools in order to make sure that this
        /// particular interaction does not generate a tap
        /// </summary>
        internal bool SuppressTap { get; set; }
        internal void PointerPressedHandler(Object sender, PointerRoutedEventArgs e)
        {
            SuppressTap = false;
            CancelCurrentSingleTap();

            // double click checking
            _IsDoubleClick = false;
            if (_MouseFirstClick && e.Pointer.PointerDeviceType == Windows.Devices.Input.PointerDeviceType.Mouse)
            {
                Windows.UI.Input.PointerPoint scDownPoint = e.GetCurrentPoint(mPDFView);
                double sqDist = (scDownPoint.Position.X - _scDoubleClickFirstDownPoint.X) * (scDownPoint.Position.X - _scDoubleClickFirstDownPoint.X)
                    + (scDownPoint.Position.Y - _scDoubleClickFirstDownPoint.Y) * (scDownPoint.Position.Y - _scDoubleClickFirstDownPoint.Y);
                TimeSpan timeDifference = DateTime.Now.Subtract(_MouseClickTime);
                if (scDownPoint.Properties.IsLeftButtonPressed && sqDist < 25 && timeDifference.TotalSeconds < 0.3) // we have a double click
                {
                    _IsDoubleClick = true;
                    _MouseFirstClick = false;
                    DoubleClickHandler(sender, e);
                    return;
                }
            }

            _MouseFirstClick = false;

            // regular touch handling
            PDFViewCtrlEventLoop(sender, e, (s, args) =>
            {
                return _CurrentTool.PointerPressedHandler(s, args);
            });
            CurrentTool.AddPointer(e.Pointer);
        }

        internal void PointerReleasedHandler(Object sender, PointerRoutedEventArgs e)
        {
            PDFViewCtrlEventLoop(sender, e, (s, args) =>
            {
                return _CurrentTool.PointerReleasedHandler(s, args);
            });
            CurrentTool.RemovePointer(e.Pointer);
        }

        internal void PointerMovedHandler(Object sender, PointerRoutedEventArgs e)
        {
            PDFViewCtrlEventLoop(sender, e, (s, args) =>
            {
                return _CurrentTool.PointerMovedHandler(s, args);
            });
        }

        internal void PointerCanceledHandler(Object sender, PointerRoutedEventArgs e)
        {
            PDFViewCtrlEventLoop(sender, e, (s, args) =>
            {
                return _CurrentTool.PointerCanceledHandler(s, args);
            });
            CurrentTool.RemovePointer(e.Pointer);
        }

        internal void PageNumberChangedHandler(int current_page, int num_pages)
        {
            if (current_page != LastPageNumberRaised)
            {
                PDFViewCtrlEventLoop(current_page, num_pages, (s, args) =>
                {
                    return _CurrentTool.OnPageNumberChanged(s, args);
                });
            }
            LastPageNumberRaised = current_page;
        }

        internal void PointerWheelChangedHandler(Object sender, PointerRoutedEventArgs e)
        {
            PDFViewCtrlEventLoop(sender, e, (s, args) =>
            {
                return _CurrentTool.PointerWheelChangedHandler(s, args);
            });
        }

        internal void PointerCaptureLostHandler(Object sender, PointerRoutedEventArgs e)
        {
            PDFViewCtrlEventLoop(sender, e, (s, args) =>
            {
                return _CurrentTool.PointerCaptureLostHandler(s, args);
            });
            CurrentTool.RemovePointer(e.Pointer);
        }

        internal void PointerEnteredHandler(Object sender, PointerRoutedEventArgs e)
        {
            PDFViewCtrlEventLoop(sender, e, (s, args) =>
            {
                return _CurrentTool.PointerEnteredHandler(s, args);
            });
        }

        internal void PointerExitedHandler(Object sender, PointerRoutedEventArgs e)
        {
            PDFViewCtrlEventLoop(sender, e, (s, args) =>
            {
                return _CurrentTool.PointerExitedHandler(s, args);
            });
            CurrentTool.RemovePointer(e.Pointer);
        }

        internal void SetDocHandler()
        {
            _LastPageNumberRaised = mPDFView.GetCurrentPage();
            if (_CurrentTool != null && _IsEnabled)
            {
                _CurrentTool.OnClose();
                CreateDefaultTool();
                _CurrentTool.OnSetDoc();
                Action open_action = mPDFView.GetDoc().GetOpenAction();
                if (open_action.IsValid() && open_action.GetType() == ActionType.e_JavaScript)
                {
                    ActionParameter action_param = new ActionParameter(open_action);
                    _CurrentTool.ExecuteAction(action_param);
                }
            }

            SetUndoRedoManager();
        }

        internal void OnSizeHandler()
        {
            PDFViewCtrlEventLoop(0, 0, (s, args) =>
            {
                _CurrentTool.OnSize();
                return true;
            });
        }

        internal void OnFinishedRenderingHandler()
        {
            PDFViewCtrlEventLoop(0, 0, (s, args) =>
            {
                _CurrentTool.OnFinishedRendering();
                return true;
            });
        }

        internal void ManipulationStartedHandler(Object sender, ManipulationStartedRoutedEventArgs e)
        {
            e.Handled = PDFViewCtrlEventLoop(sender, e, (s, args) =>
            {
                return _CurrentTool.ManipulationStartedEventHandler(s, args);
            });
        }

        internal void ManipulationCompletedHandler(Object sender, ManipulationCompletedRoutedEventArgs e)
        {
            e.Handled = PDFViewCtrlEventLoop(sender, e, (s, args) =>
            {
                return _CurrentTool.ManipulationCompletedEventHandler(s, args);
            });
        }

        internal void ManipulationDeltaHandler(Object sender, ManipulationDeltaRoutedEventArgs e)
        {
            e.Handled = PDFViewCtrlEventLoop(sender, e, (s, args) =>
            {
                return _CurrentTool.ManipulationDeltaEventHandler(s, args);
            });
        }

        internal void ManipulationInertiaStartingHandler(Object sender, ManipulationInertiaStartingRoutedEventArgs e)
        {
            e.Handled = PDFViewCtrlEventLoop(sender, e, (s, args) =>
            {
                return _CurrentTool.ManipulationInertiaStartingEventHandler(s, args);
            });
        }

        internal void ManipulationStartingHandler(Object sender, ManipulationStartingRoutedEventArgs e)
        {
            e.Handled = PDFViewCtrlEventLoop(sender, e, (s, args) =>
            {
                return _CurrentTool.ManipulationStartingEventHandler(s, args);
            });
        }

        internal void TappedHandler(Object sender, TappedRoutedEventArgs e)
        {
            if (e.PointerDeviceType == Windows.Devices.Input.PointerDeviceType.Touch)
            {
                SingleTap?.Invoke(SingleTapEvents.BeforeTool, e);
                if (e.Handled)
                {
                    return;
                }
            }
            if (_IsDoubleClick || SuppressTap)
            {
                e.Handled = true;
                return;
            }
            if (e.PointerDeviceType == Windows.Devices.Input.PointerDeviceType.Mouse)
            {
                _MouseFirstClick = true;
                _MouseClickTime = DateTime.Now;
                _scDoubleClickFirstDownPoint = e.GetPosition(mPDFView);
            }
            e.Handled = PDFViewCtrlEventLoop(sender, e, (s, args) =>
            {
                return _CurrentTool.TappedHandler(s, args);
            });
            if (!e.Handled && e.PointerDeviceType == Windows.Devices.Input.PointerDeviceType.Touch)
            {
                _SingleTapCanceller = new SingleTapCanceller();
                DelaySingleTap(e, _SingleTapCanceller);
            }
        }


        internal void HoldingHandler(Object sender, HoldingRoutedEventArgs e)
        {
            e.Handled = PDFViewCtrlEventLoop(sender, e, (s, args) =>
            {
                return _CurrentTool.HoldingHandler(s, args);
            });
        }

        internal void RightTappedHandler(Object sender, RightTappedRoutedEventArgs e)

        {
            e.Handled = PDFViewCtrlEventLoop(sender, e, (s, args) =>
            {
                return _CurrentTool.RightTappedHandler(s, args);
            });
        }

        internal void DoubleTappedHandler(object sender, DoubleTappedRoutedEventArgs e)
        {
            e.Handled = PDFViewCtrlEventLoop(sender, e, (s, args) =>
            {
                return _CurrentTool.DoubleTappedHandler(s, args);
            });
        }

        internal void ViewChangedHandler(Object sender, Windows.UI.Xaml.Controls.ScrollViewerViewChangedEventArgs e)
        {
            PDFViewCtrlEventLoop(sender, e, (s, args) =>
            {
                return _CurrentTool.OnViewChanged(s, args);
            });
        }

        public void KeyDownHandler(object sender, KeyRoutedEventArgs e)
        {
            bool handling = false;
            handling = PDFViewCtrlEventLoop(sender, e, (s, args) =>
            {
                return _CurrentTool.KeyDownHandler(s, args);
            });
            if (handling)
            {
                e.Handled = true;
                mPDFView.Focus(FocusState.Programmatic);
            }
        }

        public void KeyUpHandler(object sender, KeyRoutedEventArgs e)
        {
            e.Handled = PDFViewCtrlEventLoop(sender, e, (s, args) =>
            {
                return _CurrentTool.KeyUpHandler(s, args);
            });
        }

        private bool _MouseFirstClick = false;
        private bool _IsDoubleClick = false;
        private DateTime _MouseClickTime = DateTime.Now;
        private Windows.Foundation.Point _scDoubleClickFirstDownPoint;
        private void DoubleClickHandler(Object sender, PointerRoutedEventArgs e)
        {
            e.Handled = PDFViewCtrlEventLoop(sender, e, (s, args) =>
            {
                return _CurrentTool.DoubleClickedHandler(s, args);
            });
        }

        bool PDFViewCtrlEventLoop<S, T>(S sender, T args, Func<S, T, bool> act)
        {
            bool handled = false;
            if (_CurrentTool != null && _IsEnabled)
            {
                ToolType prev_tm = _CurrentTool.ToolMode;
                ToolType next_tm;
                while (true)
                {
                    handled |= act(sender, args);
                    next_tm = _CurrentTool.NextToolMode;

                    if (prev_tm != next_tm)
                    {
                        _ToolChangedInternally = true;
                        _CurrentTool = CreateTool(next_tm, _CurrentTool);
                        _ToolChangedInternally = false;
                    }
                    if (_CurrentTool.ToolMode != prev_tm)
                    {
                        prev_tm = _CurrentTool.ToolMode;
                    }
                    else
                    {
                        _CurrentTool.JustSwitchedFromAnotherTool = false;
                        break;
                    }
                }
            }
            return handled;
        }

        #endregion Event Handlers

        #region Redaction
        /// <summary>
        /// Apply redaction to all Redaction Annotations present in the document
        /// </summary>
        /// <param name="overlayText">Turn on/off displaying text overlay</param>
        public void ApplyRedaction(bool overlayText)
        {
            try
            {
                mPDFView.DocLock(true);

                var pdfDoc = mPDFView.GetDoc();

                IList<RedactorRedaction> rarr = new List<RedactorRedaction>();
                for (PageIterator itr = pdfDoc.GetPageIterator(); itr.HasNext(); itr.Next())
                {
                    Page page = itr.Current();
                    int num_annots = page.GetNumAnnots();
                    for (int i = 0; i < num_annots; ++i)
                    {
                        var annot = page.GetAnnot(i);
                        if (!annot.IsValid()) continue;
                        switch (annot.GetAnnotType())
                        {
                            case AnnotType.e_Redact:
                                {
                                    var redact = new pdftron.PDF.Annots.Redaction(annot.GetSDFObj());
                                    var quadPointCount = redact.GetQuadPointCount();
                                    var textOverlay = redact.GetOverlayText();
                                    if (quadPointCount != 0)
                                    {
                                        for (int j = 0; j < quadPointCount; j++)
                                        {
                                            var initialRect = QuadPointToRect(redact.GetQuadPoint(j));
                                            var redaction = new RedactorRedaction(page.GetIndex(), initialRect, false, textOverlay);
                                            rarr.Add(redaction);
                                        }
                                    }
                                    else
                                    {
                                        var redaction = new RedactorRedaction(page.GetIndex(), redact.GetRect(), false, textOverlay);
                                        rarr.Add(redaction);
                                    }
                                    break;
                                }
                            default:
                                break;
                        }
                    }
                }

                if (rarr.Count == 0)
                    return;

                RedactorAppearance appearance = new RedactorAppearance
                {
                    PositiveOverlayColor = Windows.UI.Colors.Black,
                    NegativeOverlayColor = Windows.UI.Colors.WhiteSmoke,
                    TextColor = Windows.UI.Colors.White,
                    OverlayText = true,
                    HorizTextAlignment = 0, // Center alligment
                    VertTextAlignment = 0   // Center alligment
                };

                Redactor.Redact(pdfDoc, rarr, appearance);
                UndoRedoAction.TakeSnapshot("Apply Redaction", true);
                mPDFView.Update(true);
            }
            catch(Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(
                    Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY,
                    Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                mPDFView.DocUnlock();
            }
        }

        private Rect QuadPointToRect(QuadPoint qp)
        {
            double rectX1 = Math.Min(Math.Min(Math.Min(qp.p1.x, qp.p2.x), qp.p3.x), qp.p4.x);
            double rectX2 = Math.Max(Math.Max(Math.Max(qp.p1.x, qp.p2.x), qp.p3.x), qp.p4.x);
            double rectY1 = Math.Min(Math.Min(Math.Min(qp.p1.y, qp.p2.y), qp.p3.y), qp.p4.y);
            double rectY2 = Math.Max(Math.Max(Math.Max(qp.p1.y, qp.p2.y), qp.p3.y), qp.p4.y);

            return new Rect(rectX1, rectY1, rectX2, rectY2);
        }
        #endregion

    }
}
