﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using pdftron.SDF;
using Windows.Data.Json;
using pdftron.PDF.Tools;

namespace pdftron.PDF.Tools
{
    public delegate void UndoRedoStatus(object sender, UndoRedoStatusEventArgs e);

    public class UndoRedoStatusEventArgs : EventArgs
    {
        public bool CanUndo { get; set; } = false;
        public bool CanRedo { get; set; } = false;

        public string UndoInfo { get; set; } = "Undo";

        public string RedoInfo { get; set; } = "Redo";

        public int ModifiedPage { get; set; }
    }

    public class UndoRedoManager : IUndoRedoManager, IDisposable
    {
        private bool mIsUndoRedoEnabled = false;

        private PDFViewCtrl mPdfViewCtrl = null;
        private ToolManager mToolManager = null;

        public event UndoRedoStatus OnUndoRedoStatusChanged;

        const string JSON_DESCRIPTION = "description";
        const string JSON_UPDATE_LAYOUT = "update_layout";
        const string JSON_MODIFIED_PAGE = "modified_page";
        const string JSON_INITIAL_STATE = "label";

        public void InitializeUndoRedoActionOnDocumentLoad(PDFViewCtrl ctrl, ToolManager toolManager)
        {
            if (ctrl != null)
                mPdfViewCtrl = ctrl;

            if (toolManager != null)
            {
                mToolManager = toolManager;
                // Keep track of all annotation change events
                mToolManager.AnnotationAdded += MToolManager_AnnotationAdded;
                mToolManager.AnnotationRemoved += MToolManager_AnnotationRemoved;
                mToolManager.AnnotationGroupAdded += MToolManager_AnnotationGroupAdded;
                mToolManager.AnnotationGroupRemoved += MToolManager_AnnotationGroupRemoved;
                mToolManager.AnnotationEdited += MToolManager_AnnotationEdited;
                mToolManager.AnnotationGroupEdited += MToolManager_AnnotationGroupEdited;
            }

            if (mPdfViewCtrl != null)
            {
                mPdfViewCtrl.EnableUndoRedo();
                mIsUndoRedoEnabled = true;
            }
        }

        /// <summary>
        /// Takes snapshot of the current document state.
        /// </summary>
        /// <param name="metaInfo"> Brief snap shot description </param>
        /// <param name="requiresPageLayoutUpdate"> Updates the page layout withing the view.
        /// This function must be called after document page sequence is modified (such as when a page is being added to or removed from a document)
        /// or after changes to page dimensions (e.g. after a page is rotated or resized). </param>
        public void TakeSnapshot(string metaInfo = "", bool requiresPageLayoutUpdate = false)
        {
            if (mPdfViewCtrl == null || mToolManager == null || !mIsUndoRedoEnabled)
                return;

            // JSON object to store all snapshot data
            var snapData = new JsonObject();

            var jsonValue = JsonValue.CreateBooleanValue(requiresPageLayoutUpdate);
            snapData.Add(JSON_UPDATE_LAYOUT, jsonValue);

            jsonValue = JsonValue.CreateNumberValue(mPdfViewCtrl.GetCurrentPage());
            snapData.Add(JSON_MODIFIED_PAGE, jsonValue);

            jsonValue = JsonValue.CreateStringValue(metaInfo);
            snapData.Add(JSON_DESCRIPTION, jsonValue);

            // get all keys and values in a string so we can store into the snapshot
            var snapInfo = snapData.Stringify();

            bool shouldUnlock = false;
            try
            {
                mPdfViewCtrl.DocLock(false);
                shouldUnlock = true;

                mPdfViewCtrl.TakeSnapshot(snapInfo);
                mPdfViewCtrl.RequestRendering();

                RaiseUndoRedoStatusChangeEvent(snapData);
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(
                    Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY,
                    Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }

            finally
            {
                if (shouldUnlock)
                {
                    mPdfViewCtrl.DocUnlock();
                }
            }
        }

        /// <summary>
        /// Reverts to previous document state.
        /// </summary>
        public void DoUndo()
        {
            PerformUndoRedo(true);
        }

        /// <summary>
        /// Reverts to the document state that was undone with call to DoUndo.
        /// </summary>
        public void DoRedo()
        {
            PerformUndoRedo(false);
        }

        /// <summary>
        /// Perform Undo or Redo action
        /// </summary>
        /// <param name="isUndo"> True - Undo, False - Redo </param>
        /// <returns> Stringified JSON data </returns>
        private string PerformUndoRedo(bool isUndo)
        {
            if (!mIsUndoRedoEnabled || mPdfViewCtrl.GetDoc() == null || mPdfViewCtrl == null)
                return null;

            string info = "";
            bool isPageLayoutUpdateRequired = false;
            JsonObject metaInfo;

            try
            {
                if (isUndo)
                {
                    if (!CanUndo)
                        return null;

                    mPdfViewCtrl.CancelRendering();

                    info = mPdfViewCtrl.Undo();

                    metaInfo = GetJsonFromString(info);

                    isPageLayoutUpdateRequired = metaInfo.GetNamedBoolean(JSON_UPDATE_LAYOUT);
                }
                else
                {
                    if (!CanRedo)
                        return null;

                    mPdfViewCtrl.CancelRendering();

                    info = mPdfViewCtrl.Redo();

                    metaInfo = GetJsonFromString(info);
                    isPageLayoutUpdateRequired = metaInfo.GetNamedBoolean(JSON_UPDATE_LAYOUT);
                }

                // Update Page Layout if an Edit action was made (edit page, rotate, deleting, etc)
                if (isPageLayoutUpdateRequired)
                    mPdfViewCtrl.UpdatePageLayout();

                RaiseUndoRedoStatusChangeEvent(metaInfo);

                // Raise proper Local change event for Annotation Manager
                if (mToolManager.IsAnnotManagerEnabled)
                {
                    mToolManager.GetAnnotManager().RaiseLocalChange(
                        isUndo ? AnnotManager.AnnotationAction.UNDO : AnnotManager.AnnotationAction.REDO);
                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(
                    Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY,
                    Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }

            return info;
        }

        /// <summary>
        /// It will notify the Tool Bar that Undo or Redo is allowed and send the next action info
        /// </summary>
        /// <param name="keyValuePairs"></param>
        private void RaiseUndoRedoStatusChangeEvent(JsonObject keyValuePairs)
        {
            try
            {
                var undoRedoStatusEventArgs = new UndoRedoStatusEventArgs();

                if (undoRedoStatusEventArgs != null)
                {
                    undoRedoStatusEventArgs.CanUndo = CanUndo;
                    undoRedoStatusEventArgs.UndoInfo += ": " + GetNextUndoRedoActionDescr(true);

                    undoRedoStatusEventArgs.CanRedo = CanRedo;
                    undoRedoStatusEventArgs.RedoInfo += ": " + GetNextUndoRedoActionDescr(false);

                    undoRedoStatusEventArgs.ModifiedPage = (int)keyValuePairs.GetNamedNumber(JSON_MODIFIED_PAGE);

                    OnUndoRedoStatusChanged?.Invoke(this, undoRedoStatusEventArgs);
                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(
                    Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY,
                    Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
        }

        public bool CanUndo { get { return GetNextUndoRedoActionDescr(true) == null ? false : true; } }
        public bool CanRedo { get { return GetNextUndoRedoActionDescr(false) == null ? false : true; } }

        /// <summary>
        /// Get the next undo or redo action description if available
        /// </summary>
        /// <param name="isUndo"> </param>
        /// <returns> Snapshot description if valid otherwise null </returns>
        private string GetNextUndoRedoActionDescr(bool isUndo)
        {
            if (!mIsUndoRedoEnabled || mPdfViewCtrl.GetDoc() == null)
                return null;

            string info = string.Empty;

            if (isUndo)
                info = mPdfViewCtrl.GetNextUndoInfo();
            else
                info = mPdfViewCtrl.GetNextRedoInfo();

            if (IsValidJson(info) == false)
                return null;

            var jsonObj = GetJsonFromString(info);

            if (jsonObj.ContainsKey(JSON_INITIAL_STATE))
                return null;

            if (jsonObj.ContainsKey(JSON_DESCRIPTION))
                return jsonObj.GetNamedString(JSON_DESCRIPTION);
            else
                return null;
        }

        private JsonObject GetJsonFromString(string data)
        {
            JsonObject keyValuePairs = null;

            string formated_data = data;

            /* HACK: Make sure the string uses double quotes so Json can parse it properly.
             * When using AnnotManager the xfdf string content might use single quote
             * TODO: remove it */
            if (data.Contains("'"))
                formated_data = data.Replace("'", "\"");

            JsonObject.TryParse(formated_data, out keyValuePairs);

            return keyValuePairs;
        }

        private bool IsValidJson(string data)
        {
            JsonObject keyValuePairs = null;
            return JsonObject.TryParse(data, out keyValuePairs);
        }


        #region Notify ToolManager Events

        private void MToolManager_AnnotationAdded(IAnnot annotation, int pageNumber)
        {
            if (annotation != null)
            {
                string annotName = UtilityFunctions.GetAnnotDescriptionFromEnum(annotation.GetAnnotType());

                TakeSnapshot("Add " + annotName + " Annotation");

                if (mToolManager?.IsAnnotManagerEnabled == true)
                    mToolManager.GetAnnotManager().RaiseLocalChange(AnnotManager.AnnotationAction.ADD);
            }
        }

        private void MToolManager_AnnotationEdited(IAnnot annotation, int pageNumber)
        {
            string annotName = UtilityFunctions.GetAnnotDescriptionFromEnum(annotation.GetAnnotType());

            TakeSnapshot("Modify " + annotName + " Annotation");

            if (mToolManager.IsAnnotManagerEnabled)
                mToolManager.GetAnnotManager().RaiseLocalChange(AnnotManager.AnnotationAction.MODIFY);
        }

        private void MToolManager_AnnotationRemoved(IAnnot annotation, int pageNumber)
        {
            string annotName = UtilityFunctions.GetAnnotDescriptionFromEnum(annotation.GetAnnotType());

            TakeSnapshot("Delete " + annotName + " Annotation");

            if (mToolManager.IsAnnotManagerEnabled)
                mToolManager.GetAnnotManager().RaiseLocalChange(AnnotManager.AnnotationAction.DELETE);
        }

        private void MToolManager_AnnotationGroupAdded(Dictionary<IAnnot, int> annotationGroup)
        {
            TakeSnapshot("Add Annotation Group");

            if (mToolManager.IsAnnotManagerEnabled)
                mToolManager.GetAnnotManager().RaiseLocalChange(AnnotManager.AnnotationAction.ADD);
        }

        private void MToolManager_AnnotationGroupEdited(Dictionary<IAnnot, int> annotationGroup)
        {
            TakeSnapshot("Edit Annotation Group");

            if (mToolManager.IsAnnotManagerEnabled)
                mToolManager.GetAnnotManager().RaiseLocalChange(AnnotManager.AnnotationAction.MODIFY);
        }

        private void MToolManager_AnnotationGroupRemoved(Dictionary<IAnnot, int> annotationGroup)
        {
            TakeSnapshot("Delete Annotation Group");

            if (mToolManager.IsAnnotManagerEnabled)
                mToolManager.GetAnnotManager().RaiseLocalChange(AnnotManager.AnnotationAction.DELETE);
        }

        #endregion

        public void Dispose()
        {
            // If initializing it again just cleanup events
            if (mToolManager != null)
            {
                mToolManager.AnnotationAdded -= MToolManager_AnnotationAdded;
                mToolManager.AnnotationRemoved -= MToolManager_AnnotationRemoved;
                mToolManager.AnnotationGroupAdded -= MToolManager_AnnotationGroupAdded;
                mToolManager.AnnotationGroupRemoved -= MToolManager_AnnotationGroupRemoved;
                mToolManager.AnnotationEdited -= MToolManager_AnnotationEdited;
                mToolManager.AnnotationGroupEdited -= MToolManager_AnnotationGroupEdited;
            }
        }
    }
}
