﻿using pdftron.FDF;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Windows.ApplicationModel.DataTransfer;
using Windows.Data.Xml.Dom;
using UIPoint = Windows.Foundation.Point;
using PDFDouble = pdftron.Common.DoubleRef;

namespace pdftron.PDF.Tools.Utilities
{
    /// <summary>
    /// Helper class for annotation copy/paste
    /// </summary>
    public class AnnotationClipboardHelper
    {
        static bool mIsAnnotToPaste;

        private static string XFDF_ANNOTS_KEY = "annots";
        private static string CLIPBOARD_ANNOT_COPY_KEY = "pdftron_annot_copied"; // single annot
        private static string CLIPBOARD_ANNOTS_COPY_KEY = "pdftron_annots_copied"; // multiple annots
        private static int DEFAULT_PASTE_ANNOT_OFFSET = 25;

        /// <summary>
        /// Flag to check if an annotation was copied to the Clipboard
        /// Note: this flag is only set to False after a clipboard request is performed by PasteAnnotFromClipboardAsync()
        /// </summary>
        public static bool IsAnnotToPaste
        {
            get { return mIsAnnotToPaste; }
        }

        /// <summary>
        /// Copy the annotation and store it into the Clipboard
        /// </summary>
        /// <param name="pdfView">The PDFViewCtrl source</param>
        /// <param name="annot">The annotation to be copied from the PDFViewCtrl</param>
        /// <returns>True if the Clipboard stored it, False otherwise</returns>
        public static bool CopyAnnotToClipboard(PDFViewCtrl pdfView, IAnnot annot)
        {
            if (pdfView == null || annot == null)
                return false;

            bool isDocLocked = false;

            try
            {
                pdfView.DocLock(true);
                isDocLocked = true;

                if (annot.IsValid() == false)
                    return false;                

                var doc = pdfView.GetDoc();
                var fdfExtracted = doc.FDFExtract(new List<IAnnot> { annot });

                // TODO: At this point we check if <annots> is available inside the XFDF, if not, the annot cannot be copied
                // add proper handling to copy and paste unavailble XFDF annot
                string xfdfString = fdfExtracted.SaveAsXFDFString();
                if (xfdfString.Contains("<annots>") == false)
                    return false;

                Tuple<FDFDoc, IAnnot> fdfDocAnnot = new Tuple<FDFDoc, IAnnot>(fdfExtracted, annot);

                DataPackage dataPackage = new DataPackage();
                dataPackage.RequestedOperation = DataPackageOperation.Copy;                
                dataPackage.SetText(annot.GetContents()); // We set text with annotation content
                dataPackage.Properties.Add(CLIPBOARD_ANNOT_COPY_KEY, fdfDocAnnot);

                if (UtilityFunctions.CopyToClipBoard(dataPackage))
                {
                    mIsAnnotToPaste = true;
                }
                else
                {
                    return false;
                }
            }
            catch (Exception)
            {
                // swallon exception
            }
            finally
            {
                if (isDocLocked)
                    pdfView.DocUnlock();
            }

            return true;
        }

        /// <summary>
        /// Paste the annotation in the Clipboard to selected page of the PDFViewCtrl
        /// </summary>
        /// <param name="pdfViewPasteTo">PDFViewCtrl to Paste to</param>
        /// <param name="pageNoToPaste">The page number to Paste to</param>
        /// <param name="target">Optional: destination on the page for better accuracy / positioning</param>
        /// <returns></returns>
        public static async Task<bool> PasteAnnotFromClipboardAsync(PDFViewCtrl pdfViewPasteTo, int pageNoToPaste, Point target = null) // PDFViewCtrl pdfViewCopyFrom, PDFViewCtrl pdfViewPasteTo, int pageNoToPaste, Point target
        {
            DataPackageView dataPackageView = Clipboard.GetContent();
            if (dataPackageView.Properties.Count > 0)
            {
                object o = new object();

                dataPackageView.Properties.TryGetValue(CLIPBOARD_ANNOT_COPY_KEY, out o);

                var result = o as Tuple<FDFDoc, IAnnot>;
                if (result != null)
                {
                    dataPackageView.ReportOperationCompleted(DataPackageOperation.None);

                    return await PasteAnnotImplAsync(result, pdfViewPasteTo, pageNoToPaste, target);
                }
            }

            if (mIsAnnotToPaste)
                mIsAnnotToPaste = false;

            return false;
        }

        /// <summary>
        /// Perform paste annotation to target PDFDoc / PDFViewCtrl and at the end, if success, copy the new annotation back to the clipboard
        /// </summary>
        /// <param name="clipBoardTuple"></param>
        /// <param name="pdfViewPasteTo"></param>
        /// <param name="pageNoToPaste">The page number to paste 1-based</param>
        /// <param name="target"></param>
        /// <returns></returns>
        private static async Task<bool> PasteAnnotImplAsync(Tuple<FDFDoc, IAnnot> clipBoardTuple, PDFViewCtrl pdfViewPasteTo, int pageNoToPaste, Point target) 
        {
            if (clipBoardTuple == null)
                return false;

            var fdfDoc = clipBoardTuple.Item1;
            var annot = clipBoardTuple.Item2;

            if (annot == null)
                return false;

            bool isDocLocked = false;

            try
            {
                pdfViewPasteTo.DocLock(true);
                isDocLocked = true;

                if (annot.IsValid() == false)
                    return false;

                var pageNoZeroBased = pageNoToPaste - 1; // XFDF takes 0-based for page index

                var doc = pdfViewPasteTo.GetDoc();
                var destPage = doc.GetPage(pageNoToPaste);

                // TODO: var newRectOffset = GetUpdatedAnnotBounds(annot, destPage);

                var stringFDF = fdfDoc.SaveAsXFDFString();

                var xmlDoc = new XmlDocument();
                xmlDoc.LoadXml(stringFDF);
                var annotsNode = xmlDoc.GetElementsByTagName(XFDF_ANNOTS_KEY);
                for (int i = 0; i < annotsNode[0].ChildNodes.Count; i++)
                {
                    var childNode = annotsNode[0].ChildNodes[i];
                    if (childNode.Attributes == null)
                        continue;

                    // TODO: need to ensure name and rect exist
                    var nameAttribute = childNode.Attributes.FirstOrDefault(x => x.NodeName == "name");
                    if (nameAttribute != null)
                        childNode.Attributes.GetNamedItem("name").InnerText = Guid.NewGuid().ToString();

                    var pageAttribute = childNode.Attributes.FirstOrDefault(x => x.NodeName == "page");
                    if (pageAttribute != null)
                        childNode.Attributes.GetNamedItem("page").InnerText = pageNoZeroBased.ToString();

                    // NOTE: if annotation has extra styles (opacity, free text) the node "apref" will be added
                    // with a specific obj address from the original annotation which needs to be removed, otherwise
                    // the copied annot will set the orignal annot style 
                    var aprefNode = childNode.ChildNodes.FirstOrDefault(x => x.LocalName != null && x.LocalName.ToString() == "apref");
                    if (aprefNode != null)
                        childNode.RemoveChild(aprefNode);
                }
                var xmlResult = xmlDoc.GetXml();
                var bytes = Encoding.UTF8.GetBytes(xmlResult);
                FDFDoc fDFDoc = await FDFDoc.CreateFromXFDFAsync(bytes);
                
                doc.FDFMerge(fDFDoc);                

                // If everything goes well we need to update the annotation with a new offset position
                var newAnnot = destPage.GetAnnot(destPage.GetNumAnnots() - 1);
                if (newAnnot != null && newAnnot.IsValid())
                {
                    var newRect = GetAnnotRectWithOffset(newAnnot, destPage);
                    newAnnot.Resize(newRect);
                    newAnnot.RefreshAppearance();

                    // now copy the newly added annotation to the clipboard to replace old one
                    if (newAnnot.GetAnnotType() == annot.GetAnnotType())
                        CopyAnnotToClipboard(pdfViewPasteTo, newAnnot);
                }

                pdfViewPasteTo.Update();
            }
            catch (Exception)
            {
                return false;
            }
            finally
            {
                if (isDocLocked)
                    pdfViewPasteTo.DocUnlock();
            }

            return true;
        }

        /// <summary>
        /// Get an updated rect with added offset so the pasted annotation does not overlap original one
        /// </summary>
        /// <param name="annot">Annotation which should be updated with an offset</param>
        /// <param name="pageToPaste">Destination page to ensure BBox</param>
        /// <returns>React with updated values</returns>
        private static Rect GetAnnotRectWithOffset(IAnnot annot, Page pageToPaste)
        {
            var pageCropBox = pageToPaste.GetCropBox();
            var annotRect = annot.GetRect();

            // Need to define the bounds here considering the page bounding
            var x_offset = DEFAULT_PASTE_ANNOT_OFFSET;
            var y_offset = -DEFAULT_PASTE_ANNOT_OFFSET;

            // Check and update bounds
            if (annotRect.y1 + y_offset <= pageCropBox.y1) // bottom
            {
                y_offset = DEFAULT_PASTE_ANNOT_OFFSET;
            }
            else if (annotRect.y2 + y_offset >= pageCropBox.y2) // Right top corner
            {
                y_offset = -DEFAULT_PASTE_ANNOT_OFFSET;
                x_offset = -DEFAULT_PASTE_ANNOT_OFFSET;
            }
            else if (annotRect.x2 + x_offset > pageCropBox.x2) // right bottom corner
            {
                x_offset = -DEFAULT_PASTE_ANNOT_OFFSET;
            }

            return new Rect(annotRect.x1 + x_offset, annotRect.y1 + y_offset, annotRect.x2 + x_offset, annotRect.y2 + y_offset);
        }
    }
}
