﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text.RegularExpressions;
using Windows.UI;
using Windows.UI.Xaml.Markup;

namespace pdftron.PDF.Tools
{
    /// <summary>
    /// HTML Tags
    /// </summary>
    internal static class HTMLTags
    {
        public const string span = "span";
        public const string paragraph = "p";
    }

    /// <summary>
    /// CSS Styles
    /// </summary>
    internal static class CSSTypes
    {
        public const string Bold = "font-weight:bold";
        public const string Italics = "font-style:italic";
        public const string Underline = "text-decoration:underline";
        public const string Strikethrough = "text-decoration:line-through";
        public const string UnderlineAndStrikethrough = "text-decoration:line-through underline";
        public const string ColorStyle = "color:";
        public static string Color(string color)
        {
            return $"color:#{color}";
        }
    }

    /// <summary>
    /// Command/Control Words for the RTF language
    /// </summary>
    internal static class RTFCommands
    {
        public const string BoldStart = "\\b";
        public const string BoldEnd = "\\b0";
        public const string ItalicsStart = "\\i";
        public const string ItalicsEnd = "\\i0";
        public const string UnderlineStart = "\\ul";
        public const string UnderlineEnd = "\\ulnone";
        public const string StrikethroughStart = "\\strike";
        public const string StrikethroughEnd = "\\strike0";
        public const string ParagraphStart = "\\pard";
        public const string NewLine = "\\par";
        public const string NewLine2 = "\\line";
        public const string ColorTable = "\\colortbl";
        public const string ColorCommand = "\\cf";
        public const string Red = "\\red";
        public const string Green = "\\green";
        public const string Blue = "\\blue";
    }

    /// <summary>
    /// Utility class for updating the "RC" value for Rich Text Support for FreeText Annotation
    /// </summary>
    internal class HTMLUtilities
    {
        public string GetHexColorFromStyle(string style)
        {
            if (string.IsNullOrEmpty(style))
            {
                return "";
            }

            string color = "";
            int index = style.IndexOf(CSSTypes.ColorStyle, StringComparison.OrdinalIgnoreCase);
            int colorStart = 0;
            for (int i = index; i < style.Length; i++)
            {
                if (style[i] == '#')
                {
                    colorStart = i;
                    break;
                }
            }
            if (colorStart + 7 < style.Length)
            {
                color = style.Substring(colorStart, 7);
            }
            return color;
        }

        public string GetInlineStyle(string inputHTML)
        {
            if (string.IsNullOrEmpty(inputHTML))
            {
                return "";
            }

            string style = "";

            if (inputHTML.Contains("style=\"", StringComparison.OrdinalIgnoreCase))
            {
                int startIndex = inputHTML.IndexOf("style", StringComparison.OrdinalIgnoreCase);
                int countQuotes = 0;
                int endIndex = 0;
                for (endIndex = startIndex; endIndex < inputHTML.Length && countQuotes < 2; endIndex++)
                {
                    if (inputHTML[endIndex] == '"')
                    {
                        countQuotes++;
                    }
                }

                style = inputHTML.Substring(startIndex, endIndex - startIndex);
            }

            return style;
        }
        /// <summary>
        /// Return new RC after remove all color styles and add color style to "body" tag
        /// </summary>
        /// <param name="strRC"></param>
        /// <returns></returns>
        public string ReplacePrimaryColor(string strRC, Color color)
        {
            if (string.IsNullOrEmpty(strRC))
            {
                return null;
            }
            string newRC = strRC;
            MatchCollection matchCollection = Regex.Matches(strRC, @"color:\s?#\w\w\w\w\w\w;?");
            string colorStr = color.ToString();
            if (colorStr.Length >= 9)
            {
                colorStr = colorStr.Substring(3, 6);
                foreach (var match in matchCollection)
                {
                        newRC = newRC.Replace(match.ToString(), $"color:#{colorStr};");
                }
            }
            return newRC;
        }
    }

    /// <summary>
    /// This does not parse and convert all RTF to HTML, this is purely for Rich Text Support for FreeText Annotation
    /// </summary>
    internal class ConvertRTFAndHTML
    {
        private bool isBolded;
        private bool isItalics;
        private bool isUnderline;
        private bool isStrikethrough;

        private bool hasColorTable = false;
        private int colorIndex = 0;
        private List<Color> listColors;

        #region RTF to HTML
        private bool IsStyle()
        {
            if (isBolded || isItalics || isUnderline || isStrikethrough)
            {
                return true;
            }
            return false;
        }

        private string GetCSSStyle()
        {
            string style = "";

            if (isBolded)
            {
                style += CSSTypes.Bold;
                if (style.Length > 0)
                {
                    style += ";";
                }
            }
            if (isItalics)
            {
                style += CSSTypes.Italics;
                if (style.Length > 0)
                {
                    style += ";";
                }
            }
            if (isUnderline && isStrikethrough)
            {
                style += CSSTypes.UnderlineAndStrikethrough;
                if (style.Length > 0)
                {
                    style += ";";
                }
            }
            else if (isUnderline)
            {
                style += CSSTypes.Underline;
                if (style.Length > 0)
                {
                    style += ";";
                }
            }
            else if (isStrikethrough)
            {
                style += CSSTypes.Strikethrough;
                if (style.Length > 0)
                {
                    style += ";";
                }
            }
            if (hasColorTable && listColors.Count > 0)
            {
                if (colorIndex > 0)
                {
                    string substringColor = listColors[colorIndex - 1].ToString().Substring(3, 6);
                    style += CSSTypes.Color(substringColor);
                }
                else
                {
                    string substringColor = listColors[0].ToString().Substring(3, 6);
                    style += CSSTypes.Color(substringColor);
                }

                if (style.Length > 0)
                {
                    style += ";";
                }
            }

            return style;
        }

        private string AddSpan(string contents, string style)
        {
            string htmlSpan = "";
            if (string.IsNullOrEmpty(style))
            {
                htmlSpan = @"<span>" + contents + @"</span>";
            }
            else
            {
                htmlSpan = @"<span style=""" + style + @""">" + contents + @"</span>";
            }

            return htmlSpan;
        }

        private List<string> GetRTFTokens(string textRTF)
        {
            if (string.IsNullOrEmpty(textRTF))
            {
                return null;
            }

            List<string> listTokens = new List<string>();
            string tempToken = "";
            for (int i = 0; i < textRTF.Length; i++)
            {

                // Start of Control Word
                if (textRTF[i] == '\\' || textRTF[i] == '{' || textRTF[i] == '}' || textRTF[i] == ';')
                {
                    // End of Token
                    if (tempToken.Length > 0)
                    {
                        listTokens.Add(tempToken);
                    }

                    tempToken = "\\";
                }
                else if (textRTF[i] == ' ')
                {
                    // End of Token
                    if (tempToken.Length > 0)
                    {
                        listTokens.Add(tempToken);
                    }

                    tempToken = " ";
                }
                else
                {
                    tempToken += textRTF[i];
                }
            }

            if (tempToken.Length > 0)
            {
                listTokens.Add(tempToken);
            }

            return listTokens;
        }

        private void ExtractColorsFromColorTable(string colorTable)
        {
            if (string.IsNullOrEmpty(colorTable))
            {
                return;
            }

            if (colorTable != null && colorTable.Length > 0)
            {
                var colorTableSplitElements = colorTable.Split(";");
                listColors = new List<Color>();
                hasColorTable = true;
                List<string> listTokens;
                foreach (var line in colorTableSplitElements)
                {
                    listTokens = GetRTFTokens(line);

                    byte red = 0;
                    byte blue = 0;
                    byte green = 0;
                    bool isColor = true;

                    foreach (var token in listTokens)
                    {
                        if (token.Contains("red"))
                        {
                            string newStr = token.Replace("\\red", " ");
                            byte.TryParse(newStr, out red);
                        }
                        else if (token.Contains("green"))
                        {
                            string newStr = token.Replace("\\green", " ");
                            byte.TryParse(newStr, out green);
                        }
                        else if (token.Contains("blue"))
                        {
                            string newStr = token.Replace("\\blue", " ");
                            byte.TryParse(newStr, out blue);
                        }
                        else
                        {
                            isColor = false;
                        }
                    }

                    if (isColor)
                    {
                        Color tempColor = Color.FromArgb(255, red, green, blue);
                        listColors.Add(tempColor);
                    }

                }
            }
        }

        private string BuildHTMLParagraphFromRTF(string paragraphRTF)
        {
            if (string.IsNullOrEmpty(paragraphRTF))
            {
                return "";
            }

            Regex rxValidCommand = new Regex(@"\\[a-z]+(-?[0-9]+)? ?", RegexOptions.Compiled | RegexOptions.IgnoreCase);

            string paragraphHTML = "";
            string textInfo = "";
            string plainText = "";

            paragraphHTML += "<p>";
            List<string> listTokens = new List<string>();
            isBolded = false;
            isItalics = false;
            isUnderline = false;
            isStrikethrough = false;

            listTokens = GetRTFTokens(paragraphRTF);

            // Do stuff with tokens
            bool isPrevCW = false;
            foreach (var token in listTokens)
            {
                if (rxValidCommand.IsMatch(token))
                {
                    isPrevCW = true;


                    if (hasColorTable && token.Contains(RTFCommands.ColorCommand))
                    {
                        if (!string.IsNullOrEmpty(textInfo))
                        {
                            paragraphHTML += AddSpan(textInfo, GetCSSStyle());
                            textInfo = "";
                        }

                        var colorCommand = token.Replace(RTFCommands.ColorCommand, " ").Trim();
                        int.TryParse(colorCommand, out colorIndex);
                    }

                    switch (token.Trim())
                    {
                        case RTFCommands.BoldStart:
                            if (!IsStyle())
                            {
                                paragraphHTML += AddSpan(textInfo, GetCSSStyle());
                                textInfo = "";
                            }
                            isBolded = true;
                            break;
                        case RTFCommands.BoldEnd:
                            paragraphHTML += AddSpan(textInfo, GetCSSStyle());
                            textInfo = "";
                            isBolded = false;
                            break;
                        case RTFCommands.ItalicsStart:
                            if (!IsStyle())
                            {
                                paragraphHTML += AddSpan(textInfo, GetCSSStyle());
                                textInfo = "";
                            }
                            isItalics = true;
                            break;
                        case RTFCommands.ItalicsEnd:
                            paragraphHTML += AddSpan(textInfo, GetCSSStyle());
                            textInfo = "";
                            isItalics = false;
                            break;
                        case RTFCommands.UnderlineStart:
                            if (!IsStyle())
                            {
                                paragraphHTML += AddSpan(textInfo, GetCSSStyle());
                                textInfo = "";
                            }
                            isUnderline = true;
                            break;
                        case RTFCommands.UnderlineEnd:
                            paragraphHTML += AddSpan(textInfo, GetCSSStyle());
                            textInfo = "";
                            isUnderline = false;
                            break;
                        case RTFCommands.StrikethroughStart:
                            if (!IsStyle())
                            {
                                paragraphHTML += AddSpan(textInfo, GetCSSStyle());
                                textInfo = "";
                            }
                            isStrikethrough = true;
                            break;
                        case RTFCommands.StrikethroughEnd:
                            paragraphHTML += AddSpan(textInfo, GetCSSStyle());
                            textInfo = "";
                            isStrikethrough = false;
                            break;
                        case RTFCommands.NewLine:
                            textInfo += "&#13;";
                            break;
                        case RTFCommands.NewLine2:
                            textInfo += "&#13;";
                            break;
                        default: break;
                    }
                }
                else
                {
                    // Add text
                    if (isPrevCW)
                    {
                        textInfo += token.Trim();
                        plainText += token.Trim();
                    }
                    else
                    {
                        textInfo += token;
                        plainText += token;
                    }
                    isPrevCW = false;
                }
            }

            if (textInfo.Length > 0)
            {
                paragraphHTML += AddSpan(textInfo, GetCSSStyle());
                textInfo = "";
            }
            paragraphHTML += "</p>";

            return paragraphHTML;
        }

        /// <summary>
        /// This does not convert all RTF to HTML, this is purely for Rich Text Support for FreeText Annotation
        /// This takes in RTF string from the RichEditBox, and parses it and converts for HTML data to be stored with the annotation
        /// </summary>
        /// <param name="inputRTF"></param>
        /// <returns></returns>
        public string ConvertRTFToHTML(string inputRTF)
        {
            if (string.IsNullOrEmpty(inputRTF))
            {
                return "";
            }

            string outputHTML = "<?xml version=\"1.0\"?><body xmlns=\"http://www.w3.org/1999/xhtml\" xmlns:xfa=\"http://www.xfa.org/schema/xfa-data/1.0/\">";

            // the RTF from MS splits nicely by new lines (though I should find another way to handle this in case new lines are not split properly)
            string[] strArray = inputRTF.Split("\n");

            var pardLines = strArray.Where(str => str.Contains(RTFCommands.ParagraphStart) || str.Contains(RTFCommands.NewLine));

            // Building color table
            var colorTable = strArray.Where(str => str.Contains(RTFCommands.ColorTable)).FirstOrDefault();
            ExtractColorsFromColorTable(colorTable);

            foreach (string line in pardLines)
            {
                outputHTML += BuildHTMLParagraphFromRTF(line);
            }

            outputHTML = outputHTML + "</body>";
            return outputHTML;
        }

        #endregion

        #region HTML to RTF
        private int GetColorIndex(Color color)
        {
            if (color == null)
            {
                return -1;
            }

            for (int i = 0; i < listColors.Count; i++)
            {
                if (listColors[i] == color)
                {
                    return i;
                }
            }

            return -1;
        }

        private void BuildColorTable(string[] arrHTMLSpans)
        {
            if (arrHTMLSpans == null || arrHTMLSpans.Length == 0)
            {
                return;
            }

            bool hasColor = false;

            if (!hasColor)
            {
                foreach (var line in arrHTMLSpans)
                {
                    if (line.Contains(CSSTypes.ColorStyle))
                    {
                        int index = line.IndexOf(CSSTypes.ColorStyle, StringComparison.OrdinalIgnoreCase);
                        int colorStart = 0;
                        for (int i = index; i < line.Length; i++)
                        {
                            if (line[i] == '#')
                            {
                                colorStart = i;
                                break;
                            }
                        }

                        if (colorStart + 7 < line.Length)
                        {
                            hasColor = true;
                            string colorStr = line.Substring(colorStart, 7);

                            Windows.UI.Color color = (Color)XamlBindingHelper.ConvertValue(typeof(Color), colorStr);
                            if (!listColors.Contains(color))
                            {
                                listColors.Add(color);
                            }
                        }
                    }
                }
            }
        }

        /// <summary>
        /// This does not convert all HTML to RTF, this is purely for Rich Text Support for FreeText Annotation
        /// This takes in HTML string from the annotation, and parses and converts it for RTF data to be stored in the RichEditBox
        /// </summary>
        /// <param name="inputHTML"></param>
        /// <returns></returns>
        public string ConvertHTMLToRTF(string inputHTML)
        {
            if (string.IsNullOrEmpty(inputHTML))
            {
                return "";
            }

            HTMLUtilities htmlUtil = new HTMLUtilities();
            string outputRTF = @"{\rtf1\ansi";

            string[] splitHTMLParagraph = inputHTML.Split("</p>", StringSplitOptions.RemoveEmptyEntries);
            listColors = new List<Color>();
            foreach (var paragraph in splitHTMLParagraph)
            {
                string newParagraph = paragraph + "</p>";

                string[] splitHTMLSpan = newParagraph.Split("</span>", StringSplitOptions.RemoveEmptyEntries);

                foreach (var line in splitHTMLSpan)
                {
                    BuildColorTable(splitHTMLSpan);
                }
            }

            if (listColors.Count > 0)
            {
                string colorTable = "{\\colortbl ;";
                string colorStr = "";
                foreach (Color colorElement in listColors)
                {
                    colorStr = $"\\red{colorElement.R}\\green{colorElement.G}\\blue{colorElement.B}; ";
                    colorTable += colorStr;
                }

                colorTable += "}";
                outputRTF += "\n" + colorTable + "\n";
            }

            foreach (var paragraph in splitHTMLParagraph)
            {
                string newParagraph = paragraph + "</p>";

                string[] splitHTMLSpan = newParagraph.Split("</span>", StringSplitOptions.RemoveEmptyEntries);

                foreach (var line in splitHTMLSpan)
                {
                    if (line.Contains("span"))
                    {
                        string newLine = line + "</span>";
                        string inlineStyle = htmlUtil.GetInlineStyle(newLine);
                        inlineStyle = inlineStyle.Replace(" ", string.Empty);

                        outputRTF += "\n" + @"\pard";

                        // Add CW Start
                        if (inlineStyle.Contains(CSSTypes.Bold, StringComparison.OrdinalIgnoreCase))
                        {
                            outputRTF += RTFCommands.BoldStart;
                        }

                        if (inlineStyle.Contains(CSSTypes.Italics, StringComparison.OrdinalIgnoreCase))
                        {
                            outputRTF += RTFCommands.ItalicsStart;
                        }

                        if (inlineStyle.Contains(CSSTypes.UnderlineAndStrikethrough.Replace(" ", string.Empty), StringComparison.OrdinalIgnoreCase))
                        {
                            outputRTF += RTFCommands.UnderlineStart;
                            outputRTF += RTFCommands.StrikethroughStart;
                        }
                        else if (inlineStyle.Contains(CSSTypes.Underline, StringComparison.OrdinalIgnoreCase))
                        {
                            outputRTF += RTFCommands.UnderlineStart;
                        }
                        else if (inlineStyle.Contains(CSSTypes.Strikethrough, StringComparison.OrdinalIgnoreCase))
                        {
                            outputRTF += RTFCommands.StrikethroughStart;
                        }

                        if (line.Contains(CSSTypes.ColorStyle))
                        {
                            string colorStr = htmlUtil.GetHexColorFromStyle(line);
                            Windows.UI.Color color = (Color)XamlBindingHelper.ConvertValue(typeof(Color), colorStr);

                            // Color index +1 to ignore the 0th index, the 0th index is typically ignored and reserved as a default color. The 0th index does not need to be defined.
                            int colorIndex = GetColorIndex(color);
                            if (colorIndex >= 0)
                            {
                                outputRTF += RTFCommands.ColorCommand + (colorIndex + 1);
                            }
                            else
                            {
                                outputRTF += RTFCommands.ColorCommand + "0";
                            }
                        }

                        // Add Text
                        if (outputRTF.Last() != ' ')
                        {
                            outputRTF += " ";
                        }
                        MatchCollection matchCollection = Regex.Matches(newLine, @"(?:<span[\s\S]*>)([\s\S]*)(?:<\/span>)");
                        if (matchCollection != null && matchCollection.Count() > 0)
                        {
                            var queryGroupsCollection = matchCollection.Select(groups => groups.Groups);
                            var queryGroups = queryGroupsCollection.SelectMany(group => group);
                            var queryGroupWithoutSpan = queryGroups.Where(group => group.Success && !group.Value.Contains("</span>")).FirstOrDefault();
                            string spanContents = queryGroupWithoutSpan.ToString();
                            spanContents = spanContents.Replace("&#13;", "\\par ");
                            outputRTF += spanContents;
                        }

                        // Add CW End
                        if (inlineStyle.Contains(CSSTypes.Bold, StringComparison.OrdinalIgnoreCase))
                        {
                            outputRTF += RTFCommands.BoldEnd;
                        }

                        if (inlineStyle.Contains(CSSTypes.Italics, StringComparison.OrdinalIgnoreCase))
                        {
                            outputRTF += RTFCommands.ItalicsEnd;
                        }

                        if (inlineStyle.Contains(CSSTypes.UnderlineAndStrikethrough.Replace(" ", string.Empty), StringComparison.OrdinalIgnoreCase))
                        {
                            outputRTF += RTFCommands.UnderlineEnd;
                            outputRTF += RTFCommands.StrikethroughEnd;
                        }
                        else if (inlineStyle.Contains(CSSTypes.Underline, StringComparison.OrdinalIgnoreCase))
                        {
                            outputRTF += RTFCommands.UnderlineEnd;
                        }
                        else if (inlineStyle.Contains(CSSTypes.Strikethrough, StringComparison.OrdinalIgnoreCase))
                        {
                            outputRTF += RTFCommands.StrikethroughEnd;
                        }
                    }
                }

            }
            outputRTF += "}";
            return outputRTF;
        }
        #endregion
    }
}
