﻿using pdftron.PDF.Tools.Collections;
using pdftron.PDF.Tools.Controls.ViewModels.Viewer.Tab;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Diagnostics.CodeAnalysis;
using System.IO;
using System.Threading;
using System.Threading.Tasks;
using System.Xml.Serialization;
using Windows.ApplicationModel.Core;
using Windows.Storage;
using Windows.Storage.Streams;
using Windows.UI.Xaml;

namespace pdftron.PDF.Tools.Utilities
{
    /// <summary>
    /// DocumentManager is there to aid in creating temporary files for opening documents, and to make sure that these temporary documents work with the recent list.
    /// In short, the DocumentManager will generate temporary files and keeps a mapping of standard file paths to temporary file paths.
    /// </summary>
    public class DocumentManager
    {
        private const string TEMP_FILE_FOLDER = "OpenDocuments";
        private const string FILE_NAME_MAP_FILE = "FileNameMapping.txt";
        public const string TabPreviewFolderName = "Tab_Preview_Images";
        private const string TEMP_FOLDER_SOUND_FILES = "Temp_Sound_Files";

        private Dictionary<string, string> _fileNameMap;
        private List<string> _openFiles;
        private SemaphoreSlim _loadingSemaphore;
        private SemaphoreSlim _guard;
        private StorageFolder _temporaryFilesFolder;
        DispatcherTimer _cleanupTimer;
        private int CLEANUP_TIMER_INTERVAL_MIN = 30;
        private int CLEANUP_TIMER_INTERVAL_SEC = 0;
        Windows.UI.Core.CoreWindow _coreWindow;

        private static readonly DocumentManager _current = new DocumentManager();

        public async Task InitializeInstanceAsync()
        {
            await LoadingSemaphore.WaitAsync().ConfigureAwait(false);
            try
            {
                if (!IsReady)
                {
                    if (_fileNameMap == null)
                        await LoadFileMappingAsync().ConfigureAwait(false);

                    if (_current._guard == null)
                        _current._guard = new SemaphoreSlim(1);

                    IsReady = true;
                }
            }
            catch (Exception e)
            {
                Debug.WriteLine("Error: " + e.ToString());
            }
            finally
            {
                LoadingSemaphore.Release();
            }
        }

        /// <summary>
        /// This call will perform a full cleanup of temporary files created on runtime
        /// </summary>
        /// <remarks>Calling this on startup might slow down the whole initialization
        /// process since accessing disk relies on OS processes</remarks>
        public async Task PerformFullCleanupAsync()
        {
            await CleanTemporaryFolderAsync();
            await CleanupTabPreviewImgsAsync();
        }

        [SuppressMessage("Simplification", "RCS1085:Use auto-implemented property.", Justification = "<Pending>")]
        public static DocumentManager Instance => _current;

        /// <summary>
        /// Flag to indicate when DocumentManager has been initialized
        /// </summary>
        public static bool IsReady { get; private set; }

        private SemaphoreSlim LoadingSemaphore => _loadingSemaphore ?? (_loadingSemaphore = new SemaphoreSlim(1));

        private async Task LoadFileMappingAsync()
        {
            _fileNameMap = null;
            _openFiles = null;

            try
            {
                StorageFile file = await ApplicationData.Current.LocalFolder.CreateFileAsync(FILE_NAME_MAP_FILE, CreationCollisionOption.OpenIfExists);
                if (file != null)
                {
                    string dictString = await FileIO.ReadTextAsync(file);

                    _fileNameMap = DeSerializeFileDict(dictString);
                    _openFiles = new List<string>();
                    _temporaryFilesFolder = await ApplicationData.Current.LocalFolder.CreateFolderAsync(TEMP_FILE_FOLDER, CreationCollisionOption.OpenIfExists);
                }
            }
            catch (Exception ex)
            {
                Debug.WriteLine("Error creating file dictionary: " + ex.ToString());
            }
        }

        private async Task SaveFileMappingAsync()
        {
            if (_fileNameMap == null)
                return;

            try
            {
                StorageFile file = await ApplicationData.Current.LocalFolder.GetFileAsync(FILE_NAME_MAP_FILE);
                string dictString = SerializeFileDictionary(_fileNameMap);
                await FileIO.WriteTextAsync(file, dictString);
            }
            catch (Exception ex)
            {
                Debug.WriteLine("Error opening file dictionary: " + ex.ToString());
            }
        }

        private static string SerializeFileDictionary(Dictionary<string, string> fileDict)
        {
            try
            {
                List<KeyValuePair> tuples = new List<KeyValuePair>();
                foreach (KeyValuePair<string, string> keyValPair in fileDict)
                {
                    tuples.Add(new KeyValuePair(keyValPair.Key, keyValPair.Value));
                }

                XmlSerializer xmlIzer = new XmlSerializer(typeof(List<KeyValuePair>));
                StringWriter writer = new StringWriter();
                xmlIzer.Serialize(writer, tuples);
                return writer.ToString();
            }
            catch (Exception exc)
            {
                Debug.WriteLine(exc);
            }

            return string.Empty;
        }

        private Dictionary<string, string> DeSerializeFileDict(string fileDictString)
        {
            if (!string.IsNullOrWhiteSpace(fileDictString))
            {
                Dictionary<string, string> fileDict = new Dictionary<string, string>();
                try
                {
                    List<KeyValuePair> tuples;
                    XmlSerializer xmlIzer = new XmlSerializer(typeof(List<KeyValuePair>));
                    using (StringReader sr = new StringReader(fileDictString))
                    {
                        tuples = (xmlIzer.Deserialize(sr)) as List<KeyValuePair>;
                        if (tuples != null)
                        {
                            foreach (KeyValuePair tuple in tuples)
                            {
                                fileDict[tuple.Key] = tuple.Value;
                            }
                            return fileDict;
                        }
                    }
                }
                catch (Exception exc)
                {
                    Debug.WriteLine(exc);
                }
            }

            return new Dictionary<string, string>();
        }

        /// <summary>
        /// Start timer to execute the cleanup of temporary files
        /// </summary>
        public async Task StartAutoCleanAsync()
        {

            await UtilityFunctions.RunAsyncSaferAsync(GetCoreWindow(), () =>
            {
                _cleanupTimer = new DispatcherTimer
                {
                    Interval = new TimeSpan(0, CLEANUP_TIMER_INTERVAL_MIN, CLEANUP_TIMER_INTERVAL_SEC)
                };

                _cleanupTimer.Tick += _cleanupTimer_Tick;
                _cleanupTimer.Start();
            });
        }

        private async void _cleanupTimer_Tick(object sender, object e)
        {
            await PerformFullCleanupAsync();
        }

        /// <summary>
        /// Stop timer to execute the cleanup of temporary files
        /// </summary>
        public async Task StopAutoCleanAsync()
        {
            await UtilityFunctions.RunAsyncSaferAsync(GetCoreWindow(), () =>
            {
                if (_cleanupTimer != null)
                {
                    _cleanupTimer.Stop();
                    _cleanupTimer = null;
                }
            });
        }

        /// <summary>
        /// Gets the path at which the document would have been stored in the recently used cache.
        /// </summary>
        /// <param name="fileName"></param>
        /// <returns></returns>
        public string GetRecentThumbnailPath(string fileName)
        {
            _guard.Wait();

            try
            {
                if (_fileNameMap.ContainsKey(fileName))
                    return _fileNameMap[fileName];
            }
            finally
            {
                _guard.Release();
            }

            return null;
        }

        public async Task<string> GetRecentThumbnailPathAsync(string fileName)
        {
            await _guard.WaitAsync();

            try
            {
                if (_fileNameMap.ContainsKey(fileName))
                    return _fileNameMap[fileName];
            }
            finally
            {
                _guard.Release();
            }

            return null;
        }

        /// <summary>
        /// Copies the file to the temporary folder and returns the new file
        /// The same file will always come back with the same path, provided it is in the recent list.
        /// </summary>
        /// <param name="file">The StorageFile to create a temp copy for</param>
        /// <param name="generateNew">Whether we should always create a new temp file, even if it's already in the list.</param>
        /// <returns></returns>
        public async Task<StorageFile> OpenTemporaryCopyAsync(StorageFile file, bool generateNew = false)
        {
            await _guard.WaitAsync();
            try
            {
                return await OpenTempCopyAsync(file, generateNew);
            }
            finally
            {
                _guard.Release();
            }
        }

        /// <summary>
        /// Call this when a StorageFile is no longer needed.
        /// </summary>
        /// <param name="file"></param>
        public void CloseFile(StorageFile file)
        {
            _guard.Wait();

            try
            {
                if (!string.IsNullOrWhiteSpace(file.Path) && _openFiles.Contains(file.Path))
                    _openFiles.Remove(file.Path);
            }
            finally
            {
                _guard.Release();
            }
        }

        public async Task AddChangesToOriginalAsync(StorageFile original, StorageFile temporaryFile, bool isDropBox = false)
        {
            await _guard.WaitAsync();
            Exception exc = null;
            try
            {
                await temporaryFile.CopyAndReplaceAsync(original);
            }
            catch (Exception e)
            {
                exc = e;
            }
            finally
            {
                _guard.Release();
            }

            if (exc != null)
                throw exc;
        }

        private async Task<StorageFile> OpenTempCopyAsync(StorageFile file, bool generateNew)
        {
            if (!TabControlViewModel.IsReady)
                return null;

            StorageFile newFile = null;
            Tuple<string, string> tempFileName = GetFileName(file.Path, generateNew);
            StorageFolder tempFolder = null;
            try
            {
                tempFolder = await _temporaryFilesFolder.CreateFolderAsync(tempFileName.Item1, CreationCollisionOption.OpenIfExists);
            }
            catch
            {
                //Swallow exception
            }

            if (tempFolder == null)
            {
                try
                {
                    tempFolder = await _temporaryFilesFolder.CreateFolderAsync(tempFileName.Item1, CreationCollisionOption.GenerateUniqueName);
                }
                catch
                {
                    //Swallow exception
                }
            }

            if (tempFolder == null)
                return null;

            bool noTempFile = false;

            if (TabControlViewModel.Instance.ContainsFile(file))
            {
                try
                {
                    IStorageItem item = await tempFolder.TryGetItemAsync(tempFileName.Item2);
                    if (item is StorageFile)
                        newFile = item as StorageFile;

                    if (UtilityFunctions.AreFilesEqual(newFile, file))
                    {
                        newFile = null;
                        noTempFile = true;
                    }
                }
                catch (Exception ex)
                {
                    Debug.WriteLine(ex);
                }
            }

            if (!noTempFile)
            {
                if (newFile == null)
                {
                    bool canShortenName = true;
                    string nameOfTempFile = tempFileName.Item2;
                    string ext = Path.GetExtension(nameOfTempFile);
                    if (!ext.Equals(".pdf", StringComparison.OrdinalIgnoreCase))
                        nameOfTempFile = Path.GetFileNameWithoutExtension(nameOfTempFile) + ".pdf";

                    while (canShortenName)
                    {
                        try
                        {
                            newFile = await file.CopyAsync(tempFolder, nameOfTempFile, NameCollisionOption.ReplaceExisting);
                            canShortenName = false;
                        }
                        catch (PathTooLongException)
                        {
                            nameOfTempFile = UtilityFunctions.ShortenFileName(nameOfTempFile);
                        }
                        catch (Exception ex)
                        {
                            canShortenName = false;
                            Debug.WriteLine(ex);
                        }
                    }
                }

                if (newFile == null)
                {
                    bool canShortenName = true;
                    string nameOfTempFile = tempFileName.Item2;
                    while (canShortenName)
                    {
                        try
                        {
                            newFile = await file.CopyAsync(tempFolder, tempFileName.Item2, NameCollisionOption.GenerateUniqueName);
                            canShortenName = false;
                        }
                        catch (PathTooLongException)
                        {
                            nameOfTempFile = UtilityFunctions.ShortenFileName(nameOfTempFile);
                        }
                        catch (Exception ex)
                        {
                            canShortenName = false;
                            AnalyticsHandlerBase.CURRENT.SendEvent(ex.ToString());
                        }
                    }
                }
            }

            if (newFile != null)
            {
                if (!string.IsNullOrWhiteSpace(file.Path))
                {
                    try
                    {
                        await AddFileNameToListAsync(file.Path, newFile.Path).ConfigureAwait(false);
                    }
                    catch
                    {
                        //Swallow exception
                    }

                    _openFiles.Add(file.Path);
                }

                return newFile;
            }
            else
            {
                if (!string.IsNullOrEmpty(file.Path))
                    await AddFileNameToListAsync(file.Path, file.Path);
            }

            return null;
        }

        private Tuple<string, string> GetFileName(string fileName, bool generateNew)
        {
            if (!string.IsNullOrWhiteSpace(fileName) && !generateNew && _fileNameMap.ContainsKey(fileName))
            {
                string dirName = Path.GetDirectoryName(_fileNameMap[fileName]);
                string[] dirs = dirName.Split("\\".ToCharArray());
                return new Tuple<string, string>(dirs[dirs.Length - 1], Path.GetFileName(_fileNameMap[fileName]));
            }

            Guid guid = Guid.NewGuid();
            return new Tuple<string, string>(guid.ToString(), Path.GetFileName(fileName));
        }

        private async Task AddFileNameToListAsync(string fileName, string tempFileName)
        {
            try
            {
                CleanUpFileMappingList();
            }
            catch
            {
                //Swallow exception
            }

            _fileNameMap[fileName] = tempFileName;
            await SaveFileMappingAsync();
        }

        /// <summary>
        /// Save a PDFDoc to disk as a StorageFile
        /// </summary>
        /// <param name="doc">The PDFDoc to be saved</param>
        /// <param name="newName">File name</param>
        /// <param name="cleanTempfolder"></param>
        /// <returns></returns>
        public async Task<StorageFile> SaveToTemporaryFileAsync(PDFDoc doc, string newName, bool cleanTempfolder = true)
        {
            StorageFile file = await pdftron.PDF.Tools.UtilityFunctions.CreateTemporaryAccessFileAsync(newName, cleanTempfolder);
            if (file != null)
            {
                try
                {
                    await doc.SaveAsync(file, pdftron.SDF.SDFDocSaveOptions.e_incremental);
                    return file;
                }
                catch (Exception ex)
                {
                    AppCenterHandlerBase.CURRENT.TrackError(ex);
                }
            }

            return null;
        }

        /// <summary>
        /// Perform a full cleanup of all unused temporary documents
        /// </summary>
        private async Task CleanTemporaryFolderAsync()
        {
            if (_guard == null)
                return;

            try
            {
                await _guard.WaitAsync();

                if (!TabControlViewModel.IsReady)
                    return;

                // Note: Keep all tasks in a list so we can run them in parallel
                List<Task> tasksCleanup = new List<Task>();

                IReadOnlyList<StorageFile> files = await _temporaryFilesFolder.GetFilesAsync();

                foreach (StorageFile file in files)
                {
                    if (!TabControlViewModel.Instance.ContainsFile(file))
                        tasksCleanup.Add(Task.Run(() => file.DeleteAsync()));
                }

                IReadOnlyList<StorageFolder> folders = await _temporaryFilesFolder.GetFoldersAsync();
                foreach (StorageFolder folder in folders)
                {
                    files = await folder.GetFilesAsync();
                    foreach (StorageFile file in files)
                    {
                        if (!TabControlViewModel.Instance.ContainsFile(file))
                            tasksCleanup.Add(Task.Run(() => folder.DeleteAsync()));
                    }
                }

                // Note: here just wait for all parallel tasks to be completed and then we proceed.
                await Task.WhenAll(tasksCleanup);
            }
            catch(Exception ex)
            {
                AppCenterHandlerBase.CURRENT.TrackError(ex);
            }
            finally
            {
                _guard.Release();
            }
        }

        private void CleanUpFileMappingList()
        {
            if (!RecentItemsData.IsReady || !TabControlViewModel.IsReady)
                return;

            Dictionary<string, string> itemsToDelete = new Dictionary<string, string>();
            foreach (string key in _fileNameMap.Keys)
            {
                if (!TabControlViewModel.Instance.ContainsFilePath(key))
                    itemsToDelete.Add(key, null);
            }

            foreach (RecentItem item in RecentItemsData.Instance.RecentFiles)
            {
                if ((_fileNameMap.ContainsKey(item.DocumentPath) || _openFiles.Contains(item.DocumentPath)) && itemsToDelete.ContainsKey(item.DocumentPath))
                {
                    itemsToDelete.Remove(item.DocumentPath);
                }
            }
            foreach (string key in itemsToDelete.Keys)
            {
                _fileNameMap.Remove(key);
            }
        }

        /// <summary>
        /// Helper class to cleanup Tab Preview Images stored in the app's local folder
        /// </summary>
        /// <param name="tabs"></param>
        /// <returns></returns>
        private async Task CleanupTabPreviewImgsAsync()
        {
            if (!TabControlViewModel.IsReady)
                return;

            try
            {
                await _guard.WaitAsync();

                IList<PdfViewCtrlTabInfo>tabs = new List<PdfViewCtrlTabInfo>(TabControlViewModel.Instance.Tabs);

                IStorageItem storageItem = await ApplicationData.Current.LocalFolder.TryGetItemAsync(TabPreviewFolderName);
                if (storageItem == null)
                    return;

                StorageFolder folder = null;
                if (storageItem is StorageFolder)
                {
                    folder = storageItem as StorageFolder;
                }
                else
                {
                    return;
                }

                IReadOnlyList<StorageFile> fileList = await folder.GetFilesAsync();
                if (fileList.Count > 0)
                {
                    List<StorageFile> filesToDelete = new List<StorageFile>();
                    foreach (StorageFile file in fileList)
                    {
                        bool toDelete = true;
                        foreach (PdfViewCtrlTabInfo tab in tabs)
                        {
                            if (!string.IsNullOrEmpty(tab.MetaData.ThumbnailLocation) && tab.MetaData.ThumbnailLocation.Equals(file.Name))
                                toDelete = false;
                        }

                        if (toDelete)
                            filesToDelete.Add(file);
                    }

                    foreach (StorageFile file in filesToDelete)
                    {
                        try
                        {
                            await file.DeleteAsync();
                        }
                        catch (Exception ex)
                        {
                            AppCenterHandlerBase.CURRENT.TrackError(ex);
                        }
                    }
                }
            }
            catch (Exception ex)
            {
                AppCenterHandlerBase.CURRENT.TrackError(ex);
            }
            finally
            {
                _guard.Release();
            }
        }

        #region Multimedia

        /// <summary>
        /// Saves the audio stream to disk using a WAVE format
        /// </summary>
        /// <remarks>
        /// Temp files will be saved to the temporary folder
        /// </remarks>
        public static async Task<StorageFile> SaveAudioStreamToDiskAsync(IRandomAccessStream stream)
        {
            if (stream == null)
                return null;

            StorageFolder tempFolder = await ApplicationData.Current.TemporaryFolder.CreateFolderAsync(
                TEMP_FOLDER_SOUND_FILES,
                CreationCollisionOption.OpenIfExists);

            if (tempFolder == null)
                return null;

            var audioStream = stream.CloneStream();

            string filename = Guid.NewGuid().ToString() + ".wav"; // save in WAVE format since it is a PCM audio
            StorageFile storageFile = await tempFolder.CreateFileAsync(
                filename,
                CreationCollisionOption.GenerateUniqueName);

            using (IRandomAccessStream fileStream = await storageFile.OpenAsync(FileAccessMode.ReadWrite))
            {
                await RandomAccessStream.CopyAndCloseAsync(audioStream.GetInputStreamAt(0), fileStream.GetOutputStreamAt(0));
                await audioStream.FlushAsync();
                audioStream.Dispose();
            }

            return storageFile;
        }

        /// <summary>
        /// Saves the raw audio bytes in wave format to disk
        /// </summary>
        /// <remarks>
        /// Temp files will be saved to the temporary folder
        /// </remarks>
        public static async Task<StorageFile> SaveAudioWaveToDiskAsync(byte[] bytes)
        {
            if (bytes == null || bytes.Length <= 0)
                return null;

            StorageFolder tempFolder = await ApplicationData.Current.TemporaryFolder.CreateFolderAsync(
                TEMP_FOLDER_SOUND_FILES,
                CreationCollisionOption.OpenIfExists);

            if (tempFolder == null)
                return null;

            string filename = Guid.NewGuid().ToString() + ".wav"; // save in WAVE format since it is a PCM audio
            StorageFile storageFile = await tempFolder.CreateFileAsync(
                filename,
                CreationCollisionOption.GenerateUniqueName);

            var waveFormat = new WaveFormat(
                (int)UtilityFunctions.PCM_AUDIO_SAMPLE_RATE,
                (int)UtilityFunctions.PCM_AUDIO_BITS_PER_SAMPLE,
                (int)UtilityFunctions.PCM_AUDIO_NUM_CHANNELS);

            var s = new RawSourceWaveStream(bytes, 0, bytes.Length, waveFormat);

            return await WaveFileWriterRT.CreateWaveFileAsync(storageFile, s);
        }

        /// <summary>
        /// Delete all temporary sound annotation files created during recording
        /// </summary>
        public static async Task CleanupTempAudioFilesFromDiskAsync()
        {
            var tempSoundFolder = await ApplicationData.Current.TemporaryFolder.TryGetItemAsync(TEMP_FOLDER_SOUND_FILES);
            var folder = tempSoundFolder as StorageFolder;
            if (folder == null)
                return;

            IReadOnlyList<StorageFile> filesToDelete = await folder.GetFilesAsync();
            if (filesToDelete.Count > 0)
            {
                foreach (StorageFile file in filesToDelete)
                {
                    try
                    {
                        await file.DeleteAsync();
                    }
                    catch (Exception ex)
                    {
                        AppCenterHandlerBase.CURRENT.TrackError(ex);
                    }
                }
            }
        }

        #endregion

        /// <summary>
        /// Get cached CoreWindow
        /// </summary>
        private Windows.UI.Core.CoreWindow GetCoreWindow()
        {
            if (_coreWindow == null)
                _coreWindow = CoreApplication.MainView.CoreWindow;

            return _coreWindow;
        }
    }
}
