﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

using Windows.Graphics.Imaging;
using Windows.Storage;
using Windows.Storage.Streams;

using pdftron.Common;
using pdftron.PDF.Tools.Controls.Viewer;
using pdftron.PDF.Tools.Controls.ViewModels.FileOpening;
using pdftron.SDF;


namespace pdftron.PDF.Tools.Utilities
{
    public class PDFDocCreator
    {
        protected const double A0_Width = 2383.2;
        protected const double A0_Height = 3369.6;
        protected const double A1_Width = 1684.8;
        protected const double A1_Height = 2383.2;
        protected const double A2_Width = 1188.0;
        protected const double A2_Height = 1684.8;
        protected const double A3_Width = 842.4;
        protected const double A3_Height = 1188.0;
        protected const double A4_Width = 595.44;
        protected const double A4_Height = 842.4;
        protected const double A5_Width = 419.76;
        protected const double A5_Height = 595.44;

        protected const double B0_Width = 2835.8;
        protected const double B0_Height = 4010.4;
        protected const double B1_Width = 2005.2;
        protected const double B1_Height = 2835.8;
        protected const double B2_Width = 1417.9;
        protected const double B2_Height = 2005.2;
        protected const double B3_Width = 1002.6;
        protected const double B3_Height = 1417.9;
        protected const double B4_Width = 708.9;
        protected const double B4_Height = 1002.6;
        protected const double B5_Width = 499.0;
        protected const double B5_Height = 708.9;

        protected const double Letter_Width = 612.0;
        protected const double Letter_Height = 792.0;
        protected const double Legal_Width = 612.0;
        protected const double Legal_Height = 1008;

        static double LINE_SHADE = 0.35;
        static double MARGIN_RED = 1.0;
        static double MARGIN_GREEN = 0.5;
        static double MARGIN_BLUE = 0.5;

        static uint WHITE_COLOR = 0xFFFFFFFF;

        private static Dictionary<PaperSizes, Tuple<double, double>> _PaperSizeMap;

        /// <summary>
        /// Paper size map
        /// </summary>
        public static Dictionary<PaperSizes, Tuple<double, double>> PaperSizeMap
        {
            get
            {
                return _PaperSizeMap ?? (_PaperSizeMap = new Dictionary<PaperSizes, Tuple<double, double>>
                {
                    [PaperSizes.Legal] = new Tuple<double, double>(Legal_Width, Legal_Height),
                    [PaperSizes.Letter] = new Tuple<double, double>(Letter_Width, Letter_Height),
                    [PaperSizes.A0] = new Tuple<double, double>(A0_Width, A0_Height),
                    [PaperSizes.A1] = new Tuple<double, double>(A1_Width, A1_Height),
                    [PaperSizes.A2] = new Tuple<double, double>(A2_Width, A2_Height),
                    [PaperSizes.A3] = new Tuple<double, double>(A3_Width, A3_Height),
                    [PaperSizes.A4] = new Tuple<double, double>(A4_Width, A4_Height),
                    [PaperSizes.A5] = new Tuple<double, double>(A5_Width, A5_Height),
                    [PaperSizes.B0] = new Tuple<double, double>(B0_Width, B0_Height),
                    [PaperSizes.B1] = new Tuple<double, double>(B1_Width, B1_Height),
                    [PaperSizes.B2] = new Tuple<double, double>(B2_Width, B2_Height),
                    [PaperSizes.B3] = new Tuple<double, double>(B3_Width, B3_Height),
                    [PaperSizes.B4] = new Tuple<double, double>(B4_Width, B4_Height),
                    [PaperSizes.B5] = new Tuple<double, double>(B5_Width, B5_Height)
                });
            }
        }

        /// <summary>
        /// Create a PDF document of a specific type e number of pages
        /// </summary>
        /// <param name="options">Document creation options</param>
        /// <returns>Generated PDFDoc</returns>
        public static PDFDoc Create(CreateDocumentOptions options)
        {
            double width, height;
            PDFDocCreator.GetPaperDimensionsFromSize(options.PaperSize, out width, out height);

            if (options.IsLandscape)
            {
                double temp = width;
                width = height;
                height = temp;
            }

            return Create(options.PaperType, width, height, options.Pages);
        }

        /// <summary>
        /// Create a PDF document of a specific type e number of pages
        /// </summary>
        /// <param name="paperType">Type of paper</param>
        /// <param name="width">The width in page points</param>
        /// <param name="height">The height in page points</param>
        /// <param name="pages">Number of pages</param>
        /// <returns>Generated PDFDoc</returns>
        public static PDFDoc Create(PaperType paperType, double width, double height, int pages)
        {
            double width_in = PagePointToInches(width);
            double height_in = PagePointToInches(height);

            double lineShade = LINE_SHADE;

            double bgR = ((WHITE_COLOR & 0x00FF0000) >> 16) / 255.0;
            double bgG = ((WHITE_COLOR & 0x0000FF00) >> 8) / 255.0;
            double bgB = (WHITE_COLOR & 0x000000FF) / 255.0;

            PDFDoc doc;

            switch (paperType)
            {
                case PaperType.Dotted:
                    doc = PDFDocGenerator.GenerateDottedPaperDoc(width_in, height_in, 0.25, 2.0, lineShade, lineShade, lineShade, bgR, bgG, bgB);
                    break;

                case PaperType.Graph:
                    doc = PDFDocGenerator.GenerateGraphPaperDoc(width_in, height_in, 0.25, 0.45, 1.7, 5, lineShade, lineShade, lineShade, bgR, bgG, bgB);
                    break;

                case PaperType.Grid:
                    doc = PDFDocGenerator.GenerateGridPaperDoc(width_in, height_in, 0.25, 0.45, lineShade, lineShade, lineShade, bgR, bgG, bgB);
                    break;

                case PaperType.Iso:
                    doc = PDFDocGenerator.GenerateIsometricDottedPaperDoc(width_in, height_in, 0.25, 2.0, lineShade, lineShade, lineShade, bgR, bgG, bgB);
                    break;

                case PaperType.Lined:
                    double leftMarginR = MARGIN_RED * 0.7;
                    double leftMarginG = MARGIN_GREEN * 0.7;
                    double leftMarginB = MARGIN_BLUE * 0.7;

                    double rightMarginR = MARGIN_RED * 0.45;
                    double rightMarginG = MARGIN_GREEN * 0.45;
                    double rightMarginB = MARGIN_BLUE * 0.45;

                    doc = PDFDocGenerator.GenerateLinedPaperDoc(width_in, height_in, 0.25, 0.45, lineShade,
                        lineShade, lineShade, 1.2, leftMarginR, leftMarginG, leftMarginB,
                        rightMarginR, rightMarginG, rightMarginB, bgR, bgG, bgB, 0.85, 0.35);
                    break;

                case PaperType.Music:
                    doc = PDFDocGenerator.GenerateMusicPaperDoc(width_in, height_in, 0.5, 10, 6.5, 0.25, lineShade, lineShade, lineShade, bgR, bgG, bgB);
                    break;

                default:
                case PaperType.Blank:
                    doc = PDFDocGenerator.GenerateBlankPaperDoc(width_in, height_in, bgR, bgG, bgB);
                    break;
            }

            for (int i = 1; i < pages; i++)
                doc.PagePushBack(doc.GetPage(1));

            return doc;
        }

        /// <summary>
        /// Create a PDF document from an image file respecting it's dimensions (height and width)
        /// </summary>
        /// <param name="file">The image from a StorageFile instance</param>
        /// <returns>A PDFDoc instance if success, null otherwise</returns>
        public static async Task<PDFDoc> CreateFromImageAsync(StorageFile file)
        {
            if (ViewerSettings.ImagesTypesForConversion.Contains(file.FileType) == false)
                return null;

            PDFDoc doc = new PDFDoc();

            if (file.FileType == ".xps" || file.FileType == ".oxps")
            {
                StorageFile newfile = await file.CopyAsync(ApplicationData.Current.TemporaryFolder, file.Name, NameCollisionOption.ReplaceExisting).AsTask().ConfigureAwait(false);

                pdftron.PDF.Convert.FromXps(doc, newfile.Path);
                return doc;
            }

            SDFDoc sdoc = doc.GetSDFDoc();

            ElementBuilder builder = new ElementBuilder();		// ElementBuilder is used to build new Element objects
            ElementWriter writer = new ElementWriter();	// ElementWriter is used to write Elements to the page

            IRandomAccessStream ras = await file.OpenReadAsync();
            Page page = doc.PageCreate();
            writer.Begin(page); // begin writing to this page

            Image img;
            try
            {
                img = await Image.CreateAsync(sdoc, ras);
            }
            catch (Exception)
            {
                img = null;
            }

            if (img == null)
                return null;

            try
            {
                BitmapDecoder decoder = await BitmapDecoder.CreateAsync(ras);

                uint height = decoder.PixelHeight;
                uint width = decoder.PixelWidth;

                PixelDataProvider pdp2 = await decoder.GetPixelDataAsync(
                    BitmapPixelFormat.Rgba8,
                    BitmapAlphaMode.Straight,
                    new BitmapTransform(),
                    ExifOrientationMode.RespectExifOrientation,
                    ColorManagementMode.DoNotColorManage);

                byte[] pixels = pdp2.DetachPixelData();
                byte[] noAlphaPixels = new byte[(pixels.Length * 3) / 4];

                int j = -1;
                for (int i = 0; i < pixels.Length; i += 4)
                {
                    noAlphaPixels[++j] = pixels[i];
                    noAlphaPixels[++j] = pixels[i + 1];
                    noAlphaPixels[++j] = pixels[i + 2];
                }

                try
                {
                    img = Image.Create(sdoc, noAlphaPixels, (int)width, (int)height, 8, ColorSpace.CreateDeviceRGB(), ImageInputFilter.e_none);
                }
                catch (Exception)
                {
                    img = null;
                }
            }
            catch
            {
                //Swallow exception
            }

            if (img == null)
                return null;

            page.SetCropBox(new Rect(0, 0, img.GetImageWidth(), img.GetImageHeight()));
            page.SetMediaBox(new Rect(0, 0, img.GetImageWidth(), img.GetImageHeight()));

            Element element = builder.CreateImage(img, new Matrix2D(img.GetImageWidth(), 0, 0, img.GetImageHeight(), 0, 0));

            writer.WritePlacedElement(element);

            writer.End();  // save changes to the current page
            doc.PagePushBack(page);

            return doc;
        }

        public static double PagePointToInches(double points)
        {
            return points / 72;
        }

        public static void GetPaperDimensionsFromSize(PaperSizes size, out double width, out double height)
        {
            width = PaperSizeMap[size].Item1;
            height = PaperSizeMap[size].Item2;
        }
    }

    public enum PaperType
    {
        Blank,
        Lined,
        Grid,
        Graph,
        Dotted,
        Iso,
        Music
    }

    public enum PaperSizes
    {
        Letter,
        Legal,
        A0,
        A1,
        A2,
        A3,
        A4,
        A5,
        B0,
        B1,
        B2,
        B3,
        B4,
        B5
    }

    public class PaperTypeHolder
    {
        public PaperType PaperType { get; set; }

        public override string ToString() => PaperType.ToString("g");

        public PaperTypeHolder(PaperType type)
        {
            PaperType = type;
        }
    }

    public class PaperSizeHolder
    {
        public PaperSizes PaperSize { get; }

        public double Width { get; } = 0;

        public double Height { get; } = 0;

        private readonly string _units;

        public override string ToString() => string.Format("{0} ({1:0.00} x {2:0.00} {3})", PaperSize, Width, Height, _units);

        public PaperSizeHolder(PaperSizes size, double w, double h, string units)
        {
            PaperSize = size;
            Width = w;
            Height = h;
            _units = units;
        }
    }
}
