﻿//
// Copyright (c) 2001-2022 by PDFTron Systems Inc. All Rights Reserved.
//

using System;
using System.Collections.Generic;
using System.IO;
using System.Threading.Tasks;
using System.Runtime.InteropServices.WindowsRuntime;
using Windows.Foundation;

using pdftron.PDF;
using pdftron.PDF.Annots;
using pdftron.SDF;

using Buffer = Windows.Storage.Streams.Buffer;
using Convert = System.Convert;
using PDFNetUniversalSamples.ViewModels;
using pdftron.Crypto;

namespace PDFNetSamples
{
    public class DigitalSignaturesTest : Sample
    {
        public DigitalSignaturesTest() :
            base("DigitalSignatures", "Demonstrates the basic usage of high-level digital signature API to digitally sign and/or certify PDF documents.")
        {
        }

        public override IAsyncAction RunAsync()
        {
            return Task.Run(new System.Action(async () => {
                WriteLine("--------------------------------");
                WriteLine("Starting DigitalSignatures Test...");
                WriteLine("--------------------------------\n");

                try
                {
                    bool error = false;

                    //////////////////// TEST 0: 
                    /* Create an approval signature field that we can sign after certifying.
                    (Must be done before calling CertifyOnNextSave/SignOnNextSave/WithCustomHandler.) */
                    try
                    {
                        PDFDoc document = new PDFDoc(Path.Combine(InputPath, "waiver.pdf"));
                        DigitalSignatureField approvalSignatureField = document.CreateDigitalSignatureField("PDFTronApprovalSig");
                        SignatureWidget widgetAnnotApproval = SignatureWidget.Create(document, new pdftron.PDF.Rect(300, 300, 500, 200), approvalSignatureField);
                        Page page1 = document.GetPage(1);
                        page1.AnnotPushBack(widgetAnnotApproval);
                        await document.SaveAsync(Path.Combine(OutputPath, "waiver_withApprovalField_output.pdf"), SDFDocSaveOptions.e_remove_unused);
                        await AddFileToOutputList(Path.Combine(OutputPath, "waiver_withApprovalField_output.pdf")).ConfigureAwait(false);
                    }
                    catch (Exception e)
                    {
                        WriteLine(GetExceptionMessage(e));
                        error = true;
                    }

                    //////////////////// TEST 1: certify a PDF.
                    try
                    {
                        await CertifyPDF(Path.Combine(InputPath, "waiver_withApprovalField.pdf"),
                            "PDFTronCertificationSig",
                            Path.Combine(InputPath, "pdftron.pfx"),
                            "password",
                            Path.Combine(InputPath, "pdftron.bmp"),
                            Path.Combine(OutputPath, "waiver_withApprovalField_certified_output.pdf"));

                        PrintSignaturesInfo(Path.Combine(OutputPath, "waiver_withApprovalField_certified_output.pdf"));
                        
                    }
                    catch (Exception e)
                    {
                        WriteLine(GetExceptionMessage(e));
                        error = true;
                    }

                    //////////////////// TEST 2: sign a PDF with a certification and an unsigned signature field in it.
                    try
                    {
                        await SignPDF(Path.Combine(InputPath, "waiver_withApprovalField_certified.pdf"),
                            "PDFTronApprovalSig",
                            Path.Combine(InputPath, "pdftron.pfx"),
                            "password",
                            Path.Combine(InputPath, "signature.jpg"),
                            Path.Combine(OutputPath, "waiver_withApprovalField_certified_approved_output.pdf"));

                        PrintSignaturesInfo(Path.Combine(OutputPath, "waiver_withApprovalField_certified_approved_output.pdf"));
                    }
                    catch (Exception e)
                    {
                        WriteLine(GetExceptionMessage(e));
                        error = true;
                    }

                    //////////////////// TEST 3: Clear a certification from a document that is certified and has two approval signatures.
                    try
                    {
                        await ClearSignature(Path.Combine(InputPath, "waiver_withApprovalField_certified_approved.pdf"),
                            "PDFTronCertificationSig",
                            Path.Combine(OutputPath, "waiver_withApprovalField_certified_approved_certcleared_output.pdf"));

                        PrintSignaturesInfo(Path.Combine(OutputPath, "waiver_withApprovalField_certified_approved_certcleared_output.pdf"));
                    }
                    catch (Exception e)
                    {
                        WriteLine(GetExceptionMessage(e));
                        error = true;
                    }

                    //////////////////// TEST 4: Verify a document's digital signatures.
                    try
                    {
                        VerifyAllAndPrint(Path.Combine(InputPath, "waiver_withApprovalField_certified_approved.pdf"),
                            Path.Combine(InputPath, "pdftron.cer"));
                    }
                    catch (Exception e)
                    {
                        WriteLine(GetExceptionMessage(e));
                        error = true;
                    }

                    //////////////////// TEST 5: Verify a document's digital signatures in a simple fashion using the document API.
                    try
                    {
                        VerifySimple(Path.Combine(InputPath, "waiver_withApprovalField_certified_approved.pdf"),
                            Path.Combine(InputPath, "pdftron.cer"));
                    }
                    catch (Exception e)
                    {
                        WriteLine(GetExceptionMessage(e));
                        error = true;
                    }

                    //////////////////// TEST 6: Timestamp a document, then add Long Term Validation (LTV) information for the DocTimeStamp.
                    //try
                    //{
                    //    error = !(await TimestampAndEnableLTVAsync(
                    //        Path.Combine(InputPath, "waiver.pdf"),
                    //            Path.Combine(InputPath, "SHA-2 Root USERTrust RSA CA Sectigo timestamping.crt"),
                    //            Path.Combine(InputPath, "signature.jpg"),
                    //            Path.Combine(OutputPath, "waiver_DocTimeStamp_LTV.pdf")));

                    //}
                    //catch (Exception e)
                    //{
                    //    WriteLine(GetExceptionMessage(e));
                    //    error = true;
                    //}

                    //////////////////// End of tests. ////////////////////

                    if (error)
                    {
                       WriteLine("Tests FAILED!!!\n==========");                       
                    }
                    else
                    {
                        WriteLine("Tests successful.\n==========");
                    }
                }
                catch (Exception e)
                {
                    WriteLine(GetExceptionMessage(e));
                }

                WriteLine("\n--------------------------------");
                WriteLine("Done DigitalSignatures Test.");
                WriteLine("--------------------------------\n");
            })).AsAsyncAction();
        }

        private async Task SignPDF(string inputDocumentPath,
            string inputApprovalFieldName,
            string inputPrivateKeyFilePath,
            string inputKeyFilePassword,
            string inputAppearanceImagePath,
            string inputOutPath)
        {
            WriteLine("================================================================================");
            WriteLine("Signing PDF document");

            // Open an existing PDF
            PDFDoc document = new PDFDoc(inputDocumentPath);

            // Sign the approval signatures.
            Field foundApprovalField = document.GetField(inputApprovalFieldName);
            DigitalSignatureField foundApprovalSignatureDigitalSignatureField = new DigitalSignatureField(foundApprovalField);
            Image image = Image.Create(document.GetSDFDoc(), inputAppearanceImagePath);
            SignatureWidget foundApprovalSignatureWidget = new SignatureWidget(foundApprovalField.GetSDFObj());
            foundApprovalSignatureWidget.CreateSignatureAppearance(image);

            foundApprovalSignatureDigitalSignatureField.SignOnNextSave(inputPrivateKeyFilePath, inputKeyFilePassword);

            await document.SaveAsync(inputOutPath, SDFDocSaveOptions.e_incremental);
            WriteLine("Done. Result saved in " + inputOutPath);
            await AddFileToOutputList(inputOutPath).ConfigureAwait(false);
            WriteLine("================================================================================");
        }

        private async Task CertifyPDF(string inputDocumentPath,
            string inputCertificateFieldName,
            string inputPrivateKeyFilePath,
            string inputKeyFilePassword,
            string inputAppearanceImagePath,
            string inputOutPath)
        {
            WriteLine("================================================================================");
            WriteLine("Certifying PDF document");

            // Open an existing PDF
            PDFDoc document = new PDFDoc(inputDocumentPath);
            WriteLine("PDFDoc has " + (document.HasSignatures() ? "signatures" : "no signatures"));

            Page page1 = document.GetPage(1);

            // Create a random text field that we can lock using the field permissions feature.
            TextWidget annotation = TextWidget.Create(document, new pdftron.PDF.Rect(50, 550, 350, 600), "asdf_test_field");
            page1.AnnotPushBack(annotation);

            /* Create new signature form field in the PDFDoc. The name argument is optional;
            leaving it empty causes it to be auto-generated. However, you may need the name for later.
            Acrobat doesn't show digsigfield in side panel if it's without a widget. Using a
            Rect with 0 width and 0 height, or setting the NoPrint/Invisible flags makes it invisible. */
            DigitalSignatureField certificationSignatureField = document.CreateDigitalSignatureField(inputCertificateFieldName);
            SignatureWidget widgetAnnot = SignatureWidget.Create(document, new pdftron.PDF.Rect(0, 100, 200, 150), certificationSignatureField);
            page1.AnnotPushBack(widgetAnnot);

            // (OPTIONAL) Add an appearance.

            // Widget AP from image
            Image image = Image.Create(document.GetSDFDoc(), inputAppearanceImagePath);
            widgetAnnot.CreateSignatureAppearance(image);
            // End of optional appearance-adding code.

            // Add permissions. Lock the random text field.
            WriteLine("Adding document permissions.");
            certificationSignatureField.SetDocumentPermissions(DigitalSignatureFieldDocumentPermissions.e_annotating_formfilling_signing_allowed);
            WriteLine("Adding field permissions.");
            string[] fieldsToLock = new string[1];
            fieldsToLock[0] = "asdf_test_field";
            certificationSignatureField.SetFieldPermissions(DigitalSignatureFieldFieldPermissions.e_include, fieldsToLock);

            certificationSignatureField.CertifyOnNextSave(inputPrivateKeyFilePath, inputKeyFilePassword);

            ///// (OPTIONAL) Add more information to the signature dictionary.
            certificationSignatureField.SetLocation("Vancouver, BC");
            certificationSignatureField.SetReason("Document certification.");
            certificationSignatureField.SetContactInfo("www.pdftron.com");
            ///// End of optional sig info code.

            // Save the PDFDoc. Once the method below is called, PDFNetC will also sign the document using the information provided.
            await document.SaveAsync(inputOutPath, 0);
            WriteLine("Done. Result saved in " + inputOutPath);
            await AddFileToOutputList(inputOutPath).ConfigureAwait(false);
            WriteLine("================================================================================");
        }

        private void PrintSignaturesInfo(string inputDocumentPath)
        {
            try
            {
                WriteLine("================================================================================");
                WriteLine("Reading and printing digital signature information");

                PDFDoc document = new PDFDoc(inputDocumentPath);

                if (!document.HasSignatures())
                {
                    WriteLine("Doc has no signatures.");
                    WriteLine("================================================================================");
                    return;
                }
                else
                {
                    WriteLine("Doc has signatures.");
                }


                for (FieldIterator fieldIterator = document.GetFieldIterator(); fieldIterator.HasNext(); fieldIterator.Next())
                {
                    if (fieldIterator.Current().IsLockedByDigitalSignature())
                    {
                        WriteLine("==========\nField locked by a digital signature");
                    }
                    else
                    {
                        WriteLine("==========\nField not locked by a digital signature");
                    }

                    WriteLine("Field name: " + fieldIterator.Current().GetName());
                    WriteLine("==========");
                }

                WriteLine("====================\nNow iterating over digital signatures only.\n====================");

                DigitalSignatureFieldIterator digitalSignatureFieldIterator = document.GetDigitalSignatureFieldIterator();
                for (; digitalSignatureFieldIterator.HasNext(); digitalSignatureFieldIterator.Next())
                {
                    WriteLine("==========");
                    WriteLine("Field name of digital signature: " + new Field(digitalSignatureFieldIterator.Current().GetSDFObj()).GetName());

                    DigitalSignatureField digitalSignatureField = digitalSignatureFieldIterator.Current();
                    if (!digitalSignatureField.HasCryptographicSignature())
                    {
                        WriteLine("Either digital signature field lacks a digital signature dictionary, " +
                            "or digital signature dictionary lacks a cryptographic Contents entry. " +
                            "Digital signature field is not presently considered signed.\n" +
                            "==========");
                        continue;
                    }

                    int certificateCount = Convert.ToInt32(digitalSignatureField.GetCertCount());
                    WriteLine("Cert count: " + certificateCount);
                    for (int i = 0; i < certificateCount; ++i)
                    {
                        byte[] certificateByteArray = digitalSignatureField.GetCert(Convert.ToUInt32(i));
                        WriteLine("Cert #" + i + " size: " + certificateByteArray.Length);
                    }

                    DigitalSignatureFieldSubFilterType subFilter = digitalSignatureField.GetSubFilter();

                    WriteLine("Subfilter type: " + Convert.ToInt32(subFilter));

                    if (subFilter != DigitalSignatureFieldSubFilterType.e_ETSI_RFC3161)
                    {
                        WriteLine("Signature's signer: " + digitalSignatureField.GetSignatureName());

                        Date signingTime = digitalSignatureField.GetSigningTime();
                        if (signingTime.IsValid())
                        {
                            WriteLine("Signing day: " + Convert.ToInt32(signingTime.day));
                        }

                        WriteLine("Location: " + digitalSignatureField.GetLocation());
                        WriteLine("Reason: " + digitalSignatureField.GetReason());
                        WriteLine("Contact info: " + digitalSignatureField.GetContactInfo());
                    }
                    else
                    {
                        WriteLine("SubFilter == e_ETSI_RFC3161 (DocTimeStamp; no signing info)\n");
                    }

                    WriteLine(((digitalSignatureField.HasVisibleAppearance()) ? "Visible" : "Not visible"));

                    DigitalSignatureFieldDocumentPermissions digitalSignatureDocumentPermissions = digitalSignatureField.GetDocumentPermissions();

                    var lockedFields = digitalSignatureField.GetLockedFields();
                    if (lockedFields != null)
                    {
                        foreach (string fieldName in lockedFields)
                        {
                            WriteLine("This digital signature locks a field named: " + fieldName);
                        }
                    }

                    switch (digitalSignatureDocumentPermissions)
                    {
                        case DigitalSignatureFieldDocumentPermissions.e_no_changes_allowed:
                            WriteLine("No changes to the document can be made without invalidating this digital signature.");
                            break;
                        case DigitalSignatureFieldDocumentPermissions.e_formfilling_signing_allowed:
                            WriteLine("Page template instantiation, form filling, and signing digital signatures are allowed without invalidating this digital signature.");
                            break;
                        case DigitalSignatureFieldDocumentPermissions.e_annotating_formfilling_signing_allowed:
                            WriteLine("Annotating, page template instantiation, form filling, and signing digital signatures are allowed without invalidating this digital signature.");
                            break;
                        case DigitalSignatureFieldDocumentPermissions.e_unrestricted:
                            WriteLine("Document not restricted by this digital signature.");
                            break;
                        default:
                            throw new Exception("Unrecognized digital signature document permission level.");
                    }
                    WriteLine("==========");
                }

                WriteLine("================================================================================");
            }
            catch (Exception e)
            {
                WriteLine(GetExceptionMessage(e));
            }
        }

        private async Task ClearSignature(string inputDocumentPath,
            string inputDigitalSignatureFieldName,
            string inputOutPath)
        {
            WriteLine("================================================================================");
            WriteLine("Clearing certification signature");

            PDFDoc document = new PDFDoc(inputDocumentPath);

            DigitalSignatureField digitalSignatureField = new DigitalSignatureField(document.GetField(inputDigitalSignatureFieldName));

            WriteLine("Clearing signature: " + inputDigitalSignatureFieldName);
            digitalSignatureField.ClearSignature();

            if (!digitalSignatureField.HasCryptographicSignature())
            {
                WriteLine("Cryptographic signature cleared properly.");
            }

            // Save incrementally so as to not invalidate other signatures from previous saves.
            await document.SaveAsync(inputOutPath, SDFDocSaveOptions.e_incremental);
            WriteLine("Done. Result saved in " + inputOutPath);
            await AddFileToOutputList(inputOutPath).ConfigureAwait(false);
            WriteLine("================================================================================");
        }

        private bool VerifyAllAndPrint(string in_docpath, string in_public_key_file_path)
        {
            WriteLine("================================================================================");
            WriteLine("Verify all fields in PDF document");

            PDFDoc doc = new PDFDoc(in_docpath);
            VerificationOptions opts = new VerificationOptions(VerificationOptionsSecurityLevel.e_compatibility_and_archiving);

            // Add trust root to store of trusted certificates contained in VerificationOptions.
            opts.AddTrustedCertificate(in_public_key_file_path,
                (ulong)VerificationOptionsCertificateTrustFlag.e_default_trust | (ulong)VerificationOptionsCertificateTrustFlag.e_certification_trust);

            // Iterate over the signatures and verify all of them.
            bool verification_status = true;

            for (DigitalSignatureFieldIterator digsig_fitr = doc.GetDigitalSignatureFieldIterator(); digsig_fitr.HasNext(); digsig_fitr.Next())
            {
                DigitalSignatureField curr = digsig_fitr.Current(); //digsig_fitr.next();
                VerificationResult result = curr.Verify(opts);
                if (result.GetVerificationStatus())
                {
                    WriteLine("Signature verified, ");
                }
                else
                {
                    WriteLine("Signature verification failed, ");
                    verification_status = false;
                }
                WriteLine(string.Format("objnum: {0}", curr.GetSDFObj().GetObjNum()));

                switch (result.GetDigestAlgorithm())
                {
                    case DigestAlgorithm.e_sha1:
                        WriteLine("Digest algorithm: SHA-1");
                        break;
                    case DigestAlgorithm.e_sha256:
                       WriteLine("Digest algorithm: SHA-256");
                        break;
                    case DigestAlgorithm.e_sha384:
                        WriteLine("Digest algorithm: SHA-384");
                        break;
                    case DigestAlgorithm.e_sha512:
                        WriteLine("Digest algorithm: SHA-512");
                        break;
                    case DigestAlgorithm.e_ripemd160:
                        WriteLine("Digest algorithm: RIPEMD-160");
                        break;
                    case DigestAlgorithm.e_unknown_digest_algorithm:
                        WriteLine("Digest algorithm: unknown");
                        break;
                    default:
                        WriteLine("ERROR - unrecognized digest algorithm");
                        return false;
                }
                WriteLine(String.Format("Detailed verification result: \n\t{0}\n\t{1}\n\t{2}\n\t{3}",
                    result.GetDocumentStatusAsString(),
                    result.GetDigestStatusAsString(),
                    result.GetTrustStatusAsString(),
                    result.GetPermissionsStatusAsString()));

                DisallowedChange[] changes = result.GetDisallowedChanges();

                if (changes != null)
                {
                    foreach (DisallowedChange it2 in changes)
                    {
                        WriteLine(string.Format("\tDisallowed change: {0}, objnum: {1}",
                            it2.GetTypeAsString(),
                            it2.GetObjNum()));
                    }
                }

                // Get and print all the detailed trust-related results, if they are available.
                if (result.HasTrustVerificationResult())
                {
                    TrustVerificationResult trust_verification_result = result.GetTrustVerificationResult();
                    WriteLine(trust_verification_result.WasSuccessful() ? "Trust verified." : "Trust not verifiable.");
                    WriteLine(trust_verification_result.GetResultString());

                    long time_of_verification = trust_verification_result.GetTimeOfTrustVerification();
                    switch (trust_verification_result.GetTimeOfTrustVerificationEnum())
                    {
                        case VerificationOptionsTimeMode.e_current:
                            WriteLine(string.Format("Trust verification attempted with respect to current time (as epoch time): {0}", time_of_verification));
                            break;
                        case VerificationOptionsTimeMode.e_signing:
                            WriteLine(string.Format("Trust verification attempted with respect to signing time (as epoch time): {0}", time_of_verification));
                            break;
                        case VerificationOptionsTimeMode.e_timestamp:
                            WriteLine(string.Format("Trust verification attempted with respect to secure embedded timestamp (as epoch time): {0}", time_of_verification));
                            break;
                        default:
                            WriteLine("unrecognized time enum value");
                            return false;
                    }

                    if (trust_verification_result.GetCertPath().Length == 0)
                    {
                        WriteLine("Could not print certificate path.");
                    }
                    else
                    {
                        WriteLine("Certificate path:");
                        X509Certificate[] cert_path = trust_verification_result.GetCertPath();
                        for (int j = 0; j < cert_path.Length; j++)
                        {
                            WriteLine("\tCertificate:");
                            X509Certificate full_cert = cert_path[j];
                            WriteLine("\t\tIssuer names:");

                            X501AttributeTypeAndValue[] issuer_dn = full_cert.GetIssuerField().GetAllAttributesAndValues();
                            for (int i = 0; i < issuer_dn.Length; i++)
                            {
                                WriteLine("\t\t\t" + issuer_dn[i].GetStringValue());
                            }
                            WriteLine("\t\tSubject names:");
                            X501AttributeTypeAndValue[] subject_dn = full_cert.GetSubjectField().GetAllAttributesAndValues();
                            for (int i = 0; i < subject_dn.Length; i++)
                            {
                                WriteLine("\t\t\t" + subject_dn[i].GetStringValue());
                            }
                            WriteLine("\t\tExtensions:");
                            for (int i = 0; i < full_cert.GetExtensions().Length; i++)
                            {
                                WriteLine("\t\t\t" + full_cert.GetExtensions()[i].ToString());
                            }
                        }
                    }
                }
                else
                {
                    WriteLine("No detailed trust verification result available.");
                }

                string[] unsupported_features = result.GetUnsupportedFeatures();
                if (unsupported_features != null)
                {
                    if (unsupported_features.Length > 0)
                    {
                        WriteLine("Unsupported features:");

                        foreach (String unsupported_feature in unsupported_features)
                        {
                            WriteLine("\t" + unsupported_feature);
                        }
                    }
                }
                WriteLine("================================================================================");
            }

            return verification_status;
        }

        private bool VerifySimple(string in_docpath, string in_public_key_file_path)
        {
            WriteLine("================================================================================");
            WriteLine("Verify PDF document");

            bool result = false;

            PDFDoc doc = new PDFDoc(in_docpath);
            VerificationOptions opts = new VerificationOptions(VerificationOptionsSecurityLevel.e_compatibility_and_archiving);

            // Add trust root to store of trusted certificates contained in VerificationOptions.
            opts.AddTrustedCertificate(in_public_key_file_path,
                (ulong)VerificationOptionsCertificateTrustFlag.e_default_trust | (ulong)VerificationOptionsCertificateTrustFlag.e_certification_trust);

            SignaturesVerificationStatus verifResult = doc.VerifySignedDigitalSignatures(opts);

            switch (verifResult)
            {
                case SignaturesVerificationStatus.e_unsigned:
                    WriteLine("Document has no signed signature fields.");
                    result = false;
                    break;
                /*e_failure == bad doc status, digest status, or permissions status
                (i.e. does not include trust issues, because those are flaky due to being network/config-related)*/
                case SignaturesVerificationStatus.e_failure:
                    WriteLine("Hard failure in verification on at least one signature.");
                    result = false;
                    break;
                case SignaturesVerificationStatus.e_untrusted:
                    WriteLine("Could not verify trust for at least one signature.");
                    result = false;
                    break;
                case SignaturesVerificationStatus.e_unsupported:
                    /*If necessary, call GetUnsupportedFeatures on VerificationResult to check which
                    unsupported features were encountered (requires verification using 'detailed' APIs) */
                    WriteLine("At least one signature contains unsupported features.");
                    result = false;
                    break;
                // unsigned sigs skipped; parts of document may be unsigned (check GetByteRanges on signed sigs to find out)
                case SignaturesVerificationStatus.e_verified:
                    WriteLine("All signed signatures in document verified.");
                    result = true;
                    break;
                default:
                    WriteLine("unrecognized document verification status");
                    result = false;
                    break;
            }
            WriteLine("================================================================================");
            return result;
        }

        async Task<bool> TimestampAndEnableLTVAsync(string in_docpath,
            string in_trusted_cert_path,
            string in_appearance_img_path,
            string inputOutPath)
        {
            PDFDoc doc = new PDFDoc(in_docpath);

            DigitalSignatureField doctimestamp_signature_field = doc.CreateDigitalSignatureField();
            TimestampingConfiguration tst_config = new TimestampingConfiguration("http://timestamp.sectigo.com");
            VerificationOptions opts = new VerificationOptions(VerificationOptionsSecurityLevel.e_compatibility_and_archiving);

            /* It is necessary to add to the VerificationOptions a trusted root certificate corresponding to 
            the chain used by the timestamp authority to sign the timestamp token, in order for the timestamp
            response to be verifiable during DocTimeStamp signing. It is also necessary in the context of this 
            function to do this for the later LTV section, because one needs to be able to verify the DocTimeStamp 
            in order to enable LTV for it, and we re-use the VerificationOptions opts object in that part. */
            opts.AddTrustedCertificate(in_trusted_cert_path);

            /* By default, we only check online for revocation of certificates using the newer and lighter 
            OCSP protocol as opposed to CRL, due to lower resource usage and greater reliability. However, 
            it may be necessary to enable online CRL revocation checking in order to verify some timestamps
            (i.e. those that do not have an OCSP responder URL for all non-trusted certificates). */
            opts.EnableOnlineCRLRevocationChecking(true);

            SignatureWidget widgetAnnot = SignatureWidget.Create(doc, new pdftron.PDF.Rect(0, 100, 200, 150), doctimestamp_signature_field);
            doc.GetPage(1).AnnotPushBack(widgetAnnot);

            // (OPTIONAL) Add an appearance to the signature field.
            pdftron.PDF.Image img = pdftron.PDF.Image.Create(doc.GetSDFDoc(), in_appearance_img_path);
            widgetAnnot.CreateSignatureAppearance(img);

            WriteLine("Testing timestamping configuration.");
            TimestampingResult config_result = tst_config.TestConfiguration(opts);
            if (config_result.GetStatus())
            {
                WriteLine("Success: timestamping configuration usable. Attempting to timestamp.");
            }
            else
            {
                // Print details of timestamping failure.
                WriteLine(config_result.GetString());
                if (config_result.HasResponseVerificationResult())
                {
                    EmbeddedTimestampVerificationResult tst_result = config_result.GetResponseVerificationResult();

                    WriteLine(string.Format("CMS digest status: {0}\n", tst_result.GetCMSDigestStatusAsString()));
                    WriteLine(string.Format("Message digest status: {0}\n", tst_result.GetMessageImprintDigestStatusAsString()));
                    WriteLine(string.Format("Trust status: {0}\n", tst_result.GetTrustStatusAsString()));
                }

                doc.Dispose();
                return false;
            }

            doctimestamp_signature_field.TimestampOnNextSave(tst_config, opts);

            //// Save/signing throws if timestamping fails.
            await doc.SaveAsync(inputOutPath, SDFDocSaveOptions.e_incremental);

            WriteLine("Timestamping successful. Adding LTV information for DocTimeStamp signature.");

            // Add LTV information for timestamp signature to document.
            VerificationResult timestamp_verification_result = doctimestamp_signature_field.Verify(opts);
            if (!doctimestamp_signature_field.EnableLTVOfflineVerification(timestamp_verification_result))
            {
                WriteLine("Could not enable LTV for DocTimeStamp.");
                return false;
            }

            await doc.SaveAsync(inputOutPath, SDFDocSaveOptions.e_incremental);
            WriteLine("Added LTV information for DocTimeStamp signature successfully.");

            WriteLine("Done. Result saved in " + inputOutPath);
            await AddFileToOutputList(inputOutPath).ConfigureAwait(false);
            WriteLine("================================================================================");

            return true;
        }
    }
}
